@@ -1,7 +1,7 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
  */
 /*
@@ -38,14 +38,12 @@
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
+ *
+ * $Id: bt_delete.c,v 12.13 2005/10/20 18:14:59 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: bt_delete.c,v 11.46 2003/06/30 17:19:29 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -222,12 +220,13 @@ __bam_adjindx(dbc, h, indx, indx_copy, is_insert)
  * __bam_dpages --
  *	Delete a set of locked pages.
  *
- * PUBLIC: int __bam_dpages __P((DBC *, EPG *));
+ * PUBLIC: int __bam_dpages __P((DBC *, int, int));
  */
 int
-__bam_dpages(dbc, stack_epg)
+__bam_dpages(dbc, use_top, update)
 	DBC *dbc;
-	EPG *stack_epg;
+	int use_top;
+	int update;
 {
 	BTREE_CURSOR *cp;
 	BINTERNAL *bi;
@@ -235,7 +234,7 @@ __bam_dpages(dbc, stack_epg)
 	DBT a, b;
 	DB_LOCK c_lock, p_lock;
 	DB_MPOOLFILE *mpf;
-	EPG *epg;
+	EPG *epg, *save_sp, *stack_epg;
 	PAGE *child, *parent;
 	db_indx_t nitems;
 	db_pgno_t pgno, root_pgno;
@@ -245,29 +244,27 @@ __bam_dpages(dbc, stack_epg)
 	dbp = dbc->dbp;
 	mpf = dbp->mpf;
 	cp = (BTREE_CURSOR *)dbc->internal;
+	nitems = 0;
+	pgno = PGNO_INVALID;
 
 	/*
 	 * We have the entire stack of deletable pages locked.
 	 *
-	 * Btree calls us with a pointer to the beginning of a stack, where
-	 * the first page in the stack is to have a single item deleted, and
-	 * the rest of the pages are to be removed.
+	 * Btree calls us with the first page in the stack is to have a
+	 * single item deleted, and the rest of the pages are to be removed.
 	 *
-	 * Recno calls us with a pointer into the middle of the stack, where
-	 * the referenced page is to have a single item deleted, and pages
-	 * after the stack reference are to be removed.
-	 *
-	 * First, discard any pages that we don't care about.
+	 * Recno always has a stack to the root and __bam_merge operations
+	 * may have unneeded items in the sack.  We find the lowest page
+	 * in the stack that has more than one record in it and start there.
 	 */
 	ret = 0;
-	for (epg = cp->sp; epg < stack_epg; ++epg) {
-		if ((t_ret = __memp_fput(mpf, epg->page, 0)) != 0 && ret == 0)
-			ret = t_ret;
-		(void)__TLPUT(dbc, epg->lock);
-	}
-	if (ret != 0)
-		goto err;
-
+	if (use_top)
+		stack_epg = cp->sp;
+	else
+		for (stack_epg = cp->csp; stack_epg > cp->sp; --stack_epg)
+			if (NUM_ENT(stack_epg->page) > 1)
+				break;
+	epg = stack_epg;
 	/*
 	 * !!!
 	 * There is an interesting deadlock situation here.  We have to relink
@@ -277,8 +274,9 @@ __bam_dpages(dbc, stack_epg)
 	 * It will deadlock here.  Before we unlink the subtree, we relink the
 	 * leaf page chain.
 	 */
-	if ((ret = __db_relink(dbc, DB_REM_PAGE, cp->csp->page, NULL, 1)) != 0)
-		goto err;
+	if (LEVEL(cp->csp->page) == 1 &&
+	    (ret = __bam_relink(dbc, cp->csp->page, PGNO_INVALID)) != 0)
+		goto discard;
 
 	/*
 	 * Delete the last item that references the underlying pages that are
@@ -289,16 +287,38 @@ __bam_dpages(dbc, stack_epg)
 	 * immediately.
 	 */
 	if ((ret = __bam_ditem(dbc, epg->page, epg->indx)) != 0)
-		goto err;
+		goto discard;
 	if ((ret = __bam_ca_di(dbc, PGNO(epg->page), epg->indx, -1)) != 0)
-		goto err;
+		goto discard;
+
+	if (update && epg->indx == 0) {
+		save_sp = cp->csp;
+		cp->csp = epg;
+		ret = __bam_pupdate(dbc, epg->page);
+		cp->csp = save_sp;
+		if (ret != 0)
+			goto discard;
+	}
 
 	pgno = PGNO(epg->page);
 	nitems = NUM_ENT(epg->page);
 
-	if ((ret = __memp_fput(mpf, epg->page, 0)) != 0)
+	ret = __memp_fput(mpf, epg->page, 0);
+	if ((t_ret = __TLPUT(dbc, epg->lock)) != 0 && ret == 0)
+		ret = t_ret;
+	if (ret != 0)
 		goto err_inc;
-	(void)__TLPUT(dbc, epg->lock);
+
+	/* Then, discard any pages that we don't care about. */
+discard: for (epg = cp->sp; epg < stack_epg; ++epg) {
+		if ((t_ret = __memp_fput(mpf, epg->page, 0)) != 0 && ret == 0)
+			ret = t_ret;
+		epg->page = NULL;
+		if ((t_ret = __TLPUT(dbc, epg->lock)) != 0 && ret == 0)
+			ret = t_ret;
+	}
+	if (ret != 0)
+		goto err;
 
 	/* Free the rest of the pages in the stack. */
 	while (++epg <= cp->csp) {
@@ -309,17 +329,29 @@ __bam_dpages(dbc, stack_epg)
 		 * be referenced by a cursor.
 		 */
 		if (NUM_ENT(epg->page) != 0) {
-			DB_ASSERT(NUM_ENT(epg->page) == 1);
+			DB_ASSERT(LEVEL(epg->page) != 1);
 
 			if ((ret = __bam_ditem(dbc, epg->page, epg->indx)) != 0)
 				goto err;
+			/*
+			 * Sheer paranoia: if we find any pages that aren't
+			 * emptied by the delete, someone else added an item
+			 * while we were walking the tree, and we discontinue
+			 * the delete.  Shouldn't be possible, but we check
+			 * regardless.
+			 */
+			if (NUM_ENT(epg->page) != 0)
+				goto err;
 		}
 
-		if ((ret = __db_free(dbc, epg->page)) != 0) {
-			epg->page = NULL;
+		ret = __db_free(dbc, epg->page);
+		if (cp->page == epg->page)
+			cp->page = NULL;
+		epg->page = NULL;
+		if ((t_ret = __TLPUT(dbc, epg->lock)) != 0 && ret == 0)
+			ret = t_ret;
+		if (ret != 0)
 			goto err_inc;
-		}
-		(void)__TLPUT(dbc, epg->lock);
 	}
 
 	if (0) {
@@ -447,11 +479,13 @@ err:		for (; epg <= cp->csp; ++epg) {
 		if (0) {
 stop:			done = 1;
 		}
-		(void)__TLPUT(dbc, p_lock);
+		if ((t_ret = __TLPUT(dbc, p_lock)) != 0 && ret == 0)
+			ret = t_ret;
 		if (parent != NULL &&
 		    (t_ret = __memp_fput(mpf, parent, 0)) != 0 && ret == 0)
 			ret = t_ret;
-		(void)__TLPUT(dbc, c_lock);
+		if ((t_ret = __TLPUT(dbc, c_lock)) != 0 && ret == 0)
+			ret = t_ret;
 		if (child != NULL &&
 		    (t_ret = __memp_fput(mpf, child, 0)) != 0 && ret == 0)
 			ret = t_ret;
@@ -459,3 +493,151 @@ stop:			done = 1;
 
 	return (ret);
 }
+
+/*
+ * __bam_relink --
+ *	Relink around a deleted page.
+ *
+ * PUBLIC: int __bam_relink __P((DBC *, PAGE *, db_pgno_t));
+ */
+int
+__bam_relink(dbc, pagep, new_pgno)
+	DBC *dbc;
+	PAGE *pagep;
+	db_pgno_t new_pgno;
+{
+	DB *dbp;
+	PAGE *np, *pp;
+	DB_LOCK npl, ppl;
+	DB_LSN *nlsnp, *plsnp, ret_lsn;
+	DB_MPOOLFILE *mpf;
+	int ret, t_ret;
+
+	dbp = dbc->dbp;
+	np = pp = NULL;
+	LOCK_INIT(npl);
+	LOCK_INIT(ppl);
+	nlsnp = plsnp = NULL;
+	mpf = dbp->mpf;
+	ret = 0;
+
+	/*
+	 * Retrieve and lock the one/two pages.  For a remove, we may need
+	 * two pages (the before and after).  For an add, we only need one
+	 * because, the split took care of the prev.
+	 */
+	if (pagep->next_pgno != PGNO_INVALID) {
+		if ((ret = __db_lget(dbc,
+		    0, pagep->next_pgno, DB_LOCK_WRITE, 0, &npl)) != 0)
+			goto err;
+		if ((ret = __memp_fget(mpf, &pagep->next_pgno, 0, &np)) != 0) {
+			ret = __db_pgerr(dbp, pagep->next_pgno, ret);
+			goto err;
+		}
+		nlsnp = &np->lsn;
+	}
+	if (pagep->prev_pgno != PGNO_INVALID) {
+		if ((ret = __db_lget(dbc,
+		    0, pagep->prev_pgno, DB_LOCK_WRITE, 0, &ppl)) != 0)
+			goto err;
+		if ((ret = __memp_fget(mpf, &pagep->prev_pgno, 0, &pp)) != 0) {
+			ret = __db_pgerr(dbp, pagep->prev_pgno, ret);
+			goto err;
+		}
+		plsnp = &pp->lsn;
+	}
+
+	/* Log the change. */
+	if (DBC_LOGGING(dbc)) {
+		if ((ret = __bam_relink_log(dbp, dbc->txn, &ret_lsn, 0,
+		    pagep->pgno, new_pgno, pagep->prev_pgno, plsnp,
+		    pagep->next_pgno, nlsnp)) != 0)
+			goto err;
+	} else
+		LSN_NOT_LOGGED(ret_lsn);
+	if (np != NULL)
+		np->lsn = ret_lsn;
+	if (pp != NULL)
+		pp->lsn = ret_lsn;
+
+	/*
+	 * Modify and release the two pages.
+	 */
+	if (np != NULL) {
+		if (new_pgno == PGNO_INVALID)
+			np->prev_pgno = pagep->prev_pgno;
+		else
+			np->prev_pgno = new_pgno;
+		ret = __memp_fput(mpf, np, DB_MPOOL_DIRTY);
+		if ((t_ret = __TLPUT(dbc, npl)) != 0 && ret == 0)
+			ret = t_ret;
+		if (ret != 0)
+			goto err;
+	}
+
+	if (pp != NULL) {
+		if (new_pgno == PGNO_INVALID)
+			pp->next_pgno = pagep->next_pgno;
+		else
+			pp->next_pgno = new_pgno;
+		ret = __memp_fput(mpf, pp, DB_MPOOL_DIRTY);
+		if ((t_ret = __TLPUT(dbc, ppl)) != 0 && ret == 0)
+			ret = t_ret;
+		if (ret != 0)
+			goto err;
+	}
+	return (0);
+
+err:	if (np != NULL)
+		(void)__memp_fput(mpf, np, 0);
+	(void)__TLPUT(dbc, npl);
+	if (pp != NULL)
+		(void)__memp_fput(mpf, pp, 0);
+	(void)__TLPUT(dbc, ppl);
+	return (ret);
+}
+
+/*
+ * __bam_pupdate --
+ *	Update parent key pointers up the tree.
+ *
+ * PUBLIC: int __bam_pupdate __P((DBC *, PAGE *));
+ */
+int
+__bam_pupdate(dbc, lpg)
+	DBC *dbc;
+	PAGE *lpg;
+{
+	BTREE_CURSOR *cp;
+	DB_ENV *dbenv;
+	EPG *epg;
+	int ret;
+
+	dbenv = dbc->dbp->dbenv;
+	cp = (BTREE_CURSOR *)dbc->internal;
+	ret = 0;
+
+	/*
+	 * Update the parents up the tree.  __bam_pinsert only looks at the
+	 * left child if is a leaf page, so we don't need to change it.  We
+	 * just do a delete and insert; a replace is possible but reusing
+	 * pinsert is better.
+	 */
+	for (epg = &cp->csp[-1]; epg >= cp->sp; epg--) {
+		if ((ret = __bam_ditem(dbc, epg->page, epg->indx)) != 0)
+			return (ret);
+		epg->indx--;
+		if ((ret = __bam_pinsert(dbc, epg,
+		    lpg, epg[1].page, BPI_NORECNUM)) != 0) {
+			if (ret == DB_NEEDSPLIT) {
+				/* This should not happen. */
+				__db_err(dbenv,
+				     "Not enough room in parent: %s: page %lu",
+				     dbc->dbp->fname, (u_long)PGNO(epg->page));
+				ret = __db_panic(dbenv, EINVAL);
+			}
+			return (ret);
+		}
+	}
+	return (ret);
+}
