@@ -1,7 +1,7 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
  */
 /*
@@ -38,14 +38,12 @@
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
+ *
+ * $Id: bt_put.c,v 12.10 2005/10/20 18:57:00 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: bt_put.c,v 11.78 2003/10/31 15:07:40 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -60,7 +58,9 @@ static const char revid[] = "$Id: bt_put.c,v 11.78 2003/10/31 15:07:40 bostic Ex
 
 static int __bam_build
 	       __P((DBC *, u_int32_t, DBT *, PAGE *, u_int32_t, u_int32_t));
-static int __bam_dup_convert __P((DBC *, PAGE *, u_int32_t));
+static int __bam_dup_check __P((DBC *, u_int32_t,
+		PAGE *, u_int32_t, u_int32_t, db_indx_t *));
+static int __bam_dup_convert __P((DBC *, PAGE *, u_int32_t, u_int32_t));
 static int __bam_ovput
 	       __P((DBC *, u_int32_t, db_pgno_t, PAGE *, u_int32_t, DBT *));
 static u_int32_t
@@ -86,11 +86,12 @@ __bam_iitem(dbc, key, data, op, flags)
 	DBT bk_hdr, tdbt;
 	DB_MPOOLFILE *mpf;
 	PAGE *h;
-	db_indx_t indx;
-	u_int32_t data_size, have_bytes, need_bytes, needed;
+	db_indx_t cnt, indx;
+	u_int32_t data_size, have_bytes, need_bytes, needed, pages, pagespace;
 	int cmp, bigkey, bigdata, dupadjust, padrec, replace, ret, was_deleted;
 
 	COMPQUIET(bk, NULL);
+	COMPQUIET(cnt, 0);
 
 	dbp = dbc->dbp;
 	dbenv = dbp->dbenv;
@@ -219,13 +220,39 @@ __bam_iitem(dbc, key, data, op, flags)
 		return (__db_unknown_flag(dbenv, "DB->put", op));
 	}
 
-	/*
-	 * If there's not enough room, or the user has put a ceiling on the
-	 * number of keys permitted in the page, split the page.
-	 */
+	/* Split the page if there's not enough room. */
 	if (P_FREESPACE(dbp, h) < needed)
 		return (DB_NEEDSPLIT);
 
+	/*
+	 * Check to see if we will convert to off page duplicates -- if
+	 * so, we'll need a page.
+	 */
+	if (F_ISSET(dbp, DB_AM_DUP) &&
+	    TYPE(h) == P_LBTREE && op != DB_KEYFIRST &&
+	    P_FREESPACE(dbp, h) - needed <= dbp->pgsize / 2 &&
+	    __bam_dup_check(dbc, op, h, indx, needed, &cnt)) {
+		pages = 1;
+		dupadjust = 1;
+	} else
+		pages = 0;
+
+	/*
+	 * If we are not using transactions and there is a page limit
+	 * set on the file, then figure out if things will fit before
+	 * taking action.
+	 */
+	if (dbc->txn == NULL && dbp->mpf->mfp->maxpgno != 0) {
+		pagespace = P_MAXSPACE(dbp, dbp->pgsize);
+		if (bigdata)
+			pages += ((data_size - 1) / pagespace) + 1;
+		if (bigkey)
+			pages += ((key->size - 1) / pagespace) + 1;
+
+		if (pages > (dbp->mpf->mfp->maxpgno - dbp->mpf->mfp->last_pgno))
+			return (__db_space_err(dbp));
+	}
+
 	/*
 	 * The code breaks it up into five cases:
 	 *
@@ -261,7 +288,6 @@ __bam_iitem(dbc, key, data, op, flags)
 				return (ret);
 
 			indx += 3;
-			dupadjust = 1;
 
 			cp->indx += 2;
 		} else {
@@ -278,7 +304,6 @@ __bam_iitem(dbc, key, data, op, flags)
 				return (ret);
 
 			++indx;
-			dupadjust = 1;
 		}
 		break;
 	case DB_CURRENT:
@@ -289,11 +314,11 @@ __bam_iitem(dbc, key, data, op, flags)
 		  * will try and remove the item because the cursor's delete
 		  * flag is set.
 		  */
-		(void)__bam_ca_delete(dbp, PGNO(h), indx, 0);
+		if ((ret = __bam_ca_delete(dbp, PGNO(h), indx, 0, NULL)) != 0)
+			return (ret);
 
 		if (TYPE(h) == P_LBTREE) {
 			++indx;
-			dupadjust = 1;
 		}
 
 		/*
@@ -382,10 +407,9 @@ __bam_iitem(dbc, key, data, op, flags)
 	 * up at least 25% of the space on the page.  If it does, move it onto
 	 * its own page.
 	 */
-	if (dupadjust && P_FREESPACE(dbp, h) <= dbp->pgsize / 2) {
-		if ((ret = __bam_dup_convert(dbc, h, indx - O_INDX)) != 0)
-			return (ret);
-	}
+	if (dupadjust &&
+	    (ret = __bam_dup_convert(dbc, h, indx - O_INDX, cnt)) != 0)
+		return (ret);
 
 	/* If we've modified a recno file, set the flag. */
 	if (dbc->dbtype == DB_RECNO)
@@ -666,26 +690,22 @@ __bam_ritem(dbc, h, indx, data)
 }
 
 /*
- * __bam_dup_convert --
+ * __bam_dup_check --
  *	Check to see if the duplicate set at indx should have its own page.
- *	If it should, create it.
  */
 static int
-__bam_dup_convert(dbc, h, indx)
+__bam_dup_check(dbc, op, h, indx, sz, cntp)
 	DBC *dbc;
+	u_int32_t op;
 	PAGE *h;
-	u_int32_t indx;
+	u_int32_t indx, sz;
+	db_indx_t *cntp;
 {
 	BKEYDATA *bk;
 	DB *dbp;
-	DBT hdr;
-	DB_MPOOLFILE *mpf;
-	PAGE *dp;
-	db_indx_t cnt, cpindx, dindx, first, *inp, sz;
-	int ret;
+	db_indx_t cnt, first, *inp;
 
 	dbp = dbc->dbp;
-	mpf = dbp->mpf;
 	inp = P_INP(dbp, h);
 
 	/*
@@ -697,12 +717,22 @@ __bam_dup_convert(dbc, h, indx)
 
 	/* Count the key once. */
 	bk = GET_BKEYDATA(dbp, h, indx);
-	sz = B_TYPE(bk->type) == B_KEYDATA ?
+	sz += B_TYPE(bk->type) == B_KEYDATA ?
 	    BKEYDATA_PSIZE(bk->len) : BOVERFLOW_PSIZE;
 
 	/* Sum up all the data items. */
-	for (cnt = 0, first = indx;
-	    inp[first] == inp[indx] && indx < NUM_ENT(h);
+	first = indx;
+
+	/*
+	 * Account for the record being inserted.  If we are replacing it,
+	 * don't count it twice.
+	 *
+	 * We execute the loop with first == indx to get the size of the
+	 * first record.
+	 */
+	cnt = op == DB_CURRENT ? 0 : 1;
+	for (first = indx;
+	    indx < NUM_ENT(h) && inp[first] == inp[indx];
 	    ++cnt, indx += P_INDX) {
 		bk = GET_BKEYDATA(dbp, h, indx + O_INDX);
 		sz += B_TYPE(bk->type) == B_KEYDATA ?
@@ -728,6 +758,36 @@ __bam_dup_convert(dbc, h, indx)
 	if (sz < dbp->pgsize / 4)
 		return (0);
 
+	*cntp = cnt;
+	return (1);
+}
+
+/*
+ * __bam_dup_convert --
+ *	Move a set of duplicates off-page and into their own tree.
+ */
+static int
+__bam_dup_convert(dbc, h, indx, cnt)
+	DBC *dbc;
+	PAGE *h;
+	u_int32_t indx, cnt;
+{
+	BKEYDATA *bk;
+	DB *dbp;
+	DBT hdr;
+	DB_MPOOLFILE *mpf;
+	PAGE *dp;
+	db_indx_t cpindx, dindx, first, *inp;
+	int ret;
+
+	dbp = dbc->dbp;
+	mpf = dbp->mpf;
+	inp = P_INP(dbp, h);
+
+	/* Move to the beginning of the dup set. */
+	while (indx > 0 && inp[indx] == inp[indx - P_INDX])
+		indx -= P_INDX;
+
 	/* Get a new page. */
 	if ((ret = __db_new(dbc,
 	    dbp->dup_compare == NULL ? P_LRECNO : P_LDUP, &dp)) != 0)
@@ -741,8 +801,8 @@ __bam_dup_convert(dbc, h, indx)
 	 * we're dealing with.
 	 */
 	memset(&hdr, 0, sizeof(hdr));
-	dindx = first;
-	indx = first;
+	first = indx;
+	dindx = indx;
 	cpindx = 0;
 	do {
 		/* Move cursors referencing the old entry to the new entry. */
