@@ -1,21 +1,17 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2002
+ * Copyright (c) 1997-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: bt_recno.c,v 12.6 2005/08/08 14:27:59 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: bt_recno.c,v 11.106 2002/08/16 04:56:30 ubell Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
-#include <limits.h>
-#include <stdio.h>
 #include <string.h>
 #endif
 
@@ -58,7 +54,7 @@ static int  __ram_update __P((DBC *, db_recno_t, int));
 	}								\
 }
 #define	CD_ISSET(cp)							\
-	(F_ISSET(cp, C_RENUMBER) && F_ISSET(cp, C_DELETED))
+	(F_ISSET(cp, C_RENUMBER) && F_ISSET(cp, C_DELETED) ? 1 : 0)
 
 /*
  * Macros for comparing the ordering of two cursors.
@@ -91,11 +87,13 @@ static int  __ram_update __P((DBC *, db_recno_t, int));
  * After a search, copy the found page into the cursor, discarding any
  * currently held lock.
  */
-#define	STACK_TO_CURSOR(cp) {						\
+#define	STACK_TO_CURSOR(cp, ret) {					\
+	int __t_ret;							\
 	(cp)->page = (cp)->csp->page;					\
 	(cp)->pgno = (cp)->csp->page->pgno;				\
 	(cp)->indx = (cp)->csp->indx;					\
-	(void)__TLPUT(dbc, (cp)->lock);					\
+	if ((__t_ret = __TLPUT(dbc, (cp)->lock)) != 0 && (ret) == 0)	\
+		ret = __t_ret;						\
 	(cp)->lock = (cp)->csp->lock;					\
 	(cp)->lock_mode = (cp)->csp->lock_mode;				\
 }
@@ -122,9 +120,6 @@ __ram_open(dbp, txn, name, base_pgno, flags)
 	COMPQUIET(name, NULL);
 	t = dbp->bt_internal;
 
-	/* Initialize the remaining fields/methods of the DB. */
-	dbp->stat = __bam_stat;
-
 	/* Start up the tree. */
 	if ((ret = __bam_read_root(dbp, txn, base_pgno, flags)) != 0)
 		return (ret);
@@ -143,7 +138,7 @@ __ram_open(dbp, txn, name, base_pgno, flags)
 	/* If we're snapshotting an underlying source file, do it now. */
 	if (F_ISSET(dbp, DB_AM_SNAPSHOT)) {
 		/* Allocate a cursor. */
-		if ((ret = dbp->cursor(dbp, NULL, &dbc, 0)) != 0)
+		if ((ret = __db_cursor(dbp, NULL, &dbc, 0)) != 0)
 			return (ret);
 
 		/* Do the snapshot. */
@@ -152,7 +147,7 @@ __ram_open(dbp, txn, name, base_pgno, flags)
 			ret = 0;
 
 		/* Discard the cursor. */
-		if ((t_ret = dbc->c_close(dbc)) != 0 && ret == 0)
+		if ((t_ret = __db_c_close(dbc)) != 0 && ret == 0)
 			ret = t_ret;
 	}
 
@@ -208,8 +203,7 @@ __ram_c_del(dbc)
 	DB *dbp;
 	DB_LSN lsn;
 	DBT hdr, data;
-	EPG *epg;
-	int exact, ret, stack;
+	int exact, ret, stack, t_ret;
 
 	dbp = dbc->dbp;
 	cp = (BTREE_CURSOR *)dbc->internal;
@@ -240,7 +234,9 @@ __ram_c_del(dbc)
 	stack = 1;
 
 	/* Copy the page into the cursor. */
-	STACK_TO_CURSOR(cp);
+	STACK_TO_CURSOR(cp, ret);
+	if (ret != 0)
+		goto err;
 
 	/*
 	 * If re-numbering records, the on-page deleted flag can only mean
@@ -262,7 +258,8 @@ __ram_c_del(dbc)
 		/* Delete the item, adjust the counts, adjust the cursors. */
 		if ((ret = __bam_ditem(dbc, cp->page, cp->indx)) != 0)
 			goto err;
-		__bam_adjust(dbc, -1);
+		if ((ret = __bam_adjust(dbc, -1)) != 0)
+			goto err;
 		if (__ram_ca(dbc, CA_DELETE) > 0 &&
 		    CURADJ_LOG(dbc) && (ret = __bam_rcuradj_log(dbp, dbc->txn,
 		    &lsn, 0, CA_DELETE, cp->root, cp->recno, cp->order)) != 0)
@@ -282,21 +279,11 @@ __ram_c_del(dbc)
 		 * are closed, and then clean it up.
 		 */
 		if (NUM_ENT(cp->page) == 0 && PGNO(cp->page) != cp->root) {
-			/*
-			 * We already have a locked stack of pages.  However,
-			 * there are likely entries in the stack that aren't
-			 * going to be emptied by removing the single reference
-			 * to the emptied page (or one of its parents).
-			 */
-			for (epg = cp->csp; epg >= cp->sp; --epg)
-				if (NUM_ENT(epg->page) > 1)
-					break;
-
 			/*
 			 * We want to delete a single item out of the last page
 			 * that we're not deleting.
 			 */
-			ret = __bam_dpages(dbc, epg);
+			ret = __bam_dpages(dbc, 0, 0);
 
 			/*
 			 * Regardless of the return from __bam_dpages, it will
@@ -325,8 +312,8 @@ __ram_c_del(dbc)
 
 	t->re_modified = 1;
 
-err:	if (stack)
-		__bam_stkrel(dbc, STK_CLRDBC);
+err:	if (stack && (t_ret = __bam_stkrel(dbc, STK_CLRDBC)) != 0 && ret == 0)
+		ret = t_ret;
 
 	return (ret);
 }
@@ -388,8 +375,13 @@ retry:	switch (flags) {
 		 * we have to avoid incrementing the record number so that we
 		 * return the right record by virtue of renumbering the tree.
 		 */
-		if (CD_ISSET(cp))
+		if (CD_ISSET(cp)) {
+			/*
+			 * Clear the flag, we've moved off the deleted record.
+			 */
+			CD_CLR(cp);
 			break;
+		}
 
 		if (cp->recno != RECNO_OOB) {
 			++cp->recno;
@@ -494,7 +486,9 @@ retry:	switch (flags) {
 		}
 
 		/* Copy the page into the cursor. */
-		STACK_TO_CURSOR(cp);
+		STACK_TO_CURSOR(cp, ret);
+		if (ret != 0)
+			goto err;
 
 		/*
 		 * If re-numbering records, the on-page deleted flag means this
@@ -607,9 +601,11 @@ __ram_c_put(dbc, key, data, flags, pgnop)
 				return (ret);
 			if (CURADJ_LOG(dbc) &&
 			    (ret = __bam_rcuradj_log(dbp, dbc->txn, &lsn, 0,
-			    CA_ICURRENT, cp->root, cp->recno, cp->order)))
+			    CA_ICURRENT, cp->root, cp->recno, cp->order)) != 0)
 				return (ret);
 			return (0);
+		default:
+			break;
 		}
 
 	/*
@@ -650,7 +646,9 @@ split:	if ((ret = __bam_rsearch(dbc, &cp->recno, S_INSERT, 1, &exact)) != 0)
 	DB_ASSERT(exact || CD_ISSET(cp));
 
 	/* Copy the page into the cursor. */
-	STACK_TO_CURSOR(cp);
+	STACK_TO_CURSOR(cp, ret);
+	if (ret != 0)
+		goto err;
 
 	ret = __bam_iitem(dbc, key, data, iiflags, 0);
 	t_ret = __bam_stkrel(dbc, STK_CLRDBC);
@@ -708,6 +706,8 @@ split:	if ((ret = __bam_rsearch(dbc, &cp->recno, S_INSERT, 1, &exact)) != 0)
 		    CA_ICURRENT, cp->root, cp->recno, cp->order)) != 0)
 			goto err;
 		break;
+	default:
+		break;
 	}
 
 	/* Return the key if we've created a new record. */
@@ -753,7 +753,7 @@ __ram_ca(dbc_arg, op)
 	 */
 	DB_ASSERT(F_ISSET(cp_arg, C_RENUMBER));
 
-	MUTEX_THREAD_LOCK(dbenv, dbenv->dblist_mutexp);
+	MUTEX_LOCK(dbenv, dbenv->mtx_dblist);
 	/*
 	 * Adjust the cursors.  See the comment in __bam_ca_delete().
 	 */
@@ -769,7 +769,7 @@ __ram_ca(dbc_arg, op)
 		for (ldbp = __dblist_get(dbenv, dbp->adj_fileid);
 		    ldbp != NULL && ldbp->adj_fileid == dbp->adj_fileid;
 		    ldbp = LIST_NEXT(ldbp, dblistlinks)) {
-			MUTEX_THREAD_LOCK(dbenv, dbp->mutexp);
+			MUTEX_LOCK(dbenv, dbp->mutex);
 			for (dbc = TAILQ_FIRST(&ldbp->active_queue);
 			    dbc != NULL; dbc = TAILQ_NEXT(dbc, links)) {
 				cp = (BTREE_CURSOR *)dbc->internal;
@@ -778,7 +778,7 @@ __ram_ca(dbc_arg, op)
 				    order <= cp->order)
 					order = cp->order + 1;
 			}
-			MUTEX_THREAD_UNLOCK(dbenv, dbp->mutexp);
+			MUTEX_UNLOCK(dbenv, dbp->mutex);
 		}
 	} else
 		order = INVALID_ORDER;
@@ -787,7 +787,7 @@ __ram_ca(dbc_arg, op)
 	for (ldbp = __dblist_get(dbenv, dbp->adj_fileid);
 	    ldbp != NULL && ldbp->adj_fileid == dbp->adj_fileid;
 	    ldbp = LIST_NEXT(ldbp, dblistlinks)) {
-		MUTEX_THREAD_LOCK(dbenv, dbp->mutexp);
+		MUTEX_LOCK(dbenv, dbp->mutex);
 		for (dbc = TAILQ_FIRST(&ldbp->active_queue);
 		    dbc != NULL; dbc = TAILQ_NEXT(dbc, links)) {
 			cp = (BTREE_CURSOR *)dbc->internal;
@@ -857,9 +857,9 @@ iafter:				if (!adjusted && C_LESSTHAN(cp_arg, cp)) {
 				break;
 			}
 		}
-		MUTEX_THREAD_UNLOCK(dbp->dbenv, dbp->mutexp);
+		MUTEX_UNLOCK(dbp->dbenv, dbp->mutex);
 	}
-	MUTEX_THREAD_UNLOCK(dbenv, dbenv->dblist_mutexp);
+	MUTEX_UNLOCK(dbenv, dbenv->mtx_dblist);
 
 	return (found);
 }
@@ -983,7 +983,7 @@ __ram_source(dbp)
 	 * when it comes time to write the database back to the source.
 	 */
 	if ((t->re_fp = fopen(t->re_source, "r")) == NULL) {
-		ret = errno;
+		ret = __os_get_errno();
 		__db_err(dbp->dbenv, "%s: %s", t->re_source, db_strerror(ret));
 		return (ret);
 	}
@@ -1026,8 +1026,20 @@ __ram_writeback(dbp)
 		return (0);
 	}
 
+	/*
+	 * We step through the records, writing each one out.  Use the record
+	 * number and the dbp->get() function, instead of a cursor, so we find
+	 * and write out "deleted" or non-existent records.  The DB handle may
+	 * be threaded, so allocate memory as we go.
+	 */
+	memset(&key, 0, sizeof(key));
+	key.size = sizeof(db_recno_t);
+	key.data = &keyno;
+	memset(&data, 0, sizeof(data));
+	F_SET(&data, DB_DBT_REALLOC);
+
 	/* Allocate a cursor. */
-	if ((ret = dbp->cursor(dbp, NULL, &dbc, 0)) != 0)
+	if ((ret = __db_cursor(dbp, NULL, &dbc, 0)) != 0)
 		return (ret);
 
 	/*
@@ -1053,58 +1065,47 @@ __ram_writeback(dbp)
 	 */
 	if ((ret =
 	    __ram_update(dbc, DB_MAX_RECORDS, 0)) != 0 && ret != DB_NOTFOUND)
-		return (ret);
+		goto err;
 
 	/*
 	 * Close any existing file handle and re-open the file, truncating it.
 	 */
 	if (t->re_fp != NULL) {
 		if (fclose(t->re_fp) != 0) {
-			ret = errno;
+			ret = __os_get_errno();
 			goto err;
 		}
 		t->re_fp = NULL;
 	}
 	if ((fp = fopen(t->re_source, "w")) == NULL) {
-		ret = errno;
+		ret = __os_get_errno();
 		__db_err(dbenv, "%s: %s", t->re_source, db_strerror(ret));
 		goto err;
 	}
 
-	/*
-	 * We step through the records, writing each one out.  Use the record
-	 * number and the dbp->get() function, instead of a cursor, so we find
-	 * and write out "deleted" or non-existent records.  The DB handle may
-	 * be threaded, so allocate memory as we go.
-	 */
-	memset(&key, 0, sizeof(key));
-	key.size = sizeof(db_recno_t);
-	key.data = &keyno;
-	memset(&data, 0, sizeof(data));
-	F_SET(&data, DB_DBT_REALLOC);
-
 	/*
 	 * We'll need the delimiter if we're doing variable-length records,
 	 * and the pad character if we're doing fixed-length records.
 	 */
 	delim = t->re_delim;
-	if (F_ISSET(dbp, DB_AM_FIXEDLEN)) {
-		if ((ret = __os_malloc(dbenv, t->re_len, &pad)) != 0)
-			goto err;
-		memset(pad, t->re_pad, t->re_len);
-	}
 	for (keyno = 1;; ++keyno) {
-		switch (ret = dbp->get(dbp, NULL, &key, &data, 0)) {
+		switch (ret = __db_get(dbp, NULL, &key, &data, 0)) {
 		case 0:
-			if (data.size != 0 && (u_int32_t)fwrite(
-			    data.data, 1, data.size, fp) != data.size)
+			if (data.size != 0 &&
+			    fwrite(data.data, 1, data.size, fp) != data.size)
 				goto write_err;
 			break;
 		case DB_KEYEMPTY:
-			if (F_ISSET(dbp, DB_AM_FIXEDLEN) &&
-			    (u_int32_t)fwrite(pad, 1, t->re_len, fp) !=
-			    t->re_len)
-				goto write_err;
+			if (F_ISSET(dbp, DB_AM_FIXEDLEN)) {
+				if (pad == NULL) {
+					if ((ret = __os_malloc(
+					    dbenv, t->re_len, &pad)) != 0)
+						goto err;
+					memset(pad, t->re_pad, t->re_len);
+				}
+				if (fwrite(pad, 1, t->re_len, fp) != t->re_len)
+					goto write_err;
+			}
 			break;
 		case DB_NOTFOUND:
 			ret = 0;
@@ -1114,8 +1115,8 @@ __ram_writeback(dbp)
 		}
 		if (!F_ISSET(dbp, DB_AM_FIXEDLEN) &&
 		    fwrite(&delim, 1, 1, fp) != 1) {
-write_err:		ret = errno;
-			__db_err(dbp->dbenv,
+write_err:		ret = __os_get_errno();
+			__db_err(dbenv,
 			    "%s: write failed to backing file: %s",
 			    t->re_source, strerror(ret));
 			goto err;
@@ -1125,13 +1126,14 @@ write_err:		ret = errno;
 err:
 done:	/* Close the file descriptor. */
 	if (fp != NULL && fclose(fp) != 0) {
+		t_ret = __os_get_errno();
 		if (ret == 0)
-			ret = errno;
-		__db_err(dbenv, "%s: %s", t->re_source, db_strerror(errno));
+			ret = t_ret;
+		__db_err(dbenv, "%s: %s", t->re_source, db_strerror(t_ret));
 	}
 
 	/* Discard the cursor. */
-	if ((t_ret = dbc->c_close(dbc)) != 0 && ret == 0)
+	if ((t_ret = __db_c_close(dbc)) != 0 && ret == 0)
 		ret = t_ret;
 
 	/* Discard memory allocated to hold the data items. */
@@ -1259,7 +1261,7 @@ __ram_add(dbc, recnop, data, flags, bi_flags)
 	u_int32_t flags, bi_flags;
 {
 	BTREE_CURSOR *cp;
-	int exact, ret, stack;
+	int exact, ret, stack, t_ret;
 
 	cp = (BTREE_CURSOR *)dbc->internal;
 
@@ -1270,7 +1272,9 @@ retry:	/* Find the slot for insertion. */
 	stack = 1;
 
 	/* Copy the page into the cursor. */
-	STACK_TO_CURSOR(cp);
+	STACK_TO_CURSOR(cp, ret);
+	if (ret != 0)
+		goto err;
 
 	/*
 	 * The application may modify the data based on the selected record
@@ -1320,8 +1324,8 @@ retry:	/* Find the slot for insertion. */
 		goto err;
 	}
 
-err:	if (stack)
-		__bam_stkrel(dbc, STK_CLRDBC);
+err:	if (stack && (t_ret = __bam_stkrel(dbc, STK_CLRDBC)) != 0 && ret == 0)
+		ret = t_ret;
 
 	return (ret);
 }
