@@ -1,7 +1,7 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
  */
 /*
@@ -35,14 +35,12 @@
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
+ *
+ * $Id: bt_rsearch.c,v 12.5 2005/08/08 03:37:05 ubell Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: bt_rsearch.c,v 11.37 2003/06/30 17:19:34 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 #endif
@@ -78,11 +76,12 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 	db_lockmode_t lock_mode;
 	db_pgno_t pg;
 	db_recno_t recno, t_recno, total;
-	int ret, stack;
+	int ret, stack, t_ret;
 
 	dbp = dbc->dbp;
 	mpf = dbp->mpf;
 	cp = (BTREE_CURSOR *)dbc->internal;
+	h = NULL;
 
 	BT_STK_CLR(cp);
 
@@ -101,41 +100,14 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 	 *
 	 * Retrieve the root page.
 	 */
-	pg = cp->root;
-	stack = LF_ISSET(S_STACK) ? 1 : 0;
-	lock_mode = stack ? DB_LOCK_WRITE : DB_LOCK_READ;
-	if ((ret = __db_lget(dbc, 0, pg, lock_mode, 0, &lock)) != 0)
-		return (ret);
-	if ((ret = __memp_fget(mpf, &pg, 0, &h)) != 0) {
-		/* Did not read it, so we can release the lock */
-		(void)__LPUT(dbc, lock);
-		return (ret);
-	}
 
-	/*
-	 * Decide if we need to save this page; if we do, write lock it.
-	 * We deliberately don't lock-couple on this call.  If the tree
-	 * is tiny, i.e., one page, and two threads are busily updating
-	 * the root page, we're almost guaranteed deadlocks galore, as
-	 * each one gets a read lock and then blocks the other's attempt
-	 * for a write lock.
-	 */
-	if (!stack &&
-	    ((LF_ISSET(S_PARENT) && (u_int8_t)(stop + 1) >= h->level) ||
-	    (LF_ISSET(S_WRITE) && h->level == LEAFLEVEL))) {
-		(void)__memp_fput(mpf, h, 0);
-		(void)__LPUT(dbc, lock);
-		lock_mode = DB_LOCK_WRITE;
-		if ((ret = __db_lget(dbc, 0, pg, lock_mode, 0, &lock)) != 0)
-			return (ret);
-		if ((ret = __memp_fget(mpf, &pg, 0, &h)) != 0) {
-			/* Did not read it, so we can release the lock */
-			(void)__LPUT(dbc, lock);
-			return (ret);
-		}
-		stack = 1;
-	}
+	if ((ret = __bam_get_root(dbc, cp->root, stop, flags, &stack)) != 0)
+		return (ret);
+	lock_mode = cp->csp->lock_mode;
+	lock = cp->csp->lock;
+	h = cp->csp->page;
 
+	BT_STK_CLR(cp);
 	/*
 	 * If appending to the tree, set the record number now -- we have the
 	 * root page locked.
@@ -167,9 +139,11 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 				 * eliminate any concurrency.  A possible fix
 				 * would be to lock the last leaf page instead.
 				 */
-				(void)__memp_fput(mpf, h, 0);
-				(void)__TLPUT(dbc, lock);
-				return (DB_NOTFOUND);
+				ret = __memp_fput(mpf, h, 0);
+				if ((t_ret =
+				    __TLPUT(dbc, lock)) != 0 && ret == 0)
+					ret = t_ret;
+				return (ret == 0 ? DB_NOTFOUND : ret);
 			}
 		}
 	}
@@ -201,9 +175,13 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 					*exactp = 0;
 					if (!LF_ISSET(S_PAST_EOF) ||
 					    recno > t_recno + 1) {
-						(void)__memp_fput(mpf, h, 0);
-						(void)__TLPUT(dbc, lock);
-						ret = DB_NOTFOUND;
+						ret = __memp_fput(mpf, h, 0);
+						h = NULL;
+						if ((t_ret = __TLPUT(dbc,
+						    lock)) != 0 && ret == 0)
+							ret = t_ret;
+						if (ret == 0)
+							ret = DB_NOTFOUND;
 						goto err;
 					}
 				}
@@ -252,19 +230,20 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 		}
 		--indx;
 
+		/* Return if this is the lowest page wanted. */
+		if (stop == LEVEL(h)) {
+			BT_STK_ENTER(dbp->dbenv,
+			    cp, h, indx, lock, lock_mode, ret);
+			if (ret != 0)
+				goto err;
+			return (0);
+		}
 		if (stack) {
-			/* Return if this is the lowest page wanted. */
-			if (LF_ISSET(S_PARENT) && stop == h->level) {
-				BT_STK_ENTER(dbp->dbenv,
-				    cp, h, indx, lock, lock_mode, ret);
-				if (ret != 0)
-					goto err;
-				return (0);
-			}
 			BT_STK_PUSH(dbp->dbenv,
 			    cp, h, indx, lock, lock_mode, ret);
 			if (ret != 0)
 				goto err;
+			h = NULL;
 
 			lock_mode = DB_LOCK_WRITE;
 			if ((ret =
@@ -277,11 +256,13 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 			 * never unlock it.
 			 */
 			if ((LF_ISSET(S_PARENT) &&
-			    (u_int8_t)(stop + 1) >= (u_int8_t)(h->level - 1)) ||
-			    (h->level - 1) == LEAFLEVEL)
+			    (u_int8_t)(stop + 1) >= (u_int8_t)(LEVEL(h) - 1)) ||
+			    (LEVEL(h) - 1) == LEAFLEVEL)
 				stack = 1;
 
-			(void)__memp_fput(mpf, h, 0);
+			if ((ret = __memp_fput(mpf, h, 0)) != 0)
+				goto err;
+			h = NULL;
 
 			lock_mode = stack &&
 			    LF_ISSET(S_WRITE) ? DB_LOCK_WRITE : DB_LOCK_READ;
@@ -292,7 +273,7 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 				 * is OK because this only happens when we are
 				 * descending the tree holding read-locks.
 				 */
-				__LPUT(dbc, lock);
+				(void)__LPUT(dbc, lock);
 				goto err;
 			}
 		}
@@ -302,8 +283,12 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 	}
 	/* NOTREACHED */
 
-err:	BT_STK_POP(cp);
+err:	if (h != NULL && (t_ret = __memp_fput(mpf, h, 0)) != 0 && ret == 0)
+		ret = t_ret;
+
+	BT_STK_POP(cp);
 	__bam_stkrel(dbc, 0);
+
 	return (ret);
 }
 
@@ -378,7 +363,7 @@ __bam_nrecs(dbc, rep)
 	DB_MPOOLFILE *mpf;
 	PAGE *h;
 	db_pgno_t pgno;
-	int ret;
+	int ret, t_ret;
 
 	dbp = dbc->dbp;
 	mpf = dbp->mpf;
@@ -391,10 +376,11 @@ __bam_nrecs(dbc, rep)
 
 	*rep = RE_NREC(h);
 
-	(void)__memp_fput(mpf, h, 0);
-	(void)__TLPUT(dbc, lock);
+	ret = __memp_fput(mpf, h, 0);
+	if ((t_ret = __TLPUT(dbc, lock)) != 0 && ret == 0)
+		ret = t_ret;
 
-	return (0);
+	return (ret);
 }
 
 /*
