@@ -1,7 +1,7 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
  */
 /*
@@ -38,14 +38,12 @@
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
+ *
+ * $Id: bt_search.c,v 12.17 2005/11/10 21:17:13 ubell Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: bt_search.c,v 11.47 2003/06/30 17:19:35 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -60,50 +58,32 @@ static const char revid[] = "$Id: bt_search.c,v 11.47 2003/06/30 17:19:35 bostic
 #include "dbinc/mp.h"
 
 /*
- * __bam_search --
- *	Search a btree for a key.
+ * __bam_get_root --
+ *	Fetch the root of a tree and see if we want to keep
+ * it in the stack.
  *
- * PUBLIC: int __bam_search __P((DBC *, db_pgno_t,
- * PUBLIC:     const DBT *, u_int32_t, int, db_recno_t *, int *));
+ * PUBLIC: int __bam_get_root __P((DBC *, db_pgno_t, int, u_int32_t, int *));
  */
 int
-__bam_search(dbc, root_pgno, key, flags, stop, recnop, exactp)
+__bam_get_root(dbc, pg, slevel, flags, stack)
 	DBC *dbc;
-	db_pgno_t root_pgno;
-	const DBT *key;
+	db_pgno_t pg;
+	int slevel;
 	u_int32_t flags;
-	int stop, *exactp;
-	db_recno_t *recnop;
+	int *stack;
 {
-	BTREE *t;
 	BTREE_CURSOR *cp;
 	DB *dbp;
 	DB_LOCK lock;
 	DB_MPOOLFILE *mpf;
 	PAGE *h;
-	db_indx_t base, i, indx, *inp, lim;
 	db_lockmode_t lock_mode;
-	db_pgno_t pg;
-	db_recno_t recno;
-	int adjust, cmp, deloffset, ret, stack;
-	int (*func) __P((DB *, const DBT *, const DBT *));
+	int ret, t_ret;
 
 	dbp = dbc->dbp;
 	mpf = dbp->mpf;
 	cp = (BTREE_CURSOR *)dbc->internal;
-	t = dbp->bt_internal;
-	recno = 0;
-
-	BT_STK_CLR(cp);
-
 	/*
-	 * There are several ways we search a btree tree.  The flags argument
-	 * specifies if we're acquiring read or write locks, if we position
-	 * to the first or last item in a set of duplicates, if we return
-	 * deleted items, and if we are locking pairs of pages.  In addition,
-	 * if we're modifying record numbers, we have to lock the entire tree
-	 * regardless.  See btree.h for more details.
-	 *
 	 * If write-locking pages, we need to know whether or not to acquire a
 	 * write lock on a page before getting it.  This depends on how deep it
 	 * is in tree, which we don't know until we acquire the root page.  So,
@@ -113,9 +93,12 @@ __bam_search(dbc, root_pgno, key, flags, stop, recnop, exactp)
 	 * Retrieve the root page.
 	 */
 try_again:
-	pg = root_pgno == PGNO_INVALID ? cp->root : root_pgno;
-	stack = LF_ISSET(S_STACK) && F_ISSET(cp, C_RECNUM);
-	lock_mode = stack ? DB_LOCK_WRITE : DB_LOCK_READ;
+	*stack = LF_ISSET(S_STACK) &&
+	      (dbc->dbtype == DB_RECNO || F_ISSET(cp, C_RECNUM));
+	lock_mode = DB_LOCK_READ;
+	if (*stack ||
+	    LF_ISSET(S_DEL) || (LF_ISSET(S_NEXT) && LF_ISSET(S_WRITE)))
+		lock_mode = DB_LOCK_WRITE;
 	if ((ret = __db_lget(dbc, 0, pg, lock_mode, 0, &lock)) != 0)
 		return (ret);
 	if ((ret = __memp_fget(mpf, &pg, 0, &h)) != 0) {
@@ -132,11 +115,17 @@ try_again:
 	 * each one gets a read lock and then blocks the other's attempt
 	 * for a write lock.
 	 */
-	if (!stack &&
-	    ((LF_ISSET(S_PARENT) && (u_int8_t)(stop + 1) >= h->level) ||
-	    (LF_ISSET(S_WRITE) && h->level == LEAFLEVEL))) {
-		(void)__memp_fput(mpf, h, 0);
-		(void)__LPUT(dbc, lock);
+	if (!*stack &&
+	    ((LF_ISSET(S_PARENT) && (u_int8_t)(slevel + 1) >= LEVEL(h)) ||
+	    (LF_ISSET(S_WRITE) && LEVEL(h) == LEAFLEVEL) ||
+	    (LF_ISSET(S_START) && slevel == LEVEL(h)))) {
+		if (!STD_LOCKING(dbc))
+			goto no_relock;
+		ret = __memp_fput(mpf, h, 0);
+		if ((t_ret = __LPUT(dbc, lock)) != 0 && ret == 0)
+			ret = t_ret;
+		if (ret != 0)
+			return (ret);
 		lock_mode = DB_LOCK_WRITE;
 		if ((ret = __db_lget(dbc, 0, pg, lock_mode, 0, &lock)) != 0)
 			return (ret);
@@ -146,15 +135,80 @@ try_again:
 			return (ret);
 		}
 		if (!((LF_ISSET(S_PARENT) &&
-		    (u_int8_t)(stop + 1) >= h->level) ||
-		    (LF_ISSET(S_WRITE) && h->level == LEAFLEVEL))) {
+		    (u_int8_t)(slevel + 1) >= LEVEL(h)) ||
+		    (LF_ISSET(S_WRITE) && LEVEL(h) == LEAFLEVEL) ||
+		    (LF_ISSET(S_START) && slevel == LEVEL(h)))) {
 			/* Someone else split the root, start over. */
-			(void)__memp_fput(mpf, h, 0);
-			(void)__LPUT(dbc, lock);
+			ret = __memp_fput(mpf, h, 0);
+			if ((t_ret = __LPUT(dbc, lock)) != 0 && ret == 0)
+				ret = t_ret;
+			if (ret != 0)
+				return (ret);
 			goto try_again;
 		}
-		stack = 1;
+no_relock:	*stack = 1;
 	}
+	BT_STK_ENTER(dbp->dbenv, cp, h, 0, lock, lock_mode, ret);
+
+	return (ret);
+}
+
+/*
+ * __bam_search --
+ *	Search a btree for a key.
+ *
+ * PUBLIC: int __bam_search __P((DBC *, db_pgno_t,
+ * PUBLIC:     const DBT *, u_int32_t, int, db_recno_t *, int *));
+ */
+int
+__bam_search(dbc, root_pgno, key, flags, slevel, recnop, exactp)
+	DBC *dbc;
+	db_pgno_t root_pgno;
+	const DBT *key;
+	u_int32_t flags;
+	int slevel, *exactp;
+	db_recno_t *recnop;
+{
+	BTREE *t;
+	BTREE_CURSOR *cp;
+	DB *dbp;
+	DB_LOCK lock;
+	DB_MPOOLFILE *mpf;
+	PAGE *h;
+	db_indx_t base, i, indx, *inp, lim;
+	db_lockmode_t lock_mode;
+	db_pgno_t pg;
+	db_recno_t recno;
+	int adjust, cmp, deloffset, ret, stack, t_ret;
+	int (*func) __P((DB *, const DBT *, const DBT *));
+
+	dbp = dbc->dbp;
+	mpf = dbp->mpf;
+	cp = (BTREE_CURSOR *)dbc->internal;
+	h = NULL;
+	t = dbp->bt_internal;
+	recno = 0;
+
+	BT_STK_CLR(cp);
+
+	/*
+	 * There are several ways we search a btree tree.  The flags argument
+	 * specifies if we're acquiring read or write locks, if we position
+	 * to the first or last item in a set of duplicates, if we return
+	 * deleted items, and if we are locking pairs of pages.  In addition,
+	 * if we're modifying record numbers, we have to lock the entire tree
+	 * regardless.  See btree.h for more details.
+	 */
+
+	if (root_pgno == PGNO_INVALID)
+		root_pgno = cp->root;
+	if ((ret = __bam_get_root(dbc, root_pgno, slevel, flags, &stack)) != 0)
+		return (ret);
+	lock_mode = cp->csp->lock_mode;
+	lock = cp->csp->lock;
+	h = cp->csp->page;
+
+	BT_STK_CLR(cp);
 
 	/* Choose a comparison function. */
 	func = F_ISSET(dbc, DBC_OPD) ?
@@ -163,6 +217,23 @@ try_again:
 
 	for (;;) {
 		inp = P_INP(dbp, h);
+		adjust = TYPE(h) == P_LBTREE ? P_INDX : O_INDX;
+		if (LF_ISSET(S_MIN | S_MAX)) {
+			if (LF_ISSET(S_MIN) || NUM_ENT(h) == 0)
+				indx = 0;
+			else if (TYPE(h) == P_LBTREE)
+				indx = NUM_ENT(h) - 2;
+			else
+				indx = NUM_ENT(h) - 1;
+
+			if (LEVEL(h) == LEAFLEVEL ||
+			     (!LF_ISSET(S_START) && LEVEL(h) == slevel)) {
+				if (LF_ISSET(S_NEXT))
+					goto get_next;
+				goto found;
+			}
+			goto next;
+		}
 		/*
 		 * Do a binary search on the current page.  If we're searching
 		 * a Btree leaf page, we have to walk the indices in groups of
@@ -170,7 +241,6 @@ try_again:
 		 * page, they're an index per page item.  If we find an exact
 		 * match on a leaf page, we're done.
 		 */
-		adjust = TYPE(h) == P_LBTREE ? P_INDX : O_INDX;
 		for (base = 0,
 		    lim = NUM_ENT(h) / (db_indx_t)adjust; lim != 0; lim >>= 1) {
 			indx = base + ((lim >> 1) * adjust);
@@ -178,8 +248,13 @@ try_again:
 			    __bam_cmp(dbp, key, h, indx, func, &cmp)) != 0)
 				goto err;
 			if (cmp == 0) {
-				if (TYPE(h) == P_LBTREE || TYPE(h) == P_LDUP)
+				if (LEVEL(h) == LEAFLEVEL ||
+				    (!LF_ISSET(S_START) &&
+				    LEVEL(h) == slevel)) {
+					if (LF_ISSET(S_NEXT))
+						goto get_next;
 					goto found;
+				}
 				goto next;
 			}
 			if (cmp > 0) {
@@ -192,21 +267,66 @@ try_again:
 		 * No match found.  Base is the smallest index greater than
 		 * key and may be zero or a last + O_INDX index.
 		 *
-		 * If it's a leaf page, return base as the "found" value.
+		 * If it's a leaf page or the stopping point,
+		 * return base as the "found" value.
 		 * Delete only deletes exact matches.
 		 */
-		if (TYPE(h) == P_LBTREE || TYPE(h) == P_LDUP) {
+		if (LEVEL(h) == LEAFLEVEL ||
+		    (!LF_ISSET(S_START) && LEVEL(h) == slevel)) {
 			*exactp = 0;
 
-			if (LF_ISSET(S_EXACT))
-				goto notfound;
+			if (LF_ISSET(S_EXACT)) {
+				ret = DB_NOTFOUND;
+				goto err;
+			}
 
 			if (LF_ISSET(S_STK_ONLY)) {
 				BT_STK_NUM(dbp->dbenv, cp, h, base, ret);
-				__LPUT(dbc, lock);
-				(void)__memp_fput(mpf, h, 0);
+				if ((t_ret =
+				    __LPUT(dbc, lock)) != 0 && ret == 0)
+					ret = t_ret;
+				if ((t_ret =
+				    __memp_fput(mpf, h, 0)) != 0 && ret == 0)
+					ret = t_ret;
 				return (ret);
 			}
+			if (LF_ISSET(S_NEXT)) {
+get_next:			/*
+				 * The caller could have asked for a NEXT
+				 * at the root if the tree recently collapsed.
+				 */
+				if (PGNO(h) == root_pgno) {
+					ret = DB_NOTFOUND;
+					goto err;
+				}
+				/*
+				 * Save the root of the subtree
+				 * and drop the rest of the subtree
+				 * and search down again starting at
+				 * the next child.
+				 */
+				if ((ret = __LPUT(dbc, lock)) != 0)
+					goto err;
+				if ((ret = __memp_fput(mpf, h, 0)) != 0)
+					goto err;
+				h = NULL;
+				LF_SET(S_MIN);
+				LF_CLR(S_NEXT);
+				indx = cp->sp->indx + 1;
+				if (indx == NUM_ENT(cp->sp->page)) {
+					ret = DB_NOTFOUND;
+					cp->csp++;
+					goto err;
+				}
+				h = cp->sp->page;
+				cp->sp->page = NULL;
+				lock = cp->sp->lock;
+				LOCK_INIT(cp->sp->lock);
+				if ((ret = __bam_stkrel(dbc, STK_NOLOCK)) != 0)
+					goto err;
+				stack = 1;
+				goto next;
+			}
 
 			/*
 			 * !!!
@@ -216,6 +336,8 @@ try_again:
 			 * to find an undeleted record.  This is handled by the
 			 * calling routine.
 			 */
+			if (LF_ISSET(S_DEL) && cp->csp == cp->sp)
+				cp->csp++;
 			BT_STK_ENTER(dbp->dbenv,
 			    cp, h, base, lock, lock_mode, ret);
 			if (ret != 0)
@@ -241,15 +363,24 @@ next:		if (recnop != NULL)
 
 		pg = GET_BINTERNAL(dbp, h, indx)->pgno;
 
+		/* See if we are at the level to start stacking. */
+		if (LF_ISSET(S_START) && slevel == LEVEL(h))
+			stack = 1;
+
 		if (LF_ISSET(S_STK_ONLY)) {
-			if (stop == h->level) {
+			if (slevel == LEVEL(h)) {
 				BT_STK_NUM(dbp->dbenv, cp, h, indx, ret);
-				__LPUT(dbc, lock);
-				(void)__memp_fput(mpf, h, 0);
+				if ((t_ret =
+				    __LPUT(dbc, lock)) != 0 && ret == 0)
+					ret = t_ret;
+				if ((t_ret =
+				    __memp_fput(mpf, h, 0)) != 0 && ret == 0)
+					ret = t_ret;
 				return (ret);
 			}
 			BT_STK_NUMPUSH(dbp->dbenv, cp, h, indx, ret);
 			(void)__memp_fput(mpf, h, 0);
+			h = NULL;
 			if ((ret = __db_lget(dbc,
 			    LCK_COUPLE_ALWAYS, pg, lock_mode, 0, &lock)) != 0) {
 				/*
@@ -257,22 +388,34 @@ next:		if (recnop != NULL)
 				 * is OK because it only happens when descending
 				 * the tree holding read-locks.
 				 */
-				__LPUT(dbc, lock);
+				(void)__LPUT(dbc, lock);
 				return (ret);
 			}
 		} else if (stack) {
 			/* Return if this is the lowest page wanted. */
-			if (LF_ISSET(S_PARENT) && stop == h->level) {
+			if (LF_ISSET(S_PARENT) && slevel == LEVEL(h)) {
 				BT_STK_ENTER(dbp->dbenv,
 				    cp, h, indx, lock, lock_mode, ret);
 				if (ret != 0)
 					goto err;
 				return (0);
 			}
+			if (LF_ISSET(S_DEL) && NUM_ENT(h) > 1) {
+				/*
+				 * There was a page with a singleton pointer
+				 * to a non-empty subtree.
+				 */
+				cp->csp--;
+				if ((ret = __bam_stkrel(dbc, STK_NOLOCK)) != 0)
+					goto err;
+				stack = 0;
+				goto do_del;
+			}
 			BT_STK_PUSH(dbp->dbenv,
 			    cp, h, indx, lock, lock_mode, ret);
 			if (ret != 0)
 				goto err;
+			h = NULL;
 
 			lock_mode = DB_LOCK_WRITE;
 			if ((ret =
@@ -285,14 +428,69 @@ next:		if (recnop != NULL)
 			 * unlock it.
 			 */
 			if ((LF_ISSET(S_PARENT) &&
-			    (u_int8_t)(stop + 1) >= (u_int8_t)(h->level - 1)) ||
-			    (h->level - 1) == LEAFLEVEL)
+			    (u_int8_t)(slevel + 1) >= (LEVEL(h) - 1)) ||
+			    (LEVEL(h) - 1) == LEAFLEVEL)
 				stack = 1;
 
-			(void)__memp_fput(mpf, h, 0);
+			/*
+			 * Returning a subtree.  See if we have hit the start
+			 * point if so save the parent and set stack.
+			 * Otherwise free the parent and temporarily
+			 * save this one.
+			 * For S_DEL we need to find a page with 1 entry.
+			 * For S_NEXT we want find the minimal subtree
+			 * that contains the key and the next page.
+			 * We save pages as long as we are at the right
+			 * edge of the subtree.  When we leave the right
+			 * edge, then drop the subtree.
+			 */
+			if (!LF_ISSET(S_DEL | S_NEXT)) {
+				if ((ret = __memp_fput(mpf, h, 0)) != 0)
+					goto err;
+				goto lock_next;
+			}
 
-			lock_mode = stack &&
-			    LF_ISSET(S_WRITE) ? DB_LOCK_WRITE : DB_LOCK_READ;
+			if ((LF_ISSET(S_DEL) && NUM_ENT(h) == 1)) {
+				stack = 1;
+				LF_SET(S_WRITE);
+				/* Push the parent. */
+				cp->csp++;
+				/* Push this node. */
+				BT_STK_PUSH(dbp->dbenv, cp, h,
+				     indx, lock, lock_mode, ret);
+				if (ret != 0)
+					goto err;
+				LOCK_INIT(lock);
+			} else {
+			/*
+			 * See if we want to save the tree so far.
+			 * If we are looking for the next key,
+			 * then we must save this node if we are
+			 * at the end of the page.  If not then
+			 * discard anything we have saved so far.
+			 * For delete only keep one node until
+			 * we find a singleton.
+			 */
+do_del:				if (cp->csp->page != NULL) {
+					if (LF_ISSET(S_NEXT) &&
+					     indx == NUM_ENT(h) - 1)
+						cp->csp++;
+					else if ((ret =
+					    __bam_stkrel(dbc, STK_NOLOCK)) != 0)
+						goto err;
+				}
+				/* Save this node. */
+				BT_STK_ENTER(dbp->dbenv, cp,
+				    h, indx, lock, lock_mode, ret);
+				if (ret != 0)
+					goto err;
+				LOCK_INIT(lock);
+			}
+
+lock_next:		h = NULL;
+
+			if (stack && LF_ISSET(S_WRITE))
+				lock_mode = DB_LOCK_WRITE;
 			if ((ret = __db_lget(dbc,
 			    LCK_COUPLE_ALWAYS, pg, lock_mode, 0, &lock)) != 0) {
 				/*
@@ -300,7 +498,9 @@ next:		if (recnop != NULL)
 				 * is OK because this only happens when we are
 				 * descending the tree holding read-locks.
 				 */
-				__LPUT(dbc, lock);
+				(void)__LPUT(dbc, lock);
+				if (LF_ISSET(S_DEL | S_NEXT))
+					cp->csp++;
 				goto err;
 			}
 		}
@@ -321,12 +521,12 @@ found:	*exactp = 1;
 	 * all duplicate sets that are not on overflow pages exist on a
 	 * single leaf page.
 	 */
-	if (TYPE(h) == P_LBTREE) {
+	if (TYPE(h) == P_LBTREE && NUM_ENT(h) > P_INDX) {
 		if (LF_ISSET(S_DUPLAST))
 			while (indx < (db_indx_t)(NUM_ENT(h) - P_INDX) &&
 			    inp[indx] == inp[indx + P_INDX])
 				indx += P_INDX;
-		else
+		else if (LF_ISSET(S_DUPFIRST))
 			while (indx > 0 &&
 			    inp[indx] == inp[indx - P_INDX])
 				indx -= P_INDX;
@@ -357,8 +557,10 @@ found:	*exactp = 1;
 		 * If we weren't able to find a non-deleted duplicate, return
 		 * DB_NOTFOUND.
 		 */
-		if (B_DISSET(GET_BKEYDATA(dbp, h, indx + deloffset)->type))
-			goto notfound;
+		if (B_DISSET(GET_BKEYDATA(dbp, h, indx + deloffset)->type)) {
+			ret = DB_NOTFOUND;
+			goto err;
+		}
 
 		/*
 		 * Increment the record counter to point to the found element.
@@ -381,23 +583,30 @@ found:	*exactp = 1;
 
 	if (LF_ISSET(S_STK_ONLY)) {
 		BT_STK_NUM(dbp->dbenv, cp, h, indx, ret);
-		__LPUT(dbc, lock);
-		(void)__memp_fput(mpf, h, 0);
+		if ((t_ret = __LPUT(dbc, lock)) != 0 && ret == 0)
+			ret = t_ret;
+		if ((t_ret = __memp_fput(mpf, h, 0)) != 0 && ret == 0)
+			ret = t_ret;
 	} else {
+		if (LF_ISSET(S_DEL) && cp->csp == cp->sp)
+			cp->csp++;
 		BT_STK_ENTER(dbp->dbenv, cp, h, indx, lock, lock_mode, ret);
-		if (ret != 0)
-			goto err;
 	}
+	if (ret != 0)
+		goto err;
+
 	return (0);
 
-notfound:
-	/* Keep the page locked for serializability. */
-	(void)__memp_fput(mpf, h, 0);
-	(void)__TLPUT(dbc, lock);
-	ret = DB_NOTFOUND;
+err:	if (h != NULL && (t_ret = __memp_fput(mpf, h, 0)) != 0 && ret == 0)
+		ret = t_ret;
 
-err:	BT_STK_POP(cp);
+	/* Keep any not-found page locked for serializability. */
+	if ((t_ret = __TLPUT(dbc, lock)) != 0 && ret == 0)
+		ret = t_ret;
+
+	BT_STK_POP(cp);
 	__bam_stkrel(dbc, 0);
+
 	return (ret);
 }
 
@@ -446,14 +655,24 @@ __bam_stkrel(dbc, flags)
 			 */
 			epg->page = NULL;
 		}
-		if (LF_ISSET(STK_NOLOCK))
-			(void)__LPUT(dbc, epg->lock);
-		else
-			(void)__TLPUT(dbc, epg->lock);
+		/*
+		 * We set this if we need to release our pins,
+		 * but are not logically ready to have the pages
+		 * visible.
+		 */
+		if (LF_ISSET(STK_PGONLY))
+			continue;
+		if (LF_ISSET(STK_NOLOCK)) {
+			if ((t_ret = __LPUT(dbc, epg->lock)) != 0 && ret == 0)
+				ret = t_ret;
+		} else
+			if ((t_ret = __TLPUT(dbc, epg->lock)) != 0 && ret == 0)
+				ret = t_ret;
 	}
 
 	/* Clear the stack, all pages have been released. */
-	BT_STK_CLR(cp);
+	if (!LF_ISSET(STK_PGONLY))
+		BT_STK_CLR(cp);
 
 	return (ret);
 }
