@@ -1,7 +1,7 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
  */
 /*
@@ -35,18 +35,15 @@
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
+ *
+ * $Id: bt_split.c,v 12.4 2005/06/16 20:20:22 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: bt_split.c,v 11.60 2003/06/30 17:19:35 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
-#include <limits.h>
 #include <string.h>
 #endif
 
@@ -59,7 +56,6 @@ static const char revid[] = "$Id: bt_split.c,v 11.60 2003/06/30 17:19:35 bostic
 
 static int __bam_broot __P((DBC *, PAGE *, PAGE *, PAGE *));
 static int __bam_page __P((DBC *, EPG *, EPG *));
-static int __bam_pinsert __P((DBC *, EPG *, PAGE *, PAGE *, int));
 static int __bam_psplit __P((DBC *, EPG *, PAGE *, PAGE *, db_indx_t *));
 static int __bam_root __P((DBC *, EPG *));
 static int __ram_root __P((DBC *, PAGE *, PAGE *, PAGE *));
@@ -120,7 +116,7 @@ __bam_split(dbc, arg, root_pgnop)
 			arg, S_WRPAIR, level, NULL, &exact) :
 		    __bam_rsearch(dbc,
 			(db_recno_t *)arg, S_WRPAIR, level, &exact))) != 0)
-			return (ret);
+			break;
 
 		if (root_pgnop != NULL)
 			*root_pgnop = cp->csp[0].page->pgno == root_pgno ?
@@ -134,7 +130,7 @@ __bam_split(dbc, arg, root_pgnop)
 		if (2 * B_MAXSIZEONPAGE(cp->ovflsize)
 		    <= (db_indx_t)P_FREESPACE(dbc->dbp, cp->csp[0].page)) {
 			__bam_stkrel(dbc, STK_NOLOCK);
-			return (0);
+			break;
 		}
 		ret = cp->csp[0].page->pgno == root_pgno ?
 		    __bam_root(dbc, &cp->csp[0]) :
@@ -162,10 +158,13 @@ __bam_split(dbc, arg, root_pgnop)
 				dir = UP;
 			break;
 		default:
-			return (ret);
+			goto err;
 		}
 	}
-	/* NOTREACHED */
+
+err:	if (root_pgnop != NULL)
+		*root_pgnop = cp->root;
+	return (ret);
 }
 
 /*
@@ -184,10 +183,11 @@ __bam_root(dbc, cp)
 	PAGE *lp, *rp;
 	db_indx_t split;
 	u_int32_t opflags;
-	int ret;
+	int ret, t_ret;
 
 	dbp = dbc->dbp;
 	mpf = dbp->mpf;
+	lp = rp = NULL;
 
 	/* Yeah, right. */
 	if (cp->page->level >= MAXBTREELEVEL) {
@@ -198,7 +198,6 @@ __bam_root(dbc, cp)
 	}
 
 	/* Create new left and right pages for the split. */
-	lp = rp = NULL;
 	if ((ret = __db_new(dbc, TYPE(cp->page), &lp)) != 0 ||
 	    (ret = __db_new(dbc, TYPE(cp->page), &rp)) != 0)
 		goto err;
@@ -238,24 +237,21 @@ __bam_root(dbc, cp)
 		goto err;
 
 	/* Adjust any cursors. */
-	if ((ret = __bam_ca_split(dbc,
-	    cp->page->pgno, lp->pgno, rp->pgno, split, 1)) != 0)
-		goto err;
+	ret = __bam_ca_split(dbc, cp->page->pgno, lp->pgno, rp->pgno, split, 1);
 
-	/* Success -- write the real pages back to the store. */
-	(void)__memp_fput(mpf, cp->page, DB_MPOOL_DIRTY);
-	(void)__TLPUT(dbc, cp->lock);
-	(void)__memp_fput(mpf, lp, DB_MPOOL_DIRTY);
-	(void)__memp_fput(mpf, rp, DB_MPOOL_DIRTY);
+	/* Success or error: release pages and locks. */
+err:	if ((t_ret =
+	    __memp_fput(mpf, cp->page, DB_MPOOL_DIRTY)) != 0 && ret == 0)
+		ret = t_ret;
+	if ((t_ret = __TLPUT(dbc, cp->lock)) != 0 && ret == 0)
+		ret = t_ret;
+	if (lp != NULL &&
+	    (t_ret = __memp_fput(mpf, lp, DB_MPOOL_DIRTY)) != 0 && ret == 0)
+		ret = t_ret;
+	if (rp != NULL &&
+	    (t_ret = __memp_fput(mpf, rp, DB_MPOOL_DIRTY)) != 0 && ret == 0)
+		ret = t_ret;
 
-	return (0);
-
-err:	if (lp != NULL)
-		(void)__memp_fput(mpf, lp, 0);
-	if (rp != NULL)
-		(void)__memp_fput(mpf, rp, 0);
-	(void)__memp_fput(mpf, cp->page, 0);
-	(void)__TLPUT(dbc, cp->lock);
 	return (ret);
 }
 
@@ -341,7 +337,7 @@ __bam_page(dbc, pp, cp)
 	 * page can't hold the new keys, and has to be split in turn, in which
 	 * case we want to release all the locks we can.
 	 */
-	if ((ret = __bam_pinsert(dbc, pp, lp, rp, 1)) != 0)
+	if ((ret = __bam_pinsert(dbc, pp, lp, rp, BPI_SPACEONLY)) != 0)
 		goto err;
 
 	/*
@@ -352,7 +348,7 @@ __bam_page(dbc, pp, cp)
 	 * a page that's not in our direct ancestry.  Consider a cursor walking
 	 * backward through the leaf pages, that has our following page locked,
 	 * and is waiting on a lock for the page we're splitting.  In that case
-	 * we're going to deadlock here .  It's probably OK, stepping backward
+	 * we're going to deadlock here.  It's probably OK, stepping backward
 	 * through the tree isn't a common operation.
 	 */
 	if (ISLEAF(cp->page) && NEXT_PGNO(cp->page) != PGNO_INVALID) {
@@ -371,9 +367,13 @@ __bam_page(dbc, pp, cp)
 		goto err;
 
 	/*
-	 * Lock the new page.  We need to do this because someone
-	 * could get here through bt_lpgno if this page was recently
-	 * dealocated.  They can't look at it before we commit.
+	 * Lock the new page.  We need to do this for two reasons: first, the
+	 * fast-lookup code might have a reference to this page in bt_lpgno if
+	 * the page was recently deleted from the tree, and that code doesn't
+	 * walk the tree and so won't encounter the parent's page lock.
+	 * Second, a dirty reader could get to this page via the parent or old
+	 * page after the split is done but before the transaction is committed
+	 * or aborted.
 	 */
 	if ((ret = __db_lget(dbc,
 	    0, PGNO(alloc_rp), DB_LOCK_WRITE, 0, &rplock)) != 0)
@@ -460,20 +460,24 @@ __bam_page(dbc, pp, cp)
 	if ((t_ret =
 	    __memp_fput(mpf, alloc_rp, DB_MPOOL_DIRTY)) != 0 && ret == 0)
 		ret = t_ret;
-	(void)__TLPUT(dbc, rplock);
+	if ((t_ret = __TLPUT(dbc, rplock)) != 0 && ret == 0)
+		ret = t_ret;
 	if ((t_ret =
 	    __memp_fput(mpf, pp->page, DB_MPOOL_DIRTY)) != 0 && ret == 0)
 		ret = t_ret;
-	(void)__TLPUT(dbc, pp->lock);
+	if ((t_ret = __TLPUT(dbc, pp->lock)) != 0 && ret == 0)
+		ret = t_ret;
 	if ((t_ret =
 	    __memp_fput(mpf, cp->page, DB_MPOOL_DIRTY)) != 0 && ret == 0)
 		ret = t_ret;
-	(void)__TLPUT(dbc, cp->lock);
+	if ((t_ret = __TLPUT(dbc, cp->lock)) != 0 && ret == 0)
+		ret = t_ret;
 	if (tp != NULL) {
 		if ((t_ret =
 		    __memp_fput(mpf, tp, DB_MPOOL_DIRTY)) != 0 && ret == 0)
 			ret = t_ret;
-		(void)__TLPUT(dbc, tplock);
+		if ((t_ret = __TLPUT(dbc, tplock)) != 0 && ret == 0)
+			ret = t_ret;
 	}
 	return (ret);
 
@@ -680,13 +684,15 @@ __ram_root(dbc, rootp, lp, rp)
 /*
  * __bam_pinsert --
  *	Insert a new key into a parent page, completing the split.
+ *
+ * PUBLIC: int __bam_pinsert __P((DBC *, EPG *, PAGE *, PAGE *, int));
  */
-static int
-__bam_pinsert(dbc, parent, lchild, rchild, space_check)
+int
+__bam_pinsert(dbc, parent, lchild, rchild, flags)
 	DBC *dbc;
 	EPG *parent;
 	PAGE *lchild, *rchild;
-	int space_check;
+	int flags;
 {
 	BINTERNAL bi, *child_bi;
 	BKEYDATA *child_bk, *tmp_bk;
@@ -709,7 +715,7 @@ __bam_pinsert(dbc, parent, lchild, rchild, space_check)
 
 	/* If handling record numbers, count records split to the right page. */
 	nrecs = F_ISSET(cp, C_RECNUM) &&
-	    !space_check ? __bam_total(dbp, rchild) : 0;
+	    !LF_ISSET(BPI_SPACEONLY) ? __bam_total(dbp, rchild) : 0;
 
 	/*
 	 * Now we insert the new page's first key into the parent page, which
@@ -745,7 +751,7 @@ __bam_pinsert(dbc, parent, lchild, rchild, space_check)
 
 		if (P_FREESPACE(dbp, ppage) < nbytes)
 			return (DB_NEEDSPLIT);
-		if (space_check)
+		if (LF_ISSET(BPI_SPACEONLY))
 			return (0);
 
 		/* Add a new record for the right page. */
@@ -775,7 +781,11 @@ __bam_pinsert(dbc, parent, lchild, rchild, space_check)
 		child_bk = GET_BKEYDATA(dbp, rchild, 0);
 		switch (B_TYPE(child_bk->type)) {
 		case B_KEYDATA:
+			nbytes = BINTERNAL_PSIZE(child_bk->len);
+			nksize = child_bk->len;
+
 			/*
+			 * Prefix compression:
 			 * We set t->bt_prefix to NULL if we have a comparison
 			 * callback but no prefix compression callback.  But,
 			 * if we're splitting in an off-page duplicates tree,
@@ -787,6 +797,14 @@ __bam_pinsert(dbc, parent, lchild, rchild, space_check)
 			 * as there's no way for an application to specify a
 			 * prefix compression callback that corresponds to its
 			 * comparison callback.
+			 *
+			 * No prefix compression if we don't have a compression
+			 * function, or the key we'd compress isn't a normal
+			 * key (for example, it references an overflow page).
+			 *
+			 * Generate a parent page key for the right child page
+			 * from a comparison of the last key on the left child
+			 * page and the first key on the right child page.
 			 */
 			if (F_ISSET(dbc, DBC_OPD)) {
 				if (dbp->dup_compare == __bam_defcmp)
@@ -795,13 +813,8 @@ __bam_pinsert(dbc, parent, lchild, rchild, space_check)
 					func = NULL;
 			} else
 				func = t->bt_prefix;
-
-			nbytes = BINTERNAL_PSIZE(child_bk->len);
-			nksize = child_bk->len;
 			if (func == NULL)
 				goto noprefix;
-			if (ppage->prev_pgno == PGNO_INVALID && off <= 1)
-				goto noprefix;
 			tmp_bk = GET_BKEYDATA(dbp, lchild, NUM_ENT(lchild) -
 			    (TYPE(lchild) == P_LDUP ? O_INDX : P_INDX));
 			if (B_TYPE(tmp_bk->type) != B_KEYDATA)
@@ -816,11 +829,11 @@ __bam_pinsert(dbc, parent, lchild, rchild, space_check)
 			if ((n = BINTERNAL_PSIZE(nksize)) < nbytes)
 				nbytes = n;
 			else
-noprefix:			nksize = child_bk->len;
+				nksize = child_bk->len;
 
-			if (P_FREESPACE(dbp, ppage) < nbytes)
+noprefix:		if (P_FREESPACE(dbp, ppage) < nbytes)
 				return (DB_NEEDSPLIT);
-			if (space_check)
+			if (LF_ISSET(BPI_SPACEONLY))
 				return (0);
 
 			memset(&bi, 0, sizeof(bi));
@@ -844,7 +857,7 @@ noprefix:			nksize = child_bk->len;
 
 			if (P_FREESPACE(dbp, ppage) < nbytes)
 				return (DB_NEEDSPLIT);
-			if (space_check)
+			if (LF_ISSET(BPI_SPACEONLY))
 				return (0);
 
 			memset(&bi, 0, sizeof(bi));
@@ -878,7 +891,7 @@ noprefix:			nksize = child_bk->len;
 
 		if (P_FREESPACE(dbp, ppage) < nbytes)
 			return (DB_NEEDSPLIT);
-		if (space_check)
+		if (LF_ISSET(BPI_SPACEONLY))
 			return (0);
 
 		/* Add a new record for the right page. */
@@ -899,13 +912,13 @@ noprefix:			nksize = child_bk->len;
 	 * If a Recno or Btree with record numbers AM page, or an off-page
 	 * duplicates tree, adjust the parent page's left page record count.
 	 */
-	if (F_ISSET(cp, C_RECNUM)) {
+	if (F_ISSET(cp, C_RECNUM) && !LF_ISSET(BPI_NORECNUM)) {
 		/* Log the change. */
 		if (DBC_LOGGING(dbc)) {
-		if ((ret = __bam_cadjust_log(dbp, dbc->txn,
-		    &LSN(ppage), 0, PGNO(ppage),
-		    &LSN(ppage), parent->indx, -(int32_t)nrecs, 0)) != 0)
-			return (ret);
+			if ((ret = __bam_cadjust_log(dbp, dbc->txn,
+			    &LSN(ppage), 0, PGNO(ppage), &LSN(ppage),
+			    parent->indx, -(int32_t)nrecs, 0)) != 0)
+				return (ret);
 		} else
 			LSN_NOT_LOGGED(LSN(ppage));
 
