@@ -1,18 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1999-2003
+ * Copyright (c) 1999-2005
  *	Sleepycat Software.  All rights reserved.
  *
- * $Id: bt_verify.c,v 1.87 2003/10/06 14:09:23 bostic Exp $
+ * $Id: bt_verify.c,v 12.13 2005/11/11 20:27:49 ubell Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: bt_verify.c,v 1.87 2003/10/06 14:09:23 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -34,8 +30,6 @@ static int __bam_vrfy_treeorder __P((DB *, db_pgno_t, PAGE *, BINTERNAL *,
 static int __ram_vrfy_inp __P((DB *, VRFY_DBINFO *, PAGE *, db_pgno_t,
     db_indx_t *, u_int32_t));
 
-#define	OKFLAGS	(DB_AGGRESSIVE | DB_NOORDERCHK | DB_SALVAGE)
-
 /*
  * __bam_vrfy_meta --
  *	Verify the btree-specific part of a metadata page.
@@ -95,10 +89,8 @@ __bam_vrfy_meta(dbp, vdp, meta, pgno, flags)
 	} else
 		pip->bt_minkey = meta->minkey;
 
-	/* bt_maxkey: unsupported so no constraints. */
-	pip->bt_maxkey = meta->maxkey;
-
 	/* re_len: no constraints on this (may be zero or huge--we make rope) */
+	pip->re_pad = meta->re_pad;
 	pip->re_len = meta->re_len;
 
 	/*
@@ -185,6 +177,9 @@ __bam_vrfy_meta(dbp, vdp, meta, pgno, flags)
 
 err:	if ((t_ret = __db_vrfy_putpageinfo(dbenv, vdp, pip)) != 0 && ret == 0)
 		ret = t_ret;
+	if (LF_ISSET(DB_SALVAGE) &&
+	   (t_ret = __db_salvage_markdone(vdp, pgno)) != 0 && ret == 0)
+		ret = t_ret;
 	return ((ret == 0 && isbad == 1) ? DB_VERIFY_BAD : ret);
 }
 
@@ -216,9 +211,6 @@ __ram_vrfy_leaf(dbp, vdp, h, pgno, flags)
 	if ((ret = __db_vrfy_getpageinfo(vdp, pgno, &pip)) != 0)
 		return (ret);
 
-	if ((ret = __db_fchk(dbenv, "__ram_vrfy_leaf", flags, OKFLAGS)) != 0)
-		goto err;
-
 	if (TYPE(h) != P_LRECNO) {
 		/* We should not have been called. */
 		TYPE_ERR_PRINT(dbenv, "__ram_vrfy_leaf", pgno, TYPE(h));
@@ -510,6 +502,8 @@ err:	if ((t_ret = __db_vrfy_putpageinfo(dbenv, vdp, pip)) != 0 && ret == 0)
 	return ((ret == 0 && isbad == 1) ? DB_VERIFY_BAD : ret);
 }
 
+typedef enum { VRFY_ITEM_NOTSET=0, VRFY_ITEM_BEGIN, VRFY_ITEM_END } VRFY_ITEM;
+
 /*
  * __bam_vrfy_inp --
  *	Verify that all entries in inp[] array are reasonable;
@@ -528,11 +522,14 @@ __bam_vrfy_inp(dbp, vdp, h, pgno, nentriesp, flags)
 	BOVERFLOW *bo;
 	DB_ENV *dbenv;
 	VRFY_CHILDINFO child;
+	VRFY_ITEM *pagelayout;
 	VRFY_PAGEINFO *pip;
-	int isbad, initem, isdupitem, ret, t_ret;
-	u_int32_t himark, offset; /* These would be db_indx_ts but for algnmt.*/
+	u_int32_t himark, offset;		/*
+						 * These would be db_indx_ts
+						 * but for alignment.
+						 */
 	u_int32_t i, endoff, nentries;
-	u_int8_t *pagelayout;
+	int isbad, initem, isdupitem, ret, t_ret;
 
 	dbenv = dbp->dbenv;
 	isbad = isdupitem = 0;
@@ -573,9 +570,9 @@ __bam_vrfy_inp(dbp, vdp, h, pgno, nentriesp, flags)
 	 * it and the region immediately after it.
 	 */
 	himark = dbp->pgsize;
-	if ((ret = __os_malloc(dbenv, dbp->pgsize, &pagelayout)) != 0)
+	if ((ret = __os_calloc(
+	    dbenv, dbp->pgsize, sizeof(pagelayout[0]), &pagelayout)) != 0)
 		goto err;
-	memset(pagelayout, 0, dbp->pgsize);
 	for (i = 0; i < NUM_ENT(h); i++) {
 		switch (ret = __db_vrfy_inpitem(dbp,
 		    h, pgno, i, 1, flags, &himark, &offset)) {
@@ -600,11 +597,9 @@ __bam_vrfy_inp(dbp, vdp, h, pgno, nentriesp, flags)
 		 * items have no overlaps or gaps.
 		 */
 		bk = GET_BKEYDATA(dbp, h, i);
-#define	ITEM_BEGIN	1
-#define	ITEM_END	2
-		if (pagelayout[offset] == 0)
-			pagelayout[offset] = ITEM_BEGIN;
-		else if (pagelayout[offset] == ITEM_BEGIN) {
+		if (pagelayout[offset] == VRFY_ITEM_NOTSET)
+			pagelayout[offset] = VRFY_ITEM_BEGIN;
+		else if (pagelayout[offset] == VRFY_ITEM_BEGIN) {
 			/*
 			 * Having two inp entries that point at the same patch
 			 * of page is legal if and only if the page is
@@ -676,12 +671,12 @@ __bam_vrfy_inp(dbp, vdp, h, pgno, nentriesp, flags)
 		 * If this is an onpage duplicate key we've seen before,
 		 * the end had better coincide too.
 		 */
-		if (isdupitem && pagelayout[endoff] != ITEM_END) {
+		if (isdupitem && pagelayout[endoff] != VRFY_ITEM_END) {
 			EPRINT((dbenv, "Page %lu: duplicated item %lu",
 			    (u_long)pgno, (u_long)i));
 			isbad = 1;
-		} else if (pagelayout[endoff] == 0)
-			pagelayout[endoff] = ITEM_END;
+		} else if (pagelayout[endoff] == VRFY_ITEM_NOTSET)
+			pagelayout[endoff] = VRFY_ITEM_END;
 		isdupitem = 0;
 
 		/*
@@ -771,9 +766,9 @@ __bam_vrfy_inp(dbp, vdp, h, pgno, nentriesp, flags)
 	for (i = himark; i < dbp->pgsize; i++)
 		if (initem == 0)
 			switch (pagelayout[i]) {
-			case 0:
+			case VRFY_ITEM_NOTSET:
 				/* May be just for alignment. */
-				if (i != ALIGN(i, sizeof(u_int32_t)))
+				if (i != DB_ALIGN(i, sizeof(u_int32_t)))
 					continue;
 
 				isbad = 1;
@@ -781,13 +776,13 @@ __bam_vrfy_inp(dbp, vdp, h, pgno, nentriesp, flags)
 				    "Page %lu: gap between items at offset %lu",
 				    (u_long)pgno, (u_long)i));
 				/* Find the end of the gap */
-				for ( ; pagelayout[i + 1] == 0 &&
+				for (; pagelayout[i + 1] == VRFY_ITEM_NOTSET &&
 				    (size_t)(i + 1) < dbp->pgsize; i++)
 					;
 				break;
-			case ITEM_BEGIN:
+			case VRFY_ITEM_BEGIN:
 				/* We've found an item. Check its alignment. */
-				if (i != ALIGN(i, sizeof(u_int32_t))) {
+				if (i != DB_ALIGN(i, sizeof(u_int32_t))) {
 					isbad = 1;
 					EPRINT((dbenv,
 					    "Page %lu: offset %lu unaligned",
@@ -796,7 +791,7 @@ __bam_vrfy_inp(dbp, vdp, h, pgno, nentriesp, flags)
 				initem = 1;
 				nentries++;
 				break;
-			case ITEM_END:
+			case VRFY_ITEM_END:
 				/*
 				 * We've hit the end of an item even though
 				 * we don't think we're in one;  must
@@ -807,22 +802,17 @@ __bam_vrfy_inp(dbp, vdp, h, pgno, nentriesp, flags)
 				    "Page %lu: overlapping items at offset %lu",
 				    (u_long)pgno, (u_long)i));
 				break;
-			default:
-				/* Should be impossible. */
-				DB_ASSERT(0);
-				ret = EINVAL;
-				goto err;
 			}
 		else
 			switch (pagelayout[i]) {
-			case 0:
+			case VRFY_ITEM_NOTSET:
 				/* In the middle of an item somewhere. Okay. */
 				break;
-			case ITEM_END:
+			case VRFY_ITEM_END:
 				/* End of an item; switch to out-of-item mode.*/
 				initem = 0;
 				break;
-			case ITEM_BEGIN:
+			case VRFY_ITEM_BEGIN:
 				/*
 				 * Hit a second item beginning without an
 				 * end.  Overlap.
@@ -1401,11 +1391,11 @@ __bam_vrfy_subtree(dbp, vdp, pgno, l, r, flags, levelp, nrecsp, relenp)
 			 * page's next_pgno, and our prev_pgno.
 			 */
 			if (pip->type != vdp->leaf_type) {
+				isbad = 1;
 				EPRINT((dbenv,
 	"Page %lu: unexpected page type %lu found in leaf chain (expected %lu)",
 				    (u_long)pip->pgno, (u_long)pip->type,
 				    (u_long)vdp->leaf_type));
-				isbad = 1;
 			}
 
 			/*
@@ -1414,20 +1404,20 @@ __bam_vrfy_subtree(dbp, vdp, pgno, l, r, flags, levelp, nrecsp, relenp)
 			 */
 			if (!F_ISSET(vdp, VRFY_LEAFCHAIN_BROKEN)) {
 				if (pip->pgno != vdp->next_pgno) {
+					isbad = 1;
 					EPRINT((dbenv,
 	"Page %lu: incorrect next_pgno %lu found in leaf chain (should be %lu)",
 					    (u_long)vdp->prev_pgno,
 					    (u_long)vdp->next_pgno,
 					    (u_long)pip->pgno));
-					isbad = 1;
 				}
 				if (pip->prev_pgno != vdp->prev_pgno) {
-bad_prev:				EPRINT((dbenv,
-	"Page %lu: incorrect prev_pgno %lu found in leaf chain (should be %lu)",
+bad_prev:				isbad = 1;
+					EPRINT((dbenv,
+    "Page %lu: incorrect prev_pgno %lu found in leaf chain (should be %lu)",
 					    (u_long)pip->pgno,
 					    (u_long)pip->prev_pgno,
 					    (u_long)vdp->prev_pgno));
-					isbad = 1;
 				}
 			}
 		}
@@ -1519,11 +1509,11 @@ bad_prev:				EPRINT((dbenv,
 						    dbp, vdp, child->pgno, NULL,
 						    NULL, stflags | ST_TOPLEVEL,
 						    NULL, NULL, NULL)) != 0) {
-							if (ret !=
+							if (ret ==
 							    DB_VERIFY_BAD)
-								goto err;
-							else
 								isbad = 1;
+							else
+								goto err;
 						}
 					}
 				}
@@ -1538,10 +1528,10 @@ bad_prev:				EPRINT((dbenv,
 				 */
 				if (F_ISSET(pip, VRFY_DUPS_UNSORTED) &&
 				    LF_ISSET(ST_DUPSORT)) {
+					isbad = 1;
 					EPRINT((dbenv,
 		    "Page %lu: unsorted duplicate set in sorted-dup database",
 					    (u_long)pgno));
-					isbad = 1;
 				}
 			}
 		}
@@ -1602,10 +1592,10 @@ bad_prev:				EPRINT((dbenv,
 			if ((ret = __bam_vrfy_subtree(dbp, vdp, child->pgno,
 			    NULL, NULL, flags, &child_level, &child_nrecs,
 			    &child_relen)) != 0) {
-				if (ret != DB_VERIFY_BAD)
-					goto done;
-				else
+				if (ret == DB_VERIFY_BAD)
 					isbad = 1;
+				else
+					goto done;
 			}
 
 			if (LF_ISSET(ST_RELEN)) {
@@ -1626,8 +1616,17 @@ bad_prev:				EPRINT((dbenv,
 				if (relenp)
 					*relenp = relen;
 			}
-			if (LF_ISSET(ST_RECNUM))
+			if (LF_ISSET(ST_RECNUM)) {
+				if (child->nrecs != child_nrecs) {
+					isbad = 1;
+					EPRINT((dbenv,
+		"Page %lu: record count incorrect: actual %lu, in record %lu",
+					    (u_long)child->pgno,
+					    (u_long)child_nrecs,
+					    (u_long)child->nrecs));
+				}
 				nrecs += child_nrecs;
+			}
 			if (isbad == 0 && level != child_level + 1) {
 				isbad = 1;
 				EPRINT((dbenv,
@@ -1662,10 +1661,10 @@ bad_prev:				EPRINT((dbenv,
 			 * shouldn't happen.
 			 */
 			if (child->refcnt > 2) {
+				isbad = 1;
 				EPRINT((dbenv,
     "Page %lu: overflow page %lu referenced more than twice from internal page",
 				    (u_long)pgno, (u_long)child->pgno));
-				isbad = 1;
 			} else
 				for (j = 0; j < child->refcnt; j++)
 					if ((ret = __db_vrfy_ovfl_structure(dbp,
@@ -1701,7 +1700,7 @@ bad_prev:				EPRINT((dbenv,
 	for (i = 0; i < pip->entries; i += O_INDX) {
 		li = GET_BINTERNAL(dbp, h, i);
 		ri = (i + O_INDX < pip->entries) ?
-		    GET_BINTERNAL(dbp, h, i + O_INDX) : NULL;
+		    GET_BINTERNAL(dbp, h, i + O_INDX) : rp;
 
 		/*
 		 * The leftmost key is forcibly sorted less than all entries,
@@ -1710,10 +1709,10 @@ bad_prev:				EPRINT((dbenv,
 		if ((ret = __bam_vrfy_subtree(dbp, vdp, li->pgno,
 		    i == 0 ? NULL : li, ri, flags, &child_level,
 		    &child_nrecs, NULL)) != 0) {
-			if (ret != DB_VERIFY_BAD)
-				goto done;
-			else
+			if (ret == DB_VERIFY_BAD)
 				isbad = 1;
+			else
+				goto done;
 		}
 
 		if (LF_ISSET(ST_RECNUM)) {
@@ -1792,10 +1791,10 @@ done:	if (F_ISSET(pip, VRFY_INCOMPLETE) && isbad == 0 && ret == 0) {
 			goto err;
 
 		if (NUM_ENT(h) == 0 && ISINTERNAL(h)) {
+			isbad = 1;
 			EPRINT((dbenv,
 		    "Page %lu: internal page is empty and should not be",
 			    (u_long)pgno));
-			isbad = 1;
 			goto err;
 		}
 	}
@@ -1862,9 +1861,9 @@ done:	if (F_ISSET(pip, VRFY_INCOMPLETE) && isbad == 0 && ret == 0) {
 		 * PGNO_INVALID.
 		 */
 		if (vdp->next_pgno != PGNO_INVALID) {
+			isbad = 1;
 			EPRINT((dbenv, "Page %lu: unterminated leaf chain",
 			    (u_long)vdp->prev_pgno));
-			isbad = 1;
 		}
 
 err:	if (toplevel) {
@@ -1968,7 +1967,7 @@ __bam_vrfy_treeorder(dbp, pgno, h, lp, rp, func, flags)
 			return (EINVAL);
 		}
 
-		/* On error, fall through, free if neeeded, and return. */
+		/* On error, fall through, free if needed, and return. */
 		if ((ret = __bam_cmp(dbp, &dbt, h, 0, func, &cmp)) == 0) {
 			if (cmp > 0) {
 				EPRINT((dbenv,
@@ -2004,7 +2003,7 @@ __bam_vrfy_treeorder(dbp, pgno, h, lp, rp, func, flags)
 			return (EINVAL);
 		}
 
-		/* On error, fall through, free if neeeded, and return. */
+		/* On error, fall through, free if needed, and return. */
 		if ((ret = __bam_cmp(dbp, &dbt, h, last, func, &cmp)) == 0) {
 			if (cmp < 0) {
 				EPRINT((dbenv,
@@ -2045,203 +2044,223 @@ __bam_salvage(dbp, vdp, pgno, pgtype, h, handle, callback, key, flags)
 	DBT *key;
 	u_int32_t flags;
 {
-	DBT dbt, unkdbt;
-	DB_ENV *dbenv;
 	BKEYDATA *bk;
 	BOVERFLOW *bo;
-	db_indx_t i, beg, end, *inp;
+	DBT dbt, unknown_key, unknown_data;
+	DB_ENV *dbenv;
+	VRFY_ITEM *pgmap;
+	db_indx_t i, last, beg, end, *inp;
 	u_int32_t himark;
-	u_int8_t *pgmap;
 	void *ovflbuf;
-	int t_ret, ret, err_ret;
+	int ret, t_ret, t2_ret;
 
 	dbenv = dbp->dbenv;
-
-	/* Shut up lint. */
-	COMPQUIET(end, 0);
-
 	ovflbuf = pgmap = NULL;
-	err_ret = ret = 0;
 	inp = P_INP(dbp, h);
 
 	memset(&dbt, 0, sizeof(DBT));
 	dbt.flags = DB_DBT_REALLOC;
 
-	memset(&unkdbt, 0, sizeof(DBT));
-	unkdbt.size = (u_int32_t)(strlen("UNKNOWN") + 1);
-	unkdbt.data = "UNKNOWN";
+	memset(&unknown_key, 0, sizeof(DBT));
+	unknown_key.size = (u_int32_t)strlen("UNKNOWN_KEY");
+	unknown_key.data = "UNKNOWN_KEY";
+	memset(&unknown_data, 0, sizeof(DBT));
+	unknown_data.size = (u_int32_t)strlen("UNKNOWN_DATA");
+	unknown_data.data = "UNKNOWN_DATA";
 
 	/*
 	 * Allocate a buffer for overflow items.  Start at one page;
 	 * __db_safe_goff will realloc as needed.
 	 */
 	if ((ret = __os_malloc(dbenv, dbp->pgsize, &ovflbuf)) != 0)
-		return (ret);
+		goto err;
 
-	if (LF_ISSET(DB_AGGRESSIVE)) {
-		if ((ret =
-		    __os_malloc(dbenv, dbp->pgsize, &pgmap)) != 0)
-			goto err;
-		memset(pgmap, 0, dbp->pgsize);
-	}
+	if (LF_ISSET(DB_AGGRESSIVE) && (ret =
+	    __os_calloc(dbenv, dbp->pgsize, sizeof(pgmap[0]), &pgmap)) != 0)
+		goto err;
 
 	/*
 	 * Loop through the inp array, spitting out key/data pairs.
 	 *
-	 * If we're salvaging normally, loop from 0 through NUM_ENT(h).
-	 * If we're being aggressive, loop until we hit the end of the page--
+	 * If we're salvaging normally, loop from 0 through NUM_ENT(h).  If
+	 * we're being aggressive, loop until we hit the end of the page --
 	 * NUM_ENT() may be bogus.
 	 */
 	himark = dbp->pgsize;
-	for (i = 0;; i += O_INDX) {
+	for (i = 0, last = UINT16_MAX;; i += O_INDX) {
 		/* If we're not aggressive, break when we hit NUM_ENT(h). */
 		if (!LF_ISSET(DB_AGGRESSIVE) && i >= NUM_ENT(h))
 			break;
 
 		/* Verify the current item. */
-		ret = __db_vrfy_inpitem(dbp,
-		    h, pgno, i, 1, flags, &himark, NULL);
-		/* If this returned a fatality, it's time to break. */
-		if (ret == DB_VERIFY_FATAL) {
+		t_ret =
+		    __db_vrfy_inpitem(dbp, h, pgno, i, 1, flags, &himark, NULL);
+
+		if (t_ret != 0) {
 			/*
-			 * Don't return DB_VERIFY_FATAL;  it's private
-			 * and means only that we can't go on with this
-			 * page, not with the whole database.  It's
-			 * not even an error if we've run into it
-			 * after NUM_ENT(h).
+			 * If this is a btree leaf and we've printed out a key
+			 * but not its associated data item, fix this imbalance
+			 * by printing an "UNKNOWN_DATA".
 			 */
-			ret = (i < NUM_ENT(h)) ? DB_VERIFY_BAD : 0;
-			break;
+			if (pgtype == P_LBTREE && i % P_INDX == 1 &&
+			    last == i - 1 && (t2_ret = __db_vrfy_prdbt(
+			    &unknown_data,
+			    0, " ", handle, callback, 0, vdp)) != 0) {
+				if (ret == 0)
+					ret = t2_ret;
+				goto err;
+			}
+
+			/*
+			 * Don't return DB_VERIFY_FATAL; it's private and means
+			 * only that we can't go on with this page, not with
+			 * the whole database.  It's not even an error if we've
+			 * run into it after NUM_ENT(h).
+			 */
+			if (t_ret == DB_VERIFY_FATAL) {
+				if (i < NUM_ENT(h) && ret == 0)
+					ret = DB_VERIFY_BAD;
+				break;
+			}
+			continue;
 		}
 
 		/*
 		 * If this returned 0, it's safe to print or (carefully)
 		 * try to fetch.
+		 *
+		 * We only print deleted items if DB_AGGRESSIVE is set.
 		 */
-		if (ret == 0) {
-			/*
-			 * We only want to print deleted items if
-			 * DB_AGGRESSIVE is set.
-			 */
-			bk = GET_BKEYDATA(dbp, h, i);
-			if (!LF_ISSET(DB_AGGRESSIVE) && B_DISSET(bk->type))
-				continue;
+		bk = GET_BKEYDATA(dbp, h, i);
+		if (!LF_ISSET(DB_AGGRESSIVE) && B_DISSET(bk->type))
+			continue;
 
-			/*
-			 * We're going to go try to print the next item.  If
-			 * key is non-NULL, we're a dup page, so we've got to
-			 * print the key first, unless SA_SKIPFIRSTKEY is set
-			 * and we're on the first entry.
-			 */
-			if (key != NULL &&
-			    (i != 0 || !LF_ISSET(SA_SKIPFIRSTKEY)))
-				if ((ret = __db_prdbt(key,
-				    0, " ", handle, callback, 0, vdp)) != 0)
-					err_ret = ret;
+		/*
+		 * If this is a btree leaf and we're about to print out a data
+		 * item for which we didn't print out a key, fix this imbalance
+		 * by printing an "UNKNOWN_KEY".
+		 */
+		if (pgtype == P_LBTREE && i % P_INDX == 1 &&
+		    last != i - 1 && (t_ret = __db_vrfy_prdbt(
+		    &unknown_key, 0, " ", handle, callback, 0, vdp)) != 0) {
+			if (ret == 0)
+				ret = t_ret;
+			goto err;
+		}
+		last = i;
 
-			beg = inp[i];
-			switch (B_TYPE(bk->type)) {
-			case B_DUPLICATE:
-				end = beg + BOVERFLOW_SIZE - 1;
-				/*
-				 * If we're not on a normal btree leaf page,
-				 * there shouldn't be off-page
-				 * dup sets.  Something's confused;  just
-				 * drop it, and the code to pick up unlinked
-				 * offpage dup sets will print it out
-				 * with key "UNKNOWN" later.
-				 */
-				if (pgtype != P_LBTREE)
-					break;
-
-				bo = (BOVERFLOW *)bk;
-
-				/*
-				 * If the page number is unreasonable, or
-				 * if this is supposed to be a key item,
-				 * just spit out "UNKNOWN"--the best we
-				 * can do is run into the data items in the
-				 * unlinked offpage dup pass.
-				 */
-				if (!IS_VALID_PGNO(bo->pgno) ||
-				    (i % P_INDX == 0)) {
-					/* Not much to do on failure. */
-					if ((ret = __db_prdbt(&unkdbt, 0, " ",
-					    handle, callback, 0, vdp)) != 0)
-						err_ret = ret;
-					break;
-				}
-
-				if ((ret = __db_salvage_duptree(dbp,
-				    vdp, bo->pgno, &dbt, handle, callback,
-				    flags | SA_SKIPFIRSTKEY)) != 0)
-					err_ret = ret;
-
-				break;
-			case B_KEYDATA:
-				end =
-				    ALIGN(beg + bk->len, sizeof(u_int32_t)) - 1;
-				dbt.data = bk->data;
-				dbt.size = bk->len;
-				if ((ret = __db_prdbt(&dbt,
-				    0, " ", handle, callback, 0, vdp)) != 0)
-					err_ret = ret;
-				break;
-			case B_OVERFLOW:
-				end = beg + BOVERFLOW_SIZE - 1;
-				bo = (BOVERFLOW *)bk;
-				if ((ret = __db_safe_goff(dbp, vdp,
-				    bo->pgno, &dbt, &ovflbuf, flags)) != 0) {
-					err_ret = ret;
-					/* We care about err_ret more. */
-					(void)__db_prdbt(&unkdbt, 0, " ",
-					    handle, callback, 0, vdp);
-					break;
-				}
-				if ((ret = __db_prdbt(&dbt,
-				    0, " ", handle, callback, 0, vdp)) != 0)
-					err_ret = ret;
-				break;
-			default:
-				/*
-				 * We should never get here;  __db_vrfy_inpitem
-				 * should not be returning 0 if bk->type
-				 * is unrecognizable.
-				 */
-				DB_ASSERT(0);
-				return (EINVAL);
+		/*
+		 * We're going to go try to print the next item.  If key is
+		 * non-NULL, we're a dup page, so we've got to print the key
+		 * first, unless SA_SKIPFIRSTKEY is set and we're on the first
+		 * entry.
+		 */
+		if (key != NULL && (i != 0 || !LF_ISSET(SA_SKIPFIRSTKEY)))
+			if ((t_ret = __db_vrfy_prdbt(key,
+			    0, " ", handle, callback, 0, vdp)) != 0) {
+				if (ret == 0)
+					ret = t_ret;
+				goto err;
 			}
 
+		beg = inp[i];
+		switch (B_TYPE(bk->type)) {
+		case B_DUPLICATE:
+			end = beg + BOVERFLOW_SIZE - 1;
 			/*
-			 * If we're being aggressive, mark the beginning
-			 * and end of the item;  we'll come back and print
-			 * whatever "junk" is in the gaps in case we had
-			 * any bogus inp elements and thereby missed stuff.
+			 * If we're not on a normal btree leaf page, there
+			 * shouldn't be off-page dup sets.  Something's
+			 * confused; just drop it, and the code to pick up
+			 * unlinked offpage dup sets will print it out
+			 * with key "UNKNOWN" later.
 			 */
-			if (LF_ISSET(DB_AGGRESSIVE)) {
-				pgmap[beg] = ITEM_BEGIN;
-				pgmap[end] = ITEM_END;
+			if (pgtype != P_LBTREE)
+				break;
+
+			bo = (BOVERFLOW *)bk;
+
+			/*
+			 * If the page number is unreasonable, or if this is
+			 * supposed to be a key item, output "UNKNOWN_KEY" --
+			 * the best we can do is run into the data items in
+			 * the unlinked offpage dup pass.
+			 */
+			if (!IS_VALID_PGNO(bo->pgno) || (i % P_INDX == 0)) {
+				/* Not much to do on failure. */
+				if ((t_ret = __db_vrfy_prdbt(&unknown_key,
+				    0, " ", handle, callback, 0, vdp)) != 0) {
+					if (ret == 0)
+						ret = t_ret;
+					goto err;
+				}
+				break;
 			}
+
+			/* Don't stop on error. */
+			if ((t_ret = __db_salvage_duptree(dbp,
+			    vdp, bo->pgno, &dbt, handle, callback,
+			    flags | SA_SKIPFIRSTKEY)) != 0 && ret == 0)
+				ret = t_ret;
+
+			break;
+		case B_KEYDATA:
+			end = (db_indx_t)DB_ALIGN(
+			    beg + bk->len, sizeof(u_int32_t)) - 1;
+			dbt.data = bk->data;
+			dbt.size = bk->len;
+			if ((t_ret = __db_vrfy_prdbt(&dbt,
+			    0, " ", handle, callback, 0, vdp)) != 0) {
+				if (ret == 0)
+					ret = t_ret;
+				goto err;
+			}
+			break;
+		case B_OVERFLOW:
+			end = beg + BOVERFLOW_SIZE - 1;
+			bo = (BOVERFLOW *)bk;
+
+			/* Don't stop on error. */
+			if ((t_ret = __db_safe_goff(dbp, vdp,
+			    bo->pgno, &dbt, &ovflbuf, flags)) != 0 && ret == 0)
+				ret = t_ret;
+			if ((t_ret = __db_vrfy_prdbt(
+			    t_ret == 0 ? &dbt : &unknown_key,
+			    0, " ", handle, callback, 0, vdp)) != 0 && ret == 0)
+				ret = t_ret;
+			break;
+		default:
+			/*
+			 * We should never get here; __db_vrfy_inpitem should
+			 * not be returning 0 if bk->type is unrecognizable.
+			 */
+			DB_ASSERT(0);
+			if (ret == 0)
+				ret = EINVAL;
+			goto err;
+		}
+
+		/*
+		 * If we're being aggressive, mark the beginning and end of
+		 * the item; we'll come back and print whatever "junk" is in
+		 * the gaps in case we had any bogus inp elements and thereby
+		 * missed stuff.
+		 */
+		if (LF_ISSET(DB_AGGRESSIVE)) {
+			pgmap[beg] = VRFY_ITEM_BEGIN;
+			pgmap[end] = VRFY_ITEM_END;
 		}
 	}
 
-	/*
-	 * If i is odd and this is a btree leaf, we've printed out a key but not
-	 * a datum; fix this imbalance by printing an "UNKNOWN".
-	 */
-	if (pgtype == P_LBTREE && (i % P_INDX == 1) && ((ret =
-	    __db_prdbt(&unkdbt, 0, " ", handle, callback, 0, vdp)) != 0))
-		err_ret = ret;
-
 err:	if (pgmap != NULL)
 		__os_free(dbenv, pgmap);
-	__os_free(dbenv, ovflbuf);
+	if (ovflbuf != NULL)
+		__os_free(dbenv, ovflbuf);
 
 	/* Mark this page as done. */
-	if ((t_ret = __db_salvage_markdone(vdp, pgno)) != 0)
-		return (t_ret);
+	if ((t_ret = __db_salvage_markdone(vdp, pgno)) != 0 && ret == 0)
+		ret = t_ret;
 
-	return ((err_ret != 0) ? err_ret : ret);
+	return (ret);
 }
 
 /*
