@@ -1,17 +1,17 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: db_archive.c,v 12.4 2005/09/09 12:38:30 bostic Exp $
  */
 
 #include "db_config.h"
 
 #ifndef lint
 static const char copyright[] =
-    "Copyright (c) 1996-2003\nSleepycat Software Inc.  All rights reserved.\n";
-static const char revid[] =
-    "$Id: db_archive.c,v 11.42 2003/08/13 19:57:04 ubell Exp $";
+    "Copyright (c) 1996-2005\nSleepycat Software Inc.  All rights reserved.\n";
 #endif
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -27,7 +27,9 @@ static const char revid[] =
 
 int db_archive_main __P((int, char *[]));
 int db_archive_usage __P((void));
-int db_archive_version_check __P((const char *));
+int db_archive_version_check __P((void));
+
+const char *progname;
 
 int
 db_archive(args)
@@ -50,13 +52,17 @@ db_archive_main(argc, argv)
 {
 	extern char *optarg;
 	extern int optind, __db_getopt_reset;
-	const char *progname = "db_archive";
 	DB_ENV	*dbenv;
 	u_int32_t flags;
 	int ch, exitval, ret, verbose;
 	char **file, *home, **list, *passwd;
 
-	if ((ret = db_archive_version_check(progname)) != 0)
+	if ((progname = strrchr(argv[0], '/')) == NULL)
+		progname = argv[0];
+	else
+		++progname;
+
+	if ((ret = db_archive_version_check()) != 0)
 		return (ret);
 
 	dbenv = NULL;
@@ -123,9 +129,6 @@ db_archive_main(argc, argv)
 	dbenv->set_errfile(dbenv, stderr);
 	dbenv->set_errpfx(dbenv, progname);
 
-	if (verbose)
-		(void)dbenv->set_verbose(dbenv, DB_VERB_CHKPOINT, 1);
-
 	if (passwd != NULL && (ret = dbenv->set_encrypt(dbenv,
 	    passwd, DB_ENCRYPT_AES)) != 0) {
 		dbenv->err(dbenv, ret, "set_passwd");
@@ -135,11 +138,11 @@ db_archive_main(argc, argv)
 	 * If attaching to a pre-existing environment fails, create a
 	 * private one and try again.
 	 */
-	if ((ret = dbenv->open(dbenv,
-	    home, DB_JOINENV | DB_USE_ENVIRON, 0)) != 0 &&
+	if ((ret = dbenv->open(dbenv, home, DB_USE_ENVIRON, 0)) != 0 &&
+	    (ret == DB_VERSION_MISMATCH ||
 	    (ret = dbenv->open(dbenv, home, DB_CREATE |
-	    DB_INIT_LOG | DB_INIT_TXN | DB_PRIVATE | DB_USE_ENVIRON, 0)) != 0) {
-		dbenv->err(dbenv, ret, "open");
+	    DB_INIT_LOG | DB_PRIVATE | DB_USE_ENVIRON, 0)) != 0)) {
+		dbenv->err(dbenv, ret, "DB_ENV->open");
 		goto shutdown;
 	}
 
@@ -178,13 +181,12 @@ int
 db_archive_usage()
 {
 	(void)fprintf(stderr,
-	    "usage: db_archive [-adlsVv] [-h home] [-P password]\n");
+	    "usage: %s [-adlsVv] [-h home] [-P password]\n", progname);
 	return (EXIT_FAILURE);
 }
 
 int
-db_archive_version_check(progname)
-	const char *progname;
+db_archive_version_check()
 {
 	int v_major, v_minor, v_patch;
 
