@@ -1,17 +1,17 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: db_load.c,v 12.8 2005/06/16 20:21:23 bostic Exp $
  */
 
 #include "db_config.h"
 
 #ifndef lint
 static const char copyright[] =
-    "Copyright (c) 1996-2003\nSleepycat Software Inc.  All rights reserved.\n";
-static const char revid[] =
-    "$Id: db_load.c,v 11.88 2003/10/16 17:51:08 bostic Exp $";
+    "Copyright (c) 1996-2005\nSleepycat Software Inc.  All rights reserved.\n";
 #endif
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -57,7 +57,9 @@ int	db_load_load __P((DB_ENV *, char *, DBTYPE, char **, u_int, LDG *, int *));
 int	db_load_main __P((int, char *[]));
 int	db_load_rheader __P((DB_ENV *, DB *, DBTYPE *, char **, int *, int *));
 int	db_load_usage __P((void));
-int	db_load_version_check __P((const char *));
+int	db_load_version_check __P((void));
+
+const char *progname;
 
 #define	G(f)	((LDG *)dbenv->app_private)->f
 
@@ -85,16 +87,25 @@ db_load_main(argc, argv)
 	int argc;
 	char *argv[];
 {
+	enum { NOTSET, FILEID_RESET, LSN_RESET, STANDARD_LOAD } mode;
 	extern char *optarg;
 	extern int optind, __db_getopt_reset;
 	DBTYPE dbtype;
 	DB_ENV	*dbenv;
 	LDG ldg;
-	u_int32_t ldf;
+	u_int ldf;
 	int ch, existed, exitval, ret;
 	char **clist, **clp;
 
-	ldg.progname = "db_load";
+	if ((progname = strrchr(argv[0], '/')) == NULL)
+		progname = argv[0];
+	else
+		++progname;
+
+	if ((ret = db_load_version_check()) != 0)
+		return (ret);
+
+	ldg.progname = progname;
 	ldg.lineno = 0;
 	ldg.endodata = ldg.endofile = 0;
 	ldg.version = 1;
@@ -103,26 +114,40 @@ db_load_main(argc, argv)
 	ldg.home = NULL;
 	ldg.passwd = NULL;
 
-	if ((ret = db_load_version_check(ldg.progname)) != 0)
-		return (ret);
-
+	mode = NOTSET;
 	ldf = 0;
 	exitval = existed = 0;
 	dbtype = DB_UNKNOWN;
 
 	/* Allocate enough room for configuration arguments. */
-	if ((clp = clist = calloc((size_t)argc + 1, sizeof(char *))) == NULL) {
+	if ((clp = clist =
+	    (char **)calloc((size_t)argc + 1, sizeof(char *))) == NULL) {
 		fprintf(stderr, "%s: %s\n", ldg.progname, strerror(ENOMEM));
 		return (EXIT_FAILURE);
 	}
 
+	/*
+	 * There are two modes for db_load: -r and everything else.  The -r
+	 * option zeroes out the database LSN's or resets the file ID, it
+	 * doesn't really "load" a new database.  The functionality is in
+	 * db_load because we don't have a better place to put it, and we
+	 * don't want to create a new utility for just that functionality.
+	 */
 	__db_getopt_reset = 1;
-	while ((ch = getopt(argc, argv, "c:f:h:nP:Tt:V")) != EOF)
+	while ((ch = getopt(argc, argv, "c:f:h:nP:r:Tt:V")) != EOF)
 		switch (ch) {
 		case 'c':
+			if (mode != NOTSET && mode != STANDARD_LOAD)
+				return (db_load_usage());
+			mode = STANDARD_LOAD;
+
 			*clp++ = optarg;
 			break;
 		case 'f':
+			if (mode != NOTSET && mode != STANDARD_LOAD)
+				return (db_load_usage());
+			mode = STANDARD_LOAD;
+
 			if (freopen(optarg, "r", stdin) == NULL) {
 				fprintf(stderr, "%s: %s: reopen: %s\n",
 				    ldg.progname, optarg, strerror(errno));
@@ -133,6 +158,10 @@ db_load_main(argc, argv)
 			ldg.home = optarg;
 			break;
 		case 'n':
+			if (mode != NOTSET && mode != STANDARD_LOAD)
+				return (db_load_usage());
+			mode = STANDARD_LOAD;
+
 			ldf |= LDF_NOOVERWRITE;
 			break;
 		case 'P':
@@ -145,10 +174,28 @@ db_load_main(argc, argv)
 			}
 			ldf |= LDF_PASSWORD;
 			break;
+		case 'r':
+			if (mode == STANDARD_LOAD)
+				return (db_load_usage());
+			if (strcmp(optarg, "lsn") == 0)
+				mode = LSN_RESET;
+			else if (strcmp(optarg, "fileid") == 0)
+				mode = FILEID_RESET;
+			else
+				return (db_load_usage());
+			break;
 		case 'T':
+			if (mode != NOTSET && mode != STANDARD_LOAD)
+				return (db_load_usage());
+			mode = STANDARD_LOAD;
+
 			ldf |= LDF_NOHEADER;
 			break;
 		case 't':
+			if (mode != NOTSET && mode != STANDARD_LOAD)
+				return (db_load_usage());
+			mode = STANDARD_LOAD;
+
 			if (strcmp(optarg, "btree") == 0) {
 				dbtype = DB_BTREE;
 				break;
@@ -189,10 +236,24 @@ db_load_main(argc, argv)
 	if (db_load_env_create(&dbenv, &ldg) != 0)
 		goto shutdown;
 
-	while (!ldg.endofile)
-		if (db_load_load(dbenv, argv[0], dbtype, clist, ldf,
-		    &ldg, &existed) != 0)
-			goto shutdown;
+	/* If we're resetting the LSNs, that's an entirely separate path. */
+	switch (mode) {
+	case FILEID_RESET:
+		exitval = dbenv->fileid_reset(
+		    dbenv, argv[0], ldf & LDF_PASSWORD ? DB_ENCRYPT : 0);
+		break;
+	case LSN_RESET:
+		exitval = dbenv->lsn_reset(
+		    dbenv, argv[0], ldf & LDF_PASSWORD ? DB_ENCRYPT : 0);
+		break;
+	case NOTSET:
+	case STANDARD_LOAD:
+		while (!ldg.endofile)
+			if (db_load_load(dbenv, argv[0], dbtype, clist, ldf,
+			    &ldg, &existed) != 0)
+				goto shutdown;
+		break;
+	}
 
 	if (0) {
 shutdown:	exitval = 1;
@@ -362,13 +423,12 @@ retry_db:
 	/* Open the DB file. */
 	if ((ret = dbp->open(dbp, NULL, name, subdb, dbtype,
 	    DB_CREATE | (TXN_ON(dbenv) ? DB_AUTO_COMMIT : 0),
-	    __db_omode("rwrwrw"))) != 0) {
+	    __db_omode("rw-rw-rw-"))) != 0) {
 		dbp->err(dbp, ret, "DB->open: %s", name);
 		goto err;
 	}
 	if (ldg->private != 0) {
-		if ((ret =
-		    __db_util_cache(dbenv, dbp, &ldg->cache, &resize)) != 0)
+		if ((ret = __db_util_cache(dbp, &ldg->cache, &resize)) != 0)
 			goto err;
 		if (resize) {
 			if ((ret = dbp->close(dbp, 0)) != 0)
@@ -464,8 +524,8 @@ retry:		if (txn != NULL)
 			    name,
 			    !keyflag ? recno : recno * 2 - 1);
 
-			(void)__db_prdbt(&key, checkprint, 0, stderr,
-			    __db_pr_callback, 0, NULL);
+			(void)dbenv->prdbt(&key,
+			    checkprint, 0, stderr, __db_pr_callback, 0);
 			break;
 		case DB_LOCK_DEADLOCK:
 			/* If we have a child txn, retry--else it's fatal. */
@@ -575,8 +635,10 @@ db_load_db_init(dbenv, home, cache, is_private)
 	/* We may be loading into a live environment.  Try and join. */
 	flags = DB_USE_ENVIRON |
 	    DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_MPOOL | DB_INIT_TXN;
-	if (dbenv->open(dbenv, home, flags, 0) == 0)
+	if ((ret = dbenv->open(dbenv, home, flags, 0)) == 0)
 		return (0);
+	if (ret == DB_VERSION_MISMATCH)
+		goto err;
 
 	/*
 	 * We're trying to load a database.
@@ -601,7 +663,7 @@ db_load_db_init(dbenv, home, cache, is_private)
 		return (0);
 
 	/* An environment is required. */
-	dbenv->err(dbenv, ret, "DB_ENV->open");
+err:	dbenv->err(dbenv, ret, "DB_ENV->open");
 	return (1);
 }
 
@@ -612,23 +674,22 @@ db_load_db_init(dbenv, home, cache, is_private)
 			if ((ret = dbp->set_flags(dbp, flag)) != 0) {	\
 				dbp->err(dbp, ret, "%s: set_flags: %s",	\
 				    G(progname), name);			\
-				return (1);				\
+				goto err;				\
 			}						\
 			break;						\
 		case '0':						\
 			break;						\
 		default:						\
 			db_load_badnum(dbenv);					\
-			return (1);					\
+			goto err;					\
 		}							\
 		continue;						\
 	}
 #define	NUMBER(name, value, keyword, func, t)				\
 	if (strcmp(name, keyword) == 0) {				\
-		if (__db_getlong(dbenv,					\
-		    NULL, value, 1, LONG_MAX, &val) != 0)		\
-			return (1);					\
-		if ((ret = dbp->func(dbp, (t)val)) != 0)		\
+		if ((ret = __db_getlong(dbenv,				\
+		    NULL, value, 0, LONG_MAX, &val)) != 0 ||		\
+		    (ret = dbp->func(dbp, (t)val)) != 0)		\
 			goto nameerr;					\
 		continue;						\
 	}
@@ -685,9 +746,6 @@ db_load_configure(dbenv, dbp, clp, subdbp, keysp)
 			continue;
 		}
 
-#ifdef notyet
-		NUMBER(name, value, "bt_maxkey", set_bt_maxkey, u_int32_t);
-#endif
 		NUMBER(name, value, "bt_minkey", set_bt_minkey, u_int32_t);
 		NUMBER(name, value, "db_lorder", set_lorder, int);
 		NUMBER(name, value, "db_pagesize", set_pagesize, u_int32_t);
@@ -709,7 +767,7 @@ db_load_configure(dbenv, dbp, clp, subdbp, keysp)
 
 nameerr:
 	dbp->err(dbp, ret, "%s: %s=%s", G(progname), name, value);
-	return (1);
+err:	return (1);
 }
 
 /*
@@ -731,7 +789,7 @@ db_load_rheader(dbenv, dbp, dbtypep, subdbp, checkprintp, keysp)
 
 	*dbtypep = DB_UNKNOWN;
 	*checkprintp = 0;
-	name = p = NULL;
+	name = NULL;
 
 	/*
 	 * We start with a smallish buffer;  most headers are small.
@@ -740,10 +798,8 @@ db_load_rheader(dbenv, dbp, dbtypep, subdbp, checkprintp, keysp)
 	buflen = 4096;
 	if (G(hdrbuf) == NULL) {
 		hdr = 0;
-		if ((buf = malloc(buflen)) == NULL) {
-memerr:			dbp->errx(dbp, "could not allocate buffer %d", buflen);
-			return (1);
-		}
+		if ((buf = malloc(buflen)) == NULL)
+			goto memerr;
 		G(hdrbuf) = buf;
 		G(origline) = G(lineno);
 	} else {
@@ -768,14 +824,10 @@ memerr:			dbp->errx(dbp, "could not allocate buffer %d", buflen);
 					break;
 				}
 
-				if (ch == '\n')
-					break;
-
 				/*
-				 * If the buffer is too small, double it.  The
-				 * +1 is for the nul byte inserted below.
+				 * If the buffer is too small, double it.
 				 */
-				if (linelen + start + 1 == buflen) {
+				if (linelen + start == buflen) {
 					G(hdrbuf) =
 					    realloc(G(hdrbuf), buflen *= 2);
 					if (G(hdrbuf) == NULL)
@@ -783,6 +835,9 @@ memerr:			dbp->errx(dbp, "could not allocate buffer %d", buflen);
 					buf = &G(hdrbuf)[start];
 				}
 
+				if (ch == '\n')
+					break;
+
 				buf[linelen++] = ch;
 			}
 			if (G(endofile) == 1)
@@ -793,7 +848,6 @@ memerr:			dbp->errx(dbp, "could not allocate buffer %d", buflen);
 		start += linelen;
 
 		if (name != NULL) {
-			*p = '=';
 			free(name);
 			name = NULL;
 		}
@@ -900,19 +954,21 @@ memerr:			dbp->errx(dbp, "could not allocate buffer %d", buflen);
 		goto err;
 	}
 	ret = 0;
+
 	if (0) {
 nameerr:	dbp->err(dbp, ret, "%s: %s=%s", G(progname), name, value);
-err:		ret = 1;
+		ret = 1;
 	}
 	if (0) {
 badfmt:		dbp->errx(dbp, "line %lu: unexpected format", G(lineno));
 		ret = 1;
 	}
-	if (name != NULL) {
-		if (p != NULL)
-			*p = '=';
-		free(name);
+	if (0) {
+memerr:		dbp->errx(dbp, "unable to allocate memory");
+err:		ret = 1;
 	}
+	if (name != NULL)
+		free(name);
 	return (ret);
 }
 
@@ -1125,6 +1181,7 @@ db_load_dbt_rrecno(dbenv, dbtp, ishex)
 	int ishex;
 {
 	char buf[32], *p, *q;
+	u_long recno;
 
 	++G(lineno);
 
@@ -1163,12 +1220,12 @@ db_load_dbt_rrecno(dbenv, dbtp, ishex)
 		*p = '\0';
 	}
 
-	if (__db_getulong(dbenv,
-	    G(progname), buf + 1, 0, 0, (u_long *)dbtp->data)) {
+	if (__db_getulong(dbenv, G(progname), buf + 1, 0, 0, &recno)) {
 bad:		db_load_badend(dbenv);
 		return (1);
 	}
 
+	*((db_recno_t *)dbtp->data) = recno;
 	dbtp->size = sizeof(db_recno_t);
 	return (0);
 }
@@ -1253,15 +1310,16 @@ db_load_badend(dbenv)
 int
 db_load_usage()
 {
-	(void)fprintf(stderr, "%s\n\t%s\n",
-	    "usage: db_load [-nTV] [-c name=value] [-f file]",
+	(void)fprintf(stderr, "usage: %s %s\n\t%s\n", progname,
+	    "[-nTV] [-c name=value] [-f file]",
     "[-h home] [-P password] [-t btree | hash | recno | queue] db_file");
+	(void)fprintf(stderr, "usage: %s %s\n",
+	    progname, "-r lsn | fileid [-h home] [-P password] db_file");
 	return (EXIT_FAILURE);
 }
 
 int
-db_load_version_check(progname)
-	const char *progname;
+db_load_version_check()
 {
 	int v_major, v_minor, v_patch;
 
