@@ -1,17 +1,17 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: db_verify.c,v 12.3 2005/06/16 20:21:37 bostic Exp $
  */
 
 #include "db_config.h"
 
 #ifndef lint
 static const char copyright[] =
-    "Copyright (c) 1996-2003\nSleepycat Software Inc.  All rights reserved.\n";
-static const char revid[] =
-    "$Id: db_verify.c,v 1.45 2003/08/13 19:57:09 ubell Exp $";
+    "Copyright (c) 1996-2005\nSleepycat Software Inc.  All rights reserved.\n";
 #endif
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -27,7 +27,9 @@ static const char revid[] =
 
 int db_verify_main __P((int, char *[]));
 int db_verify_usage __P((void));
-int db_verify_version_check __P((const char *));
+int db_verify_version_check __P((void));
+
+const char *progname;
 
 int
 db_verify(args)
@@ -50,24 +52,29 @@ db_verify_main(argc, argv)
 {
 	extern char *optarg;
 	extern int optind, __db_getopt_reset;
-	const char *progname = "db_verify";
 	DB *dbp, *dbp1;
 	DB_ENV *dbenv;
-	u_int32_t cache;
-	int ch, exitval, nflag, oflag, private;
+	u_int32_t flags, cache;
+	int ch, exitval, nflag, private;
 	int quiet, resize, ret;
 	char *home, *passwd;
 
-	if ((ret = db_verify_version_check(progname)) != 0)
+	if ((progname = strrchr(argv[0], '/')) == NULL)
+		progname = argv[0];
+	else
+		++progname;
+
+	if ((ret = db_verify_version_check()) != 0)
 		return (ret);
 
 	dbenv = NULL;
 	dbp = NULL;
 	cache = MEGABYTE;
-	exitval = nflag = oflag = quiet = 0;
+	exitval = nflag = quiet = 0;
+	flags = 0;
 	home = passwd = NULL;
 	__db_getopt_reset = 1;
-	while ((ch = getopt(argc, argv, "h:NoP:qV")) != EOF)
+	while ((ch = getopt(argc, argv, "h:NoP:quV")) != EOF)
 		switch (ch) {
 		case 'h':
 			home = optarg;
@@ -85,11 +92,14 @@ db_verify_main(argc, argv)
 			}
 			break;
 		case 'o':
-			oflag = 1;
+			LF_SET(DB_NOORDERCHK);
 			break;
 		case 'q':
 			quiet = 1;
 			break;
+		case 'u':			/* Undocumented. */
+			LF_SET(DB_UNREF);
+			break;
 		case 'V':
 			printf("%s\n", db_version(NULL, NULL, NULL));
 			return (EXIT_SUCCESS);
@@ -145,14 +155,18 @@ retry:	if ((ret = db_env_create(&dbenv, 0)) != 0) {
 	private = 0;
 	if ((ret =
 	    dbenv->open(dbenv, home, DB_INIT_MPOOL | DB_USE_ENVIRON, 0)) != 0) {
-		if ((ret = dbenv->set_cachesize(dbenv, 0, cache, 1)) != 0) {
-			dbenv->err(dbenv, ret, "set_cachesize");
-			goto shutdown;
+		if (ret != DB_VERSION_MISMATCH) {
+			if ((ret =
+			    dbenv->set_cachesize(dbenv, 0, cache, 1)) != 0) {
+				dbenv->err(dbenv, ret, "set_cachesize");
+				goto shutdown;
+			}
+			private = 1;
+			ret = dbenv->open(dbenv, home, DB_CREATE |
+			    DB_INIT_MPOOL | DB_PRIVATE | DB_USE_ENVIRON, 0);
 		}
-		private = 1;
-		if ((ret = dbenv->open(dbenv, home,
-	    DB_CREATE | DB_INIT_MPOOL | DB_PRIVATE | DB_USE_ENVIRON, 0)) != 0) {
-			dbenv->err(dbenv, ret, "open");
+		if (ret != 0) {
+			dbenv->err(dbenv, ret, "DB_ENV->open");
 			goto shutdown;
 		}
 	}
@@ -166,6 +180,10 @@ retry:	if ((ret = db_env_create(&dbenv, 0)) != 0) {
 		/*
 		 * We create a 2nd dbp to this database to get its pagesize
 		 * because the dbp we're using for verify cannot be opened.
+		 *
+		 * If the database is corrupted, we may not be able to open
+		 * it, of course.  In that case, just continue, using the
+		 * cache size we have.
 		 */
 		if (private) {
 			if ((ret = db_create(&dbp1, dbenv, 0)) != 0) {
@@ -174,12 +192,9 @@ retry:	if ((ret = db_env_create(&dbenv, 0)) != 0) {
 				goto shutdown;
 			}
 
-			if ((ret = dbp1->open(dbp1, NULL,
-			    argv[0], NULL, DB_UNKNOWN, DB_RDONLY, 0)) != 0) {
-				dbenv->err(dbenv, ret, "DB->open: %s", argv[0]);
-				(void)dbp1->close(dbp1, 0);
-				goto shutdown;
-			}
+			ret = dbp1->open(dbp1,
+			    NULL, argv[0], NULL, DB_UNKNOWN, DB_RDONLY, 0);
+
 			/*
 			 * If we get here, we can check the cache/page.
 			 * !!!
@@ -188,29 +203,26 @@ retry:	if ((ret = db_env_create(&dbenv, 0)) != 0) {
 			 * will still be working on the same argv when we
 			 * get back into the for-loop.
 			 */
-			ret = __db_util_cache(dbenv, dbp1, &cache, &resize);
-			(void)dbp1->close(dbp1, 0);
-			if (ret != 0)
-				goto shutdown;
+			if (ret == 0) {
+				if (__db_util_cache(
+				    dbp1, &cache, &resize) == 0 && resize) {
+					(void)dbp1->close(dbp1, 0);
+					(void)dbp->close(dbp, 0);
+					dbp = NULL;
 
-			if (resize) {
-				(void)dbp->close(dbp, 0);
-				dbp = NULL;
-
-				(void)dbenv->close(dbenv, 0);
-				dbenv = NULL;
-				goto retry;
+					(void)dbenv->close(dbenv, 0);
+					dbenv = NULL;
+					goto retry;
+				}
 			}
+			(void)dbp1->close(dbp1, 0);
 		}
 
 		/* The verify method is a destructor. */
-		ret = dbp->verify(dbp,
-		    argv[0], NULL, NULL, oflag ? DB_NOORDERCHK : 0);
+		ret = dbp->verify(dbp, argv[0], NULL, NULL, flags);
 		dbp = NULL;
-		if (ret != 0) {
-			dbenv->err(dbenv, ret, "DB->verify: %s", argv[0]);
+		if (ret != 0)
 			goto shutdown;
-		}
 	}
 
 	if (0) {
@@ -239,14 +251,13 @@ shutdown:	exitval = 1;
 int
 db_verify_usage()
 {
-	fprintf(stderr, "%s\n",
-	    "usage: db_verify [-NoqV] [-h home] [-P password] db_file ...");
+	fprintf(stderr, "usage: %s %s\n", progname,
+	    "[-NoqV] [-h home] [-P password] db_file ...");
 	return (EXIT_FAILURE);
 }
 
 int
-db_verify_version_check(progname)
-	const char *progname;
+db_verify_version_check()
 {
 	int v_major, v_minor, v_patch;
 
