@@ -1,16 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2001
+ * Copyright (c) 1997-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: raise.c,v 12.2 2005/06/16 20:20:50 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: raise.c,v 11.4 2001/01/25 18:22:37 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <signal.h>
 #include <unistd.h>
