@@ -1,16 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2000-2003
+ * Copyright (c) 2000-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: util_sig.c,v 12.1 2005/06/16 20:20:55 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: util_sig.c,v 1.8 2003/01/08 04:07:58 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
