@@ -1,19 +1,17 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
  *
  * Some parts of this code originally written by Adam Stubblefield
- * - astubble@rice.edu
+ * -- astubble@rice.edu
+ *
+ * $Id: crypto.c,v 12.5 2005/07/20 16:50:56 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: crypto.c,v 1.23 2003/06/30 17:19:41 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <string.h>
 #endif
@@ -42,35 +40,31 @@ __crypto_region_init(dbenv)
 	ret = 0;
 	infop = dbenv->reginfo;
 	renv = infop->primary;
-	MUTEX_LOCK(dbenv, &renv->mutex);
 	if (renv->cipher_off == INVALID_ROFF) {
 		if (!CRYPTO_ON(dbenv))
-			goto out;
+			return (0);
 		if (!F_ISSET(infop, REGION_CREATE)) {
 			__db_err(dbenv,
     "Joining non-encrypted environment with encryption key");
-			ret = EINVAL;
-			goto out;
+			return (EINVAL);
 		}
 		if (F_ISSET(db_cipher, CIPHER_ANY)) {
 			__db_err(dbenv, "Encryption algorithm not supplied");
-			ret = EINVAL;
-			goto out;
+			return (EINVAL);
 		}
 		/*
-		 * Must create the shared information.  We need:
-		 * Shared cipher information that contains the passwd.
-		 * After we copy the passwd, we smash and free the one in the
-		 * dbenv.
+		 * Must create the shared information.  We need: Shared cipher
+		 * information that contains the passwd.  After we copy the
+		 * passwd, we smash and free the one in the dbenv.
 		 */
-		if ((ret = __db_shalloc(infop->addr,
-		    sizeof(CIPHER), MUTEX_ALIGN, &cipher)) != 0)
-			goto out;
+		if ((ret =
+		    __db_shalloc(infop, sizeof(CIPHER), 0, &cipher)) != 0)
+			return (ret);
 		memset(cipher, 0, sizeof(*cipher));
-		if ((ret = __db_shalloc(infop->addr,
-		    dbenv->passwd_len, 0, &sh_passwd)) != 0) {
-			__db_shalloc_free(infop->addr, cipher);
-			goto out;
+		if ((ret = __db_shalloc(
+		    infop, dbenv->passwd_len, 0, &sh_passwd)) != 0) {
+			__db_shalloc_free(infop, cipher);
+			return (ret);
 		}
 		memset(sh_passwd, 0, dbenv->passwd_len);
 		cipher->passwd = R_OFFSET(infop, sh_passwd);
@@ -81,51 +75,44 @@ __crypto_region_init(dbenv)
 	} else {
 		if (!CRYPTO_ON(dbenv)) {
 			__db_err(dbenv,
-    "Encrypted environment: no encryption key supplied");
-			ret = EINVAL;
-			goto out;
+		    "Encrypted environment: no encryption key supplied");
+			return (EINVAL);
 		}
 		cipher = R_ADDR(infop, renv->cipher_off);
 		sh_passwd = R_ADDR(infop, cipher->passwd);
 		if ((cipher->passwd_len != dbenv->passwd_len) ||
 		    memcmp(dbenv->passwd, sh_passwd, cipher->passwd_len) != 0) {
 			__db_err(dbenv, "Invalid password");
-			ret = EPERM;
-			goto out;
+			return (EPERM);
 		}
 		if (!F_ISSET(db_cipher, CIPHER_ANY) &&
 		    db_cipher->alg != cipher->flags) {
 			__db_err(dbenv,
     "Environment encrypted using a different algorithm");
-			ret = EINVAL;
-			goto out;
+			return (EINVAL);
 		}
 		if (F_ISSET(db_cipher, CIPHER_ANY))
 			/*
-			 * We have CIPHER_ANY and we are joining the
-			 * existing env.  Setup our cipher structure
-			 * for whatever algorithm this env has.
+			 * We have CIPHER_ANY and we are joining the existing
+			 * env.  Setup our cipher structure for whatever
+			 * algorithm this env has.
 			 */
 			if ((ret = __crypto_algsetup(dbenv, db_cipher,
 			    cipher->flags, 0)) != 0)
-				goto out;
+				return (ret);
 	}
-	MUTEX_UNLOCK(dbenv, &renv->mutex);
 	ret = db_cipher->init(dbenv, db_cipher);
 
 	/*
-	 * On success, no matter if we allocated it or are using the
-	 * already existing one, we are done with the passwd in the dbenv.
-	 * We smash N-1 bytes so that we don't overwrite the nul.
+	 * On success, no matter if we allocated it or are using the already
+	 * existing one, we are done with the passwd in the dbenv.  We smash
+	 * N-1 bytes so that we don't overwrite the nul.
 	 */
 	memset(dbenv->passwd, 0xff, dbenv->passwd_len-1);
 	__os_free(dbenv, dbenv->passwd);
 	dbenv->passwd = NULL;
 	dbenv->passwd_len = 0;
 
-	if (0) {
-out:		MUTEX_UNLOCK(dbenv, &renv->mutex);
-	}
 	return (ret);
 }
 
@@ -157,6 +144,30 @@ __crypto_dbenv_close(dbenv)
 	return (ret);
 }
 
+/*
+ * __crypto_region_destroy --
+ *	Destroy any system resources allocated in the primary region.
+ *
+ * PUBLIC: int __crypto_region_destroy __P((DB_ENV *));
+ */
+int
+__crypto_region_destroy(dbenv)
+	DB_ENV *dbenv;
+{
+	CIPHER *cipher;
+	REGENV *renv;
+	REGINFO *infop;
+
+	infop = dbenv->reginfo;
+	renv = infop->primary;
+	if (renv->cipher_off != INVALID_ROFF) {
+		cipher = R_ADDR(infop, renv->cipher_off);
+		__db_shalloc_free(infop, R_ADDR(infop, cipher->passwd));
+		__db_shalloc_free(infop, cipher);
+	}
+	return (0);
+}
+
 /*
  * __crypto_algsetup --
  *	Given a db_cipher structure and a valid algorithm flag, call
@@ -214,33 +225,54 @@ __crypto_decrypt_meta(dbenv, dbp, mbuf, do_metachk)
 	u_int8_t *iv;
 
 	/*
-	 * If we weren't given a dbp, we just want to decrypt the page
-	 * on behalf of some internal subsystem, not on behalf of a user
-	 * with a dbp.  Therefore, set up a dummy dbp so that the call
-	 * to P_OVERHEAD below works.
+	 * If we weren't given a dbp, we just want to decrypt the page on
+	 * behalf of some internal subsystem, not on behalf of a user with
+	 * a dbp.  Therefore, set up a dummy dbp so that the call to
+	 * P_OVERHEAD below works.
 	 */
 	if (dbp == NULL) {
 		memset(&dummydb, 0, sizeof(DB));
 		dbp = &dummydb;
 	}
-	/*
-	 * Meta-pages may be encrypted for DBMETASIZE bytes.  If
-	 * we have a non-zero IV (that is written after encryption)
-	 * then we decrypt (or error if the user isn't set up for
-	 * security).  We guarantee that the IV space on non-encrypted
-	 * pages will be zero and a zero-IV is illegal for encryption.
-	 * Therefore any non-zero IV means an encrypted database.
-	 * This basically checks the passwd on the file
-	 * if we cannot find a good magic number.
-	 * We walk through all the algorithms we know about attempting
-	 * to decrypt (and possibly byteswap).
-	 *
-	 * !!!
-	 * All method meta pages have the IV and checksum at the
-	 * exact same location, but not in DBMETA, use BTMETA.
-	 */
+
 	ret = 0;
 	meta = (DBMETA *)mbuf;
+
+	/*
+	 * !!!
+	 * We used an "unused" field in the meta-data page to flag whether or
+	 * not the database is encrypted.  Unfortunately, that unused field
+	 * was used in Berkeley DB releases before 3.0 (for example, 2.7.7).
+	 * It would have been OK, except encryption doesn't follow the usual
+	 * rules of "upgrade before doing anything else", we check encryption
+	 * before checking for old versions of the database.
+	 *
+	 * We don't have to check Btree databases -- before 3.0, the field of
+	 * interest was the bt_maxkey field (which was never supported and has
+	 * since been removed).
+	 *
+	 * Ugly check to jump out if this format is older than what we support.
+	 * It assumes no encrypted page will have an unencrypted magic number,
+	 * but that seems relatively safe.  [#10920]
+	 */
+	if (meta->magic == DB_HASHMAGIC && meta->version <= 5)
+		return (0);
+
+	/*
+	 * Meta-pages may be encrypted for DBMETASIZE bytes.  If we have a
+	 * non-zero IV (that is written after encryption) then we decrypt (or
+	 * error if the user isn't set up for security).  We guarantee that
+	 * the IV space on non-encrypted pages will be zero and a zero-IV is
+	 * illegal for encryption.  Therefore any non-zero IV means an
+	 * encrypted database.  This basically checks the passwd on the file
+	 * if we cannot find a good magic number.  We walk through all the
+	 * algorithms we know about attempting to decrypt (and possibly
+	 * byteswap).
+	 *
+	 * !!!
+	 * All method meta pages have the IV and checksum at the exact same
+	 * location, but not in DBMETA, use BTMETA.
+	 */
 	if (meta->encrypt_alg != 0) {
 		db_cipher = (DB_CIPHER *)dbenv->crypto_handle;
 		if (!F_ISSET(dbp, DB_AM_ENCRYPT)) {
@@ -250,11 +282,10 @@ __crypto_decrypt_meta(dbenv, dbp, mbuf, do_metachk)
 				return (EINVAL);
 			}
 			/*
-			 * User has a correct, secure env, but has
-			 * encountered a database in that env that is
-			 * secure, but user didn't dbp->set_flags.  Since
-			 * it is existing, use encryption if it is that
-			 * way already.
+			 * User has a correct, secure env, but has encountered
+			 * a database in that env that is secure, but user
+			 * didn't dbp->set_flags.  Since it is existing, use
+			 * encryption if it is that way already.
 			 */
 			F_SET(dbp, DB_AM_ENCRYPT|DB_AM_CHKSUM);
 		}
@@ -272,9 +303,9 @@ __crypto_decrypt_meta(dbenv, dbp, mbuf, do_metachk)
 		DB_ASSERT(F_ISSET(dbp, DB_AM_CHKSUM));
 		iv = ((BTMETA *)mbuf)->iv;
 		/*
-		 * For ALL pages, we do not encrypt the beginning
-		 * of the page that contains overhead information.
-		 * This is true of meta and all other pages.
+		 * For ALL pages, we do not encrypt the beginning of the page
+		 * that contains overhead information.  This is true of meta
+		 * and all other pages.
 		 */
 		pg_off = P_OVERHEAD(dbp);
 alg_retry:
@@ -308,10 +339,10 @@ alg_retry:
 		goto alg_retry;
 	} else if (F_ISSET(dbp, DB_AM_ENCRYPT)) {
 		/*
-		 * They gave us a passwd, but the database is not
-		 * encrypted.  This is an error.  We do NOT want to
-		 * silently allow them to write data in the clear when
-		 * the user set up and expects encrypted data.
+		 * They gave us a passwd, but the database is not encrypted.
+		 * This is an error.  We do NOT want to silently allow them
+		 * to write data in the clear when the user set up and expects
+		 * encrypted data.
 		 *
 		 * This covers at least the following scenario.
 		 * 1.  User creates and sets up an encrypted database.
@@ -359,5 +390,5 @@ __crypto_set_passwd(dbenv_src, dbenv_dest)
 
 	cipher = R_ADDR(infop, renv->cipher_off);
 	sh_passwd = R_ADDR(infop, cipher->passwd);
-	return (__dbenv_set_encrypt(dbenv_dest, sh_passwd, DB_ENCRYPT_AES));
+	return (__env_set_encrypt(dbenv_dest, sh_passwd, DB_ENCRYPT_AES));
 }
