@@ -1,16 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2003
+ * Copyright (c) 1997-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: cxx_lock.cpp,v 12.1 2005/06/16 20:20:59 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: cxx_lock.cpp,v 11.18 2003/01/08 04:10:36 bostic Exp $";
-#endif /* not lint */
-
 #include <errno.h>
 #include <string.h>
 
