@@ -1,16 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1998-2001
+ * Copyright (c) 1998-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: db_am.c,v 12.12 2005/11/01 00:44:09 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: db_am.c,v 11.62 2001/07/02 16:45:54 krinsky Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -18,92 +16,27 @@ static const char revid[] = "$Id: db_am.c,v 11.62 2001/07/02 16:45:54 krinsky Ex
 #endif
 
 #include "db_int.h"
-#include "db_page.h"
-#include "db_shash.h"
-#include "btree.h"
-#include "hash.h"
-#include "qam.h"
-#include "lock.h"
-#include "mp.h"
-#include "txn.h"
-#include "db_am.h"
-#include "db_ext.h"
+#include "dbinc/db_page.h"
+#include "dbinc/db_shash.h"
+#include "dbinc/btree.h"
+#include "dbinc/hash.h"
+#include "dbinc/lock.h"
+#include "dbinc/log.h"
+#include "dbinc/mp.h"
+#include "dbinc/qam.h"
 
 static int __db_append_primary __P((DBC *, DBT *, DBT *));
 static int __db_secondary_get __P((DB *, DB_TXN *, DBT *, DBT *, u_int32_t));
-static int __db_secondary_close __P((DB *, u_int32_t));
 
 /*
- * __db_cursor --
- *	Allocate and return a cursor.
+ * __db_cursor_int --
+ *	Internal routine to create a cursor.
  *
- * PUBLIC: int __db_cursor __P((DB *, DB_TXN *, DBC **, u_int32_t));
- */
-int
-__db_cursor(dbp, txn, dbcp, flags)
-	DB *dbp;
-	DB_TXN *txn;
-	DBC **dbcp;
-	u_int32_t flags;
-{
-	DB_ENV *dbenv;
-	DBC *dbc;
-	db_lockmode_t mode;
-	u_int32_t op;
-	int dirty, ret;
-
-	dbenv = dbp->dbenv;
-
-	PANIC_CHECK(dbenv);
-	DB_ILLEGAL_BEFORE_OPEN(dbp, "DB->cursor");
-
-	/* Check for invalid flags. */
-	dirty = LF_ISSET(DB_DIRTY_READ) ? 1 : 0;
-	LF_CLR(DB_DIRTY_READ);
-
-	if ((ret = __db_cursorchk(dbp, flags, F_ISSET(dbp, DB_AM_RDONLY))) != 0)
-		return (ret);
-
-	if ((ret = __db_icursor(dbp,
-	    txn, dbp->type, PGNO_INVALID, 0, DB_LOCK_INVALIDID, dbcp)) != 0)
-		return (ret);
-	dbc = *dbcp;
-
-	/*
-	 * If this is CDB, do all the locking in the interface, which is
-	 * right here.
-	 */
-	if (CDB_LOCKING(dbenv)) {
-		op = LF_ISSET(DB_OPFLAGS_MASK);
-		mode = (op == DB_WRITELOCK) ? DB_LOCK_WRITE :
-		    ((op == DB_WRITECURSOR) ? DB_LOCK_IWRITE : DB_LOCK_READ);
-		if ((ret = lock_get(dbenv, dbc->locker, 0,
-		    &dbc->lock_dbt, mode, &dbc->mylock)) != 0) {
-			(void)__db_c_close(dbc);
-			return (ret);
-		}
-		if (op == DB_WRITECURSOR)
-			F_SET(dbc, DBC_WRITECURSOR);
-		if (op == DB_WRITELOCK)
-			F_SET(dbc, DBC_WRITER);
-	}
-
-	if (dirty || (txn != NULL && F_ISSET(txn, TXN_DIRTY_READ)))
-		F_SET(dbc, DBC_DIRTY_READ);
-	return (0);
-}
-
-/*
- * __db_icursor --
- *	Internal version of __db_cursor.  If dbcp is
- *	non-NULL it is assumed to point to an area to
- *	initialize as a cursor.
- *
- * PUBLIC: int __db_icursor
+ * PUBLIC: int __db_cursor_int
  * PUBLIC:     __P((DB *, DB_TXN *, DBTYPE, db_pgno_t, int, u_int32_t, DBC **));
  */
 int
-__db_icursor(dbp, txn, dbtype, root, is_opd, lockerid, dbcp)
+__db_cursor_int(dbp, txn, dbtype, root, is_opd, lockerid, dbcp)
 	DB *dbp;
 	DB_TXN *txn;
 	DBTYPE dbtype;
@@ -112,31 +45,36 @@ __db_icursor(dbp, txn, dbtype, root, is_opd, lockerid, dbcp)
 	u_int32_t lockerid;
 	DBC **dbcp;
 {
-	DBC *dbc, *adbc;
+	DBC *dbc;
 	DBC_INTERNAL *cp;
 	DB_ENV *dbenv;
+	db_threadid_t tid;
 	int allocated, ret;
+	pid_t pid;
 
 	dbenv = dbp->dbenv;
 	allocated = 0;
 
 	/*
+	 * If dbcp is non-NULL it is assumed to point to an area to initialize
+	 * as a cursor.
+	 *
 	 * Take one from the free list if it's available.  Take only the
 	 * right type.  With off page dups we may have different kinds
 	 * of cursors on the queue for a single database.
 	 */
-	MUTEX_THREAD_LOCK(dbenv, dbp->mutexp);
+	MUTEX_LOCK(dbenv, dbp->mutex);
 	for (dbc = TAILQ_FIRST(&dbp->free_queue);
 	    dbc != NULL; dbc = TAILQ_NEXT(dbc, links))
 		if (dbtype == dbc->dbtype) {
 			TAILQ_REMOVE(&dbp->free_queue, dbc, links);
-			dbc->flags = 0;
+			F_CLR(dbc, ~DBC_OWN_LID);
 			break;
 		}
-	MUTEX_THREAD_UNLOCK(dbenv, dbp->mutexp);
+	MUTEX_UNLOCK(dbenv, dbp->mutex);
 
 	if (dbc == NULL) {
-		if ((ret = __os_calloc(dbp->dbenv, 1, sizeof(DBC), &dbc)) != 0)
+		if ((ret = __os_calloc(dbenv, 1, sizeof(DBC), &dbc)) != 0)
 			return (ret);
 		allocated = 1;
 		dbc->flags = 0;
@@ -146,25 +84,34 @@ __db_icursor(dbp, txn, dbtype, root, is_opd, lockerid, dbcp)
 		/* Set up locking information. */
 		if (LOCKING_ON(dbenv)) {
 			/*
-			 * If we are not threaded, then there is no need to
-			 * create new locker ids.  We know that no one else
-			 * is running concurrently using this DB, so we can
-			 * take a peek at any cursors on the active queue.
+			 * If we are not threaded, we share a locker ID among
+			 * all cursors opened in the environment handle,
+			 * allocating one if this is the first cursor.
+			 *
+			 * This relies on the fact that non-threaded DB handles
+			 * always have non-threaded environment handles, since
+			 * we set DB_THREAD on DB handles created with threaded
+			 * environment handles.
 			 */
-			if (!DB_IS_THREADED(dbp) &&
-			    (adbc = TAILQ_FIRST(&dbp->active_queue)) != NULL)
-				dbc->lid = adbc->lid;
-			else
-				if ((ret = lock_id(dbenv, &dbc->lid)) != 0)
+			if (!DB_IS_THREADED(dbp)) {
+				if (dbp->dbenv->env_lref == NULL &&
+				    (ret = __lock_id(dbenv, NULL,
+				    (DB_LOCKER **)&dbp->dbenv->env_lref)) != 0)
 					goto err;
+				dbc->lref = dbp->dbenv->env_lref;
+			} else {
+				if ((ret = __lock_id(dbenv, NULL,
+				    (DB_LOCKER **)&dbc->lref)) != 0)
+					goto err;
+				F_SET(dbc, DBC_OWN_LID);
+			}
 
 			/*
 			 * In CDB, secondary indices should share a lock file
-			 * ID with the primary;  otherwise we're susceptible to
-			 * deadlocks.  We also use __db_icursor rather
-			 * than sdbp->cursor to create secondary update
-			 * cursors in c_put and c_del;  these won't
-			 * acquire a new lock.
+			 * ID with the primary;  otherwise we're susceptible
+			 * to deadlocks.  We also use __db_cursor_int rather
+			 * than __db_cursor to create secondary update cursors
+			 * in c_put and c_del; these won't acquire a new lock.
 			 *
 			 * !!!
 			 * Since this is in the one-time cursor allocation
@@ -172,7 +119,7 @@ __db_icursor(dbp, txn, dbtype, root, is_opd, lockerid, dbcp)
 			 * close, all cursors in the secondary when we
 			 * associate.
 			 */
-			if (CDB_LOCKING(dbp->dbenv) &&
+			if (CDB_LOCKING(dbenv) &&
 			    F_ISSET(dbp, DB_AM_SECONDARY))
 				memcpy(dbc->lock.fileid,
 				    dbp->s_primary->fileid, DB_FILE_ID_LEN);
@@ -218,9 +165,9 @@ __db_icursor(dbp, txn, dbtype, root, is_opd, lockerid, dbcp)
 			if ((ret = __qam_c_init(dbc)) != 0)
 				goto err;
 			break;
+		case DB_UNKNOWN:
 		default:
-			ret = __db_unknown_type(dbp->dbenv,
-			    "__db_icursor", dbtype);
+			ret = __db_unknown_type(dbenv, "DB->cursor", dbtype);
 			goto err;
 		}
 
@@ -231,7 +178,9 @@ __db_icursor(dbp, txn, dbtype, root, is_opd, lockerid, dbcp)
 	dbc->dbtype = dbtype;
 	RESET_RET_MEM(dbc);
 
-	if ((dbc->txn = txn) == NULL) {
+	if ((dbc->txn = txn) != NULL)
+		dbc->locker = txn->txnid;
+	else if (LOCKING_ON(dbenv)) {
 		/*
 		 * There are certain cases in which we want to create a
 		 * new cursor with a particular locker ID that is known
@@ -249,17 +198,24 @@ __db_icursor(dbp, txn, dbtype, root, is_opd, lockerid, dbcp)
 		 * primary are subdatabases or we're using env-wide locking,
 		 * this is disastrous.
 		 *
-		 * In these cases, our caller will pass a nonzero locker ID
-		 * into this function.  Use this locker ID instead of dbc->lid
-		 * as the locker ID for our new cursor.
+		 * In these cases, our caller will pass a nonzero locker
+		 * ID into this function.  Use this locker ID instead of
+		 * the default as the locker ID for our new cursor.
 		 */
 		if (lockerid != DB_LOCK_INVALIDID)
 			dbc->locker = lockerid;
-		else
-			dbc->locker = dbc->lid;
-	} else {
-		dbc->locker = txn->txnid;
-		txn->cursors++;
+		else {
+			/*
+			 * If we are threaded then we need to set the
+			 * proper thread id into the locker.
+			 */
+			if (DB_IS_THREADED(dbp)) {
+				dbenv->thread_id(dbenv, &pid, &tid);
+				__lock_set_thread_id(
+				    (DB_LOCKER *)dbc->lref, pid, tid);
+			}
+			dbc->locker = ((DB_LOCKER *)dbc->lref)->id;
+		}
 	}
 
 	/*
@@ -270,12 +226,14 @@ __db_icursor(dbp, txn, dbtype, root, is_opd, lockerid, dbcp)
 	 * __db_c_get is used by all access methods, so this should be safe.
 	 */
 	if (F_ISSET(dbp, DB_AM_SECONDARY))
-		dbc->c_get = __db_c_secondary_get;
+		dbc->c_get = __db_c_secondary_get_pp;
 
 	if (is_opd)
 		F_SET(dbc, DBC_OPD);
 	if (F_ISSET(dbp, DB_AM_RECOVER))
 		F_SET(dbc, DBC_RECOVER);
+	if (F_ISSET(dbp, DB_AM_COMPENSATE))
+		F_SET(dbc, DBC_COMPENSATE);
 
 	/* Refresh the DBC internal structure. */
 	cp = dbc->internal;
@@ -295,193 +253,30 @@ __db_icursor(dbp, txn, dbtype, root, is_opd, lockerid, dbcp)
 	case DB_HASH:
 	case DB_QUEUE:
 		break;
+	case DB_UNKNOWN:
 	default:
-		ret = __db_unknown_type(dbp->dbenv, "__db_icursor", dbp->type);
+		ret = __db_unknown_type(dbenv, "DB->cursor", dbp->type);
 		goto err;
 	}
 
-	MUTEX_THREAD_LOCK(dbenv, dbp->mutexp);
+	/*
+	 * The transaction keeps track of how many cursors were opened within
+	 * it to catch application errors where the cursor isn't closed when
+	 * the transaction is resolved.
+	 */
+	if (txn != NULL)
+		++txn->cursors;
+
+	MUTEX_LOCK(dbenv, dbp->mutex);
 	TAILQ_INSERT_TAIL(&dbp->active_queue, dbc, links);
 	F_SET(dbc, DBC_ACTIVE);
-	MUTEX_THREAD_UNLOCK(dbenv, dbp->mutexp);
+	MUTEX_UNLOCK(dbenv, dbp->mutex);
 
 	*dbcp = dbc;
 	return (0);
 
 err:	if (allocated)
-		__os_free(dbp->dbenv, dbc, sizeof(*dbc));
-	return (ret);
-}
-
-#ifdef DEBUG
-/*
- * __db_cprint --
- *	Display the current cursor list.
- *
- * PUBLIC: int __db_cprint __P((DB *));
- */
-int
-__db_cprint(dbp)
-	DB *dbp;
-{
-	static const FN fn[] = {
-		{ DBC_ACTIVE,		"active" },
-		{ DBC_COMPENSATE,	"compensate" },
-		{ DBC_OPD,		"off-page-dup" },
-		{ DBC_RECOVER,		"recover" },
-		{ DBC_RMW,		"read-modify-write" },
-		{ DBC_TRANSIENT,	"transient" },
-		{ DBC_WRITECURSOR,	"write cursor" },
-		{ DBC_WRITEDUP,		"internally dup'ed write cursor" },
-		{ DBC_WRITER,		"short-term write cursor" },
-		{ 0,			NULL }
-	};
-	DBC *dbc;
-	DBC_INTERNAL *cp;
-	char *s;
-
-	MUTEX_THREAD_LOCK(dbp->dbenv, dbp->mutexp);
-	for (dbc = TAILQ_FIRST(&dbp->active_queue);
-	    dbc != NULL; dbc = TAILQ_NEXT(dbc, links)) {
-		switch (dbc->dbtype) {
-		case DB_BTREE:
-			s = "btree";
-			break;
-		case DB_HASH:
-			s = "hash";
-			break;
-		case DB_RECNO:
-			s = "recno";
-			break;
-		case DB_QUEUE:
-			s = "queue";
-			break;
-		default:
-			DB_ASSERT(0);
-			return (1);
-		}
-		cp = dbc->internal;
-		fprintf(stderr, "%s/%#0lx: opd: %#0lx\n",
-		    s, P_TO_ULONG(dbc), P_TO_ULONG(cp->opd));
-		fprintf(stderr, "\ttxn: %#0lx lid: %lu locker: %lu\n",
-		    P_TO_ULONG(dbc->txn),
-		    (u_long)dbc->lid, (u_long)dbc->locker);
-		fprintf(stderr, "\troot: %lu page/index: %lu/%lu",
-		    (u_long)cp->root, (u_long)cp->pgno, (u_long)cp->indx);
-		__db_prflags(dbc->flags, fn, stderr);
-		fprintf(stderr, "\n");
-
-		switch (dbp->type) {
-		case DB_BTREE:
-			__bam_cprint(dbc);
-			break;
-		case DB_HASH:
-			__ham_cprint(dbc);
-			break;
-		default:
-			break;
-		}
-	}
-	for (dbc = TAILQ_FIRST(&dbp->free_queue);
-	    dbc != NULL; dbc = TAILQ_NEXT(dbc, links))
-		fprintf(stderr, "free: %#0lx ", P_TO_ULONG(dbc));
-	fprintf(stderr, "\n");
-	MUTEX_THREAD_UNLOCK(dbp->dbenv, dbp->mutexp);
-
-	return (0);
-}
-#endif /* DEBUG */
-
-/*
- * db_fd --
- *	Return a file descriptor for flock'ing.
- *
- * PUBLIC: int __db_fd __P((DB *, int *));
- */
-int
-__db_fd(dbp, fdp)
-	DB *dbp;
-	int *fdp;
-{
-	DB_FH *fhp;
-	int ret;
-
-	PANIC_CHECK(dbp->dbenv);
-	DB_ILLEGAL_BEFORE_OPEN(dbp, "DB->fd");
-
-	/*
-	 * XXX
-	 * Truly spectacular layering violation.
-	 */
-	if ((ret = __mp_xxx_fh(dbp->mpf, &fhp)) != 0)
-		return (ret);
-
-	if (F_ISSET(fhp, DB_FH_VALID)) {
-		*fdp = fhp->fd;
-		return (0);
-	} else {
-		*fdp = -1;
-		__db_err(dbp->dbenv, "DB does not have a valid file handle.");
-		return (ENOENT);
-	}
-}
-
-/*
- * __db_get --
- *	Return a key/data pair.
- *
- * PUBLIC: int __db_get __P((DB *, DB_TXN *, DBT *, DBT *, u_int32_t));
- */
-int
-__db_get(dbp, txn, key, data, flags)
-	DB *dbp;
-	DB_TXN *txn;
-	DBT *key, *data;
-	u_int32_t flags;
-{
-	DBC *dbc;
-	int mode, ret, t_ret;
-
-	PANIC_CHECK(dbp->dbenv);
-	DB_ILLEGAL_BEFORE_OPEN(dbp, "DB->get");
-
-	if ((ret = __db_getchk(dbp, key, data, flags)) != 0)
-		return (ret);
-
-	mode = 0;
-	if (LF_ISSET(DB_DIRTY_READ)) {
-		mode = DB_DIRTY_READ;
-		LF_CLR(DB_DIRTY_READ);
-	}
-	else if (flags == DB_CONSUME || flags == DB_CONSUME_WAIT)
-		mode = DB_WRITELOCK;
-	if ((ret = dbp->cursor(dbp, txn, &dbc, mode)) != 0)
-		return (ret);
-
-	DEBUG_LREAD(dbc, txn, "__db_get", key, NULL, flags);
-
-	/*
-	 * The DBC_TRANSIENT flag indicates that we're just doing a
-	 * single operation with this cursor, and that in case of
-	 * error we don't need to restore it to its old position--we're
-	 * going to close it right away.  Thus, we can perform the get
-	 * without duplicating the cursor, saving some cycles in this
-	 * common case.
-	 *
-	 * SET_RET_MEM indicates that if key and/or data have no DBT
-	 * flags set and DB manages the returned-data memory, that memory
-	 * will belong to this handle, not to the underlying cursor.
-	 */
-	F_SET(dbc, DBC_TRANSIENT);
-	SET_RET_MEM(dbc, dbp);
-
-	if (LF_ISSET(~(DB_RMW | DB_MULTIPLE)) == 0)
-		LF_SET(DB_SET);
-	ret = dbc->c_get(dbc, key, data, flags);
-
-	if ((t_ret = __db_c_close(dbc)) != 0 && ret == 0)
-		ret = t_ret;
-
+		__os_free(dbenv, dbc);
 	return (ret);
 }
 
@@ -500,20 +295,16 @@ __db_put(dbp, txn, key, data, flags)
 {
 	DBC *dbc;
 	DBT tdata;
+	DB_ENV *dbenv;
 	int ret, t_ret;
 
-	PANIC_CHECK(dbp->dbenv);
-	DB_ILLEGAL_BEFORE_OPEN(dbp, "DB->put");
+	dbenv = dbp->dbenv;
 
-	if ((ret = __db_putchk(dbp, key, data,
-	    flags, F_ISSET(dbp, DB_AM_RDONLY),
-	    F_ISSET(dbp, DB_AM_DUP) || F_ISSET(key, DB_DBT_DUPOK))) != 0)
+	if ((ret = __db_cursor(dbp, txn, &dbc, DB_WRITELOCK)) != 0)
 		return (ret);
 
-	DB_CHECK_TXN(dbp, txn);
+	DEBUG_LWRITE(dbc, txn, "DB->put", key, data, flags);
 
-	if ((ret = dbp->cursor(dbp, txn, &dbc, DB_WRITELOCK)) != 0)
-		return (ret);
 	SET_RET_MEM(dbc, dbp);
 
 	/*
@@ -528,8 +319,6 @@ __db_put(dbp, txn, key, data, flags)
 	 */
 	F_SET(dbc, DBC_TRANSIENT);
 
-	DEBUG_LWRITE(dbc, txn, "__db_put", key, data, flags);
-
 	switch (flags) {
 	case DB_APPEND:
 		/*
@@ -547,17 +336,22 @@ __db_put(dbp, txn, key, data, flags)
 		switch (dbp->type) {
 		case DB_QUEUE:
 			if ((ret = __qam_append(dbc, key, &tdata)) != 0)
-				goto done;
+				goto err;
 			break;
 		case DB_RECNO:
 			if ((ret = __ram_append(dbc, key, &tdata)) != 0)
-				goto done;
+				goto err;
 			break;
+		case DB_BTREE:
+		case DB_HASH:
+		case DB_UNKNOWN:
 		default:
 			/* The interface should prevent this. */
-			DB_ASSERT(0);
-			ret = __db_ferr(dbp->dbenv, "__db_put", flags);
-			goto done;
+			DB_ASSERT(
+			    dbp->type == DB_QUEUE || dbp->type == DB_RECNO);
+
+			ret = __db_ferr(dbenv, "DB->put", 0);
+			goto err;
 		}
 
 		/*
@@ -595,7 +389,7 @@ __db_put(dbp, txn, key, data, flags)
 		 * If we're doing page-level locking, set the read-modify-write
 		 * flag, we're going to overwrite immediately.
 		 */
-		if ((ret = dbc->c_get(dbc, key, &tdata,
+		if ((ret = __db_c_get(dbc, key, &tdata,
 		    DB_SET | (STD_LOCKING(dbc) ? DB_RMW : 0))) == 0)
 			ret = DB_KEYEXIST;
 		else if (ret == DB_NOTFOUND || ret == DB_KEYEMPTY)
@@ -605,49 +399,43 @@ __db_put(dbp, txn, key, data, flags)
 		/* Fall through to normal cursor put. */
 		break;
 	}
-	if (ret == 0)
-		ret = dbc->c_put(dbc,
-		     key, data, flags == 0 ? DB_KEYLAST : flags);
 
-done:	if ((t_ret = __db_c_close(dbc)) != 0 && ret == 0)
+	if (ret == 0)
+		ret = __db_c_put(dbc,
+		    key, data, flags == 0 ? DB_KEYLAST : flags);
+
+err:
+done:	/* Close the cursor. */
+	if ((t_ret = __db_c_close(dbc)) != 0 && ret == 0)
 		ret = t_ret;
 
 	return (ret);
 }
 
 /*
- * __db_delete --
+ * __db_del --
  *	Delete the items referenced by a key.
  *
- * PUBLIC: int __db_delete __P((DB *, DB_TXN *, DBT *, u_int32_t));
+ * PUBLIC: int __db_del __P((DB *, DB_TXN *, DBT *, u_int32_t));
  */
 int
-__db_delete(dbp, txn, key, flags)
+__db_del(dbp, txn, key, flags)
 	DB *dbp;
 	DB_TXN *txn;
 	DBT *key;
 	u_int32_t flags;
 {
 	DBC *dbc;
-	DBT lkey;
-	DBT data;
+	DBT data, lkey;
 	u_int32_t f_init, f_next;
 	int ret, t_ret;
 
-	PANIC_CHECK(dbp->dbenv);
-	DB_ILLEGAL_BEFORE_OPEN(dbp, "DB->del");
-	DB_CHECK_TXN(dbp, txn);
-
-	/* Check for invalid flags. */
-	if ((ret =
-	    __db_delchk(dbp, key, flags, F_ISSET(dbp, DB_AM_RDONLY))) != 0)
-		return (ret);
-
 	/* Allocate a cursor. */
-	if ((ret = dbp->cursor(dbp, txn, &dbc, DB_WRITELOCK)) != 0)
-		return (ret);
+	if ((ret = __db_cursor(dbp, txn, &dbc, DB_WRITELOCK)) != 0)
+		goto err;
 
-	DEBUG_LWRITE(dbc, txn, "db_delete", key, NULL, flags);
+	DEBUG_LWRITE(dbc, txn, "DB->del", key, NULL, flags);
+	COMPQUIET(flags, 0);
 
 	/*
 	 * Walk a cursor through the key/data pairs, deleting as we go.  Set
@@ -671,53 +459,73 @@ __db_delete(dbp, txn, key, flags)
 		f_next |= DB_RMW;
 	}
 
-	/* Walk through the set of key/data pairs, deleting as we go. */
-	if ((ret = dbc->c_get(dbc, key, &data, f_init)) != 0)
-		goto done;
-
 	/*
-	 * Hash permits an optimization in DB->del:  since on-page
-	 * duplicates are stored in a single HKEYDATA structure, it's
-	 * possible to delete an entire set of them at once, and as
-	 * the HKEYDATA has to be rebuilt and re-put each time it
-	 * changes, this is much faster than deleting the duplicates
-	 * one by one.  Thus, if we're not pointing at an off-page
-	 * duplicate set, and we're not using secondary indices (in
-	 * which case we'd have to examine the items one by one anyway),
-	 * let hash do this "quick delete".
+	 * Optimize the simple cases.  For all AMs if we don't have secondaries
+	 * and are not a secondary and there are no dups then we can avoid a
+	 * bunch of overhead.  For queue we don't need to fetch the record since
+	 * we delete by direct calculation from the record number.
+	 *
+	 * Hash permits an optimization in DB->del: since on-page duplicates are
+	 * stored in a single HKEYDATA structure, it's possible to delete an
+	 * entire set of them at once, and as the HKEYDATA has to be rebuilt
+	 * and re-put each time it changes, this is much faster than deleting
+	 * the duplicates one by one.  Thus, if not pointing at an off-page
+	 * duplicate set, and we're not using secondary indices (in which case
+	 * we'd have to examine the items one by one anyway), let hash do this
+	 * "quick delete".
 	 *
 	 * !!!
 	 * Note that this is the only application-executed delete call in
 	 * Berkeley DB that does not go through the __db_c_del function.
 	 * If anything other than the delete itself (like a secondary index
 	 * update) has to happen there in a particular situation, the
-	 * conditions here should be modified not to call __ham_quick_delete.
-	 * The ordinary AM-independent alternative will work just fine with
-	 * a hash;  it'll just be slower.
+	 * conditions here should be modified not to use these optimizations.
+	 * The ordinary AM-independent alternative will work just fine;
+	 * it'll just be slower.
 	 */
-	if (dbp->type == DB_HASH) {
-		if (LIST_FIRST(&dbp->s_secondaries) == NULL &&
-		    !F_ISSET(dbp, DB_AM_SECONDARY) &&
-		    dbc->internal->opd == NULL) {
+	if (!F_ISSET(dbp, DB_AM_SECONDARY) &&
+	    LIST_FIRST(&dbp->s_secondaries) == NULL) {
+
+#ifdef HAVE_QUEUE
+		if (dbp->type == DB_QUEUE) {
+			ret = __qam_delete(dbc, key);
+			goto done;
+		}
+#endif
+
+		/* Fetch the first record. */
+		if ((ret = __db_c_get(dbc, key, &data, f_init)) != 0)
+			goto err;
+
+#ifdef HAVE_HASH
+		if (dbp->type == DB_HASH && dbc->internal->opd == NULL) {
 			ret = __ham_quick_delete(dbc);
 			goto done;
 		}
-	}
+#endif
 
+		if ((dbp->type == DB_BTREE || dbp->type == DB_RECNO) &&
+		    !F_ISSET(dbp, DB_AM_DUP)) {
+			ret = dbc->c_am_del(dbc);
+			goto done;
+		}
+	} else if ((ret = __db_c_get(dbc, key, &data, f_init)) != 0)
+		goto err;
+
+	/* Walk through the set of key/data pairs, deleting as we go. */
 	for (;;) {
-		if ((ret = dbc->c_del(dbc, 0)) != 0)
-			goto done;
-		if ((ret = dbc->c_get(dbc, &lkey, &data, f_next)) != 0) {
-			if (ret == DB_NOTFOUND) {
+		if ((ret = __db_c_del(dbc, 0)) != 0)
+			break;
+		if ((ret = __db_c_get(dbc, &lkey, &data, f_next)) != 0) {
+			if (ret == DB_NOTFOUND)
 				ret = 0;
-				break;
-			}
-			goto done;
+			break;
 		}
 	}
 
-done:	/* Discard the cursor. */
-	if ((t_ret = dbc->c_close(dbc)) != 0 && ret == 0)
+done:
+err:	/* Discard the cursor. */
+	if ((t_ret = __db_c_close(dbc)) != 0 && ret == 0)
 		ret = t_ret;
 
 	return (ret);
@@ -727,22 +535,17 @@ done:	/* Discard the cursor. */
  * __db_sync --
  *	Flush the database cache.
  *
- * PUBLIC: int __db_sync __P((DB *, u_int32_t));
+ * PUBLIC: int __db_sync __P((DB *));
  */
 int
-__db_sync(dbp, flags)
+__db_sync(dbp)
 	DB *dbp;
-	u_int32_t flags;
 {
 	int ret, t_ret;
 
-	PANIC_CHECK(dbp->dbenv);
-	DB_ILLEGAL_BEFORE_OPEN(dbp, "DB->sync");
+	ret = 0;
 
-	if ((ret = __db_syncchk(dbp, flags)) != 0)
-		return (ret);
-
-	/* Read-only trees never need to be sync'd. */
+	/* If the database was read-only, we're done. */
 	if (F_ISSET(dbp, DB_AM_RDONLY))
 		return (0);
 
@@ -750,13 +553,17 @@ __db_sync(dbp, flags)
 	if (dbp->type == DB_RECNO)
 		ret = __ram_writeback(dbp);
 
-	/* If the tree was never backed by a database file, we're done. */
+	/* If the database was never backed by a database file, we're done. */
 	if (F_ISSET(dbp, DB_AM_INMEM))
-		return (0);
+		return (ret);
+
+	if (dbp->type == DB_QUEUE)
+		ret = __qam_sync(dbp);
+	else
+		/* Flush any dirty pages from the cache to the backing file. */
+		if ((t_ret = __memp_fsync(dbp->mpf)) != 0 && ret == 0)
+			ret = t_ret;
 
-	/* Flush any dirty pages from the cache to the backing file. */
-	if ((t_ret = memp_fsync(dbp->mpf)) != 0 && ret == 0)
-		ret = t_ret;
 	return (ret);
 }
 
@@ -764,34 +571,63 @@ __db_sync(dbp, flags)
  * __db_associate --
  *	Associate another database as a secondary index to this one.
  *
- * PUBLIC: int __db_associate __P((DB *, DB *, int (*)(DB *, const DBT *,
- * PUBLIC:     const DBT *, DBT *), u_int32_t));
+ * PUBLIC: int __db_associate __P((DB *, DB_TXN *, DB *,
+ * PUBLIC:     int (*)(DB *, const DBT *, const DBT *, DBT *), u_int32_t));
  */
 int
-__db_associate(dbp, sdbp, callback, flags)
+__db_associate(dbp, txn, sdbp, callback, flags)
 	DB *dbp, *sdbp;
-	int (*callback)(DB *, const DBT *, const DBT *, DBT *);
+	DB_TXN *txn;
+	int (*callback) __P((DB *, const DBT *, const DBT *, DBT *));
 	u_int32_t flags;
 {
 	DB_ENV *dbenv;
-	DB_TXN *txn;
 	DBC *pdbc, *sdbc;
 	DBT skey, key, data;
 	int build, ret, t_ret;
 
 	dbenv = dbp->dbenv;
-
-	if ((ret = __db_associatechk(dbp, sdbp, callback, flags)) != 0)
-		return (ret);
+	pdbc = sdbc = NULL;
+	ret = 0;
 
 	/*
-	 * The assumption we follow is that secondaries may be
-	 * associated while a primary is already in active use by
-	 * another thread, but that a secondary is newly created
-	 * and thus will not be in use anywhere else.  There's
-	 * no real way to avoid this assumption without mutex-protecting
-	 * every call, so we simply document this as a restriction and
-	 * let people coredump or get screwy results if they disobey.
+	 * Check to see if the secondary is empty -- and thus if we should
+	 * build it -- before we link it in and risk making it show up in other
+	 * threads.  Do this first so that the databases remain unassociated on
+	 * error.
+	 */
+	build = 0;
+	if (LF_ISSET(DB_CREATE)) {
+		if ((ret = __db_cursor(sdbp, txn, &sdbc, 0)) != 0)
+			goto err;
+
+		/*
+		 * We don't care about key or data;  we're just doing
+		 * an existence check.
+		 */
+		memset(&key, 0, sizeof(DBT));
+		memset(&data, 0, sizeof(DBT));
+		F_SET(&key, DB_DBT_PARTIAL | DB_DBT_USERMEM);
+		F_SET(&data, DB_DBT_PARTIAL | DB_DBT_USERMEM);
+		if ((ret = __db_c_get(sdbc, &key, &data,
+		    (STD_LOCKING(sdbc) ? DB_RMW : 0) |
+		    DB_FIRST)) == DB_NOTFOUND) {
+			build = 1;
+			ret = 0;
+		}
+
+		if ((t_ret = __db_c_close(sdbc)) != 0 && ret == 0)
+			ret = t_ret;
+
+		/* Reset for later error check. */
+		sdbc = NULL;
+
+		if (ret != 0)
+			goto err;
+	}
+
+	/*
+	 * Set up the database handle as a secondary.
 	 */
 	sdbp->s_callback = callback;
 	sdbp->s_primary = dbp;
@@ -800,182 +636,81 @@ __db_associate(dbp, sdbp, callback, flags)
 	sdbp->get = __db_secondary_get;
 
 	sdbp->stored_close = sdbp->close;
-	sdbp->close = __db_secondary_close;
-
-	/*
-	 * Secondary cursors may have the primary's lock file ID, so we
-	 * need to make sure that no older cursors are lying around
-	 * when we make the transition.
-	 */
-	if (TAILQ_FIRST(&sdbp->active_queue) != NULL ||
-	    TAILQ_FIRST(&sdbp->join_queue) != NULL) {
-		__db_err(dbenv,
-    "Databases may not become secondary indices while cursors are open");
-		return (EINVAL);
-	}
-	while ((sdbc = TAILQ_FIRST(&sdbp->free_queue)) != NULL)
-		if ((ret = __db_c_destroy(sdbc)) != 0)
-			return (ret);
+	sdbp->close = __db_secondary_close_pp;
 
 	F_SET(sdbp, DB_AM_SECONDARY);
 
-	/*
-	 * Check to see if the secondary is empty--and thus if we should
-	 * build it--before we link it in and risk making it show up in
-	 * other threads.
-	 */
-	build = 0;
-	if (LF_ISSET(DB_CREATE)) {
-		if ((ret = sdbp->cursor(sdbp, NULL, &sdbc, 0)) != 0)
-			return (ret);
-
-		memset(&key, 0, sizeof(DBT));
-		memset(&data, 0, sizeof(DBT));
-
-		/*
-		 * We don't care about key or data;  we're just doing
-		 * an existence check.
-		 */
-		F_SET(&key, DB_DBT_PARTIAL | DB_DBT_USERMEM);
-		F_SET(&data, DB_DBT_PARTIAL | DB_DBT_USERMEM);
-		if ((ret = sdbc->c_real_get(sdbc,
-		    &key, &data, DB_FIRST)) == DB_NOTFOUND) {
-			build = 1;
-			ret = 0;
-		}
-
-		/*
-		 * Secondary cursors have special refcounting close
-		 * methods.  Be careful.
-		 */
-		if ((t_ret = __db_c_close(sdbc)) != 0)
-			ret = t_ret;
-		if (ret != 0)
-			return (ret);
-	}
+	if (LF_ISSET(DB_IMMUTABLE_KEY))
+		FLD_SET(sdbp->s_assoc_flags, DB_ASSOC_IMMUTABLE_KEY);
 
 	/*
 	 * Add the secondary to the list on the primary.  Do it here
 	 * so that we see any updates that occur while we're walking
 	 * the primary.
 	 */
-	MUTEX_THREAD_LOCK(dbenv, dbp->mutexp);
+	MUTEX_LOCK(dbenv, dbp->mutex);
 
 	/* See __db_s_next for an explanation of secondary refcounting. */
 	DB_ASSERT(sdbp->s_refcnt == 0);
 	sdbp->s_refcnt = 1;
 	LIST_INSERT_HEAD(&dbp->s_secondaries, sdbp, s_links);
-	MUTEX_THREAD_UNLOCK(dbenv, dbp->mutexp);
+	MUTEX_UNLOCK(dbenv, dbp->mutex);
 
 	if (build) {
 		/*
 		 * We loop through the primary, putting each item we
 		 * find into the new secondary.
 		 *
-		 * We use separate transactions to avoid locking
-		 * down the whole secondary.
+		 * If we're using CDB, opening these two cursors puts us
+		 * in a bit of a locking tangle:  CDB locks are done on the
+		 * primary, so that we stay deadlock-free, but that means
+		 * that updating the secondary while we have a read cursor
+		 * open on the primary will self-block.  To get around this,
+		 * we force the primary cursor to use the same locker ID
+		 * as the secondary, so they won't conflict.  This should
+		 * be harmless even if we're not using CDB.
 		 */
+		if ((ret = __db_cursor(sdbp, txn, &sdbc,
+		    CDB_LOCKING(sdbp->dbenv) ? DB_WRITECURSOR : 0)) != 0)
+			goto err;
+		if ((ret = __db_cursor_int(dbp,
+		    txn, dbp->type, PGNO_INVALID, 0, sdbc->locker, &pdbc)) != 0)
+			goto err;
+
+		/* Lock out other threads, now that we have a locker ID. */
+		dbp->associate_lid = sdbc->locker;
+
 		memset(&key, 0, sizeof(DBT));
 		memset(&data, 0, sizeof(DBT));
-		txn = NULL;
-		if ((ret = dbp->cursor(dbp, NULL, &pdbc, 0)) != 0)
-			return (ret);
-retry:		while ((ret = pdbc->c_get(pdbc, &key, &data, DB_NEXT)) == 0) {
+		while ((ret = __db_c_get(pdbc, &key, &data, DB_NEXT)) == 0) {
 			memset(&skey, 0, sizeof(DBT));
 			if ((ret = callback(sdbp, &key, &data, &skey)) != 0) {
 				if (ret == DB_DONOTINDEX)
 					continue;
-				else
-					goto err;
-			}
-retry2:			if (TXN_ON(dbenv) &&
-			    (ret = txn_begin(dbenv, NULL, &txn, 0)) != 0)
-				goto err;
-			if ((ret = sdbp->cursor(sdbp, txn, &sdbc, 0)) != 0) {
-				FREE_IF_NEEDED(sdbp, &skey);
-				if (TXN_ON(dbenv))
-					(void)txn_abort(txn);
 				goto err;
 			}
-			if ((ret = sdbc->c_put(sdbc,
+			SWAP_IF_NEEDED(dbp, sdbp, &key);
+			if ((ret = __db_c_put(sdbc,
 			    &skey, &key, DB_UPDATE_SECONDARY)) != 0) {
-				if (TXN_ON(dbenv))
-					(void)txn_abort(txn);
-				if (ret == DB_LOCK_DEADLOCK)
-					goto retry2;
 				FREE_IF_NEEDED(sdbp, &skey);
 				goto err;
 			}
+			SWAP_IF_NEEDED(dbp, sdbp, &key);
+
 			FREE_IF_NEEDED(sdbp, &skey);
-
-			if ((ret = sdbc->c_close(sdbc)) != 0) {
-				if (TXN_ON(dbenv))
-					(void)txn_abort(txn);
-				goto err;
-			}
-
-			if (TXN_ON(dbenv) &&
-			    (ret = txn_commit(txn, DB_TXN_NOSYNC)) != 0)
-				goto err;
 		}
-		if (ret == DB_NOTFOUND) {
+		if (ret == DB_NOTFOUND)
 			ret = 0;
-			goto err;
-		} else if (ret == DB_LOCK_DEADLOCK)
-			goto retry;
-
-		/*
-		 * We've done all our txn_commits with NOSYNC, for performance.
-		 * Now flush them.
-		 */
-		ret = log_flush(dbenv, NULL);
-
-err:		if ((t_ret = pdbc->c_close(pdbc)) != 0 && ret == 0)
-			ret = t_ret;
 	}
 
-	return (ret);
-}
-
-/*
- * __db_pget --
- *	Return a primary key/data pair given a secondary key.
- *
- * PUBLIC: int __db_pget __P((DB *, DB_TXN *, DBT *, DBT *, DBT *, u_int32_t));
- */
-int
-__db_pget(dbp, txn, skey, pkey, data, flags)
-	DB *dbp;
-	DB_TXN *txn;
-	DBT *skey, *pkey, *data;
-	u_int32_t flags;
-{
-	DBC *dbc;
-	int ret, t_ret;
-
-	PANIC_CHECK(dbp->dbenv);
-	DB_ILLEGAL_BEFORE_OPEN(dbp, "DB->pget");
-
-	if ((ret = __db_pgetchk(dbp, skey, pkey, data, flags)) != 0)
-		return (ret);
-
-	if ((ret = dbp->cursor(dbp, txn, &dbc, 0)) != 0)
-		return (ret);
-	SET_RET_MEM(dbc, dbp);
-
-	DEBUG_LREAD(dbc, txn, "__db_pget", skey, NULL, flags);
-
-	/*
-	 * The cursor is just a perfectly ordinary secondary database
-	 * cursor.  Call its c_pget() method to do the dirty work.
-	 */
-	if (flags == 0 || flags == DB_RMW)
-		flags |= DB_SET;
-	if ((ret = dbc->c_pget(dbc, skey, pkey, data, flags)) != 0)
-		return (ret);
-
-	if ((t_ret = __db_c_close(dbc)) != 0 && ret == 0)
+err:	if (sdbc != NULL && (t_ret = __db_c_close(sdbc)) != 0 && ret == 0)
 		ret = t_ret;
+
+	if (pdbc != NULL && (t_ret = __db_c_close(pdbc)) != 0 && ret == 0)
+		ret = t_ret;
+
+	dbp->associate_lid = DB_LOCK_INVALIDID;
+
 	return (ret);
 }
 
@@ -993,7 +728,7 @@ __db_secondary_get(sdbp, txn, skey, data, flags)
 {
 
 	DB_ASSERT(F_ISSET(sdbp, DB_AM_SECONDARY));
-	return (sdbp->pget(sdbp, txn, skey, NULL, data, flags));
+	return (__db_pget_pp(sdbp, txn, skey, NULL, data, flags));
 }
 
 /*
@@ -1001,8 +736,10 @@ __db_secondary_get(sdbp, txn, skey, data, flags)
  *	Wrapper function for DB->close() which we use on secondaries to
  *	manage refcounting and make sure we don't close them underneath
  *	a primary that is updating.
+ *
+ * PUBLIC: int __db_secondary_close __P((DB *, u_int32_t));
  */
-static int
+int
 __db_secondary_close(sdbp, flags)
 	DB *sdbp;
 	u_int32_t flags;
@@ -1013,7 +750,7 @@ __db_secondary_close(sdbp, flags)
 	doclose = 0;
 	primary = sdbp->s_primary;
 
-	MUTEX_THREAD_LOCK(primary->dbenv, primary->mutexp);
+	MUTEX_LOCK(primary->dbenv, primary->mutex);
 	/*
 	 * Check the refcount--if it was at 1 when we were called, no
 	 * thread is currently updating this secondary through the primary,
@@ -1029,13 +766,13 @@ __db_secondary_close(sdbp, flags)
 		/* We don't want to call close while the mutex is held. */
 		doclose = 1;
 	}
-	MUTEX_THREAD_UNLOCK(primary->dbenv, primary->mutexp);
+	MUTEX_UNLOCK(primary->dbenv, primary->mutex);
 
 	/*
 	 * sdbp->close is this function;  call the real one explicitly if
 	 * need be.
 	 */
-	return (doclose ? __db_close(sdbp, flags) : 0);
+	return (doclose ? __db_close(sdbp, NULL, flags) : 0);
 }
 
 /*
@@ -1076,12 +813,12 @@ __db_append_primary(dbc, key, data)
 		 * correctly-constructed full data item from this partial
 		 * put is on the page waiting for us.
 		 */
-		if ((ret = __db_c_idup(dbc, &pdbc, DB_POSITIONI)) != 0)
+		if ((ret = __db_c_idup(dbc, &pdbc, DB_POSITION)) != 0)
 			return (ret);
 		memset(&pkey, 0, sizeof(DBT));
 		memset(&pdata, 0, sizeof(DBT));
 
-		if ((ret = pdbc->c_get(pdbc, &pkey, &pdata, DB_CURRENT)) != 0)
+		if ((ret = __db_c_get(pdbc, &pkey, &pdata, DB_CURRENT)) != 0)
 			goto err;
 
 		key = &pkey;
@@ -1097,17 +834,17 @@ __db_append_primary(dbc, key, data)
 	 * just that section into a common function, but the basic
 	 * overview is the same here.
 	 */
-	for (sdbp = __db_s_first(dbp); sdbp != NULL && ret == 0;
-	    ret = __db_s_next(&sdbp)) {
+	if ((ret = __db_s_first(dbp, &sdbp)) != 0)
+		goto err;
+	for (; sdbp != NULL && ret == 0; ret = __db_s_next(&sdbp)) {
 		memset(&skey, 0, sizeof(DBT));
 		if ((ret = sdbp->s_callback(sdbp, key, data, &skey)) != 0) {
 			if (ret == DB_DONOTINDEX)
 				continue;
-			else
-				goto err;
+			goto err;
 		}
 
-		if ((ret = __db_icursor(sdbp, dbc->txn, sdbp->type,
+		if ((ret = __db_cursor_int(sdbp, dbc->txn, sdbp->type,
 		    PGNO_INVALID, 0, dbc->locker, &sdbc)) != 0) {
 			FREE_IF_NEEDED(sdbp, &skey);
 			goto err;
@@ -1127,7 +864,7 @@ __db_append_primary(dbc, key, data)
 		if (!F_ISSET(sdbp, DB_AM_DUP)) {
 			memset(&oldpkey, 0, sizeof(DBT));
 			F_SET(&oldpkey, DB_DBT_MALLOC);
-			ret = sdbc->c_real_get(sdbc, &skey, &oldpkey,
+			ret = __db_c_get(sdbc, &skey, &oldpkey,
 			    DB_SET | (STD_LOCKING(dbc) ? DB_RMW : 0));
 			if (ret == 0) {
 				cmp = __bam_defcmp(sdbp, &oldpkey, key);
@@ -1136,12 +873,12 @@ __db_append_primary(dbc, key, data)
 				 * This needs to use the right free function
 				 * as soon as this is possible.
 				 */
-				__os_free(sdbp->dbenv,
-				    oldpkey.data, oldpkey.size);
+				__os_ufree(sdbp->dbenv,
+				    oldpkey.data);
 				if (cmp != 0) {
-					__db_err(sdbp->dbenv, "%s%s"
-			    "Append results in a non-unique secondary key in "
-			    "an index not configured to support duplicates");
+					__db_err(sdbp->dbenv, "%s%s",
+			    "Append results in a non-unique secondary key in",
+			    " an index not configured to support duplicates");
 					ret = EINVAL;
 					goto err1;
 				}
@@ -1149,15 +886,17 @@ __db_append_primary(dbc, key, data)
 				goto err1;
 		}
 
-		ret = sdbc->c_put(sdbc, &skey, key, DB_UPDATE_SECONDARY);
+		ret = __db_c_put(sdbc, &skey, key, DB_UPDATE_SECONDARY);
 
 err1:		FREE_IF_NEEDED(sdbp, &skey);
 
-		if ((t_ret = sdbc->c_close(sdbc)) != 0 && ret == 0)
+		if ((t_ret = __db_c_close(sdbc)) != 0 && ret == 0)
 			ret = t_ret;
+		if (ret != 0)
+			goto err;
 	}
 
-err:	if (pdbc != NULL && (t_ret = pdbc->c_close(pdbc)) != 0 && ret == 0)
+err:	if (pdbc != NULL && (t_ret = __db_c_close(pdbc)) != 0 && ret == 0)
 		ret = t_ret;
 	if (sdbp != NULL && (t_ret = __db_s_done(sdbp)) != 0 && ret == 0)
 		ret = t_ret;
