@@ -1,16 +1,14 @@
 /*
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1998-2003
+ * Copyright (c) 1998-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: db_join.c,v 12.6 2005/10/07 20:21:22 ubell Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: db_join.c,v 11.65 2003/10/07 18:55:39 mjc Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -68,7 +66,7 @@ static int __db_join_put __P((DBC *, DBT *, DBT *, u_int32_t));
  * cursor method of a DB, join cursors are created through an explicit
  * call to DB->join.
  *
- * The curslist is an array of existing, intialized cursors and primary
+ * The curslist is an array of existing, initialized cursors and primary
  * is the DB of the primary file.  The data item that joins all the
  * cursors in the curslist is used as the key into the primary and that
  * key and data are returned.  When no more items are left in the join
@@ -218,9 +216,9 @@ __db_join(primary, curslist, dbcp, flags)
 
 	*dbcp = dbc;
 
-	MUTEX_THREAD_LOCK(dbenv, primary->mutexp);
+	MUTEX_LOCK(dbenv, primary->mutex);
 	TAILQ_INSERT_TAIL(&primary->join_queue, dbc, links);
-	MUTEX_THREAD_UNLOCK(dbenv, primary->mutexp);
+	MUTEX_UNLOCK(dbenv, primary->mutex);
 
 	return (0);
 
@@ -229,7 +227,7 @@ err:	if (jc != NULL) {
 			__os_free(dbenv, jc->j_curslist);
 		if (jc->j_workcurs != NULL) {
 			if (jc->j_workcurs[0] != NULL)
-				__os_free(dbenv, jc->j_workcurs[0]);
+				(void)__db_c_close(jc->j_workcurs[0]);
 			__os_free(dbenv, jc->j_workcurs);
 		}
 		if (jc->j_fdupcurs != NULL)
@@ -252,24 +250,30 @@ __db_join_close_pp(dbc)
 	DBC *dbc;
 {
 	DB_ENV *dbenv;
+	DB_THREAD_INFO *ip;
 	DB *dbp;
-	int handle_check, ret;
+	int handle_check, ret, t_ret;
 
 	dbp = dbc->dbp;
 	dbenv = dbp->dbenv;
 
 	PANIC_CHECK(dbenv);
 
-	handle_check = IS_REPLICATED(dbenv, dbp);
+	ENV_ENTER(dbenv, ip);
+
+	handle_check = IS_ENV_REPLICATED(dbenv);
 	if (handle_check &&
-	    (ret = __db_rep_enter(dbp, 0, dbc->txn != NULL)) != 0)
-		return (ret);
+	    (ret = __db_rep_enter(dbp, 1, 0, dbc->txn != NULL)) != 0) {
+		handle_check = 0;
+		goto err;
+	}
 
 	ret = __db_join_close(dbc);
 
-	if (handle_check)
-		__env_rep_exit(dbenv);
+	if (handle_check && (t_ret = __env_db_rep_exit(dbenv)) != 0 && ret == 0)
+		ret = t_ret;
 
+err:	ENV_LEAVE(dbenv, ip);
 	return (ret);
 }
 
@@ -311,7 +315,9 @@ __db_join_get_pp(dbc, key, data, flags)
 {
 	DB *dbp;
 	DB_ENV *dbenv;
-	u_int32_t handle_check, ret, save_flags;
+	DB_THREAD_INFO *ip;
+	u_int32_t handle_check, save_flags;
+	int ret, t_ret;
 
 	dbp = dbc->dbp;
 	dbenv = dbp->dbenv;
@@ -321,11 +327,10 @@ __db_join_get_pp(dbc, key, data, flags)
 
 	PANIC_CHECK(dbenv);
 
-	if (LF_ISSET(DB_DIRTY_READ | DB_RMW)) {
+	if (LF_ISSET(DB_READ_COMMITTED | DB_READ_UNCOMMITTED | DB_RMW)) {
 		if (!LOCKING_ON(dbp->dbenv))
 			return (__db_fnl(dbp->dbenv, "DBcursor->c_get"));
-
-		LF_CLR(DB_DIRTY_READ | DB_RMW);
+		LF_CLR(DB_READ_COMMITTED | DB_READ_UNCOMMITTED | DB_RMW);
 	}
 
 	switch (flags) {
@@ -353,19 +358,24 @@ __db_join_get_pp(dbc, key, data, flags)
 		return (EINVAL);
 	}
 
-	handle_check = IS_REPLICATED(dbp->dbenv, dbp);
+	ENV_ENTER(dbenv, ip);
+
+	handle_check = IS_ENV_REPLICATED(dbp->dbenv);
 	if (handle_check &&
-	    (ret = __db_rep_enter(dbp, 1, dbc->txn != NULL)) != 0)
-		return (ret);
+	    (ret = __db_rep_enter(dbp, 1, 0, dbc->txn != NULL)) != 0) {
+		handle_check = 0;
+		goto err;
+	}
 
 	/* Restore the original flags value. */
 	flags = save_flags;
 
 	ret = __db_join_get(dbc, key, data, flags);
 
-	if (handle_check)
-		__db_rep_exit(dbenv);
+	if (handle_check && (t_ret = __env_db_rep_exit(dbenv)) != 0 && ret == 0)
+		ret = t_ret;
 
+err:	ENV_LEAVE(dbenv, ip);
 	return (ret);
 }
 
@@ -391,7 +401,7 @@ __db_join_get(dbc, key_arg, data_arg, flags)
 	 * If the set of flags here changes, check that __db_join_primget
 	 * is updated to handle them properly.
 	 */
-	opmods = LF_ISSET(DB_RMW | DB_DIRTY_READ);
+	opmods = LF_ISSET(DB_READ_COMMITTED | DB_READ_UNCOMMITTED | DB_RMW);
 
 	/*
 	 * Since we are fetching the key as a datum in the secondary indices,
@@ -423,7 +433,7 @@ __db_join_get(dbc, key_arg, data_arg, flags)
 retry:	ret = __db_c_get(jc->j_workcurs[0], &jc->j_key, key_n,
 	    opmods | (jc->j_exhausted[0] ? DB_NEXT_DUP : DB_CURRENT));
 
-	if (ret == ENOMEM) {
+	if (ret == DB_BUFFER_SMALL) {
 		jc->j_key.ulen <<= 1;
 		if ((ret = __os_realloc(dbp->dbenv,
 		    jc->j_key.ulen, &jc->j_key.data)) != 0)
@@ -473,7 +483,7 @@ retry:	ret = __db_c_get(jc->j_workcurs[0], &jc->j_key, key_n,
 		if (jc->j_workcurs[i] == NULL)
 			/* If this is NULL, we need to dup curslist into it. */
 			if ((ret = __db_c_dup(jc->j_curslist[i],
-			    jc->j_workcurs + i, DB_POSITION)) != 0)
+			    &jc->j_workcurs[i], DB_POSITION)) != 0)
 				goto err;
 
 retry2:		cp = jc->j_workcurs[i];
@@ -560,18 +570,17 @@ retry2:		cp = jc->j_workcurs[i];
 				    __db_c_close(jc->j_workcurs[j])) != 0)
 					goto err;
 				jc->j_exhausted[j] = 0;
-				if (jc->j_fdupcurs[j] != NULL &&
-				    (ret = __db_c_dup(jc->j_fdupcurs[j],
+				if (jc->j_fdupcurs[j] == NULL)
+					jc->j_workcurs[j] = NULL;
+				else if ((ret = __db_c_dup(jc->j_fdupcurs[j],
 				    &jc->j_workcurs[j], DB_POSITION)) != 0)
 					goto err;
-				else
-					jc->j_workcurs[j] = NULL;
 			}
 			goto retry2;
 			/* NOTREACHED */
 		}
 
-		if (ret == ENOMEM) {
+		if (ret == DB_BUFFER_SMALL) {
 			jc->j_key.ulen <<= 1;
 			if ((ret = __os_realloc(dbp->dbenv, jc->j_key.ulen,
 			    &jc->j_key.data)) != 0) {
@@ -611,7 +620,6 @@ mem_err:			__db_err(dbp->dbenv,
 		if (SORTED_SET(jc, i) && jc->j_fdupcurs[i] == NULL && (ret =
 		    __db_c_dup(cp, &jc->j_fdupcurs[i], DB_POSITION)) != 0)
 			goto err;
-
 	}
 
 err:	if (ret != 0)
@@ -729,9 +737,9 @@ __db_join_close(dbc)
 	 * must happen before any action that can fail and return, or else
 	 * __db_close may loop indefinitely.
 	 */
-	MUTEX_THREAD_LOCK(dbenv, dbp->mutexp);
+	MUTEX_LOCK(dbenv, dbp->mutex);
 	TAILQ_REMOVE(&dbp->join_queue, dbc, links);
-	MUTEX_THREAD_UNLOCK(dbenv, dbp->mutexp);
+	MUTEX_UNLOCK(dbenv, dbp->mutex);
 
 	PANIC_CHECK(dbenv);
 
@@ -875,26 +883,31 @@ __db_join_primget(dbp, txn, lockerid, key, data, flags)
 	u_int32_t flags;
 {
 	DBC *dbc;
-	int dirty, ret, rmw, t_ret;
-
-	/*
-	 * The only allowable flags here are the two flags copied into
-	 * "opmods" in __db_join_get, DB_RMW and DB_DIRTY_READ.  The former
-	 * is an op on the c_get call, the latter on the cursor call.
-	 * It's a DB bug if we allow any other flags down in here.
-	 */
-	rmw = LF_ISSET(DB_RMW);
-	dirty = LF_ISSET(DB_DIRTY_READ);
-	LF_CLR(DB_RMW | DB_DIRTY_READ);
-	DB_ASSERT(flags == 0);
+	u_int32_t rmw;
+	int ret, t_ret;
 
 	if ((ret = __db_cursor_int(dbp,
 	    txn, dbp->type, PGNO_INVALID, 0, lockerid, &dbc)) != 0)
 		return (ret);
 
-	if (dirty ||
-	    (txn != NULL && F_ISSET(txn, TXN_DIRTY_READ)))
-		F_SET(dbc, DBC_DIRTY_READ);
+	/*
+	 * The only allowable flags here are the two flags copied into "opmods"
+	 * in __db_join_get, DB_RMW and DB_READ_UNCOMMITTED.  The former is an
+	 * op on the c_get call, the latter on the cursor call.  It's a DB bug
+	 * if we allow any other flags down in here.
+	 */
+	rmw = LF_ISSET(DB_RMW);
+	if (LF_ISSET(DB_READ_UNCOMMITTED) ||
+	    (txn != NULL && F_ISSET(txn, TXN_READ_UNCOMMITTED)))
+		F_SET(dbc, DBC_READ_UNCOMMITTED);
+
+	if (LF_ISSET(DB_READ_COMMITTED) ||
+	    (txn != NULL && F_ISSET(txn, TXN_READ_COMMITTED)))
+		F_SET(dbc, DBC_READ_COMMITTED);
+
+	LF_CLR(DB_READ_COMMITTED | DB_READ_UNCOMMITTED | DB_RMW);
+	DB_ASSERT(flags == 0);
+
 	F_SET(dbc, DBC_TRANSIENT);
 
 	/*
