@@ -1,16 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1999-2003
+ * Copyright (c) 1999-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: db_method.c,v 12.15 2005/11/08 03:24:58 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: db_method.c,v 11.99 2003/07/08 20:14:17 ubell Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -21,6 +19,10 @@ static const char revid[] = "$Id: db_method.c,v 11.99 2003/07/08 20:14:17 ubell
 #include <string.h>
 #endif
 
+#ifdef HAVE_RPC
+#include "db_server.h"
+#endif
+
 #include "db_int.h"
 #include "dbinc/crypto.h"
 #include "dbinc/db_page.h"
@@ -30,21 +32,19 @@ static const char revid[] = "$Id: db_method.c,v 11.99 2003/07/08 20:14:17 ubell
 #include "dbinc/lock.h"
 #include "dbinc/mp.h"
 #include "dbinc/qam.h"
-#include "dbinc/xa.h"
-#include "dbinc_auto/xa_ext.h"
+#include "dbinc/txn.h"
 
 #ifdef HAVE_RPC
-#include "dbinc_auto/db_server.h"
 #include "dbinc_auto/rpc_client_ext.h"
 #endif
 
 static int  __db_get_byteswapped __P((DB *, int *));
 static int  __db_get_dbname __P((DB *, const char **, const char **));
-static int  __db_get_env __P((DB *, DB_ENV **));
-static int  __db_get_lorder __P((DB *, int *));
-static int  __db_get_transactional __P((DB *, int *));
+static DB_ENV *__db_get_env __P((DB *));
+static DB_MPOOLFILE *__db_get_mpf __P((DB *));
+static int  __db_get_transactional __P((DB *));
 static int  __db_get_type __P((DB *, DBTYPE *dbtype));
-static int  __db_init __P((DB *, u_int32_t));
+static int  __db_init __P((DB_ENV *, DB *, u_int32_t));
 static int  __db_set_alloc __P((DB *, void *(*)(size_t),
 		void *(*)(void *, size_t), void (*)(void *)));
 static int  __db_set_append_recno __P((DB *, int (*)(DB *, DBT *, db_recno_t)));
@@ -56,21 +56,21 @@ static int  __db_get_encrypt_flags __P((DB *, u_int32_t *));
 static int  __db_set_encrypt __P((DB *, const char *, u_int32_t));
 static int  __db_set_feedback __P((DB *, void (*)(DB *, int, int)));
 static void __db_map_flags __P((DB *, u_int32_t *, u_int32_t *));
-static int  __db_get_flags __P((DB *, u_int32_t *));
 static int  __db_get_pagesize __P((DB *, u_int32_t *));
 static int  __db_set_paniccall __P((DB *, void (*)(DB_ENV *, int)));
-static void __db_set_errcall __P((DB *, void (*)(const char *, char *)));
+static void __db_set_errcall
+	      __P((DB *, void (*)(const DB_ENV *, const char *, const char *)));
 static void __db_get_errfile __P((DB *, FILE **));
 static void __db_set_errfile __P((DB *, FILE *));
 static void __db_get_errpfx __P((DB *, const char **));
 static void __db_set_errpfx __P((DB *, const char *));
+static void __db_set_msgcall
+	      __P((DB *, void (*)(const DB_ENV *, const char *)));
+static void __db_get_msgfile __P((DB *, FILE **));
+static void __db_set_msgfile __P((DB *, FILE *));
 static void __dbh_err __P((DB *, int, const char *, ...));
 static void __dbh_errx __P((DB *, const char *, ...));
 
-#ifdef HAVE_RPC
-static int  __dbcl_init __P((DB *, DB_ENV *, u_int32_t));
-#endif
-
 /*
  * db_create --
  *	DB constructor.
@@ -84,14 +84,14 @@ db_create(dbpp, dbenv, flags)
 	u_int32_t flags;
 {
 	DB *dbp;
+	DB_THREAD_INFO *ip;
+	DB_REP *db_rep;
 	int ret;
 
 	/* Check for invalid function flags. */
 	switch (flags) {
 	case 0:
 		break;
-	case DB_REP_CREATE:
-		break;
 	case DB_XA_CREATE:
 		if (dbenv != NULL) {
 			__db_err(dbenv,
@@ -111,16 +111,17 @@ db_create(dbpp, dbenv, flags)
 		return (__db_ferr(dbenv, "db_create", 0));
 	}
 
+	ip = NULL;
+	if (dbenv != NULL)
+		ENV_ENTER(dbenv, ip);
 	/* Allocate the DB. */
-	if ((ret = __os_calloc(dbenv, 1, sizeof(*dbp), &dbp)) != 0)
+	if ((ret = __os_calloc(dbenv, 1, sizeof(*dbp), &dbp)) != 0) {
+		if (dbenv != NULL)
+			ENV_LEAVE(dbenv, ip);
 		return (ret);
-#ifdef HAVE_RPC
-	if (dbenv != NULL && RPC_ON(dbenv))
-		ret = __dbcl_init(dbp, dbenv, flags);
-	else
-#endif
-		ret = __db_init(dbp, flags);
-	if (ret != 0)
+	}
+
+	if ((ret = __db_init(dbenv, dbp, flags)) != 0)
 		goto err;
 
 	/* If we don't have an environment yet, allocate a local one. */
@@ -128,22 +129,33 @@ db_create(dbpp, dbenv, flags)
 		if ((ret = db_env_create(&dbenv, 0)) != 0)
 			goto err;
 		F_SET(dbenv, DB_ENV_DBLOCAL);
+		ENV_ENTER(dbenv, ip);
 	}
 	dbp->dbenv = dbenv;
-	MUTEX_THREAD_LOCK(dbenv, dbenv->dblist_mutexp);
+	MUTEX_LOCK(dbenv, dbenv->mtx_dblist);
 	++dbenv->db_ref;
-	MUTEX_THREAD_UNLOCK(dbenv, dbenv->dblist_mutexp);
+	MUTEX_UNLOCK(dbenv, dbenv->mtx_dblist);
 
 	/*
 	 * Set the replication timestamp; it's 0 if we're not in a replicated
-	 * environment.
+	 * environment.  Don't acquire a lock to read the value, even though
+	 * it's opaque: all we check later is value equality, nothing else.
 	 */
-	dbp->timestamp =
-	    (F_ISSET(dbenv, DB_ENV_DBLOCAL) || !REP_ON(dbenv)) ? 0 :
-	    ((DB_REP *)dbenv->rep_handle)->region->timestamp;
+	dbp->timestamp = REP_ON(dbenv) ?
+	    ((REGENV *)((REGINFO *)dbenv->reginfo)->primary)->rep_timestamp : 0;
+	/*
+	 * Set the replication generation number for fid management; valid
+	 * replication generations start at 1.  Don't acquire a lock to
+	 * read the value.  All we check later is value equality.
+	 */
+	db_rep = dbenv->rep_handle;
+	dbp->fid_gen =
+	    (REP_ON(dbenv) && db_rep->region != NULL) ?
+	    ((REP *)db_rep->region)->gen : 0;
 
-	/* Open a backing DB_MPOOLFILE handle in the memory pool. */
-	if ((ret = __memp_fcreate(dbenv, &dbp->mpf)) != 0)
+	/* If not RPC, open a backing DB_MPOOLFILE handle in the memory pool. */
+	if (!RPC_ON(dbenv) &&
+	    (ret = __memp_fcreate(dbenv, &dbp->mpf)) != 0)
 		goto err;
 
 	dbp->type = DB_UNKNOWN;
@@ -154,9 +166,10 @@ db_create(dbpp, dbenv, flags)
 err:	if (dbp->mpf != NULL)
 		(void)__memp_fclose(dbp->mpf, 0);
 	if (dbenv != NULL && F_ISSET(dbenv, DB_ENV_DBLOCAL))
-		(void)__dbenv_close(dbenv, 0);
+		(void)__env_close(dbenv, 0);
 	__os_free(dbenv, dbp);
 	*dbpp = NULL;
+	ENV_LEAVE(dbenv, ip);
 	return (ret);
 }
 
@@ -165,7 +178,8 @@ err:	if (dbp->mpf != NULL)
  *	Initialize a DB structure.
  */
 static int
-__db_init(dbp, flags)
+__db_init(dbenv, dbp, flags)
+	DB_ENV *dbenv;
 	DB *dbp;
 	u_int32_t flags;
 {
@@ -182,18 +196,30 @@ __db_init(dbp, flags)
 	FLD_SET(dbp->am_ok,
 	    DB_OK_BTREE | DB_OK_HASH | DB_OK_QUEUE | DB_OK_RECNO);
 
+	/* DB PUBLIC HANDLE LIST BEGIN */
 	dbp->associate = __db_associate_pp;
 	dbp->close = __db_close_pp;
+	dbp->compact = __db_compact_pp;
 	dbp->cursor = __db_cursor_pp;
 	dbp->del = __db_del_pp;
+	dbp->dump = __db_dump_pp;
 	dbp->err = __dbh_err;
 	dbp->errx = __dbh_errx;
 	dbp->fd = __db_fd_pp;
 	dbp->get = __db_get_pp;
 	dbp->get_byteswapped = __db_get_byteswapped;
+	dbp->get_cachesize = __db_get_cachesize;
 	dbp->get_dbname = __db_get_dbname;
+	dbp->get_encrypt_flags = __db_get_encrypt_flags;
 	dbp->get_env = __db_get_env;
+	dbp->get_errfile = __db_get_errfile;
+	dbp->get_errpfx = __db_get_errpfx;
+	dbp->get_flags = __db_get_flags;
+	dbp->get_lorder = __db_get_lorder;
+	dbp->get_mpf = __db_get_mpf;
+	dbp->get_msgfile = __db_get_msgfile;
 	dbp->get_open_flags = __db_get_open_flags;
+	dbp->get_pagesize = __db_get_pagesize;
 	dbp->get_transactional = __db_get_transactional;
 	dbp->get_type = __db_get_type;
 	dbp->join = __db_join_pp;
@@ -203,31 +229,28 @@ __db_init(dbp, flags)
 	dbp->put = __db_put_pp;
 	dbp->remove = __db_remove_pp;
 	dbp->rename = __db_rename_pp;
-	dbp->truncate = __db_truncate_pp;
 	dbp->set_alloc = __db_set_alloc;
 	dbp->set_append_recno = __db_set_append_recno;
-	dbp->get_cachesize = __db_get_cachesize;
 	dbp->set_cachesize = __db_set_cachesize;
 	dbp->set_dup_compare = __db_set_dup_compare;
-	dbp->get_encrypt_flags = __db_get_encrypt_flags;
 	dbp->set_encrypt = __db_set_encrypt;
 	dbp->set_errcall = __db_set_errcall;
-	dbp->get_errfile = __db_get_errfile;
 	dbp->set_errfile = __db_set_errfile;
-	dbp->get_errpfx = __db_get_errpfx;
 	dbp->set_errpfx = __db_set_errpfx;
 	dbp->set_feedback = __db_set_feedback;
-	dbp->get_flags = __db_get_flags;
 	dbp->set_flags = __db_set_flags;
-	dbp->get_lorder = __db_get_lorder;
 	dbp->set_lorder = __db_set_lorder;
-	dbp->get_pagesize = __db_get_pagesize;
+	dbp->set_msgcall = __db_set_msgcall;
+	dbp->set_msgfile = __db_set_msgfile;
 	dbp->set_pagesize = __db_set_pagesize;
 	dbp->set_paniccall = __db_set_paniccall;
 	dbp->stat = __db_stat_pp;
+	dbp->stat_print = __db_stat_print_pp;
 	dbp->sync = __db_sync_pp;
+	dbp->truncate = __db_truncate_pp;
 	dbp->upgrade = __db_upgrade_pp;
 	dbp->verify = __db_verify_pp;
+	/* DB PUBLIC HANDLE LIST END */
 
 					/* Access method specific. */
 	if ((ret = __bam_db_create(dbp)) != 0)
@@ -244,8 +267,25 @@ __db_init(dbp, flags)
 	if (LF_ISSET(DB_XA_CREATE) && (ret = __db_xa_create(dbp)) != 0)
 		return (ret);
 
-	if (LF_ISSET(DB_REP_CREATE))
-		F_SET(dbp, DB_AM_REPLICATION);
+#ifdef HAVE_RPC
+	/*
+	 * RPC specific: must be last, as we replace methods set by the
+	 * access methods.
+	 */
+	if (dbenv != NULL && RPC_ON(dbenv)) {
+		__dbcl_dbp_init(dbp);
+		/*
+		 * !!!
+		 * We wrap the DB->open method for RPC, and the rpc.src file
+		 * can't handle that.
+		 */
+		dbp->open = __dbcl_db_open_wrap;
+		if ((ret = __dbcl_db_create(dbp, dbenv, flags)) != 0)
+			return (ret);
+	}
+#else
+	COMPQUIET(dbenv, NULL);
+#endif
 
 	return (0);
 }
@@ -351,13 +391,22 @@ __db_get_dbname(dbp, fnamep, dnamep)
  * __db_get_env --
  *	Get the DB_ENV handle that was passed to db_create.
  */
-static int
-__db_get_env(dbp, dbenvp)
+static DB_ENV *
+__db_get_env(dbp)
 	DB *dbp;
-	DB_ENV **dbenvp;
 {
-	*dbenvp = dbp->dbenv;
-	return (0);
+	return (dbp->dbenv);
+}
+
+/*
+ * __db_get_mpf --
+ *	Get the underlying DB_MPOOLFILE handle.
+ */
+static DB_MPOOLFILE *
+__db_get_mpf(dbp)
+	DB *dbp;
+{
+	return (dbp->mpf);
 }
 
 /*
@@ -365,14 +414,10 @@ __db_get_env(dbp, dbenvp)
  *	Get whether this database was created in a transaction.
  */
 static int
-__db_get_transactional(dbp, istxnp)
+__db_get_transactional(dbp)
 	DB *dbp;
-	int *istxnp;
 {
-	DB_ILLEGAL_BEFORE_OPEN(dbp, "DB->get_transactional");
-
-	*istxnp = F_ISSET(dbp, DB_AM_TXN) ? 1 : 0;
-	return (0);
+	return (F_ISSET(dbp, DB_AM_TXN) ? 1 : 0);
 }
 
 /*
@@ -472,7 +517,7 @@ __db_get_encrypt_flags(dbp, flagsp)
 {
 	DB_ILLEGAL_IN_ENV(dbp, "DB->get_encrypt_flags");
 
-	return (__dbenv_get_encrypt_flags(dbp->dbenv, flagsp));
+	return (__env_get_encrypt_flags(dbp->dbenv, flagsp));
 }
 
 /*
@@ -491,7 +536,7 @@ __db_set_encrypt(dbp, passwd, flags)
 	DB_ILLEGAL_IN_ENV(dbp, "DB->set_encrypt");
 	DB_ILLEGAL_AFTER_OPEN(dbp, "DB->set_encrypt");
 
-	if ((ret = __dbenv_set_encrypt(dbp->dbenv, passwd, flags)) != 0)
+	if ((ret = __env_set_encrypt(dbp->dbenv, passwd, flags)) != 0)
 		return (ret);
 
 	/*
@@ -509,9 +554,9 @@ __db_set_encrypt(dbp, passwd, flags)
 static void
 __db_set_errcall(dbp, errcall)
 	DB *dbp;
-	void (*errcall) __P((const char *, char *));
+	void (*errcall) __P((const DB_ENV *, const char *, const char *));
 {
-	__dbenv_set_errcall(dbp->dbenv, errcall);
+	__env_set_errcall(dbp->dbenv, errcall);
 }
 
 static void
@@ -519,7 +564,7 @@ __db_get_errfile(dbp, errfilep)
 	DB *dbp;
 	FILE **errfilep;
 {
-	__dbenv_get_errfile(dbp->dbenv, errfilep);
+	__env_get_errfile(dbp->dbenv, errfilep);
 }
 
 static void
@@ -527,7 +572,7 @@ __db_set_errfile(dbp, errfile)
 	DB *dbp;
 	FILE *errfile;
 {
-	__dbenv_set_errfile(dbp->dbenv, errfile);
+	__env_set_errfile(dbp->dbenv, errfile);
 }
 
 static void
@@ -535,7 +580,7 @@ __db_get_errpfx(dbp, errpfxp)
 	DB *dbp;
 	const char **errpfxp;
 {
-	__dbenv_get_errpfx(dbp->dbenv, errpfxp);
+	__env_get_errpfx(dbp->dbenv, errpfxp);
 }
 
 static void
@@ -543,7 +588,7 @@ __db_set_errpfx(dbp, errpfx)
 	DB *dbp;
 	const char *errpfx;
 {
-	__dbenv_set_errpfx(dbp->dbenv, errpfx);
+	__env_set_errpfx(dbp->dbenv, errpfx);
 }
 
 static int
@@ -581,7 +626,13 @@ __db_map_flags(dbp, inflagsp, outflagsp)
 	}
 }
 
-static int
+/*
+ * __db_get_flags --
+ *	The DB->get_flags method.
+ *
+ * PUBLIC: int __db_get_flags __P((DB *, u_int32_t *));
+ */
+int
 __db_get_flags(dbp, flagsp)
 	DB *dbp;
 	u_int32_t *flagsp;
@@ -591,6 +642,7 @@ __db_get_flags(dbp, flagsp)
 		DB_DUP,
 		DB_DUPSORT,
 		DB_ENCRYPT,
+		DB_INORDER,
 		DB_RECNUM,
 		DB_RENUMBER,
 		DB_REVSPLITOFF,
@@ -607,6 +659,9 @@ __db_get_flags(dbp, flagsp)
 		__db_map_flags(dbp, &f, &mapped_flag);
 		__bam_map_flags(dbp, &f, &mapped_flag);
 		__ram_map_flags(dbp, &f, &mapped_flag);
+#ifdef HAVE_QUEUE
+		__qam_map_flags(dbp, &f, &mapped_flag);
+#endif
 		DB_ASSERT(f == 0);
 		if (F_ISSET(dbp, mapped_flag) == mapped_flag)
 			LF_SET(db_flags[i]);
@@ -647,6 +702,10 @@ __db_set_flags(dbp, flags)
 		return (ret);
 	if ((ret = __ram_set_flags(dbp, &flags)) != 0)
 		return (ret);
+#ifdef HAVE_QUEUE
+	if ((ret = __qam_set_flags(dbp, &flags)) != 0)
+		return (ret);
+#endif
 
 	return (flags == 0 ? 0 : __db_ferr(dbenv, "DB->set_flags", 0));
 }
@@ -654,8 +713,10 @@ __db_set_flags(dbp, flags)
 /*
  * __db_get_lorder --
  *	Get whether lorder is swapped or not.
+ *
+ * PUBLIC: int  __db_get_lorder __P((DB *, int *));
  */
-static int
+int
 __db_get_lorder(dbp, db_lorderp)
 	DB *dbp;
 	int *db_lorderp;
@@ -718,7 +779,31 @@ __db_set_alloc(dbp, mal_func, real_func, free_func)
 	DB_ILLEGAL_IN_ENV(dbp, "DB->set_alloc");
 	DB_ILLEGAL_AFTER_OPEN(dbp, "DB->set_alloc");
 
-	return (__dbenv_set_alloc(dbp->dbenv, mal_func, real_func, free_func));
+	return (__env_set_alloc(dbp->dbenv, mal_func, real_func, free_func));
+}
+
+static void
+__db_set_msgcall(dbp, msgcall)
+	DB *dbp;
+	void (*msgcall) __P((const DB_ENV *, const char *));
+{
+	__env_set_msgcall(dbp->dbenv, msgcall);
+}
+
+static void
+__db_get_msgfile(dbp, msgfilep)
+	DB *dbp;
+	FILE **msgfilep;
+{
+	__env_get_msgfile(dbp->dbenv, msgfilep);
+}
+
+static void
+__db_set_msgfile(dbp, msgfile)
+	DB *dbp;
+	FILE *msgfile;
+{
+	__env_set_msgfile(dbp->dbenv, msgfile);
 }
 
 static int
@@ -778,97 +863,5 @@ __db_set_paniccall(dbp, paniccall)
 	DB *dbp;
 	void (*paniccall) __P((DB_ENV *, int));
 {
-	return (__dbenv_set_paniccall(dbp->dbenv, paniccall));
+	return (__env_set_paniccall(dbp->dbenv, paniccall));
 }
-
-#ifdef HAVE_RPC
-/*
- * __dbcl_init --
- *	Initialize a DB structure on the server.
- */
-static int
-__dbcl_init(dbp, dbenv, flags)
-	DB *dbp;
-	DB_ENV *dbenv;
-	u_int32_t flags;
-{
-	TAILQ_INIT(&dbp->free_queue);
-	TAILQ_INIT(&dbp->active_queue);
-	/* !!!
-	 * Note that we don't need to initialize the join_queue;  it's
-	 * not used in RPC clients.  See the comment in __dbcl_db_join_ret().
-	 */
-
-	dbp->associate = __dbcl_db_associate;
-	dbp->close = __dbcl_db_close;
-	dbp->cursor = __dbcl_db_cursor;
-	dbp->del = __dbcl_db_del;
-	dbp->err = __dbh_err;
-	dbp->errx = __dbh_errx;
-	dbp->fd = __dbcl_db_fd;
-	dbp->get = __dbcl_db_get;
-	dbp->get_byteswapped = __db_get_byteswapped;
-	dbp->get_transactional = __db_get_transactional;
-	dbp->get_type = __db_get_type;
-	dbp->join = __dbcl_db_join;
-	dbp->key_range = __dbcl_db_key_range;
-	dbp->get_dbname = __dbcl_db_get_name;
-	dbp->get_open_flags = __dbcl_db_get_open_flags;
-	dbp->open = __dbcl_db_open_wrap;
-	dbp->pget = __dbcl_db_pget;
-	dbp->put = __dbcl_db_put;
-	dbp->remove = __dbcl_db_remove;
-	dbp->rename = __dbcl_db_rename;
-	dbp->set_alloc = __dbcl_db_alloc;
-	dbp->set_append_recno = __dbcl_db_set_append_recno;
-	dbp->get_cachesize = __dbcl_db_get_cachesize;
-	dbp->set_cachesize = __dbcl_db_cachesize;
-	dbp->set_dup_compare = __dbcl_db_dup_compare;
-	dbp->get_encrypt_flags = __dbcl_db_get_encrypt_flags;
-	dbp->set_encrypt = __dbcl_db_encrypt;
-	dbp->set_errcall = __db_set_errcall;
-	dbp->get_errfile = __db_get_errfile;
-	dbp->set_errfile = __db_set_errfile;
-	dbp->get_errpfx = __db_get_errpfx;
-	dbp->set_errpfx = __db_set_errpfx;
-	dbp->set_feedback = __dbcl_db_feedback;
-	dbp->get_flags = __dbcl_db_get_flags;
-	dbp->set_flags = __dbcl_db_flags;
-	dbp->get_lorder = __dbcl_db_get_lorder;
-	dbp->set_lorder = __dbcl_db_lorder;
-	dbp->get_pagesize = __dbcl_db_get_pagesize;
-	dbp->set_pagesize = __dbcl_db_pagesize;
-	dbp->set_paniccall = __dbcl_db_panic;
-	dbp->stat = __dbcl_db_stat;
-	dbp->sync = __dbcl_db_sync;
-	dbp->truncate = __dbcl_db_truncate;
-	dbp->upgrade = __dbcl_db_upgrade;
-	dbp->verify = __dbcl_db_verify;
-
-	/*
-	 * Set all the method specific functions to client funcs as well.
-	 */
-	dbp->set_bt_compare = __dbcl_db_bt_compare;
-	dbp->set_bt_maxkey = __dbcl_db_bt_maxkey;
-	dbp->get_bt_minkey = __dbcl_db_get_bt_minkey;
-	dbp->set_bt_minkey = __dbcl_db_bt_minkey;
-	dbp->set_bt_prefix = __dbcl_db_bt_prefix;
-	dbp->get_h_ffactor = __dbcl_db_get_h_ffactor;
-	dbp->set_h_ffactor = __dbcl_db_h_ffactor;
-	dbp->set_h_hash = __dbcl_db_h_hash;
-	dbp->get_h_nelem = __dbcl_db_get_h_nelem;
-	dbp->set_h_nelem = __dbcl_db_h_nelem;
-	dbp->get_q_extentsize = __dbcl_db_get_extentsize;
-	dbp->set_q_extentsize = __dbcl_db_extentsize;
-	dbp->get_re_delim = __dbcl_db_get_re_delim;
-	dbp->set_re_delim = __dbcl_db_re_delim;
-	dbp->get_re_len = __dbcl_db_get_re_len;
-	dbp->set_re_len = __dbcl_db_re_len;
-	dbp->get_re_pad = __dbcl_db_get_re_pad;
-	dbp->set_re_pad = __dbcl_db_re_pad;
-	dbp->get_re_source = __dbcl_db_get_re_source;
-	dbp->set_re_source = __dbcl_db_re_source;
-
-	return (__dbcl_db_create(dbp, dbenv, flags));
-}
-#endif
