@@ -1,17 +1,17 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: db_printlog.c,v 12.5 2005/09/09 12:38:33 bostic Exp $
  */
 
 #include "db_config.h"
 
 #ifndef lint
 static const char copyright[] =
-    "Copyright (c) 1996-2003\nSleepycat Software Inc.  All rights reserved.\n";
-static const char revid[] =
-    "$Id: db_printlog.c,v 11.59 2003/08/18 18:00:31 ubell Exp $";
+    "Copyright (c) 1996-2005\nSleepycat Software Inc.  All rights reserved.\n";
 #endif
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -33,11 +33,14 @@ static const char revid[] =
 #include "dbinc/qam.h"
 #include "dbinc/txn.h"
 
+int lsn_arg __P((char *, DB_LSN *));
 int main __P((int, char *[]));
-int usage __P((void));
-int version_check __P((const char *));
-int print_app_record __P((DB_ENV *, DBT *, DB_LSN *, db_recops));
 int open_rep_db __P((DB_ENV *, DB **, DBC **));
+int print_app_record __P((DB_ENV *, DBT *, DB_LSN *, db_recops));
+int usage __P((void));
+int version_check __P((void));
+
+const char *progname;
 
 int
 main(argc, argv)
@@ -46,31 +49,47 @@ main(argc, argv)
 {
 	extern char *optarg;
 	extern int optind;
-	const char *progname = "db_printlog";
 	DB *dbp;
 	DBC *dbc;
+	DBT data, keydbt;
 	DB_ENV	*dbenv;
 	DB_LOGC *logc;
-	int (**dtab) __P((DB_ENV *, DBT *, DB_LSN *, db_recops, void *));
+	DB_LSN key, start, stop;
 	size_t dtabsize;
-	DBT data, keydbt;
-	DB_LSN key;
-	int ch, exitval, nflag, rflag, ret, repflag;
+	u_int32_t logcflag;
+	int ch, cmp, exitval, nflag, rflag, ret, repflag;
+	int (**dtab) __P((DB_ENV *, DBT *, DB_LSN *, db_recops, void *));
 	char *home, *passwd;
 
-	if ((ret = version_check(progname)) != 0)
+	if ((progname = strrchr(argv[0], '/')) == NULL)
+		progname = argv[0];
+	else
+		++progname;
+
+	if ((ret = version_check()) != 0)
 		return (ret);
 
-	dbenv = NULL;
 	dbp = NULL;
 	dbc = NULL;
+	dbenv = NULL;
 	logc = NULL;
-	exitval = nflag = rflag = repflag = 0;
-	home = passwd = NULL;
+	ZERO_LSN(start);
+	ZERO_LSN(stop);
 	dtabsize = 0;
+	exitval = nflag = rflag = repflag = 0;
 	dtab = NULL;
-	while ((ch = getopt(argc, argv, "h:NP:rRV")) != EOF)
+	home = passwd = NULL;
+
+	while ((ch = getopt(argc, argv, "b:e:h:NP:rRV")) != EOF)
 		switch (ch) {
+		case 'b':
+			if (lsn_arg(optarg, &start))
+				return (usage());
+			break;
+		case 'e':
+			if (lsn_arg(optarg, &stop))
+				return (usage());
+			break;
 		case 'h':
 			home = optarg;
 			break;
@@ -89,7 +108,7 @@ main(argc, argv)
 		case 'r':
 			rflag = 1;
 			break;
-		case 'R':
+		case 'R':		/* Undocumented */
 			repflag = 1;
 			break;
 		case 'V':
@@ -157,28 +176,33 @@ main(argc, argv)
 	if (repflag) {
 		if ((ret = dbenv->open(dbenv, home,
 		    DB_INIT_MPOOL | DB_USE_ENVIRON, 0)) != 0 &&
+		    (ret == DB_VERSION_MISMATCH ||
 		    (ret = dbenv->open(dbenv, home,
 		    DB_CREATE | DB_INIT_MPOOL | DB_PRIVATE | DB_USE_ENVIRON, 0))
-		    != 0) {
-			dbenv->err(dbenv, ret, "open");
+		    != 0)) {
+			dbenv->err(dbenv, ret, "DB_ENV->open");
 			goto shutdown;
 		}
-	} else if ((ret = dbenv->open(dbenv, home,
-	    DB_JOINENV | DB_USE_ENVIRON, 0)) != 0 &&
+	} else if ((ret = dbenv->open(dbenv, home, DB_USE_ENVIRON, 0)) != 0 &&
+	    (ret == DB_VERSION_MISMATCH ||
 	    (ret = dbenv->open(dbenv, home,
-	    DB_CREATE | DB_INIT_LOG | DB_PRIVATE | DB_USE_ENVIRON, 0)) != 0) {
-		dbenv->err(dbenv, ret, "open");
+	    DB_CREATE | DB_INIT_LOG | DB_PRIVATE | DB_USE_ENVIRON, 0)) != 0)) {
+		dbenv->err(dbenv, ret, "DB_ENV->open");
 		goto shutdown;
 	}
 
 	/* Initialize print callbacks. */
 	if ((ret = __bam_init_print(dbenv, &dtab, &dtabsize)) != 0 ||
-	    (ret = __dbreg_init_print(dbenv, &dtab, &dtabsize)) != 0 ||
 	    (ret = __crdel_init_print(dbenv, &dtab, &dtabsize)) != 0 ||
 	    (ret = __db_init_print(dbenv, &dtab, &dtabsize)) != 0 ||
+	    (ret = __dbreg_init_print(dbenv, &dtab, &dtabsize)) != 0 ||
 	    (ret = __fop_init_print(dbenv, &dtab, &dtabsize)) != 0 ||
-	    (ret = __qam_init_print(dbenv, &dtab, &dtabsize)) != 0 ||
+#ifdef HAVE_HASH
 	    (ret = __ham_init_print(dbenv, &dtab, &dtabsize)) != 0 ||
+#endif
+#ifdef HAVE_QUEUE
+	    (ret = __qam_init_print(dbenv, &dtab, &dtabsize)) != 0 ||
+#endif
 	    (ret = __txn_init_print(dbenv, &dtab, &dtabsize)) != 0) {
 		dbenv->err(dbenv, ret, "callback: initialization");
 		goto shutdown;
@@ -193,17 +217,22 @@ main(argc, argv)
 		goto shutdown;
 	}
 
+	if (IS_ZERO_LSN(start)) {
+		memset(&keydbt, 0, sizeof(keydbt));
+		logcflag = rflag ? DB_PREV : DB_NEXT;
+	} else {
+		key = start;
+		logcflag = DB_SET;
+	}
 	memset(&data, 0, sizeof(data));
-	memset(&keydbt, 0, sizeof(keydbt));
-	while (!__db_util_interrupted()) {
+
+	for (; !__db_util_interrupted(); logcflag = rflag ? DB_PREV : DB_NEXT) {
 		if (repflag) {
-			ret = dbc->c_get(dbc,
-			    &keydbt, &data, rflag ? DB_PREV : DB_NEXT);
+			ret = dbc->c_get(dbc, &keydbt, &data, logcflag);
 			if (ret == 0)
 				key = ((REP_CONTROL *)keydbt.data)->lsn;
 		} else
-			ret = logc->get(logc,
-			    &key, &data, rflag ? DB_PREV : DB_NEXT);
+			ret = logc->get(logc, &key, &data, logcflag);
 		if (ret != 0) {
 			if (ret == DB_NOTFOUND)
 				break;
@@ -212,6 +241,15 @@ main(argc, argv)
 			goto shutdown;
 		}
 
+		/*
+		 * We may have reached the end of the range we're displaying.
+		 */
+		if (!IS_ZERO_LSN(stop)) {
+			cmp = log_compare(&key, &stop);
+			if ((rflag && cmp < 0) || (!rflag && cmp > 0))
+				break;
+		}
+
 		ret = __db_dispatch(dbenv,
 		    dtab, dtabsize, &data, &key, DB_TXN_PRINT, NULL);
 
@@ -264,14 +302,13 @@ shutdown:	exitval = 1;
 int
 usage()
 {
-	fprintf(stderr, "%s\n",
-	    "usage: db_printlog [-NrV] [-h home] [-P password]");
+	fprintf(stderr, "usage: %s %s\n", progname,
+	    "[-NrV] [-b file/offset] [-e file/offset] [-h home] [-P password]");
 	return (EXIT_FAILURE);
 }
 
 int
-version_check(progname)
-	const char *progname;
+version_check()
 {
 	int v_major, v_minor, v_patch;
 
@@ -361,3 +398,37 @@ err:	if (*dbpp != NULL)
 		(void)(*dbpp)->close(*dbpp, 0);
 	return (ret);
 }
+
+/*
+ * lsn_arg --
+ *	Parse a LSN argument.
+ */
+int
+lsn_arg(arg, lsnp)
+	char *arg;
+	DB_LSN *lsnp;
+{
+	char *p;
+	u_long uval;
+
+	/*
+	 * Expected format is: lsn.file/lsn.offset.
+	 *
+	 * Don't use getsubopt(3), some systems don't have it.
+	 */
+	if ((p = strchr(arg, '/')) == NULL)
+		return (1);
+	*p = '\0';
+
+	if (__db_getulong(NULL, progname, arg, 0, 0, &uval))
+		return (1);
+	if (uval > UINT32_MAX)
+		return (1);
+	lsnp->file = uval;
+	if (__db_getulong(NULL, progname, p + 1, 0, 0, &uval))
+		return (1);
+	if (uval > UINT32_MAX)
+		return (1);
+	lsnp->offset = uval;
+	return (0);
+}
