@@ -1,7 +1,7 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
  */
 /*
@@ -39,7 +39,7 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: btree.h,v 11.47 2003/05/18 16:54:52 bostic Exp $
+ * $Id: btree.h,v 12.8 2005/08/08 14:52:30 bostic Exp $
  */
 #ifndef	_DB_BTREE_H_
 #define	_DB_BTREE_H_
@@ -74,6 +74,7 @@ struct __recno;		typedef struct __recno RECNO;
 /* Flags for __bam_stkrel(). */
 #define	STK_CLRDBC	0x01		/* Clear dbc->page reference. */
 #define	STK_NOLOCK	0x02		/* Don't retain locks. */
+#define	STK_PGONLY	0x04
 
 /* Flags for __ram_ca(). These get logged, so make the values explicit. */
 typedef enum {
@@ -113,6 +114,11 @@ typedef enum {
 					 * return an entry one past end-of-page.
 					 */
 #define	S_STK_ONLY	0x04000		/* Just return info in the stack */
+#define	S_MAX		0x08000		/* Get the right most key */
+#define	S_MIN		0x10000		/* Get the left most key */
+#define	S_NEXT		0x20000		/* Get the page after this key */
+#define	S_DEL		0x40000		/* Get the tree to delete this key. */
+#define	S_START		0x80000		/* Level to start stack. */
 
 #define	S_DELETE	(S_WRITE | S_DUPFIRST | S_DELNO | S_EXACT | S_STACK)
 #define	S_FIND		(S_READ | S_DUPFIRST | S_DELNO)
@@ -153,7 +159,7 @@ struct __epg {
 	if ((ret = ((c)->csp == (c)->esp ?				\
 	    __bam_stkgrow(dbenv, c) : 0)) == 0) {			\
 		(c)->csp->page = pagep;					\
-		(c)->csp->indx = page_indx;				\
+		(c)->csp->indx = (page_indx);				\
 		(c)->csp->entries = NUM_ENT(pagep);			\
 		(c)->csp->lock = l;					\
 		(c)->csp->lock_mode = mode;				\
@@ -166,10 +172,10 @@ struct __epg {
 } while (0)
 
 #define	BT_STK_NUM(dbenv, c, pagep, page_indx, ret) do {		\
-	if ((ret =							\
-	    (c)->csp == (c)->esp ? __bam_stkgrow(dbenv, c) : 0) == 0) {	\
+	if ((ret = ((c)->csp ==						\
+	    (c)->esp ? __bam_stkgrow(dbenv, c) : 0)) == 0) {		\
 		(c)->csp->page = NULL;					\
-		(c)->csp->indx = page_indx;				\
+		(c)->csp->indx = (page_indx);				\
 		(c)->csp->entries = NUM_ENT(pagep);			\
 		LOCK_INIT((c)->csp->lock);				\
 		(c)->csp->lock_mode = DB_LOCK_NG;			\
@@ -237,7 +243,7 @@ struct __cursor {
  */
 #define	B_MINKEY_TO_OVFLSIZE(dbp, minkey, pgsize)			\
 	((u_int16_t)(((pgsize) - P_OVERHEAD(dbp)) / ((minkey) * P_INDX) -\
-	    (BKEYDATA_PSIZE(0) + ALIGN(1, sizeof(int32_t)))))
+	    (BKEYDATA_PSIZE(0) + DB_ALIGN(1, sizeof(int32_t)))))
 
 /*
  * The maximum space that a single item can ever take up on one page.
@@ -259,7 +265,6 @@ struct __btree {			/* Btree access method. */
 	db_pgno_t bt_meta;		/* Database meta-data page. */
 	db_pgno_t bt_root;		/* Database root page. */
 
-	u_int32_t bt_maxkey;		/* Maximum keys per page. */
 	u_int32_t bt_minkey;		/* Minimum keys per page. */
 
 					/* Btree comparison function. */
@@ -281,6 +286,7 @@ struct __btree {			/* Btree access method. */
 	 * of its information.
 	 */
 	db_pgno_t bt_lpgno;		/* Last insert location. */
+	DB_LSN	  bt_llsn;		/* Last insert LSN. */
 
 	/*
 	 * !!!
@@ -300,6 +306,7 @@ struct __btree {			/* Btree access method. */
 	FILE		*re_fp;		/* Source file handle. */
 	int		 re_eof;	/* Backing source file EOF reached. */
 	db_recno_t	 re_last;	/* Last record number read. */
+
 };
 
 /*
@@ -314,6 +321,12 @@ typedef enum {
 	DB_CA_SPLIT	= 4
 } db_ca_mode;
 
+/*
+ * Flags for __bam_pinsert.
+ */
+#define	BPI_SPACEONLY	0x01		/* Only check for space to update. */
+#define	BPI_NORECNUM	0x02		/* Not update the recnum on the left. */
+
 #include "dbinc_auto/btree_auto.h"
 #include "dbinc_auto/btree_ext.h"
 #include "dbinc/db_am.h"
