@@ -1,7 +1,7 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
  */
 /*
@@ -38,14 +38,12 @@
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
+ *
+ * $Id: dbm.c,v 12.2 2005/06/16 20:21:49 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: dbm.c,v 11.15 2003/01/08 04:34:46 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -76,14 +74,11 @@ static const char revid[] = "$Id: dbm.c,v 11.15 2003/01/08 04:34:46 bostic Exp $
  * EXTERN: int	 __db_ndbm_store __P((DBM *, datum, datum, int));
  *
  * EXTERN: int	 __db_dbm_close __P((void));
- * EXTERN: int	 __db_dbm_dbrdonly __P((void));
  * EXTERN: int	 __db_dbm_delete __P((datum));
- * EXTERN: int	 __db_dbm_dirf __P((void));
  * EXTERN: datum __db_dbm_fetch __P((datum));
  * EXTERN: datum __db_dbm_firstkey __P((void));
  * EXTERN: int	 __db_dbm_init __P((char *));
  * EXTERN: datum __db_dbm_nextkey __P((datum));
- * EXTERN: int	 __db_dbm_pagf __P((void));
  * EXTERN: int	 __db_dbm_store __P((datum, datum));
  *
  * EXTERN: #endif
@@ -101,9 +96,9 @@ __db_dbm_init(file)
 	char *file;
 {
 	if (__cur_db != NULL)
-		(void)dbm_close(__cur_db);
+		dbm_close(__cur_db);
 	if ((__cur_db =
-	    dbm_open(file, O_CREAT | O_RDWR, __db_omode("rw----"))) != NULL)
+	    dbm_open(file, O_CREAT | O_RDWR, __db_omode(OWNER_RW))) != NULL)
 		return (0);
 	if ((__cur_db = dbm_open(file, O_RDONLY, 0)) != NULL)
 		return (0);
@@ -128,7 +123,8 @@ __db_dbm_fetch(key)
 
 	if (__cur_db == NULL) {
 		__db_no_open();
-		item.dptr = 0;
+		item.dptr = NULL;
+		item.dsize = 0;
 		return (item);
 	}
 	return (dbm_fetch(__cur_db, key));
@@ -141,7 +137,8 @@ __db_dbm_firstkey()
 
 	if (__cur_db == NULL) {
 		__db_no_open();
-		item.dptr = 0;
+		item.dptr = NULL;
+		item.dsize = 0;
 		return (item);
 	}
 	return (dbm_firstkey(__cur_db));
@@ -157,7 +154,8 @@ __db_dbm_nextkey(key)
 
 	if (__cur_db == NULL) {
 		__db_no_open();
-		item.dptr = 0;
+		item.dptr = NULL;
+		item.dsize = 0;
 		return (item);
 	}
 	return (dbm_nextkey(__cur_db));
@@ -289,7 +287,7 @@ __db_ndbm_fetch(dbm, key)
 
 	memset(&_key, 0, sizeof(DBT));
 	memset(&_data, 0, sizeof(DBT));
-	_key.size = key.dsize;
+	_key.size = (u_int32_t)key.dsize;
 	_key.data = key.dptr;
 
 	/*
@@ -299,7 +297,7 @@ __db_ndbm_fetch(dbm, key)
 	 */
 	if ((ret = dbc->dbp->get(dbc->dbp, NULL, &_key, &_data, 0)) == 0) {
 		data.dptr = _data.data;
-		data.dsize = _data.size;
+		data.dsize = (int)_data.size;
 	} else {
 		data.dptr = NULL;
 		data.dsize = 0;
@@ -334,7 +332,7 @@ __db_ndbm_firstkey(dbm)
 
 	if ((ret = dbc->c_get(dbc, &_key, &_data, DB_FIRST)) == 0) {
 		key.dptr = _key.data;
-		key.dsize = _key.size;
+		key.dsize = (int)_key.size;
 	} else {
 		key.dptr = NULL;
 		key.dsize = 0;
@@ -369,7 +367,7 @@ __db_ndbm_nextkey(dbm)
 
 	if ((ret = dbc->c_get(dbc, &_key, &_data, DB_NEXT)) == 0) {
 		key.dptr = _key.data;
-		key.dsize = _key.size;
+		key.dsize = (int)_key.size;
 	} else {
 		key.dptr = NULL;
 		key.dsize = 0;
@@ -401,7 +399,7 @@ __db_ndbm_delete(dbm, key)
 
 	memset(&_key, 0, sizeof(DBT));
 	_key.data = key.dptr;
-	_key.size = key.dsize;
+	_key.size = (u_int32_t)key.dsize;
 
 	if ((ret = dbc->dbp->del(dbc->dbp, NULL, &_key, 0)) == 0)
 		return (0);
@@ -435,11 +433,11 @@ __db_ndbm_store(dbm, key, data, flags)
 
 	memset(&_key, 0, sizeof(DBT));
 	_key.data = key.dptr;
-	_key.size = key.dsize;
+	_key.size = (u_int32_t)key.dsize;
 
 	memset(&_data, 0, sizeof(DBT));
 	_data.data = data.dptr;
-	_data.size = data.dsize;
+	_data.size = (u_int32_t)data.dsize;
 
 	if ((ret = dbc->dbp->put(dbc->dbp, NULL,
 	    &_key, &_data, flags == DBM_INSERT ? DB_NOOVERWRITE : 0)) == 0)
