@@ -1,34 +1,40 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: env_region.c,v 12.13 2005/10/21 19:13:01 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: env_region.c,v 11.79 2003/10/31 01:56:10 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
+#if TIME_WITH_SYS_TIME
+#include <sys/time.h>
+#include <time.h>
+#else
+#if HAVE_SYS_TIME_H
+#include <sys/time.h>
+#else
+#include <time.h>
+#endif
+#endif
+
 #include <string.h>
 #endif
 
 #include "db_int.h"
 #include "dbinc/db_shash.h"
-#include "dbinc/lock.h"
-#include "dbinc/log.h"
+#include "dbinc/crypto.h"
 #include "dbinc/mp.h"
-#include "dbinc/txn.h"
 
-static int  __db_des_destroy __P((DB_ENV *, REGION *, int));
+static void __db_des_destroy __P((DB_ENV *, REGION *));
 static int  __db_des_get __P((DB_ENV *, REGINFO *, REGINFO *, REGION **));
 static int  __db_e_remfile __P((DB_ENV *));
 static int  __db_faultmem __P((DB_ENV *, void *, size_t, int));
-static void __db_region_destroy __P((DB_ENV *, REGINFO *));
 
 /*
  * __db_e_attach
@@ -47,36 +53,11 @@ __db_e_attach(dbenv, init_flagsp)
 	REGION *rp, tregion;
 	size_t size;
 	size_t nrw;
-	u_int32_t mbytes, bytes;
-	int retry_cnt, ret, segid;
+	u_int32_t bytes, i, mbytes, nregions;
+	u_int retry_cnt;
+	int majver, minver, patchver, ret, segid;
 	char buf[sizeof(DB_REGION_FMT) + 20];
 
-#if !defined(HAVE_MUTEX_THREADS)
-	/*
-	 * !!!
-	 * If we don't have spinlocks, we need a file descriptor for fcntl(2)
-	 * locking.  We use the file handle from the REGENV file for this
-	 * purpose.
-	 *
-	 * Since we may be using shared memory regions, e.g., shmget(2), and
-	 * not a mapped-in regular file, the backing file may be only a few
-	 * bytes in length.  So, this depends on the ability to call fcntl to
-	 * lock file offsets much larger than the actual physical file.  I
-	 * think that's safe -- besides, very few systems actually need this
-	 * kind of support, SunOS is the only one still in wide use of which
-	 * I'm aware.
-	 *
-	 * The error case is if an application lacks spinlocks and wants to be
-	 * threaded.  That doesn't work because fcntl may lock the underlying
-	 * process, including all its threads.
-	 */
-	if (F_ISSET(dbenv, DB_ENV_THREAD)) {
-		__db_err(dbenv,
-"architecture lacks fast mutexes: applications cannot be threaded");
-		return (EINVAL);
-	}
-#endif
-
 	/* Initialization */
 	retry_cnt = 0;
 
@@ -86,17 +67,17 @@ loop:	renv = NULL;
 	/* Set up the DB_ENV's REG_INFO structure. */
 	if ((ret = __os_calloc(dbenv, 1, sizeof(REGINFO), &infop)) != 0)
 		return (ret);
+	infop->dbenv = dbenv;
 	infop->type = REGION_TYPE_ENV;
 	infop->id = REGION_ID_ENV;
-	infop->mode = dbenv->db_mode;
 	infop->flags = REGION_JOIN_OK;
 	if (F_ISSET(dbenv, DB_ENV_CREATE))
 		F_SET(infop, REGION_CREATE_OK);
 
 	/*
 	 * We have to single-thread the creation of the REGENV region.  Once
-	 * it exists, we can do locking using locks in the region, but until
-	 * then we have to be the only player in the game.
+	 * it exists, we can serialize using region mutexes, but until then
+	 * we have to be the only player in the game.
 	 *
 	 * If this is a private environment, we are only called once and there
 	 * are no possible race conditions.
@@ -127,7 +108,7 @@ loop:	renv = NULL;
 	 */
 	if (F_ISSET(dbenv, DB_ENV_CREATE)) {
 		if ((ret = __os_open(dbenv, infop->name,
-		    DB_OSO_CREATE | DB_OSO_DIRECT | DB_OSO_EXCL | DB_OSO_REGION,
+		    DB_OSO_CREATE | DB_OSO_EXCL | DB_OSO_REGION,
 		    dbenv->db_mode, &dbenv->lockfhp)) == 0)
 			goto creation;
 		if (ret != EEXIST) {
@@ -141,8 +122,8 @@ loop:	renv = NULL;
 	 * If we couldn't create the file, try and open it.  (If that fails,
 	 * we're done.)
 	 */
-	if ((ret = __os_open(dbenv, infop->name, DB_OSO_REGION | DB_OSO_DIRECT,
-	    dbenv->db_mode, &dbenv->lockfhp)) != 0)
+	if ((ret = __os_open(
+	    dbenv, infop->name, DB_OSO_REGION, 0, &dbenv->lockfhp)) != 0)
 		goto err;
 
 	/* The region exists, it's not okay to recreate it. */
@@ -193,7 +174,8 @@ loop:	renv = NULL;
 	/*
 	 * If the size is less than the size of a REGENV_REF structure, the
 	 * region (or, possibly, the REGENV_REF structure) has not yet been
-	 * completely written.  Wait awhile and try again.
+	 * completely written.  Shouldn't be possible, but there's no reason
+	 * not to wait awhile and try again.
 	 *
 	 * Otherwise, if the size is the size of a REGENV_REF structure,
 	 * read it into memory and use it as a reference to the real region.
@@ -224,13 +206,12 @@ loop:	renv = NULL;
 	} else
 		segid = INVALID_REGION_SEGID;
 
+#ifndef HAVE_MUTEX_FCNTL
 	/*
-	 * If not doing thread locking, we need to save the file handle for
-	 * fcntl(2) locking.  Otherwise, discard the handle, we no longer
-	 * need it, and the less contact between the buffer cache and the VM,
-	 * the better.
+	 * If we're not doing fcntl locking, we can close the file handle.  We
+	 * no longer need it and the less contact between the buffer cache and
+	 * the VM, the better.
 	 */
-#ifdef HAVE_MUTEX_THREADS
 	 (void)__os_closehandle(dbenv, dbenv->lockfhp);
 	 dbenv->lockfhp = NULL;
 #endif
@@ -249,20 +230,29 @@ loop:	renv = NULL;
 	 * this means that all of our offsets (R_ADDR/R_OFFSET) get shifted
 	 * as well, but that should be fine.
 	 */
-	infop->primary = R_ADDR(infop, 0);
+	infop->primary = infop->addr;
 	infop->addr = (u_int8_t *)infop->addr + sizeof(REGENV);
 	renv = infop->primary;
 
-	/* Make sure the region matches our build. */
+	/*
+	 * Make sure the region matches our build.  Special case a region
+	 * that's all nul bytes, just treat it like any other corruption.
+	 *
+	 * !!!
+	 * We don't display the major/minor version from the environment,
+	 * because it may be in a different place in the two regions.
+	 */
 	if (renv->majver != DB_VERSION_MAJOR ||
 	    renv->minver != DB_VERSION_MINOR) {
-		__db_err(dbenv,
-	"Program version %d.%d doesn't match environment version",
-		    DB_VERSION_MAJOR, DB_VERSION_MINOR);
-#ifndef DIAGNOSTIC
-		ret = EINVAL;
+		if (renv->majver != 0 || renv->minver != 0) {
+			__db_err(dbenv,
+	"Program version %d.%d doesn't match environment version %d.%d",
+			    DB_VERSION_MAJOR, DB_VERSION_MINOR,
+			    renv->majver, renv->minver);
+			ret = DB_VERSION_MISMATCH;
+		} else
+			ret = EINVAL;
 		goto err;
-#endif
 	}
 
 	/*
@@ -279,33 +269,19 @@ loop:	renv = NULL;
 	 * I'd rather play permissions games using the underlying file, but I
 	 * can't because Windows/NT filesystems won't open files mode 0.
 	 */
-	if (renv->envpanic && !F_ISSET(dbenv, DB_ENV_NOPANIC)) {
+	if (renv->panic && !F_ISSET(dbenv, DB_ENV_NOPANIC)) {
 		ret = __db_panic_msg(dbenv);
 		goto err;
 	}
 	if (renv->magic != DB_REGION_MAGIC)
 		goto retry;
 
-	/* Lock the environment. */
-	MUTEX_LOCK(dbenv, &renv->mutex);
-
-	/*
-	 * Finally!  We own the environment now.  Repeat the panic check, it's
-	 * possible that it was set while we waited for the lock.
-	 */
-	if (renv->envpanic && !F_ISSET(dbenv, DB_ENV_NOPANIC)) {
-		ret = __db_panic_msg(dbenv);
-		goto err_unlock;
-	}
-
 	/*
 	 * Get a reference to the underlying REGION information for this
 	 * environment.
 	 */
-	if ((ret = __db_des_get(dbenv, infop, infop, &rp)) != 0 || rp == NULL) {
-		MUTEX_UNLOCK(dbenv, &renv->mutex);
+	if ((ret = __db_des_get(dbenv, infop, infop, &rp)) != 0 || rp == NULL)
 		goto find_err;
-	}
 	infop->rp = rp;
 
 	/*
@@ -314,25 +290,32 @@ loop:	renv = NULL;
 	 * growing as part of its creation.  We can detect this by checking the
 	 * size we originally found against the region's current size.  (The
 	 * region's current size has to be final, the creator finished growing
-	 * it before releasing the environment for us to lock.)
+	 * it before setting the magic number in the region.)
 	 */
-	if (rp->size != size) {
-err_unlock:	MUTEX_UNLOCK(dbenv, &renv->mutex);
+	if (rp->size != size)
 		goto retry;
-	}
 
 	/* Increment the reference count. */
+	MUTEX_LOCK(dbenv, renv->mtx_regenv);
 	++renv->refcnt;
+	MUTEX_UNLOCK(dbenv, renv->mtx_regenv);
 
 	/*
-	 * If our caller wants them, return the flags this environment was
-	 * initialized with.
+	 * Check our callers configuration flags, it's an error to configure
+	 * incompatible or additional subsystems in an existing environment.
+	 * Return the total set of flags to the caller so they initialize the
+	 * correct set of subsystems.
 	 */
-	if (init_flagsp != NULL)
+	if (init_flagsp != NULL) {
+		FLD_CLR(*init_flagsp, renv->init_flags);
+		if (*init_flagsp != 0) {
+			__db_err(dbenv,
+    "configured environment flags incompatible with existing environment");
+			ret = EINVAL;
+			goto err;
+		}
 		*init_flagsp = renv->init_flags;
-
-	/* Discard our lock. */
-	MUTEX_UNLOCK(dbenv, &renv->mutex);
+	}
 
 	/*
 	 * Fault the pages into memory.  Note, do this AFTER releasing the
@@ -349,15 +332,17 @@ creation:
 	F_SET(infop, REGION_CREATE);
 
 	/*
-	 * Allocate room for 50 REGION structures plus overhead (we're going
-	 * to use this space for last-ditch allocation requests), although we
-	 * should never need anything close to that.
+	 * Allocate room for REGION structures plus overhead.
 	 *
-	 * Encryption passwds are stored in the env region.  Add that in too.
+	 * XXX
+	 * Overhead is so high because encryption passwds are stored in the
+	 * base environment region, as are replication vote arrays.  This is
+	 * a bug, not a feature, replication needs its own region.
 	 */
 	memset(&tregion, 0, sizeof(tregion));
-	tregion.size = (roff_t)(50 * sizeof(REGION) +
-	    dbenv->passwd_len + 2048);
+	nregions = dbenv->mp_ncache + 10;
+	tregion.size =
+	   (roff_t)(nregions * sizeof(REGION) + dbenv->passwd_len + 16 * 1024);
 	tregion.segid = INVALID_REGION_SEGID;
 	if ((ret = __os_r_attach(dbenv, infop, &tregion)) != 0)
 		goto err;
@@ -386,21 +371,30 @@ creation:
 	 * region detach, and that all of our offsets (R_ADDR/R_OFFSET) will be
 	 * shifted as well, but that should be fine.
 	 */
-	infop->primary = R_ADDR(infop, 0);
+	infop->primary = infop->addr;
 	infop->addr = (u_int8_t *)infop->addr + sizeof(REGENV);
-	__db_shalloc_init(infop->addr, tregion.size - sizeof(REGENV));
+	__db_shalloc_init(infop, tregion.size - sizeof(REGENV));
 
 	/*
-	 * Initialize the rest of the REGENV structure, except for the magic
-	 * number which validates the file/environment.
+	 * Initialize the rest of the REGENV structure.  (Don't set the magic
+	 * number to the correct value, that would validate the environment).
 	 */
 	renv = infop->primary;
-	renv->envpanic = 0;
-	db_version(&renv->majver, &renv->minver, &renv->patch);
-	SH_LIST_INIT(&renv->regionq);
+	renv->magic = 0;
+	renv->panic = 0;
+
+	(void)db_version(&majver, &minver, &patchver);
+	renv->majver = (u_int32_t)majver;
+	renv->minver = (u_int32_t)minver;
+	renv->patchver = (u_int32_t)patchver;
+
+	(void)time(&renv->timestamp);
+	__os_unique_id(dbenv, &renv->envid);
+
+	if ((ret = __mutex_alloc(
+	    dbenv, MTX_ENV_REGION, 0, &renv->mtx_regenv)) != 0)
+		goto err;
 	renv->refcnt = 1;
-	renv->cipher_off = INVALID_ROFF;
-	renv->rep_off = INVALID_ROFF;
 
 	/*
 	 * Initialize init_flags to store the flags that any other environment
@@ -409,27 +403,27 @@ creation:
 	renv->init_flags = (init_flagsp == NULL) ? 0 : *init_flagsp;
 
 	/*
-	 * Lock the environment.
-	 *
-	 * Check the lock call return.  This is the first lock we initialize
-	 * and acquire, and we have to know if it fails.  (It CAN fail, e.g.,
-	 * SunOS, when using fcntl(2) for locking and using an in-memory
-	 * filesystem as the database home.  But you knew that, I'm sure -- it
-	 * probably wasn't even worth mentioning.)
+	 * Set up the region array.  We use an array rather than a linked list
+	 * as we have to traverse this list after failure in some cases, and
+	 * we don't want to infinitely loop should the application fail while
+	 * we're manipulating the list.
 	 */
-	if ((ret = __db_mutex_setup(dbenv, infop, &renv->mutex,
-	    MUTEX_NO_RECORD | MUTEX_NO_RLOCK)) != 0) {
-		__db_err(dbenv, "%s: unable to initialize environment lock: %s",
-		    infop->name, db_strerror(ret));
+	renv->region_cnt = nregions;
+	if ((ret =
+	    __db_shalloc(infop, nregions * sizeof(REGION), 0, &rp)) != 0) {
+		__db_err(dbenv, "unable to create new master region array: %s",
+		    db_strerror(ret));
 		goto err;
 	}
+	renv->region_off = R_OFFSET(infop, rp);
+	for (i = 0; i < nregions; ++i, ++rp)
+		rp->id = INVALID_REGION_ID;
 
-	if (!F_ISSET(&renv->mutex, MUTEX_IGNORE) &&
-	    (ret = __db_mutex_lock(dbenv, &renv->mutex)) != 0) {
-		__db_err(dbenv, "%s: unable to acquire environment lock: %s",
-		    infop->name, db_strerror(ret));
-		goto err;
-	}
+	renv->cipher_off = INVALID_ROFF;
+
+	renv->rep_off = INVALID_ROFF;
+	renv->flags = 0;
+	renv->op_timestamp = renv->rep_timestamp = 0;
 
 	/*
 	 * Get the underlying REGION structure for this environment.  Note,
@@ -438,8 +432,7 @@ creation:
 	 * the REGION structure.
 	 */
 	if ((ret = __db_des_get(dbenv, infop, infop, &rp)) != 0) {
-find_err:	__db_err(dbenv,
-		    "%s: unable to find environment", infop->name);
+find_err:	__db_err(dbenv, "%s: unable to find environment", infop->name);
 		if (ret == 0)
 			ret = EINVAL;
 		goto err;
@@ -473,25 +466,18 @@ find_err:	__db_err(dbenv,
 		}
 	}
 
+#ifndef HAVE_MUTEX_FCNTL
 	/*
-	 * If not doing thread locking, we need to save the file handle for
-	 * fcntl(2) locking.  Otherwise, discard the handle, we no longer
-	 * need it, and the less contact between the buffer cache and the VM,
-	 * the better.
+	 * If we're not doing fcntl locking, we can close the file handle.  We
+	 * no longer need it and the less contact between the buffer cache and
+	 * the VM, the better.
 	 */
-#if defined(HAVE_MUTEX_THREADS)
 	if (dbenv->lockfhp != NULL) {
 		 (void)__os_closehandle(dbenv, dbenv->lockfhp);
 		 dbenv->lockfhp = NULL;
 	}
 #endif
 
-	/* Validate the file. */
-	renv->magic = DB_REGION_MAGIC;
-
-	/* Discard our lock. */
-	MUTEX_UNLOCK(dbenv, &renv->mutex);
-
 	/* Everything looks good, we're done. */
 	dbenv->reginfo = infop;
 	return (0);
@@ -531,7 +517,7 @@ retry:	/* Close any open file handle. */
 			__db_err(dbenv, "unable to join the environment");
 			ret = EAGAIN;
 		} else {
-			(void)__os_sleep(dbenv, retry_cnt * 3, 0);
+			__os_sleep(dbenv, retry_cnt * 3, 0);
 			goto loop;
 		}
 	}
@@ -539,6 +525,35 @@ retry:	/* Close any open file handle. */
 	return (ret);
 }
 
+/*
+ * __db_e_golive --
+ *	Turn on the created environment.
+ *
+ * PUBLIC: int __db_e_golive __P((DB_ENV *));
+ */
+int
+__db_e_golive(dbenv)
+	DB_ENV *dbenv;
+{
+	REGENV *renv;
+	REGINFO *infop;
+
+	infop = dbenv->reginfo;
+	renv = infop->primary;
+
+	/* If we didn't create the region, there's no need for further work. */
+	if (!F_ISSET(infop, REGION_CREATE))
+		return (0);
+
+	/*
+	 * Validate the file.  All other threads of control are waiting
+	 * on this value to be written -- "Let slip the hounds of war!"
+	 */
+	renv->magic = DB_REGION_MAGIC;
+
+	return (0);
+}
+
 /*
  * __db_e_detach --
  *	Detach from the environment.
@@ -552,66 +567,97 @@ __db_e_detach(dbenv, destroy)
 {
 	REGENV *renv;
 	REGINFO *infop;
+	REGION rp;
+	int ret, t_ret;
 
 	infop = dbenv->reginfo;
 	renv = infop->primary;
+	ret = 0;
 
 	if (F_ISSET(dbenv, DB_ENV_PRIVATE))
 		destroy = 1;
-	/* Lock the environment. */
-	MUTEX_LOCK(dbenv, &renv->mutex);
 
 	/* Decrement the reference count. */
-	if (renv->refcnt == 0) {
-		__db_err(dbenv,
-		    "region %lu (environment): reference count went negative",
-		    (u_long)infop->rp->id);
-	} else
+	MUTEX_LOCK(dbenv, renv->mtx_regenv);
+	if (renv->refcnt == 0)
+		__db_err(dbenv, "environment reference count went negative");
+	else
 		--renv->refcnt;
-
-	/* Release the lock. */
-	MUTEX_UNLOCK(dbenv, &renv->mutex);
+	MUTEX_UNLOCK(dbenv, renv->mtx_regenv);
 
 	/* Close the locking file handle. */
 	if (dbenv->lockfhp != NULL) {
-		(void)__os_closehandle(dbenv, dbenv->lockfhp);
+		if ((t_ret =
+		    __os_closehandle(dbenv, dbenv->lockfhp)) != 0 && ret == 0)
+			ret = t_ret;
 		dbenv->lockfhp = NULL;
 	}
 
-	/* Reset the addr value that we "corrected" above. */
-	infop->addr = infop->primary;
-
-	/*
-	 * If we are destroying the environment, we need to
-	 * destroy any system resources backing the mutex, as well
-	 * as any system resources that the replication system may have
-	 * acquired and put in the main region.
-	 *
-	 * Do these now before we free the memory in __os_r_detach.
-	 */
-	if (destroy) {
-		(void)__rep_region_destroy(dbenv);
-		__db_mutex_destroy(&renv->mutex);
-		__db_mutex_destroy(&infop->rp->mutex);
-	}
-
 	/*
 	 * Release the region, and kill our reference.
-	 *
-	 * We set the DB_ENV->reginfo field to NULL here and discard its memory.
-	 * DB_ENV->remove calls __dbenv_remove to do the region remove, and
-	 * __dbenv_remove attached and then detaches from the region.  We don't
-	 * want to return to DB_ENV->remove with a non-NULL DB_ENV->reginfo
-	 * field because it will attempt to detach again as part of its cleanup.
 	 */
-	(void)__os_r_detach(dbenv, infop, destroy);
+	if (destroy) {
+#ifdef HAVE_CRYPTO
+		/*
+		 * Destroy any system resources the crypto subsystem may have
+		 * acquired.
+		 */
+		if ((t_ret = __crypto_region_destroy(dbenv)) != 0 && ret == 0)
+			ret = t_ret;
+#endif
+		/*
+		 * Destroy any system resources the replication subsystem may
+		 * have acquired.
+		 */
+		if ((t_ret = __rep_region_destroy(dbenv)) != 0 && ret == 0)
+			ret = t_ret;
 
-	if (infop->name != NULL)
-		__os_free(dbenv, infop->name);
-	__os_free(dbenv, dbenv->reginfo);
+		/*
+		 * Free the REGION array.
+		 *
+		 * The actual underlying region structure is allocated from the
+		 * primary shared region, and we're about to free it.  Save a
+		 * copy on our stack for the REGINFO to reference when it calls
+		 * down into the OS layer to release the shared memory segment.
+		 */
+		rp = *infop->rp;
+		infop->rp = &rp;
+
+		if (renv->region_off != INVALID_ROFF)
+			__db_shalloc_free(
+			   infop, R_ADDR(infop, renv->region_off));
+
+		/* Discard any mutex resources we may have acquired. */
+		if ((t_ret =
+		    __mutex_free(dbenv, &renv->mtx_regenv)) != 0 && ret == 0)
+			ret = t_ret;
+	}
+
+	/*
+	 * Set the DB_ENV->reginfo field to NULL.  First, DB_ENV->remove calls
+	 * __env_remove to do the region remove, and __envremove attached and
+	 * then detaches from the region.  We don't want to return to
+	 * DB_ENV->remove with a non-NULL DB_ENV->reginfo field because it will
+	 * attempt to detach again as part of its cleanup.
+	 *
+	 * Second, DB code uses DB_ENV->reginfo to decide if it's OK to read
+	 * the underlying region.  We're about to destroy what it references,
+	 * so it needs to be cleared.
+	 */
 	dbenv->reginfo = NULL;
 
-	return (0);
+	/* Reset the addr value that we "corrected" above. */
+	infop->addr = infop->primary;
+
+	if ((t_ret = __os_r_detach(dbenv, infop, destroy)) != 0 && ret == 0)
+		ret = t_ret;
+	if (infop->name != NULL)
+		__os_free(dbenv, infop->name);
+
+	/* Discard the DB_ENV->reginfo field's memory. */
+	__os_free(dbenv, infop);
+
+	return (ret);
 }
 
 /*
@@ -628,33 +674,39 @@ __db_e_remove(dbenv, flags)
 	REGENV *renv;
 	REGINFO *infop, reginfo;
 	REGION *rp;
-	u_int32_t db_env_reset;
-	int force, ret;
+	u_int32_t db_env_reset, i;
+	int ret;
+
+	db_env_reset = F_ISSET(dbenv, DB_ENV_NOLOCKING | DB_ENV_NOPANIC);
 
-	force = LF_ISSET(DB_FORCE) ? 1 : 0;
 	/*
 	 * This routine has to walk a nasty line between not looking into
 	 * the environment (which may be corrupted after an app or system
 	 * crash), and removing everything that needs removing.  What we
 	 * do is:
-	 *	1. Connect to the environment (so it better be OK).
+	 *	1. Connect to the environment.
 	 *	2. If the environment is in use (reference count is non-zero),
 	 *	   return EBUSY.
-	 *	3. Overwrite the magic number so that any threads of control
-	 *	   attempting to connect will backoff and retry.
-	 *	4. Walk the list of regions.  Connect to each region and then
+	 *	3. Panic it and overwrite the magic number so any threads of
+	 *	   control attempting to connect (or racing with us) backoff
+	 *	   and retry or just die.
+	 *	4. Walk the array of regions.  Connect to each region and then
 	 *	   disconnect with the destroy flag set.  This shouldn't cause
 	 *	   any problems, even if the region is corrupted, because we
-	 *	   should never be looking inside the region.
+	 *	   never look inside the region (with the single exception of
+	 *	   mutex regions on systems where we have to return resources
+	 *	   to the underlying system).
 	 *	5. Walk the list of files in the directory, unlinking any
 	 *	   files that match a region name.  Unlink the environment
 	 *	   file last.
 	 *
 	 * If the force flag is set, we do not acquire any locks during this
 	 * process.
+	 *
+	 * We're going to panic the environment, so we'll want to ignore that
+	 * flag.
 	 */
-	db_env_reset = F_ISSET(dbenv, DB_ENV_NOLOCKING | DB_ENV_NOPANIC);
-	if (force)
+	if (LF_ISSET(DB_FORCE))
 		F_SET(dbenv, DB_ENV_NOLOCKING);
 	F_SET(dbenv, DB_ENV_NOPANIC);
 
@@ -666,7 +718,7 @@ __db_e_remove(dbenv, flags)
 		 * probably isn't important.
 		 */
 		ret = 0;
-		if (force)
+		if (LF_ISSET(DB_FORCE))
 			goto remfiles;
 		goto done;
 	}
@@ -675,14 +727,14 @@ __db_e_remove(dbenv, flags)
 	renv = infop->primary;
 
 	/* Lock the environment. */
-	MUTEX_LOCK(dbenv, &renv->mutex);
+	MUTEX_LOCK(dbenv, renv->mtx_regenv);
 
 	/*
 	 * If it's in use, we're done unless we're forcing the issue or the
 	 * environment has panic'd.  (Presumably, if the environment panic'd,
 	 * the thread holding the reference count may not have cleaned up.)
 	 */
-	if (renv->refcnt == 1 || renv->envpanic == 1 || force) {
+	if (renv->refcnt == 1 || renv->panic == 1 || LF_ISSET(DB_FORCE)) {
 		/*
 		 * Set the panic flag and overwrite the magic number.
 		 *
@@ -690,42 +742,50 @@ __db_e_remove(dbenv, flags)
 		 * From this point on, there's no going back, we pretty
 		 * much ignore errors, and just whack on whatever we can.
 		 */
-		renv->envpanic = 1;
 		renv->magic = 0;
+		renv->panic = 1;
 
 		/*
-		 * Unlock the environment.  We should no longer need the lock
-		 * because we've poisoned the pool, but we can't continue to
-		 * hold it either, because other routines may want it.
+		 * Unlock the environment -- nobody should need this lock
+		 * because we've poisoned the pool.
 		 */
-		MUTEX_UNLOCK(dbenv, &renv->mutex);
+		MUTEX_UNLOCK(dbenv, renv->mtx_regenv);
 
-		/*
-		 * Attach to each sub-region and destroy it.
-		 *
-		 * !!!
-		 * The REGION_CREATE_OK flag is set for Windows/95 -- regions
-		 * are zero'd out when the last reference to the region goes
-		 * away, in which case the underlying OS region code requires
-		 * callers be prepared to create the region in order to join it.
-		 */
-		memset(&reginfo, 0, sizeof(reginfo));
-		for (rp = SH_LIST_FIRST(&renv->regionq, __db_region);
-		    rp != NULL; rp = SH_LIST_NEXT(rp, q, __db_region)) {
-			if (rp->type == REGION_TYPE_ENV)
+		/* Attach to each sub-region and destroy it. */
+		for (rp = R_ADDR(infop, renv->region_off),
+		    i = 0; i < renv->region_cnt; ++i, ++rp) {
+			if (rp->id == INVALID_REGION_ID ||
+			    rp->type == REGION_TYPE_ENV)
 				continue;
+			/*
+			 * !!!
+			 * The REGION_CREATE_OK flag is set for Windows/95 --
+			 * regions are zero'd out when the last reference to
+			 * the region goes away, in which case the underlying
+			 * OS region code requires callers be prepared to
+			 * create the region in order to join it.
+			 */
+			memset(&reginfo, 0, sizeof(reginfo));
+			reginfo.id = rp->id;
+			reginfo.flags = REGION_CREATE_OK;
 
 			/*
 			 * If we get here and can't attach and/or detach to the
 			 * region, it's a mess.  Ignore errors, there's nothing
 			 * we can do about them.
 			 */
-			reginfo.id = rp->id;
-			reginfo.flags = REGION_CREATE_OK;
-			if (__db_r_attach(dbenv, &reginfo, 0) == 0) {
-				R_UNLOCK(dbenv, &reginfo);
-				(void)__db_r_detach(dbenv, &reginfo, 1);
-			}
+			if (__db_r_attach(dbenv, &reginfo, 0) != 0)
+				continue;
+
+#ifdef  HAVE_MUTEX_SYSTEM_RESOURCES
+			/*
+			 * If destroying the mutex region, return any system
+			 * resources to the system.
+			 */
+			if (reginfo.type == REGION_TYPE_MUTEX)
+				__mutex_resource_return(dbenv, &reginfo);
+#endif
+			(void)__db_r_detach(dbenv, &reginfo, 1);
 		}
 
 		/* Destroy the environment's region. */
@@ -735,7 +795,7 @@ __db_e_remove(dbenv, flags)
 remfiles:	(void)__db_e_remfile(dbenv);
 	} else {
 		/* Unlock the environment. */
-		MUTEX_UNLOCK(dbenv, &renv->mutex);
+		MUTEX_UNLOCK(dbenv, renv->mtx_regenv);
 
 		/* Discard the environment. */
 		(void)__db_e_detach(dbenv, 0);
@@ -758,9 +818,8 @@ __db_e_remfile(dbenv)
 	DB_ENV *dbenv;
 {
 	int cnt, fcnt, lastrm, ret;
-	u_int8_t saved_byte;
 	const char *dir;
-	char *p, **names, *path, buf[sizeof(DB_REGION_FMT) + 20];
+	char saved_char, *p, **names, *path, buf[sizeof(DB_REGION_FMT) + 20];
 
 	/* Get the full path of a file in the environment. */
 	(void)snprintf(buf, sizeof(buf), "%s", DB_REGION_ENV);
@@ -770,11 +829,11 @@ __db_e_remfile(dbenv)
 	/* Get the parent directory for the environment. */
 	if ((p = __db_rpath(path)) == NULL) {
 		p = path;
-		saved_byte = *p;
+		saved_char = *p;
 
 		dir = PATH_DOT;
 	} else {
-		saved_byte = *p;
+		saved_char = *p;
 		*p = '\0';
 
 		dir = path;
@@ -785,7 +844,7 @@ __db_e_remfile(dbenv)
 		__db_err(dbenv, "%s: %s", dir, db_strerror(ret));
 
 	/* Restore the path, and free it. */
-	*p = saved_byte;
+	*p = saved_char;
 	__os_free(dbenv, path);
 
 	if (ret != 0)
@@ -804,6 +863,14 @@ __db_e_remfile(dbenv)
 		if (strncmp(names[cnt], "__dbq.", 6) == 0)
 			continue;
 
+		/* Skip registry files. */
+		if (strncmp(names[cnt], "__db.register", 13) == 0)
+			continue;
+
+		/* Skip replication files. */
+		if (strncmp(names[cnt], "__db.rep.", 9) == 0)
+			continue;
+
 		/*
 		 * Remove the primary environment region last, because it's
 		 * the key to this whole mess.
@@ -827,7 +894,7 @@ __db_e_remfile(dbenv)
 			 */
 			if (F_ISSET(dbenv, DB_ENV_OVERWRITE) &&
 			    strlen(names[cnt]) == DB_REGION_NAME_LENGTH)
-				(void)__db_overwrite(dbenv, path);
+				(void)__db_file_multi_write(dbenv, path);
 			(void)__os_unlink(dbenv, path);
 			__os_free(dbenv, path);
 		}
@@ -837,7 +904,7 @@ __db_e_remfile(dbenv)
 		if (__db_appname(dbenv,
 		    DB_APP_NONE, names[lastrm], 0, NULL, &path) == 0) {
 			if (F_ISSET(dbenv, DB_ENV_OVERWRITE))
-				(void)__db_overwrite(dbenv, path);
+				(void)__db_file_multi_write(dbenv, path);
 			(void)__os_unlink(dbenv, path);
 			__os_free(dbenv, path);
 		}
@@ -846,61 +913,6 @@ __db_e_remfile(dbenv)
 	return (0);
 }
 
-/*
- * __db_e_stat
- *	Statistics for the environment.
- *
- * PUBLIC: int __db_e_stat __P((DB_ENV *,
- * PUBLIC:       REGENV *, REGION *, int *, u_int32_t));
- */
-int
-__db_e_stat(dbenv, arg_renv, arg_regions, arg_regions_cnt, flags)
-	DB_ENV *dbenv;
-	REGENV *arg_renv;
-	REGION *arg_regions;
-	int *arg_regions_cnt;
-	u_int32_t flags;
-{
-	REGENV *renv;
-	REGINFO *infop;
-	REGION *rp;
-	int n, ret;
-
-	infop = dbenv->reginfo;
-	renv = infop->primary;
-	rp = infop->rp;
-	if ((ret = __db_fchk(dbenv,
-	    "DB_ENV->stat", flags, DB_STAT_CLEAR)) != 0)
-		return (ret);
-
-	/* Lock the environment. */
-	MUTEX_LOCK(dbenv, &rp->mutex);
-
-	*arg_renv = *renv;
-	if (LF_ISSET(DB_STAT_CLEAR)) {
-		renv->mutex.mutex_set_nowait = 0;
-		renv->mutex.mutex_set_wait = 0;
-	}
-
-	for (n = 0, rp = SH_LIST_FIRST(&renv->regionq, __db_region);
-	    n < *arg_regions_cnt && rp != NULL;
-	    ++n, rp = SH_LIST_NEXT(rp, q, __db_region)) {
-		arg_regions[n] = *rp;
-		if (LF_ISSET(DB_STAT_CLEAR)) {
-			rp->mutex.mutex_set_nowait = 0;
-			rp->mutex.mutex_set_wait = 0;
-		}
-	}
-
-	/* Release the lock. */
-	rp = infop->rp;
-	MUTEX_UNLOCK(dbenv, &rp->mutex);
-
-	*arg_regions_cnt = n == 0 ? n : n - 1;
-
-	return (0);
-}
-
 /*
  * __db_r_attach
  *	Join/create a region.
@@ -913,30 +925,26 @@ __db_r_attach(dbenv, infop, size)
 	REGINFO *infop;
 	size_t size;
 {
-	REGENV *renv;
 	REGION *rp;
 	int ret;
 	char buf[sizeof(DB_REGION_FMT) + 20];
 
-	renv = ((REGINFO *)dbenv->reginfo)->primary;
-
-	/* Lock the environment. */
-	MUTEX_LOCK(dbenv, &renv->mutex);
-
 	/*
 	 * Find or create a REGION structure for this region.  If we create
 	 * it, the REGION_CREATE flag will be set in the infop structure.
 	 */
 	F_CLR(infop, REGION_CREATE);
-	if ((ret = __db_des_get(dbenv, dbenv->reginfo, infop, &rp)) != 0) {
-		MUTEX_UNLOCK(dbenv, &renv->mutex);
+	if ((ret = __db_des_get(dbenv, dbenv->reginfo, infop, &rp)) != 0)
 		return (ret);
-	}
+	infop->dbenv = dbenv;
 	infop->rp = rp;
 	infop->type = rp->type;
 	infop->id = rp->id;
 
-	/* If we're creating the region, set the desired size. */
+	/*
+	 * __db_des_get may have created the region and reset the create
+	 * flag.  If we're creating the region, set the desired size.
+	 */
 	if (F_ISSET(infop, REGION_CREATE))
 		rp->size = (roff_t)size;
 
@@ -965,21 +973,12 @@ __db_r_attach(dbenv, infop, size)
 	 * If we created the region, initialize it for allocation.
 	 */
 	if (F_ISSET(infop, REGION_CREATE))
-		(void)__db_shalloc_init(infop->addr, rp->size);
-
-	/*
-	 * If the underlying REGION isn't the environment, acquire a lock
-	 * for it and release our lock on the environment.
-	 */
-	if (infop->type != REGION_TYPE_ENV) {
-		MUTEX_LOCK(dbenv, &rp->mutex);
-		MUTEX_UNLOCK(dbenv, &renv->mutex);
-	}
+		__db_shalloc_init(infop, rp->size);
 
 	return (0);
 
-	/* Discard the underlying region. */
-err:	if (infop->addr != NULL)
+err:	/* Discard the underlying region. */
+	if (infop->addr != NULL)
 		(void)__os_r_detach(dbenv,
 		    infop, F_ISSET(infop, REGION_CREATE));
 	infop->rp = NULL;
@@ -987,13 +986,10 @@ err:	if (infop->addr != NULL)
 
 	/* Discard the REGION structure if we created it. */
 	if (F_ISSET(infop, REGION_CREATE)) {
-		(void)__db_des_destroy(dbenv, rp, 1);
+		__db_des_destroy(dbenv, rp);
 		F_CLR(infop, REGION_CREATE);
 	}
 
-	/* Release the environment lock. */
-	MUTEX_UNLOCK(dbenv, &renv->mutex);
-
 	return (ret);
 }
 
@@ -1009,47 +1005,28 @@ __db_r_detach(dbenv, infop, destroy)
 	REGINFO *infop;
 	int destroy;
 {
-	REGENV *renv;
 	REGION *rp;
-	int ret, t_ret;
+	int ret;
 
-	renv = ((REGINFO *)dbenv->reginfo)->primary;
 	rp = infop->rp;
 	if (F_ISSET(dbenv, DB_ENV_PRIVATE))
 		destroy = 1;
 
-	/* Lock the environment. */
-	MUTEX_LOCK(dbenv, &renv->mutex);
-
-	/* Acquire the lock for the REGION. */
-	MUTEX_LOCK(dbenv, &rp->mutex);
-
 	/*
-	 * We need to call destroy on per-subsystem info before
-	 * we free the memory associated with the region.
+	 * When discarding the regions as we shut down a database environment,
+	 * discard any allocated shared memory segments.  This is the last time
+	 * we use them, and db_region_destroy is the last region-specific call
+	 * we make.
 	 */
-	if (destroy)
-		__db_region_destroy(dbenv, infop);
+	if (F_ISSET(dbenv, DB_ENV_PRIVATE) && infop->primary != NULL)
+		__db_shalloc_free(infop, infop->primary);
 
 	/* Detach from the underlying OS region. */
 	ret = __os_r_detach(dbenv, infop, destroy);
 
-	/* Release the REGION lock. */
-	MUTEX_UNLOCK(dbenv, &rp->mutex);
-
-	/*
-	 * If we destroyed the region, discard the REGION structure.  The only
-	 * time this routine is called with the destroy flag set is when the
-	 * environment is being removed, and it's likely that the only reason
-	 * the environment is being removed is because we crashed.  Don't do
-	 * any unnecessary shared memory manipulation.
-	 */
-	if (destroy &&
-	    ((t_ret = __db_des_destroy(dbenv, rp, 0)) != 0) && ret == 0)
-		ret = t_ret;
-
-	/* Release the environment lock. */
-	MUTEX_UNLOCK(dbenv, &renv->mutex);
+	/* If we destroyed the region, discard the REGION structure. */
+	if (destroy)
+		__db_des_destroy(dbenv, rp);
 
 	/* Destroy the structure. */
 	if (infop->name != NULL)
@@ -1070,14 +1047,9 @@ __db_des_get(dbenv, env_infop, infop, rpp)
 	REGION **rpp;
 {
 	REGENV *renv;
-	REGION *rp, *first_type;
-	u_int32_t maxid;
-	int ret;
+	REGION *rp, *empty_slot, *first_type;
+	u_int32_t i, maxid;
 
-	/*
-	 * !!!
-	 * Called with the environment already locked.
-	 */
 	*rpp = NULL;
 	renv = env_infop->primary;
 
@@ -1088,14 +1060,20 @@ __db_des_get(dbenv, env_infop, infop, rpp)
 	 * return the "primary" region, that is, the first region that was
 	 * created of this type.
 	 *
-	 * Track the maximum region ID so we can allocate a new region,
-	 * note that we have to start at 1 because the primary environment
-	 * uses ID == 1.
+	 * Track the first empty slot and maximum region ID for new region
+	 * allocation.
+	 *
+	 * MaxID starts at REGION_ID_ENV, the ID of the primary environment.
 	 */
 	maxid = REGION_ID_ENV;
-	for (first_type = NULL,
-	    rp = SH_LIST_FIRST(&renv->regionq, __db_region);
-	    rp != NULL; rp = SH_LIST_NEXT(rp, q, __db_region)) {
+	empty_slot = first_type = NULL;
+	for (rp = R_ADDR(env_infop, renv->region_off),
+	    i = 0; i < renv->region_cnt; ++i, ++rp) {
+		if (rp->id == INVALID_REGION_ID) {
+			if (empty_slot == NULL)
+				empty_slot = rp;
+			continue;
+		}
 		if (infop->id != INVALID_REGION_ID) {
 			if (infop->id == rp->id)
 				break;
@@ -1109,47 +1087,48 @@ __db_des_get(dbenv, env_infop, infop, rpp)
 		if (rp->id > maxid)
 			maxid = rp->id;
 	}
-	if (rp == NULL)
+
+	/* If we found a matching ID (or a matching type), return it. */
+	if (i >= renv->region_cnt)
 		rp = first_type;
+	if (rp != NULL) {
+		*rpp = rp;
+		return (0);
+	}
 
 	/*
-	 * If we didn't find a region and we can't create the region, fail.
-	 * The caller generates any error message.
+	 * If we didn't find a region and we don't have permission to create
+	 * the region, fail.  The caller generates any error message.
 	 */
-	if (rp == NULL && !F_ISSET(infop, REGION_CREATE_OK))
+	if (!F_ISSET(infop, REGION_CREATE_OK))
 		return (ENOENT);
 
 	/*
-	 * If we didn't find a region, create and initialize a REGION structure
-	 * for the caller.  If id was set, use that value, otherwise we use the
-	 * next available ID.
+	 * If we didn't find a region and don't have room to create the region
+	 * fail with an error message, there's a sizing problem.
 	 */
-	if (rp == NULL) {
-		if ((ret = __db_shalloc(env_infop->addr,
-		    sizeof(REGION), MUTEX_ALIGN, &rp)) != 0)
-			return (ret);
-
-		/* Initialize the region. */
-		memset(rp, 0, sizeof(*rp));
-		if ((ret = __db_mutex_setup(dbenv, env_infop, &rp->mutex,
-		    MUTEX_NO_RECORD | MUTEX_NO_RLOCK)) != 0) {
-			__db_shalloc_free(env_infop->addr, rp);
-			return (ret);
-		}
-		rp->segid = INVALID_REGION_SEGID;
-
-		/*
-		 * Set the type and ID; if no region ID was specified,
-		 * allocate one.
-		 */
-		rp->type = infop->type;
-		rp->id = infop->id == INVALID_REGION_ID ? maxid + 1 : infop->id;
-
-		SH_LIST_INSERT_HEAD(&renv->regionq, rp, q, __db_region);
-		F_SET(infop, REGION_CREATE);
+	if (empty_slot == NULL) {
+		__db_err(dbenv, "no room remaining for additional REGIONs");
+		return (ENOENT);
 	}
 
-	*rpp = rp;
+	/*
+	 * Initialize a REGION structure for the caller.  If id was set, use
+	 * that value, otherwise we use the next available ID.
+	 */
+	memset(empty_slot, 0, sizeof(REGION));
+	empty_slot->segid = INVALID_REGION_SEGID;
+
+	/*
+	 * Set the type and ID; if no region ID was specified,
+	 * allocate one.
+	 */
+	empty_slot->type = infop->type;
+	empty_slot->id = infop->id == INVALID_REGION_ID ? maxid + 1 : infop->id;
+
+	F_SET(infop, REGION_CREATE);
+
+	*rpp = empty_slot;
 	return (0);
 }
 
@@ -1157,38 +1136,14 @@ __db_des_get(dbenv, env_infop, infop, rpp)
  * __db_des_destroy --
  *	Destroy a reference to a REGION.
  */
-static int
-__db_des_destroy(dbenv, rp, shmem_safe)
+static void
+__db_des_destroy(dbenv, rp)
 	DB_ENV *dbenv;
 	REGION *rp;
-	int shmem_safe;
 {
-	REGINFO *infop;
+	COMPQUIET(dbenv, NULL);
 
-	/*
-	 * !!!
-	 * Called with the environment already locked.
-	 */
-	infop = dbenv->reginfo;
-
-	/*
-	 * If we're calling during recovery, it may not be safe to access the
-	 * shared memory, as the shared memory may have been corrupted during
-	 * the crash.  If the shared memory is safe, remove the REGION entry
-	 * from its linked list, destroy the mutex, and free the allocated
-	 * memory.  On systems that require system mutex support, we don't
-	 * have a choice -- safe or not, we have to destroy the mutex or we'll
-	 * leak memory.
-	 */
-	if (shmem_safe)
-		SH_LIST_REMOVE(rp, q, __db_region);
-
-	__db_mutex_destroy(&rp->mutex);
-
-	if (shmem_safe)
-		__db_shalloc_free(infop->addr, rp);
-
-	return (0);
+	rp->id = INVALID_REGION_ID;
 }
 
 /*
@@ -1205,6 +1160,10 @@ __db_faultmem(dbenv, addr, size, created)
 	int ret;
 	u_int8_t *p, *t;
 
+	/* Ignore heap regions. */
+	if (F_ISSET(dbenv, DB_ENV_PRIVATE))
+		return (0);
+
 	/*
 	 * It's sometimes significantly faster to page-fault in all of the
 	 * region's pages before we run the application, as we see nasty
@@ -1216,50 +1175,21 @@ __db_faultmem(dbenv, addr, size, created)
 	 * system can't cheat.  If we're just joining the region, we can
 	 * only read the value and try to confuse the compiler sufficiently
 	 * that it doesn't figure out that we're never really using it.
+	 *
+	 * Touch every page (assuming pages are 512B, the smallest VM page
+	 * size used in any general purpose processor).
 	 */
 	ret = 0;
 	if (F_ISSET(dbenv, DB_ENV_REGION_INIT)) {
 		if (created)
-			for (p = addr, t = (u_int8_t *)addr + size;
-			    p < t; p += OS_VMPAGESIZE)
+			for (p = addr,
+			    t = (u_int8_t *)addr + size; p < t; p += 512)
 				p[0] = 0xdb;
 		else
-			for (p = addr, t = (u_int8_t *)addr + size;
-			    p < t; p += OS_VMPAGESIZE)
+			for (p = addr,
+			    t = (u_int8_t *)addr + size; p < t; p += 512)
 				ret |= p[0];
 	}
 
 	return (ret);
 }
-
-/*
- * __db_region_destroy --
- *	Destroy per-subsystem region information.
- *	Called with the region already locked.
- */
-static void
-__db_region_destroy(dbenv, infop)
-	DB_ENV *dbenv;
-	REGINFO *infop;
-{
-	switch (infop->type) {
-	case REGION_TYPE_LOCK:
-		__lock_region_destroy(dbenv, infop);
-		break;
-	case REGION_TYPE_LOG:
-		__log_region_destroy(dbenv, infop);
-		break;
-	case REGION_TYPE_MPOOL:
-		__mpool_region_destroy(dbenv, infop);
-		break;
-	case REGION_TYPE_TXN:
-		__txn_region_destroy(dbenv, infop);
-		break;
-	case REGION_TYPE_ENV:
-	case REGION_TYPE_MUTEX:
-		break;
-	default:
-		DB_ASSERT(0);
-		break;
-	}
-}
