@@ -1,10 +1,10 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2002
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
  *
- * $Id: ex_dbclient.c,v 1.28 2002/08/06 06:11:24 bostic Exp $
+ * $Id: ex_dbclient.c,v 12.1 2005/06/16 20:22:01 bostic Exp $
  */
 
 #include <sys/types.h>
@@ -21,6 +21,7 @@
 #define	DATABASE	"access.db"
 
 int	db_clientrun __P((DB_ENV *, const char *));
+int	ex_dbclient __P((const char *));
 int	ex_dbclient_run __P((const char *, FILE *, const char *, const char *));
 int	main __P((int, char *[]));
 
@@ -80,7 +81,7 @@ ex_dbclient_run(home, errfp, host, progname)
 	 * Create an environment object and initialize it for error
 	 * reporting.
 	 */
-	if ((ret = db_env_create(&dbenv, DB_CLIENT)) != 0) {
+	if ((ret = db_env_create(&dbenv, DB_RPCCLIENT)) != 0) {
 		fprintf(errfp, "%s: %s\n", progname, db_strerror(ret));
 		return (1);
 	}
