@@ -1,10 +1,10 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2001-2003
+ * Copyright (c) 2001-2005
  *	Sleepycat Software.  All rights reserved.
  *
- * $Id: ex_rq_net.c,v 1.46 2003/09/05 00:05:34 bostic Exp $
+ * $Id: ex_rq_net.c,v 12.7 2005/11/03 18:29:21 bostic Exp $
  */
 
 #include <sys/types.h>
@@ -25,12 +25,11 @@
 #include <dbinc/queue.h>		/* !!!: for the LIST_XXX macros. */
 #endif
 
-
-int machtab_add __P((machtab_t *, int, u_int32_t, int, int *));
+int machtab_add __P((machtab_t *, socket_t, u_int32_t, int, int *));
 #ifdef DIAGNOSTIC
 void machtab_print __P((machtab_t *));
 #endif
-ssize_t readn __P((int, void *, size_t));
+ssize_t readn __P((socket_t, void *, size_t));
 
 /*
  * This file defines the communication infrastructure for the ex_repquote
@@ -81,14 +80,14 @@ struct __member {
 	u_int32_t hostaddr;	/* Host IP address. */
 	int port;		/* Port number. */
 	int eid;		/* Application-specific machine id. */
-	int fd;			/* File descriptor for the socket. */
+	socket_t fd;		/* File descriptor for the socket. */
 	LIST_ENTRY(__member) links;
 				/* For linked list of all members we know of. */
 };
 
 static int quote_send_broadcast __P((machtab_t *,
     const DBT *, const DBT *, u_int32_t));
-static int quote_send_one __P((const DBT *, const DBT *, int, u_int32_t));
+static int quote_send_one __P((const DBT *, const DBT *, socket_t, u_int32_t));
 
 /*
  * machtab_init --
@@ -105,8 +104,10 @@ machtab_init(machtabp, pri, nsites)
 	int ret;
 	machtab_t *machtab;
 
-	if ((machtab = malloc(sizeof(machtab_t))) == NULL)
+	if ((machtab = malloc(sizeof(machtab_t))) == NULL) {
+		fprintf(stderr, "can't allocate memory\n");
 		return (ENOMEM);
+	}
 
 	LIST_INIT(&machtab->machlist);
 
@@ -131,7 +132,7 @@ machtab_init(machtabp, pri, nsites)
 int
 machtab_add(machtab, fd, hostaddr, port, idp)
 	machtab_t *machtab;
-	int fd;
+	socket_t fd;
 	u_int32_t hostaddr;
 	int port, *idp;
 {
@@ -139,15 +140,19 @@ machtab_add(machtab, fd, hostaddr, port, idp)
 	member_t *m, *member;
 
 	ret = 0;
-	if ((member = malloc(sizeof(member_t))) == NULL)
+	if ((member = malloc(sizeof(member_t))) == NULL) {
+		fprintf(stderr, "can't allocate memory\n");
 		return (ENOMEM);
+	}
 
 	member->fd = fd;
 	member->hostaddr = hostaddr;
 	member->port = port;
 
-	if ((ret = mutex_lock(&machtab->mtmutex)) != 0)
+	if ((ret = mutex_lock(&machtab->mtmutex)) != 0) {
+		fprintf(stderr, "can't lock mutex");
 		return (ret);
+	}
 
 	for (m = LIST_FIRST(&machtab->machlist);
 	    m != NULL; m = LIST_NEXT(m, links))
@@ -160,7 +165,10 @@ machtab_add(machtab, fd, hostaddr, port, idp)
 	} else
 		member->eid = m->eid;
 
-	ret = mutex_unlock(&machtab->mtmutex);
+	if ((ret = mutex_unlock(&machtab->mtmutex)) != 0) {
+		fprintf(stderr, "can't unlock mutex\n");
+		return (ret);
+	}
 
 	if (idp != NULL)
 		*idp = member->eid;
@@ -193,8 +201,10 @@ machtab_getinfo(machtab, eid, hostp, portp)
 	int ret;
 	member_t *member;
 
-	if ((ret = mutex_lock(&machtab->mtmutex)) != 0)
+	if ((ret = mutex_lock(&machtab->mtmutex)) != 0) {
+		fprintf(stderr, "can't lock mutex\n");
 		return (ret);
+	}
 
 	for (member = LIST_FIRST(&machtab->machlist);
 	    member != NULL;
@@ -205,8 +215,10 @@ machtab_getinfo(machtab, eid, hostp, portp)
 			break;
 		}
 
-	if ((ret = mutex_unlock(&machtab->mtmutex)) != 0)
+	if ((ret = mutex_unlock(&machtab->mtmutex)) != 0) {
+		fprintf(stderr, "can't unlock mutex\n");
 		return (ret);
+	}
 
 	return (member != NULL ? 0 : EINVAL);
 }
@@ -227,8 +239,10 @@ machtab_rem(machtab, eid, lock)
 	member_t *member;
 
 	ret = 0;
-	if (lock && (ret = mutex_lock(&machtab->mtmutex)) != 0)
+	if (lock && (ret = mutex_lock(&machtab->mtmutex)) != 0) {
+		fprintf(stderr, "can't lock mutex\n");
 		return (ret);
+	}
 
 	for (found = 0, member = LIST_FIRST(&machtab->machlist);
 	    member != NULL;
@@ -245,8 +259,8 @@ machtab_rem(machtab, eid, lock)
 	if (LIST_FIRST(&machtab->machlist) == NULL)
 		machtab->nextid = 2;
 
-	if (lock)
-		ret = mutex_unlock(&machtab->mtmutex);
+	if (lock && (ret = mutex_unlock(&machtab->mtmutex)) != 0)
+		fprintf(stderr, "can't unlock mutex\n");
 
 #ifdef DIAGNOSTIC
 	printf("Exiting machtab_rem\n");
@@ -276,17 +290,22 @@ machtab_print(machtab)
 {
 	member_t *m;
 
-	(void)mutex_lock(&machtab->mtmutex);
+	if (mutex_lock(&machtab->mtmutex) != 0) {
+		fprintf(stderr, "can't lock mutex\n");
+		abort();
+	}
 
 	for (m = LIST_FIRST(&machtab->machlist);
 	    m != NULL; m = LIST_NEXT(m, links)) {
 
 	    printf("IP: %lx Port: %6d EID: %2d FD: %3d\n",
-	        (long)m->hostaddr, m->port, m->eid, m->fd);
+		(long)m->hostaddr, m->port, m->eid, m->fd);
 	}
 
-	(void)mutex_unlock(&machtab->mtmutex);
-
+	if (mutex_unlock(&machtab->mtmutex) != 0) {
+		fprintf(stderr, "can't unlock mutex\n");
+		abort();
+	}
 }
 #endif
 /*
@@ -295,20 +314,21 @@ machtab_print(machtab)
  *	a file descriptor for the socket, ready for an accept() call
  *	in a thread that we're happy to let block.
  */
-int
+socket_t
 listen_socket_init(progname, port)
 	const char *progname;
 	int port;
 {
-	int s, sockopt;
-	struct protoent *proto;
+	socket_t s;
+	int sockopt;
 	struct sockaddr_in si;
 
-	if ((proto = getprotobyname("tcp")) == NULL)
-		return (-1);
+	COMPQUIET(progname, NULL);
 
-	if ((s = socket(AF_INET, SOCK_STREAM, proto->p_proto)) < 0)
+	if ((s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
+		perror("can't create listen socket");
 		return (-1);
+	}
 
 	memset(&si, 0, sizeof(si));
 	si.sin_family = AF_INET;
@@ -321,18 +341,22 @@ listen_socket_init(progname, port)
 	 * in use" errors unless this option is set.
 	 */
 	sockopt = 1;
-	setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &sockopt, sizeof (sockopt));
+	setsockopt(s, SOL_SOCKET, SO_REUSEADDR,
+	    (const char *)&sockopt, sizeof (sockopt));
 
-	if (bind(s, (struct sockaddr *)&si, sizeof(si)) != 0)
+	if (bind(s, (struct sockaddr *)&si, sizeof(si)) != 0) {
+		perror("can't bind listen socket");
 		goto err;
+	}
 
-	if (listen(s, 5) != 0)
+	if (listen(s, 5) != 0) {
+		perror("can't establish listen queue");
 		goto err;
+	}
 
 	return (s);
 
-err:	fprintf(stderr, "%s: %s", progname, strerror(net_errno));
-	closesocket(s);
+err:	closesocket(s);
 	return (-1);
 }
 
@@ -341,15 +365,17 @@ err:	fprintf(stderr, "%s: %s", progname, strerror(net_errno));
  *	Accept a connection on a socket.  This is essentially just a wrapper
  *	for accept(3).
  */
-int
+socket_t
 listen_socket_accept(machtab, progname, s, eidp)
 	machtab_t *machtab;
 	const char *progname;
-	int s, *eidp;
+	socket_t s;
+	int *eidp;
 {
 	struct sockaddr_in si;
 	int si_len;
-	int host, ns, ret;
+	int host, ret;
+	socket_t ns;
 	u_int16_t port;
 
 	COMPQUIET(progname, NULL);
@@ -357,6 +383,10 @@ listen_socket_accept(machtab, progname, s, eidp)
 wait:	memset(&si, 0, sizeof(si));
 	si_len = sizeof(si);
 	ns = accept(s, (struct sockaddr *)&si, &si_len);
+	if (ns == SOCKET_CREATION_FAILURE) {
+		fprintf(stderr, "can't accept incoming connection\n");
+		return ns;
+	}
 	host = ntohl(si.sin_addr.s_addr);
 
 	/*
@@ -374,54 +404,11 @@ wait:	memset(&si, 0, sizeof(si));
 		goto wait;
 	} else if (ret != 0)
 		goto err;
-printf("Connected to host %x port %d, eid = %d\n", host, port, *eidp);
+	printf("Connected to host %x port %d, eid = %d\n", host, port, *eidp);
 	return (ns);
 
 err:	closesocket(ns);
-	return (-1);
-}
-
-/*
- * get_accepted_socket --
- *	Listen on the specified port, and return a file descriptor
- *	when we have accepted a connection on it.
- */
-int
-get_accepted_socket(progname, port)
-	const char *progname;
-	int port;
-{
-	struct protoent *proto;
-	struct sockaddr_in si;
-	int si_len;
-	int s, ns;
-
-	if ((proto = getprotobyname("tcp")) == NULL)
-		return (-1);
-
-	if ((s = socket(AF_INET, SOCK_STREAM, proto->p_proto)) < 0)
-		return (-1);
-
-	memset(&si, 0, sizeof(si));
-	si.sin_family = AF_INET;
-	si.sin_addr.s_addr = htonl(INADDR_ANY);
-	si.sin_port = htons((unsigned short)port);
-
-	if (bind(s, (struct sockaddr *)&si, sizeof(si)) != 0)
-		goto err;
-
-	if (listen(s, 5) != 0)
-		goto err;
-
-	memset(&si, 0, sizeof(si));
-	si_len = sizeof(si);
-	ns = accept(s, (struct sockaddr *)&si, &si_len);
-
-	return (ns);
-
-err:	fprintf(stderr, "%s: %s", progname, strerror(net_errno));
-	closesocket(s);
-	return (-1);
+	return SOCKET_CREATION_FAILURE;
 }
 
 /*
@@ -432,32 +419,31 @@ err:	fprintf(stderr, "%s: %s", progname, strerror(net_errno));
  *	open to this machine, then don't create another one, return the eid
  *	of the connection (in *eidp) and set is_open to 1.  Return 0.
  */
-int
+socket_t
 get_connected_socket(machtab, progname, remotehost, port, is_open, eidp)
 	machtab_t *machtab;
 	const char *progname, *remotehost;
 	int port, *is_open, *eidp;
 {
-	int ret, s;
+	int ret;
+	socket_t s;
 	struct hostent *hp;
-	struct protoent *proto;
 	struct sockaddr_in si;
 	u_int32_t addr;
 	u_int16_t nport;
 
 	*is_open = 0;
 
-	if ((proto = getprotobyname("tcp")) == NULL)
-		return (-1);
-
 	if ((hp = gethostbyname(remotehost)) == NULL) {
 		fprintf(stderr, "%s: host not found: %s\n", progname,
 		    strerror(net_errno));
 		return (-1);
 	}
 
-	if ((s = socket(AF_INET, SOCK_STREAM, proto->p_proto)) < 0)
+	if ((s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
+		perror("can't create outgoing socket");
 		return (-1);
+	}
 	memset(&si, 0, sizeof(si));
 	memcpy((char *)&si.sin_addr, hp->h_addr, hp->h_length);
 	addr = ntohl(si.sin_addr.s_addr);
@@ -501,7 +487,7 @@ get_connected_socket(machtab, progname, remotehost, port, is_open, eidp)
  */
 int
 get_next_message(fd, rec, control)
-	int fd;
+	socket_t fd;
 	DBT *rec, *control;
 {
 	size_t nr;
@@ -566,7 +552,7 @@ get_next_message(fd, rec, control)
  */
 ssize_t
 readn(fd, vptr, n)
-	int fd;
+	socket_t fd;
 	void *vptr;
 	size_t n;
 {
@@ -584,8 +570,10 @@ readn(fd, vptr, n)
 			 */
 			if (net_errno == EINTR)
 				nread = 0;
-			else
+			else {
+				perror("can't read from socket");
 				return (-1);
+			}
 		} else if (nread == 0)
 			break;  /* EOF */
 
@@ -608,7 +596,8 @@ quote_send(dbenv, control, rec, lsnp, eid, flags)
 	int eid;
 	u_int32_t flags;
 {
-	int fd, n, ret, t_ret;
+	int n, ret, t_ret;
+	socket_t fd;
 	machtab_t *machtab;
 	member_t *m;
 
@@ -627,8 +616,10 @@ quote_send(dbenv, control, rec, lsnp, eid, flags)
 		return (0);
 	}
 
-	if ((ret = mutex_lock(&machtab->mtmutex)) != 0)
+	if ((ret = mutex_lock(&machtab->mtmutex)) != 0) {
+		dbenv->errx(dbenv, "can't lock mutex");
 		return (ret);
+	}
 
 	fd = 0;
 	for (m = LIST_FIRST(&machtab->machlist); m != NULL;
@@ -645,10 +636,14 @@ quote_send(dbenv, control, rec, lsnp, eid, flags)
 		return (DB_REP_UNAVAIL);
 	}
 
-	ret = quote_send_one(rec, control, fd, flags);
+	if ((ret = quote_send_one(rec, control, fd, flags)) != 0)
+		fprintf(stderr, "socket write error in send() function\n");
 
-	if ((t_ret = mutex_unlock(&machtab->mtmutex)) != 0 && ret == 0)
-		ret = t_ret;
+	if ((t_ret = mutex_unlock(&machtab->mtmutex)) != 0) {
+	        dbenv->errx(dbenv, "can't unlock mutex");
+	        if (ret == 0)
+		        ret = t_ret;
+	}
 
 	return (ret);
 }
@@ -668,20 +663,25 @@ quote_send_broadcast(machtab, rec, control, flags)
 	int ret, sent;
 	member_t *m, *next;
 
-	if ((ret = mutex_lock(&machtab->mtmutex)) != 0)
-		return (0);
+	if ((ret = mutex_lock(&machtab->mtmutex)) != 0) {
+		fprintf(stderr, "can't lock mutex\n");
+		return (ret);
+	}
 
 	sent = 0;
 	for (m = LIST_FIRST(&machtab->machlist); m != NULL; m = next) {
 		next = LIST_NEXT(m, links);
 		if ((ret = quote_send_one(rec, control, m->fd, flags)) != 0) {
+			fprintf(stderr, "socket write error in broadcast\n");
 			(void)machtab_rem(machtab, m->eid, 0);
 		} else
 			sent++;
 	}
 
-	if (mutex_unlock(&machtab->mtmutex) != 0)
+	if (mutex_unlock(&machtab->mtmutex) != 0) {
+		fprintf(stderr, "can't unlock mutex\n");
 		return (-1);
+	}
 
 	return (sent);
 }
@@ -699,7 +699,7 @@ quote_send_broadcast(machtab, rec, control, flags)
 static int
 quote_send_one(rec, control, fd, flags)
 	const DBT *rec, *control;
-	int fd;
+	socket_t fd;
 	u_int32_t flags;
 
 {
