@@ -1,10 +1,10 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2003
+ * Copyright (c) 1997-2005
  *	Sleepycat Software.  All rights reserved.
  *
- * $Id: ex_tpcb.c,v 11.44 2003/04/24 15:44:15 bostic Exp $
+ * $Id: ex_tpcb.c,v 12.2 2005/09/22 03:53:45 mjc Exp $
  */
 
 #include <sys/types.h>
@@ -24,7 +24,7 @@ extern int getopt(int, char * const *, const char *);
 
 typedef enum { ACCOUNT, BRANCH, TELLER } FTYPE;
 
-DB_ENV	 *db_init __P((const char *, const char *, int, int, u_int32_t));
+DB_ENV	 *db_init __P((const char *, const char *, int, u_int32_t));
 int	  hpopulate __P((DB *, int, int, int, int));
 int	  populate __P((DB *, u_int32_t, u_int32_t, int, const char *));
 u_int32_t random_id __P((FTYPE, int, int, int));
@@ -102,6 +102,8 @@ typedef struct _histrec {
 	u_int8_t	pad[RECLEN - 4 * sizeof(u_int32_t)];
 } histrec;
 
+char *progname = "ex_tpcb";			/* Program name. */
+
 int
 main(argc, argv)
 	int argc;
@@ -112,10 +114,9 @@ main(argc, argv)
 	DB_ENV *dbenv;
 	int accounts, branches, seed, tellers, history;
 	int ch, iflag, mpool, ntxns, ret, txn_no_sync, verbose;
-	const char *home, *progname;
+	const char *home;
 
 	home = "TESTDIR";
-	progname = "ex_tpcb";
 	accounts = branches = history = tellers = 0;
 	iflag = mpool = ntxns = txn_no_sync = verbose = 0;
 	seed = (int)time(NULL);
@@ -173,7 +174,7 @@ main(argc, argv)
 
 	/* Initialize the database environment. */
 	if ((dbenv = db_init(home,
-	    progname, mpool, iflag, txn_no_sync ? DB_TXN_NOSYNC : 0)) == NULL)
+	    progname, mpool, txn_no_sync ? DB_TXN_NOSYNC : 0)) == NULL)
 		return (EXIT_FAILURE);
 
 	accounts = accounts == 0 ? ACCOUNTS : accounts;
@@ -236,9 +237,9 @@ usage(progname)
  *	Initialize the environment.
  */
 DB_ENV *
-db_init(home, prefix, cachesize, initializing, flags)
+db_init(home, prefix, cachesize, flags)
 	const char *home, *prefix;
-	int cachesize, initializing;
+	int cachesize;
 	u_int32_t flags;
 {
 	DB_ENV *dbenv;
@@ -258,8 +259,8 @@ db_init(home, prefix, cachesize, initializing, flags)
 		(void)dbenv->set_flags(dbenv, DB_TXN_NOSYNC, 1);
 	flags &= ~(DB_TXN_NOSYNC);
 
-	local_flags = flags | DB_CREATE | (initializing ? DB_INIT_MPOOL :
-	    DB_INIT_TXN | DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_MPOOL);
+	local_flags = flags | DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG |
+	    DB_INIT_MPOOL | DB_INIT_TXN;
 	if ((ret = dbenv->open(dbenv, home, local_flags, 0)) != 0) {
 		dbenv->err(dbenv, ret, "DB_ENV->open: %s", home);
 		(void)dbenv->close(dbenv, 0);
@@ -285,7 +286,7 @@ tp_populate(env, accounts, branches, history, tellers, verbose)
 
 	idnum = BEGID;
 	balance = 500000;
-	oflags = DB_CREATE | DB_TRUNCATE;
+	oflags = DB_CREATE;
 
 	if ((ret = db_create(&dbp, env, 0)) != 0) {
 		env->err(env, ret, "db_create");
@@ -499,12 +500,10 @@ tp_run(dbenv, n, accounts, branches, tellers, verbose)
 	int n, accounts, branches, tellers, verbose;
 {
 	DB *adb, *bdb, *hdb, *tdb;
-	double gtps, itps;
-	int failed, ifailed, ret, txns;
-	time_t starttime, curtime, lasttime;
+	int failed, ret, txns;
+	time_t start_time, end_time;
 
 	adb = bdb = hdb = tdb = NULL;
-	txns = failed = 0;
 
 	/*
 	 * Open the database files.
@@ -546,27 +545,16 @@ tp_run(dbenv, n, accounts, branches, tellers, verbose)
 		goto err;
 	}
 
-	starttime = time(NULL);
-	lasttime = starttime;
-	for (ifailed = 0; n-- > 0;) {
-		txns++;
-		ret = tp_txn(dbenv, adb, bdb, tdb, hdb,
-		    accounts, branches, tellers, verbose);
-		if (ret != 0) {
-			failed++;
-			ifailed++;
-		}
-		if (n % 5000 == 0) {
-			curtime = time(NULL);
-			gtps = (double)(txns - failed) / (curtime - starttime);
-			itps = (double)(5000 - ifailed) / (curtime - lasttime);
-			printf("%d txns %d failed ", txns, failed);
-			printf("%6.2f TPS (gross) %6.2f TPS (interval)\n",
-			   gtps, itps);
-			lasttime = curtime;
-			ifailed = 0;
-		}
-	}
+	(void)time(&start_time);
+	for (txns = n, failed = 0; n-- > 0;)
+		if ((ret = tp_txn(dbenv, adb, bdb, tdb, hdb,
+		    accounts, branches, tellers, verbose)) != 0)
+			++failed;
+	(void)time(&end_time);
+	if (end_time == start_time)
+		++end_time;
+	printf("%s: %d txns: %d failed, %.2f TPS\n", progname,
+	    txns, failed, (txns - failed) / (double)(end_time - start_time));
 
 err:	if (adb != NULL)
 		(void)adb->close(adb, 0);
@@ -576,8 +564,6 @@ err:	if (adb != NULL)
 		(void)tdb->close(tdb, 0);
 	if (hdb != NULL)
 		(void)hdb->close(hdb, 0);
-
-	printf("%ld transactions begun %ld failed\n", (long)txns, (long)failed);
 	return (ret == 0 ? 0 : 1);
 }
 
@@ -602,8 +588,9 @@ tp_txn(dbenv, adb, bdb, tdb, hdb, accounts, branches, tellers, verbose)
 	acurs = bcurs = tcurs = NULL;
 
 	/*
-	 * XXX We could move a lot of this into the driver to make this
-	 * faster.
+	 * !!!
+	 * This is sample code -- we could move a lot of this into the driver
+	 * to make it faster.
 	 */
 	account = random_id(ACCOUNT, accounts, branches, tellers);
 	branch = random_id(BRANCH, accounts, branches, tellers);
@@ -630,7 +617,14 @@ tp_txn(dbenv, adb, bdb, tdb, hdb, accounts, branches, tellers, verbose)
 	/* Request 0 bytes since we're just positioning. */
 	d_histdbt.flags = DB_DBT_PARTIAL;
 
-	/* START TIMING */
+	/*
+	 * START PER-TRANSACTION TIMING.
+	 *
+	 * Technically, TPCB requires a limit on response time, you only get
+	 * to count transactions that complete within 2 seconds.  That's not
+	 * an issue for this sample application -- regardless, here's where
+	 * the transaction begins.
+	 */
 	if (dbenv->txn_begin(dbenv, NULL, &t, 0) != 0)
 		goto err;
 
@@ -678,8 +672,8 @@ tp_txn(dbenv, adb, bdb, tdb, hdb, accounts, branches, tellers, verbose)
 	t = NULL;
 	if (ret != 0)
 		goto err;
+	/* END PER-TRANSACTION TIMING. */
 
-	/* END TIMING */
 	return (0);
 
 err:	if (acurs != NULL)
