@@ -1,7 +1,7 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
  */
 /*
@@ -38,14 +38,12 @@
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
+ *
+ * $Id: hash_page.c,v 12.7 2005/10/13 22:22:43 ubell Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: hash_page.c,v 11.93 2003/06/30 17:20:12 bostic Exp $";
-#endif /* not lint */
-
 /*
  * PACKAGE:  hashing
  *
@@ -150,7 +148,7 @@ __ham_item_reset(dbc)
 	DB *dbp;
 	DB_MPOOLFILE *mpf;
 	HASH_CURSOR *hcp;
-	int ret;
+	int ret, t_ret;
 
 	dbp = dbc->dbp;
 	mpf = dbp->mpf;
@@ -160,29 +158,33 @@ __ham_item_reset(dbc)
 	if (hcp->page != NULL)
 		ret = __memp_fput(mpf, hcp->page, 0);
 
-	__ham_item_init(dbc);
+	if ((t_ret = __ham_item_init(dbc)) != 0 && ret == 0)
+		ret = t_ret;
+
 	return (ret);
 }
 
 /*
- * PUBLIC: void __ham_item_init __P((DBC *));
+ * PUBLIC: int __ham_item_init __P((DBC *));
  */
-void
+int
 __ham_item_init(dbc)
 	DBC *dbc;
 {
 	HASH_CURSOR *hcp;
+	int ret;
 
 	hcp = (HASH_CURSOR *)dbc->internal;
-	/*
-	 * If this cursor still holds any locks, we must
-	 * release them if we are not running with transactions.
-	 */
-	(void)__TLPUT(dbc, hcp->lock);
 
 	/*
-	 * The following fields must *not* be initialized here
-	 * because they may have meaning across inits.
+	 * If this cursor still holds any locks, we must release them if
+	 * we are not running with transactions.
+	 */
+	ret = __TLPUT(dbc, hcp->lock);
+
+	/*
+	 * The following fields must *not* be initialized here because they
+	 * may have meaning across inits.
 	 *	hlock, hdr, split_buf, stats
 	 */
 	hcp->bucket = BUCKET_INVALID;
@@ -199,6 +201,8 @@ __ham_item_init(dbc)
 	hcp->pgno = PGNO_INVALID;
 	hcp->indx = NDX_INVALID;
 	hcp->page = NULL;
+
+	return (ret);
 }
 
 /*
@@ -550,8 +554,8 @@ __ham_del_pair(dbc, reclaim_page)
 	db_ham_mode op;
 	db_indx_t ndx;
 	db_pgno_t chg_pgno, pgno, tmp_pgno;
-	int ret, t_ret;
 	u_int32_t order;
+	int ret, t_ret;
 
 	dbp = dbc->dbp;
 	mpf = dbp->mpf;
@@ -571,12 +575,12 @@ __ham_del_pair(dbc, reclaim_page)
 	 * to remove the big item and then update the page to remove the
 	 * entry referring to the big item.
 	 */
-	ret = 0;
 	if (HPAGE_PTYPE(H_PAIRKEY(dbp, p, ndx)) == H_OFFPAGE) {
 		memcpy(&pgno, HOFFPAGE_PGNO(P_ENTRY(dbp, p, H_KEYINDEX(ndx))),
 		    sizeof(db_pgno_t));
 		ret = __db_doff(dbc, pgno);
-	}
+	} else
+		ret = 0;
 
 	if (ret == 0)
 		switch (HPAGE_PTYPE(H_PAIRDATA(dbp, p, ndx))) {
@@ -595,6 +599,9 @@ __ham_del_pair(dbc, reclaim_page)
 			 */
 			F_CLR(hcp, H_ISDUP);
 			break;
+		default:
+			/* No-op */
+			break;
 		}
 
 	if (ret)
@@ -834,22 +841,27 @@ __ham_replpair(dbc, dbt, make_dup)
 	DB_ENV *dbenv;
 	DB_LSN	new_lsn;
 	HASH_CURSOR *hcp;
-	int32_t change;			/* XXX: Possible overflow. */
-	u_int32_t dup_flag, len, memsize;
-	int beyond_eor, is_big, ret, type;
+	u_int32_t change;
+	u_int32_t dup_flag, len, memsize, newlen;
+	int beyond_eor, is_big, is_plus, ret, type;
 	u_int8_t *beg, *dest, *end, *hk, *src;
 	void *memp;
 
 	/*
-	 * Big item replacements are handled in generic code.
-	 * Items that fit on the current page fall into 4 classes.
-	 * 1. On-page element, same size
-	 * 2. On-page element, new is bigger (fits)
-	 * 3. On-page element, new is bigger (does not fit)
-	 * 4. On-page element, old is bigger
-	 * Numbers 1, 2, and 4 are essentially the same (and should
-	 * be the common case).  We handle case 3 as a delete and
-	 * add.
+	 * Items that were already offpage (ISBIG) were handled before
+	 * we get in here.  So, we need only handle cases where the old
+	 * key is on a regular page.  That leaves us 6 cases:
+	 * 1. Original data onpage; new data is smaller
+	 * 2. Original data onpage; new data is the same size
+	 * 3. Original data onpage; new data is bigger, but not ISBIG,
+	 *    fits on page
+	 * 4. Original data onpage; new data is bigger, but not ISBIG,
+	 *    does not fit on page
+	 * 5. Original data onpage; New data is an off-page item.
+	 * 6. Original data was offpage; new item is smaller.
+	 *
+	 * Cases 1-3 are essentially the same (and should be the common case).
+	 * We handle 4-6 as delete and add.
 	 */
 	dbp = dbc->dbp;
 	dbenv = dbp->dbenv;
@@ -864,7 +876,13 @@ __ham_replpair(dbc, dbt, make_dup)
 	 * formula doesn't work, because we are essentially adding
 	 * new bytes.
 	 */
-	change = dbt->size - dbt->dlen;
+	if (dbt->size > dbt->dlen) {
+		change = dbt->size - dbt->dlen;
+		is_plus = 1;
+	} else {
+		change = dbt->dlen - dbt->size;
+		is_plus = 0;
+	}
 
 	hk = H_PAIRDATA(dbp, hcp->page, hcp->indx);
 	is_big = HPAGE_PTYPE(hk) == H_OFFPAGE;
@@ -876,13 +894,39 @@ __ham_replpair(dbc, dbt, make_dup)
 		    dbp->pgsize, H_DATAINDEX(hcp->indx));
 
 	beyond_eor = dbt->doff + dbt->dlen > len;
-	if (beyond_eor)
-		change += dbt->doff + dbt->dlen - len;
-
-	if (change > (int32_t)P_FREESPACE(dbp, hcp->page) ||
-	    beyond_eor || is_big) {
+	if (beyond_eor) {
 		/*
-		 * Case 3 -- two subcases.
+		 * The change is beyond the end of record.  If change
+		 * is a positive number, we can simply add the extension
+		 * to it.  However, if change is negative, then we need
+		 * to figure out if the extension is larger than the
+		 * negative change.
+		 */
+		if (is_plus)
+			change += dbt->doff + dbt->dlen - len;
+		else if (dbt->doff + dbt->dlen - len > change) {
+			/* Extension bigger than change */
+			is_plus = 1;
+			change = (dbt->doff + dbt->dlen - len) - change;
+		} else /* Extension is smaller than change. */
+			change -= (dbt->doff + dbt->dlen - len);
+	}
+
+	newlen = (is_plus ? len + change : len - change);
+	if (ISBIG(hcp, newlen) ||
+	    (is_plus && change > P_FREESPACE(dbp, hcp->page)) ||
+	    beyond_eor || is_big) {
+	    	/* 
+		 * If we are in cases 4 or 5 then is_plus will be true.
+		 * If we don't have a transaction then we cannot roll back,
+		 * make sure there is enough room for the new page.
+		 */
+		if (is_plus && dbc->txn == NULL &&
+		    dbp->mpf->mfp->maxpgno != 0 &&
+		    dbp->mpf->mfp->maxpgno == dbp->mpf->mfp->last_pgno)
+		    	return (__db_space_err(dbp));
+		/*
+		 * Cases 4-6 -- two subcases.
 		 * A. This is not really a partial operation, but an overwrite.
 		 *    Simple del and add works.
 		 * B. This is a partial and we need to construct the data that
@@ -923,7 +967,7 @@ __ham_replpair(dbc, dbt, make_dup)
 			}
 
 			/* Now shift old data around to make room for new. */
-			if (change > 0) {
+			if (is_plus) {
 				if ((ret = __os_realloc(dbenv,
 				    tdata.size + change, &tdata.data)) != 0)
 					return (ret);
@@ -936,13 +980,19 @@ __ham_replpair(dbc, dbt, make_dup)
 
 			src = (u_int8_t *)tdata.data + dbt->doff + dbt->dlen;
 			if (src < end && tdata.size > dbt->doff + dbt->dlen) {
-				len = tdata.size - dbt->doff - dbt->dlen;
-				dest = src + change;
+				len = tdata.size - (dbt->doff + dbt->dlen);
+				if (is_plus)
+					dest = src + change;
+				else
+					dest = src - change;
 				memmove(dest, src, len);
 			}
 			memcpy((u_int8_t *)tdata.data + dbt->doff,
 			    dbt->data, dbt->size);
-			tdata.size += change;
+			if (is_plus)
+				tdata.size += change;
+			else
+				tdata.size -= change;
 
 			/* Now add the pair. */
 			ret = __ham_add_el(dbc, &tmp, &tdata, type);
@@ -970,7 +1020,7 @@ err:		return (ret);
 		if ((ret = __ham_replace_log(dbp,
 		    dbc->txn, &new_lsn, 0, PGNO(hcp->page),
 		    (u_int32_t)H_DATAINDEX(hcp->indx), &LSN(hcp->page),
-		    (u_int32_t)dbt->doff, &old_dbt, dbt, make_dup)) != 0)
+		    (int32_t)dbt->doff, &old_dbt, dbt, make_dup)) != 0)
 			return (ret);
 
 	} else
@@ -979,7 +1029,7 @@ err:		return (ret);
 	LSN(hcp->page) = new_lsn;	/* Structure assignment. */
 
 	__ham_onpage_replace(dbp, hcp->page, (u_int32_t)H_DATAINDEX(hcp->indx),
-	    (int32_t)dbt->doff, change, dbt);
+	    (int32_t)dbt->doff, change, is_plus, dbt);
 
 	return (0);
 }
@@ -997,15 +1047,16 @@ err:		return (ret);
  * dbt: the new data that gets written at beg.
  *
  * PUBLIC: void __ham_onpage_replace __P((DB *, PAGE *, u_int32_t,
- * PUBLIC:     int32_t, int32_t,  DBT *));
+ * PUBLIC:     int32_t, u_int32_t,  int, DBT *));
  */
 void
-__ham_onpage_replace(dbp, pagep, ndx, off, change, dbt)
+__ham_onpage_replace(dbp, pagep, ndx, off, change, is_plus, dbt)
 	DB *dbp;
 	PAGE *pagep;
 	u_int32_t ndx;
 	int32_t off;
-	int32_t change;
+	u_int32_t change;
+	int is_plus;
 	DBT *dbt;
 {
 	db_indx_t i, *inp;
@@ -1030,15 +1081,25 @@ __ham_onpage_replace(dbp, pagep, ndx, off, change, dbt)
 			len = (int32_t)(
 			    (HKEYDATA_DATA(P_ENTRY(dbp, pagep, ndx)) + off) -
 			    src);
-		dest = src - change;
-		memmove(dest, src, len);
+		if (is_plus)
+			dest = src - change;
+		else
+			dest = src + change;
+		memmove(dest, src, (size_t)len);
 		if (zero_me)
 			memset(dest + len, 0, change);
 
 		/* Now update the indices. */
-		for (i = ndx; i < NUM_ENT(pagep); i++)
-			inp[i] -= change;
-		HOFFSET(pagep) -= change;
+		for (i = ndx; i < NUM_ENT(pagep); i++) {
+			if (is_plus)
+				inp[i] -= change;
+			else
+				inp[i] += change;
+		}
+		if (is_plus)
+			HOFFSET(pagep) -= change;
+		else
+			HOFFSET(pagep) += change;
 	}
 	if (off >= 0)
 		memcpy(HKEYDATA_DATA(P_ENTRY(dbp, pagep, ndx)) + off,
@@ -1071,11 +1132,12 @@ __ham_split_page(dbc, obucket, nbucket)
 	void *big_buf;
 
 	dbp = dbc->dbp;
+	carray = NULL;
 	dbenv = dbp->dbenv;
 	mpf = dbp->mpf;
 	hcp = (HASH_CURSOR *)dbc->internal;
 	temp_pagep = old_pagep = new_pagep = NULL;
-	carray = NULL;
+	npgno = PGNO_INVALID;
 	LOCK_INIT(block);
 
 	bucket_pgno = BUCKET_TO_PAGE(hcp, obucket);
@@ -1268,8 +1330,8 @@ err:		if (old_pagep != NULL)
 		if (temp_pagep != NULL && PGNO(temp_pagep) != bucket_pgno)
 			(void)__memp_fput(mpf, temp_pagep, DB_MPOOL_DIRTY);
 	}
-	if (LOCK_ISSET(block))
-		__TLPUT(dbc, block);
+	if ((t_ret = __TLPUT(dbc, block)) != 0 && ret == 0)
+		ret = t_ret;
 	if (carray != NULL)		/* We never knew its size. */
 		__os_free(dbenv, carray);
 	return (ret);
@@ -1298,7 +1360,8 @@ __ham_add_el(dbc, key, val, type)
 	HASH_CURSOR *hcp;
 	HOFFPAGE doff, koff;
 	db_pgno_t next_pgno, pgno;
-	u_int32_t data_size, key_size, pairsize, rectype;
+	u_int32_t data_size, key_size;
+	u_int32_t pages, pagespace, pairsize, rectype;
 	int do_expand, is_keybig, is_databig, ret;
 	int key_type, data_type;
 
@@ -1349,6 +1412,24 @@ __ham_add_el(dbc, key, val, type)
 		hcp->pgno = PGNO(hcp->page);
 	}
 
+	/*
+	 * If we don't have a transaction then make sure we will not
+	 * run out of file space before updating the key or data.
+	 */
+	if (dbc->txn == NULL &&
+	    dbp->mpf->mfp->maxpgno != 0 && (is_keybig || is_databig)) {
+		pagespace = P_MAXSPACE(dbp, dbp->pgsize);
+		pages = 0;
+		if (is_databig) 
+			pages = ((data_size - 1) / pagespace) + 1;
+		if (is_keybig) {
+			pages += ((key->size - 1) / pagespace) + 1;
+			if (pages >
+			    (dbp->mpf->mfp->maxpgno - dbp->mpf->mfp->last_pgno))
+				return (__db_space_err(dbp));
+		}
+	}
+
 	/*
 	 * Update cursor.
 	 */
@@ -1554,12 +1635,19 @@ __ham_get_cpage(dbc, mode)
 	 */
 	LOCK_INIT(tmp_lock);
 	if (STD_LOCKING(dbc)) {
-		if (hcp->lbucket != hcp->bucket &&	/* Case 4 */
-		    (ret = __TLPUT(dbc, hcp->lock)) != 0)
-			return (ret);
+		if (hcp->lbucket != hcp->bucket) {	/* Case 4 */
+			if ((ret = __TLPUT(dbc, hcp->lock)) != 0)
+				return (ret);
+			LOCK_INIT(hcp->lock);
+		}
 
+		/*
+		 * See if we have the right lock.  If we are doing
+		 * dirty reads we assume the write lock has been downgraded.
+		 */
 		if ((LOCK_ISSET(hcp->lock) &&
-		    (hcp->lock_mode == DB_LOCK_READ &&
+		    ((hcp->lock_mode == DB_LOCK_READ ||
+		    F_ISSET(dbp, DB_AM_READ_UNCOMMITTED)) &&
 		    mode == DB_LOCK_WRITE))) {
 			/* Case 3. */
 			tmp_lock = hcp->lock;
@@ -1575,9 +1663,9 @@ __ham_get_cpage(dbc, mode)
 		if (ret == 0) {
 			hcp->lock_mode = mode;
 			hcp->lbucket = hcp->bucket;
-			if (LOCK_ISSET(tmp_lock))
-				/* Case 3: release the original lock. */
-				ret = __lock_put(dbp->dbenv, &tmp_lock);
+			/* Case 3: release the original lock. */
+			if ((ret = __ENV_LPUT(dbp->dbenv, tmp_lock)) != 0)
+				return (ret);
 		} else if (LOCK_ISSET(tmp_lock))
 			hcp->lock = tmp_lock;
 	}
@@ -1777,7 +1865,7 @@ __ham_c_delpg(dbc, old_pgno, new_pgno, num_ent, op, orderp)
 	my_txn = IS_SUBTRANSACTION(dbc->txn) ? dbc->txn : NULL;
 	found = 0;
 
-	MUTEX_THREAD_LOCK(dbenv, dbenv->dblist_mutexp);
+	MUTEX_LOCK(dbenv, dbenv->mtx_dblist);
 	/*
 	 * Find the highest order of any cursor our movement
 	 * may collide with.
@@ -1786,7 +1874,7 @@ __ham_c_delpg(dbc, old_pgno, new_pgno, num_ent, op, orderp)
 	for (ldbp = __dblist_get(dbenv, dbp->adj_fileid);
 	    ldbp != NULL && ldbp->adj_fileid == dbp->adj_fileid;
 	    ldbp = LIST_NEXT(ldbp, dblistlinks)) {
-		MUTEX_THREAD_LOCK(dbenv, dbp->mutexp);
+		MUTEX_LOCK(dbenv, dbp->mutex);
 		for (cp = TAILQ_FIRST(&ldbp->active_queue); cp != NULL;
 		    cp = TAILQ_NEXT(cp, links)) {
 			if (cp == dbc || cp->dbtype != DB_HASH)
@@ -1803,13 +1891,13 @@ __ham_c_delpg(dbc, old_pgno, new_pgno, num_ent, op, orderp)
 				    F_ISSET(hcp, H_DELETED)));
 			}
 		}
-		MUTEX_THREAD_UNLOCK(dbenv, dbp->mutexp);
+		MUTEX_UNLOCK(dbenv, dbp->mutex);
 	}
 
 	for (ldbp = __dblist_get(dbenv, dbp->adj_fileid);
 	    ldbp != NULL && ldbp->adj_fileid == dbp->adj_fileid;
 	    ldbp = LIST_NEXT(ldbp, dblistlinks)) {
-		MUTEX_THREAD_LOCK(dbenv, dbp->mutexp);
+		MUTEX_LOCK(dbenv, dbp->mutex);
 		for (cp = TAILQ_FIRST(&ldbp->active_queue); cp != NULL;
 		    cp = TAILQ_NEXT(cp, links)) {
 			if (cp == dbc || cp->dbtype != DB_HASH)
@@ -1854,9 +1942,9 @@ __ham_c_delpg(dbc, old_pgno, new_pgno, num_ent, op, orderp)
 					found = 1;
 			}
 		}
-		MUTEX_THREAD_UNLOCK(dbenv, dbp->mutexp);
+		MUTEX_UNLOCK(dbenv, dbp->mutex);
 	}
-	MUTEX_THREAD_UNLOCK(dbenv, dbenv->dblist_mutexp);
+	MUTEX_UNLOCK(dbenv, dbenv->mtx_dblist);
 
 	if (found != 0 && DBC_LOGGING(dbc)) {
 		if ((ret = __ham_chgpg_log(dbp, my_txn, &lsn, 0, op,
