@@ -1,7 +1,7 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
  */
 /*
@@ -38,14 +38,12 @@
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
+ *
+ * $Id: hash_rec.c,v 12.7 2005/09/28 17:44:52 margo Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: hash_rec.c,v 11.76 2003/07/04 17:45:07 margo Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -88,25 +86,37 @@ __ham_insdel_recover(dbenv, dbtp, lsnp, op, info)
 	COMPQUIET(info, NULL);
 
 	REC_PRINT(__ham_insdel_print);
-	REC_INTRO(__ham_insdel_read, 1);
+	REC_INTRO(__ham_insdel_read, 1, 0);
 
 	if ((ret = __memp_fget(mpf, &argp->pgno, 0, &pagep)) != 0) {
 		if (DB_UNDO(op)) {
-			/*
-			 * We are undoing and the page doesn't exist.  That
-			 * is equivalent to having a pagelsn of 0, so we
-			 * would not have to undo anything.  In this case,
-			 * don't bother creating a page.
-			 */
+			if (ret == DB_PAGE_NOTFOUND)
+				goto done;
+			else {
+				ret = __db_pgerr(file_dbp, argp->pgno, ret);
+				goto out;
+			}
+		}
+#ifdef HAVE_FTRUNCATE
+		/* If the page is not here then it was later truncated. */
+		if (!IS_ZERO_LSN(argp->pagelsn))
 			goto done;
-		} else if ((ret = __memp_fget(mpf,
-		    &argp->pgno, DB_MPOOL_CREATE, &pagep)) != 0)
+#endif
+		/*
+		 * This page was created by a group allocation and
+		 * the file may not have been extend yet.
+		 * Create the page if necessary.
+		 */
+		if ((ret = __memp_fget(mpf,
+		     &argp->pgno, DB_MPOOL_CREATE, &pagep)) != 0) {
+			ret = __db_pgerr(file_dbp, argp->pgno, ret);
 			goto out;
+		}
 	}
 
 	cmp_n = log_compare(lsnp, &LSN(pagep));
 	cmp_p = log_compare(&LSN(pagep), &argp->pagelsn);
-	CHECK_LSN(op, cmp_p, &LSN(pagep), &argp->pagelsn);
+	CHECK_LSN(dbenv, op, cmp_p, &LSN(pagep), &argp->pagelsn);
 	/*
 	 * Two possible things going on:
 	 * redo a delete/undo a put: delete the item from the page.
@@ -117,8 +127,8 @@ __ham_insdel_recover(dbenv, dbtp, lsnp, op, info)
 	 * We do this by calling __putitem with the type H_OFFPAGE instead
 	 * of H_KEYDATA.
 	 */
-	opcode = OPCODE_OF(argp->opcode);
 
+	opcode = OPCODE_OF(argp->opcode);
 	flags = 0;
 	if ((opcode == DELPAIR && cmp_n == 0 && DB_UNDO(op)) ||
 	    (opcode == PUTPAIR && cmp_p == 0 && DB_REDO(op))) {
@@ -142,7 +152,7 @@ __ham_insdel_recover(dbenv, dbtp, lsnp, op, info)
 				type = H_KEYDATA;
 			__ham_putitem(file_dbp, pagep, &argp->data, type);
 		} else
-			(void)__ham_reputpair(file_dbp, pagep,
+			__ham_reputpair(file_dbp, pagep,
 			    argp->ndx, &argp->key, &argp->data);
 
 		LSN(pagep) = DB_REDO(op) ? *lsnp : argp->pagelsn;
@@ -197,22 +207,9 @@ __ham_newpage_recover(dbenv, dbtp, lsnp, op, info)
 	COMPQUIET(info, NULL);
 
 	REC_PRINT(__ham_newpage_print);
-	REC_INTRO(__ham_newpage_read, 1);
+	REC_INTRO(__ham_newpage_read, 1, 0);
 
-	if ((ret = __memp_fget(mpf, &argp->new_pgno, 0, &pagep)) != 0) {
-		if (DB_UNDO(op)) {
-			/*
-			 * We are undoing and the page doesn't exist.  That
-			 * is equivalent to having a pagelsn of 0, so we
-			 * would not have to undo anything.  In this case,
-			 * don't bother creating a page.
-			 */
-			ret = 0;
-			goto ppage;
-		} else if ((ret = __memp_fget(mpf,
-		    &argp->new_pgno, DB_MPOOL_CREATE, &pagep)) != 0)
-			goto out;
-	}
+	REC_FGET(mpf, argp->new_pgno, &pagep, ppage);
 
 	/*
 	 * There are potentially three pages we need to check: the one
@@ -222,7 +219,7 @@ __ham_newpage_recover(dbenv, dbtp, lsnp, op, info)
 
 	cmp_n = log_compare(lsnp, &LSN(pagep));
 	cmp_p = log_compare(&LSN(pagep), &argp->pagelsn);
-	CHECK_LSN(op, cmp_p, &LSN(pagep), &argp->pagelsn);
+	CHECK_LSN(dbenv, op, cmp_p, &LSN(pagep), &argp->pagelsn);
 
 	flags = 0;
 	if ((cmp_p == 0 && DB_REDO(op) && argp->opcode == PUTOVFL) ||
@@ -249,25 +246,11 @@ __ham_newpage_recover(dbenv, dbtp, lsnp, op, info)
 
 	/* Now do the prev page. */
 ppage:	if (argp->prev_pgno != PGNO_INVALID) {
-		if ((ret =
-		    __memp_fget(mpf, &argp->prev_pgno, 0, &pagep)) != 0) {
-			if (DB_UNDO(op)) {
-				/*
-				 * We are undoing and the page doesn't exist.
-				 * That is equivalent to having a pagelsn of 0,
-				 * so we would not have to undo anything.  In
-				 * this case, don't bother creating a page.
-				 */
-				ret = 0;
-				goto npage;
-			} else if ((ret = __memp_fget(mpf,
-			    &argp->prev_pgno, DB_MPOOL_CREATE, &pagep)) != 0)
-				goto out;
-		}
+		REC_FGET(mpf, argp->prev_pgno, &pagep, npage);
 
 		cmp_n = log_compare(lsnp, &LSN(pagep));
 		cmp_p = log_compare(&LSN(pagep), &argp->prevlsn);
-		CHECK_LSN(op, cmp_p, &LSN(pagep), &argp->prevlsn);
+		CHECK_LSN(dbenv, op, cmp_p, &LSN(pagep), &argp->prevlsn);
 		flags = 0;
 
 		if ((cmp_p == 0 && DB_REDO(op) && argp->opcode == PUTOVFL) ||
@@ -293,24 +276,11 @@ ppage:	if (argp->prev_pgno != PGNO_INVALID) {
 
 	/* Now time to do the next page */
 npage:	if (argp->next_pgno != PGNO_INVALID) {
-		if ((ret =
-		    __memp_fget(mpf, &argp->next_pgno, 0, &pagep)) != 0) {
-			if (DB_UNDO(op)) {
-				/*
-				 * We are undoing and the page doesn't exist.
-				 * That is equivalent to having a pagelsn of 0,
-				 * so we would not have to undo anything.  In
-				 * this case, don't bother creating a page.
-				 */
-				goto done;
-			} else if ((ret = __memp_fget(mpf,
-			    &argp->next_pgno, DB_MPOOL_CREATE, &pagep)) != 0)
-				goto out;
-		}
+		REC_FGET(mpf, argp->next_pgno, &pagep, done);
 
 		cmp_n = log_compare(lsnp, &LSN(pagep));
 		cmp_p = log_compare(&LSN(pagep), &argp->nextlsn);
-		CHECK_LSN(op, cmp_p, &LSN(pagep), &argp->nextlsn);
+		CHECK_LSN(dbenv, op, cmp_p, &LSN(pagep), &argp->nextlsn);
 		flags = 0;
 
 		if ((cmp_p == 0 && DB_REDO(op) && argp->opcode == PUTOVFL) ||
@@ -365,57 +335,63 @@ __ham_replace_recover(dbenv, dbtp, lsnp, op, info)
 	DBT dbt;
 	PAGE *pagep;
 	u_int32_t flags;
-	int32_t grow;
-	int cmp_n, cmp_p, ret;
+	u_int32_t change;
+	int cmp_n, cmp_p, is_plus, ret;
 	u_int8_t *hk;
 
 	pagep = NULL;
 	COMPQUIET(info, NULL);
 
 	REC_PRINT(__ham_replace_print);
-	REC_INTRO(__ham_replace_read, 1);
+	REC_INTRO(__ham_replace_read, 1, 0);
 
-	if ((ret = __memp_fget(mpf, &argp->pgno, 0, &pagep)) != 0) {
-		if (DB_UNDO(op)) {
-			/*
-			 * We are undoing and the page doesn't exist.  That
-			 * is equivalent to having a pagelsn of 0, so we
-			 * would not have to undo anything.  In this case,
-			 * don't bother creating a page.
-			 */
-			goto done;
-		} else if ((ret = __memp_fget(mpf,
-		    &argp->pgno, DB_MPOOL_CREATE, &pagep)) != 0)
-			goto out;
-	}
+	REC_FGET(mpf, argp->pgno, &pagep, done);
 
 	cmp_n = log_compare(lsnp, &LSN(pagep));
 	cmp_p = log_compare(&LSN(pagep), &argp->pagelsn);
-	CHECK_LSN(op, cmp_p, &LSN(pagep), &argp->pagelsn);
+	CHECK_LSN(dbenv, op, cmp_p, &LSN(pagep), &argp->pagelsn);
 
 	memset(&dbt, 0, sizeof(dbt));
 	flags = 0;
-	grow = 1;
 
+	/*
+	 * Before we know the direction of the transformation we will
+	 * determine the size differential; then once we know if we are
+	 * redoing or undoing, we'll adjust the sign (is_plus) appropriately.
+	 */
+	if (argp->newitem.size > argp->olditem.size) {
+		change = argp->newitem.size - argp->olditem.size;
+		is_plus = 1;
+	} else {
+		change = argp->olditem.size - argp->newitem.size;
+		is_plus = 0;
+	}
 	if (cmp_p == 0 && DB_REDO(op)) {
 		/* Reapply the change as specified. */
 		dbt.data = argp->newitem.data;
 		dbt.size = argp->newitem.size;
-		grow = argp->newitem.size - argp->olditem.size;
 		LSN(pagep) = *lsnp;
+		/*
+		 * The is_plus flag is set properly to reflect
+		 * newitem.size - olditem.size.
+		 */
 		flags = DB_MPOOL_DIRTY;
 	} else if (cmp_n == 0 && DB_UNDO(op)) {
 		/* Undo the already applied change. */
 		dbt.data = argp->olditem.data;
 		dbt.size = argp->olditem.size;
-		grow = argp->olditem.size - argp->newitem.size;
+		/*
+		 * Invert is_plus to reflect sign of
+		 * olditem.size - newitem.size.
+		 */
+		is_plus = !is_plus;
 		LSN(pagep) = argp->pagelsn;
 		flags = DB_MPOOL_DIRTY;
 	}
 
 	if (flags) {
 		__ham_onpage_replace(file_dbp, pagep,
-		    argp->ndx, argp->off, grow, &dbt);
+		    argp->ndx, argp->off, change, is_plus, &dbt);
 		if (argp->makedup) {
 			hk = P_ENTRY(file_dbp, pagep, argp->ndx);
 			if (DB_REDO(op))
@@ -463,25 +439,37 @@ __ham_splitdata_recover(dbenv, dbtp, lsnp, op, info)
 	COMPQUIET(info, NULL);
 
 	REC_PRINT(__ham_splitdata_print);
-	REC_INTRO(__ham_splitdata_read, 1);
+	REC_INTRO(__ham_splitdata_read, 1, 0);
 
 	if ((ret = __memp_fget(mpf, &argp->pgno, 0, &pagep)) != 0) {
 		if (DB_UNDO(op)) {
-			/*
-			 * We are undoing and the page doesn't exist.  That
-			 * is equivalent to having a pagelsn of 0, so we
-			 * would not have to undo anything.  In this case,
-			 * don't bother creating a page.
-			 */
+			if (ret == DB_PAGE_NOTFOUND)
+				goto done;
+			else {
+				ret = __db_pgerr(file_dbp, argp->pgno, ret);
+				goto out;
+			}
+		}
+#ifdef HAVE_FTRUNCATE
+		/* If the page is not here then it was later truncated. */
+		if (!IS_ZERO_LSN(argp->pagelsn))
 			goto done;
-		} else if ((ret = __memp_fget(mpf,
-		    &argp->pgno, DB_MPOOL_CREATE, &pagep)) != 0)
+#endif
+		/*
+		 * This page was created by a group allocation and
+		 * the file may not have been extend yet.
+		 * Create the page if necessary.
+		 */
+		if ((ret = __memp_fget(mpf,
+		     &argp->pgno, DB_MPOOL_CREATE, &pagep)) != 0) {
+			ret = __db_pgerr(file_dbp, argp->pgno, ret);
 			goto out;
+		}
 	}
 
 	cmp_n = log_compare(lsnp, &LSN(pagep));
 	cmp_p = log_compare(&LSN(pagep), &argp->pagelsn);
-	CHECK_LSN(op, cmp_p, &LSN(pagep), &argp->pagelsn);
+	CHECK_LSN(dbenv, op, cmp_p, &LSN(pagep), &argp->pagelsn);
 
 	/*
 	 * There are two types of log messages here, one for the old page
@@ -551,29 +539,16 @@ __ham_copypage_recover(dbenv, dbtp, lsnp, op, info)
 	COMPQUIET(info, NULL);
 
 	REC_PRINT(__ham_copypage_print);
-	REC_INTRO(__ham_copypage_read, 1);
+	REC_INTRO(__ham_copypage_read, 1, 0);
 
 	flags = 0;
 
 	/* This is the bucket page. */
-	if ((ret = __memp_fget(mpf, &argp->pgno, 0, &pagep)) != 0) {
-		if (DB_UNDO(op)) {
-			/*
-			 * We are undoing and the page doesn't exist.  That
-			 * is equivalent to having a pagelsn of 0, so we
-			 * would not have to undo anything.  In this case,
-			 * don't bother creating a page.
-			 */
-			ret = 0;
-			goto donext;
-		} else if ((ret = __memp_fget(mpf,
-		    &argp->pgno, DB_MPOOL_CREATE, &pagep)) != 0)
-			goto out;
-	}
+	REC_FGET(mpf, argp->pgno, &pagep, donext);
 
 	cmp_n = log_compare(lsnp, &LSN(pagep));
 	cmp_p = log_compare(&LSN(pagep), &argp->pagelsn);
-	CHECK_LSN(op, cmp_p, &LSN(pagep), &argp->pagelsn);
+	CHECK_LSN(dbenv, op, cmp_p, &LSN(pagep), &argp->pagelsn);
 
 	if (cmp_p == 0 && DB_REDO(op)) {
 		/* Need to redo update described. */
@@ -594,25 +569,12 @@ __ham_copypage_recover(dbenv, dbtp, lsnp, op, info)
 	pagep = NULL;
 
 donext:	/* Now fix up the "next" page. */
-	if ((ret = __memp_fget(mpf, &argp->next_pgno, 0, &pagep)) != 0) {
-		if (DB_UNDO(op)) {
-			/*
-			 * We are undoing and the page doesn't exist.  That
-			 * is equivalent to having a pagelsn of 0, so we
-			 * would not have to undo anything.  In this case,
-			 * don't bother creating a page.
-			 */
-			ret = 0;
-			goto do_nn;
-		} else if ((ret = __memp_fget(mpf,
-		    &argp->next_pgno, DB_MPOOL_CREATE, &pagep)) != 0)
-			goto out;
-	}
+	REC_FGET(mpf, argp->next_pgno, &pagep, do_nn);
 
 	/* For REDO just update the LSN. For UNDO copy page back. */
 	cmp_n = log_compare(lsnp, &LSN(pagep));
 	cmp_p = log_compare(&LSN(pagep), &argp->nextlsn);
-	CHECK_LSN(op, cmp_p, &LSN(pagep), &argp->nextlsn);
+	CHECK_LSN(dbenv, op, cmp_p, &LSN(pagep), &argp->nextlsn);
 	flags = 0;
 	if (cmp_p == 0 && DB_REDO(op)) {
 		LSN(pagep) = *lsnp;
@@ -630,23 +592,11 @@ donext:	/* Now fix up the "next" page. */
 do_nn:	if (argp->nnext_pgno == PGNO_INVALID)
 		goto done;
 
-	if ((ret = __memp_fget(mpf, &argp->nnext_pgno, 0, &pagep)) != 0) {
-		if (DB_UNDO(op)) {
-			/*
-			 * We are undoing and the page doesn't exist.  That
-			 * is equivalent to having a pagelsn of 0, so we
-			 * would not have to undo anything.  In this case,
-			 * don't bother creating a page.
-			 */
-			goto done;
-		} else if ((ret = __memp_fget(mpf,
-		    &argp->nnext_pgno, DB_MPOOL_CREATE, &pagep)) != 0)
-			goto out;
-	}
+	REC_FGET(mpf, argp->nnext_pgno, &pagep, done);
 
 	cmp_n = log_compare(lsnp, &LSN(pagep));
 	cmp_p = log_compare(&LSN(pagep), &argp->nnextlsn);
-	CHECK_LSN(op, cmp_p, &LSN(pagep), &argp->nnextlsn);
+	CHECK_LSN(dbenv, op, cmp_p, &LSN(pagep), &argp->nnextlsn);
 
 	flags = 0;
 	if (cmp_p == 0 && DB_REDO(op)) {
@@ -696,17 +646,18 @@ __ham_metagroup_recover(dbenv, dbtp, lsnp, op, info)
 	PAGE *pagep;
 	db_pgno_t pgno;
 	u_int32_t flags, mmeta_flags;
-	int cmp_n, cmp_p, did_recover, groupgrow, ret;
+	int cmp_n, cmp_p, did_alloc, did_recover, groupgrow, ret;
 
 	COMPQUIET(info, NULL);
 	mmeta_flags = 0;
+	did_alloc = 0;
 	mmeta = NULL;
 	REC_PRINT(__ham_metagroup_print);
-	REC_INTRO(__ham_metagroup_read, 1);
+	REC_INTRO(__ham_metagroup_read, 1, 1);
 
 	/*
 	 * This logs the virtual create of pages pgno to pgno + bucket
-	 * Since the mpool page-allocation is not really able to be
+	 * If HAVE_FTRUNCATE is not supported the mpool page-allocation is not
 	 * transaction protected, we can never undo it.  Even in an abort,
 	 * we have to allocate these pages to the hash table if they
 	 * were actually created.  In particular, during disaster
@@ -714,42 +665,77 @@ __ham_metagroup_recover(dbenv, dbtp, lsnp, op, info)
 	 * are rolling backward.  If the file has not been extended
 	 * then the metapage could not have been updated.
 	 * The log record contains:
-	 * bucket: new bucket being allocated.
+	 * bucket: old maximum bucket
 	 * pgno: page number of the new bucket.
-	 * if bucket is a power of 2, then we allocated a whole batch of
-	 * pages; if it's not, then we simply allocated one new page.
+	 * We round up on log calculations, so we can figure out if we are
+	 * about to double the hash table if argp->bucket+1 is a power of 2.
+	 * If it is, then we are allocating an entire doubling of pages,
+	 * otherwise, we are simply allocated one new page.
 	 */
-	groupgrow = (u_int32_t)(1 << __db_log2(argp->bucket + 1)) ==
-	    argp->bucket + 1;
+	groupgrow =
+	    (u_int32_t)(1 << __db_log2(argp->bucket + 1)) == argp->bucket + 1;
 	pgno = argp->pgno;
 	if (argp->newalloc)
 		pgno += argp->bucket;
 
-	if ((ret = __memp_fget(mpf, &pgno, DB_MPOOL_CREATE, &pagep)) != 0) {
+	flags = 0;
+	pagep = NULL;
+#ifndef HAVE_FTRUNCATE
+	flags = DB_MPOOL_CREATE;
+#endif
+	ret = __memp_fget(mpf, &pgno, flags, &pagep);
+
+#ifdef HAVE_FTRUNCATE
+	/* If we are undoing, then we don't want to create the page. */
+	if (ret != 0 && DB_REDO(op))
+		ret = __memp_fget(mpf, &pgno, DB_MPOOL_CREATE, &pagep);
+	else if (ret == DB_PAGE_NOTFOUND)
+		goto do_meta;
+#endif
+	if (ret != 0) {
 		if (ret != ENOSPC)
 			goto out;
 		pgno = 0;
-		groupgrow = 0;
 		goto do_meta;
 	}
 
+	/*
+	 * When we get here then either we did not grow the file
+	 * (groupgrow == 0) or we did grow the file and the allocation
+	 * of those new pages succeeded.
+	 */
+	did_alloc = groupgrow;
+
 	cmp_n = log_compare(lsnp, &LSN(pagep));
 	cmp_p = log_compare(&LSN(pagep), &argp->pagelsn);
-	CHECK_LSN(op, cmp_p, &LSN(pagep), &argp->pagelsn);
+	CHECK_LSN(dbenv, op, cmp_p, &LSN(pagep), &argp->pagelsn);
 
 	flags = 0;
-	if ((cmp_p == 0 && DB_REDO(op)) || (cmp_n == 0 && DB_UNDO(op))) {
-		/*
-		 * We need to make sure that we redo the allocation of the
-		 * pages.
-		 */
-		if (DB_REDO(op))
-			pagep->lsn = *lsnp;
-		else
-			pagep->lsn = argp->pagelsn;
+	if (cmp_p == 0 && DB_REDO(op)) {
+		pagep->lsn = *lsnp;
 		flags = DB_MPOOL_DIRTY;
+	} else if (cmp_n == 0 && DB_UNDO(op)) {
+#ifdef HAVE_FTRUNCATE
+		/* If this record allocated the pages give them back. */
+		if (argp->newalloc) {
+			if (pagep != NULL && (ret =
+			     __memp_fput(mpf, pagep, DB_MPOOL_DISCARD)) != 0)
+				goto out;
+			pagep = NULL;
+			if ((ret = __memp_ftruncate(mpf, argp->pgno, 0)) != 0)
+				goto out;
+		} else
+#endif
+		{
+			/*
+			 * Otherwise just roll the page back to its
+			 * previous state.
+			 */
+			pagep->lsn = argp->pagelsn;
+			flags = DB_MPOOL_DIRTY;
+		}
 	}
-	if ((ret = __memp_fput(mpf, pagep, flags)) != 0)
+	if (pagep != NULL && (ret = __memp_fput(mpf, pagep, flags)) != 0)
 		goto out;
 
 do_meta:
@@ -759,7 +745,7 @@ do_meta:
 		goto out;
 	cmp_n = log_compare(lsnp, &hcp->hdr->dbmeta.lsn);
 	cmp_p = log_compare(&hcp->hdr->dbmeta.lsn, &argp->metalsn);
-	CHECK_LSN(op, cmp_p, &hcp->hdr->dbmeta.lsn, &argp->metalsn);
+	CHECK_LSN(dbenv, op, cmp_p, &hcp->hdr->dbmeta.lsn, &argp->metalsn);
 	did_recover = 0;
 	if (cmp_p == 0 && DB_REDO(op)) {
 		/* Redo the actual updating of bucket counts. */
@@ -773,9 +759,9 @@ do_meta:
 		did_recover = 1;
 	} else if (cmp_n == 0 && DB_UNDO(op)) {
 		/* Undo the actual updating of bucket counts. */
-		--hcp->hdr->max_bucket;
+		hcp->hdr->max_bucket = argp->bucket;
 		if (groupgrow) {
-			hcp->hdr->high_mask = hcp->hdr->low_mask;
+			hcp->hdr->high_mask = argp->bucket;
 			hcp->hdr->low_mask = hcp->hdr->high_mask >> 1;
 		}
 		hcp->hdr->dbmeta.lsn = argp->metalsn;
@@ -786,16 +772,26 @@ do_meta:
 	 * Now we need to fix up the spares array.  Each entry in the
 	 * spares array indicates the beginning page number for the
 	 * indicated doubling.  We need to fill this in whenever the
-	 * spares array is invalid, since we never reclaim pages from
-	 * the spares array and we have to allocate the pages to the
-	 * spares array in both the redo and undo cases.
+	 * spares array is invalid, if we never reclaim pages then
+	 * we have to allocate the pages to the spares array in both
+	 * the redo and undo cases.
 	 */
-	if (groupgrow &&
+	if (did_alloc &&
+#ifdef HAVE_FTRUNCATE
+	    !DB_UNDO(op) &&
+#endif
 	    hcp->hdr->spares[__db_log2(argp->bucket + 1) + 1] == PGNO_INVALID) {
 		hcp->hdr->spares[__db_log2(argp->bucket + 1) + 1] =
-		    argp->pgno - argp->bucket - 1;
+		    (argp->pgno - argp->bucket) - 1;
 		did_recover = 1;
 	}
+#ifdef HAVE_FTRUNCATE
+	if (cmp_n == 0 && groupgrow && DB_UNDO(op)) {
+		hcp->hdr->spares[
+		    __db_log2(argp->bucket + 1) + 1] = PGNO_INVALID;
+		did_recover = 1;
+	}
+#endif
 
 	/*
 	 * Finally, we need to potentially fix up the last_pgno field
@@ -808,16 +804,18 @@ do_meta:
 		mmeta_flags = 0;
 		cmp_n = log_compare(lsnp, &mmeta->lsn);
 		cmp_p = log_compare(&mmeta->lsn, &argp->mmetalsn);
-		if (cmp_p == 0 && DB_REDO(op)) {
+		if (cmp_p == 0 && DB_REDO(op))
 			mmeta->lsn = *lsnp;
-			mmeta_flags = DB_MPOOL_DIRTY;
-		} else if (cmp_n == 0 && DB_UNDO(op)) {
+		else if (cmp_n == 0 && DB_UNDO(op))
 			mmeta->lsn = argp->mmetalsn;
-			mmeta_flags = DB_MPOOL_DIRTY;
-		}
 	} else
 		mmeta = (DBMETA *)hcp->hdr;
 
+#ifdef HAVE_FTRUNCATE
+	if (cmp_n == 0 && DB_UNDO(op))
+		mmeta->last_pgno = argp->last_pgno;
+	else if (DB_REDO(op))
+#endif
 	if (mmeta->last_pgno < pgno)
 		mmeta->last_pgno = pgno;
 	mmeta_flags = DB_MPOOL_DIRTY;
@@ -870,23 +868,20 @@ __ham_groupalloc_recover(dbenv, dbtp, lsnp, op, info)
 	mmeta = NULL;
 	modified = 0;
 	REC_PRINT(__ham_groupalloc_print);
-	REC_INTRO(__ham_groupalloc_read, 0);
+	REC_INTRO(__ham_groupalloc_read, 0, 0);
 
 	pgno = PGNO_BASE_MD;
 	if ((ret = __memp_fget(mpf, &pgno, 0, &mmeta)) != 0) {
 		if (DB_REDO(op)) {
-			/* Page should have existed. */
 			ret = __db_pgerr(file_dbp, pgno, ret);
 			goto out;
-		} else {
-			ret = 0;
+		} else
 			goto done;
-		}
 	}
 
 	cmp_n = log_compare(lsnp, &LSN(mmeta));
 	cmp_p = log_compare(&LSN(mmeta), &argp->meta_lsn);
-	CHECK_LSN(op, cmp_p, &LSN(mmeta), &argp->meta_lsn);
+	CHECK_LSN(dbenv, op, cmp_p, &LSN(mmeta), &argp->meta_lsn);
 
 	/*
 	 * Basically, we used mpool to allocate a chunk of pages.
@@ -906,25 +901,63 @@ __ham_groupalloc_recover(dbenv, dbtp, lsnp, op, info)
 			modified = 1;
 		}
 	} else if (DB_UNDO(op)) {
+		/*
+		 * Fetch the last page and determine if it is in
+		 * the post allocation state.
+		 */
+		pagep = NULL;
+		if ((ret = __memp_fget(mpf, &pgno, 0, &pagep)) == 0) {
+			if (log_compare(&pagep->lsn, lsnp) != 0) {
+				if ((ret = __memp_fput(mpf,
+				     pagep, DB_MPOOL_DISCARD)) != 0)
+					goto out;
+				pagep = NULL;
+			}
+		} else if (ret != DB_PAGE_NOTFOUND)
+			goto out;
+#ifdef HAVE_FTRUNCATE
+		COMPQUIET(info, NULL);
+		/*
+		 * If the last page was allocated then truncate back
+		 * to the first page.
+		 */
+		if (pagep != NULL) {
+			if ((ret =
+			    __memp_fput(mpf, pagep, DB_MPOOL_DISCARD)) != 0)
+				goto out;
+			if ((ret =
+			     __memp_ftruncate(mpf, argp->start_pgno, 0)) != 0)
+				goto out;
+		}
+
+		/*
+		 * If we are rolling back the metapage, then make
+		 * sure it reflects the the correct last_pgno.
+		 */
+		if (cmp_n == 0) {
+			mmeta->last_pgno = argp->last_pgno;
+			modified = 1;
+		}
+		pgno = 0;
+#else
 		/*
 		 * Reset the last page back to its preallocation state.
 		 */
-		if ((ret = __memp_fget(mpf, &pgno, 0, &pagep)) == 0) {
-
+		if (pagep != NULL) {
 			if (log_compare(&pagep->lsn, lsnp) == 0)
 				ZERO_LSN(pagep->lsn);
 
 			if ((ret =
 			    __memp_fput(mpf, pagep, DB_MPOOL_DIRTY)) != 0)
 				goto out;
-		} else if (ret != DB_PAGE_NOTFOUND)
-			goto out;
+		}
 		/*
-		 * Always put the pages into the limbo list and free them later.
+		 * Put the pages into the limbo list and free them later.
 		 */
 		if ((ret = __db_add_limbo(dbenv,
 		    info, argp->fileid, argp->start_pgno, argp->num)) != 0)
 			goto out;
+#endif
 		if (cmp_n == 0) {
 			LSN(mmeta) = argp->meta_lsn;
 			modified = 1;
@@ -933,7 +966,8 @@ __ham_groupalloc_recover(dbenv, dbtp, lsnp, op, info)
 
 	/*
 	 * In both REDO and UNDO, we have grown the file and need to make
-	 * sure that last_pgno is correct.
+	 * sure that last_pgno is correct.  If we HAVE_FTRUNCATE pgno
+	 * will only be valid on REDO.
 	 */
 	if (pgno > mmeta->last_pgno) {
 		mmeta->last_pgno = pgno;
@@ -942,6 +976,7 @@ __ham_groupalloc_recover(dbenv, dbtp, lsnp, op, info)
 
 done:	if (ret == 0)
 		*lsnp = argp->prev_lsn;
+	ret = 0;
 
 out:	if (mmeta != NULL)
 		(void)__memp_fput(mpf, mmeta, modified ? DB_MPOOL_DIRTY : 0);
@@ -1025,15 +1060,15 @@ __ham_curadj_recover(dbenv, dbtp, lsnp, op, info)
 
 	COMPQUIET(info, NULL);
 	REC_PRINT(__ham_curadj_print);
-	REC_INTRO(__ham_curadj_read, 0);
+	REC_INTRO(__ham_curadj_read, 0, 1);
 
 	if (op != DB_TXN_ABORT)
 		goto done;
 
 	/*
-	 * Undo the adjustment by reinitializing the the cursor
-	 * to look like the one that was used to do the adustment,
-	 * then we invert the add so that undo the adjustment.
+	 * Undo the adjustment by reinitializing the the cursor to look like
+	 * the one that was used to do the adjustment, then we invert the
+	 * add so that undo the adjustment.
 	 */
 	hcp = (HASH_CURSOR *)dbc->internal;
 	hcp->pgno = argp->pgno;
@@ -1075,7 +1110,7 @@ __ham_chgpg_recover(dbenv, dbtp, lsnp, op, info)
 
 	COMPQUIET(info, NULL);
 	REC_PRINT(__ham_chgpg_print);
-	REC_INTRO(__ham_chgpg_read, 0);
+	REC_INTRO(__ham_chgpg_read, 0, 0);
 
 	if (op != DB_TXN_ABORT)
 		goto done;
@@ -1084,11 +1119,11 @@ __ham_chgpg_recover(dbenv, dbtp, lsnp, op, info)
 	indx = argp->old_indx;
 	order = argp->new_indx;
 
-	MUTEX_THREAD_LOCK(dbenv, dbenv->dblist_mutexp);
+	MUTEX_LOCK(dbenv, dbenv->mtx_dblist);
 	for (ldbp = __dblist_get(dbenv, file_dbp->adj_fileid);
 	    ldbp != NULL && ldbp->adj_fileid == file_dbp->adj_fileid;
 	    ldbp = LIST_NEXT(ldbp, dblistlinks)) {
-		MUTEX_THREAD_LOCK(dbenv, file_dbp->mutexp);
+		MUTEX_LOCK(dbenv, file_dbp->mutex);
 
 		for (cp = TAILQ_FIRST(&ldbp->active_queue); cp != NULL;
 		    cp = TAILQ_NEXT(cp, links)) {
@@ -1154,17 +1189,17 @@ __ham_chgpg_recover(dbenv, dbtp, lsnp, op, info)
 				 * the cursor we're adjusting can't be closed
 				 * under us.
 				 */
-				MUTEX_THREAD_UNLOCK(dbenv, file_dbp->mutexp);
+				MUTEX_UNLOCK(dbenv, file_dbp->mutex);
 				if ((ret = __db_c_close(lcp->opd)) != 0)
 					goto out;
-				MUTEX_THREAD_LOCK(dbenv, file_dbp->mutexp);
+				MUTEX_LOCK(dbenv, file_dbp->mutex);
 				lcp->opd = NULL;
 				break;
 			}
 		}
-		MUTEX_THREAD_UNLOCK(dbenv, file_dbp->mutexp);
+		MUTEX_UNLOCK(dbenv, file_dbp->mutex);
 	}
-	MUTEX_THREAD_UNLOCK(dbenv, dbenv->dblist_mutexp);
+	MUTEX_UNLOCK(dbenv, dbenv->mtx_dblist);
 
 done:	*lsnp = argp->prev_lsn;
 out:	REC_CLOSE;
