@@ -1,16 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: lock_region.c,v 12.6 2005/10/07 20:21:31 ubell Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: lock_region.c,v 11.73 2003/07/23 13:13:12 mjc Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -21,14 +19,10 @@ static const char revid[] = "$Id: lock_region.c,v 11.73 2003/07/23 13:13:12 mjc
 #include "dbinc/db_shash.h"
 #include "dbinc/lock.h"
 
-static int  __lock_init __P((DB_ENV *, DB_LOCKTAB *));
+static int  __lock_region_init __P((DB_ENV *, DB_LOCKTAB *));
 static size_t
 	    __lock_region_size __P((DB_ENV *));
 
-#ifdef HAVE_MUTEX_SYSTEM_RESOURCES
-static size_t __lock_region_maint __P((DB_ENV *));
-#endif
-
 /*
  * The conflict arrays are set up such that the row is the lock you are
  * holding and the column is the lock that is desired.
@@ -75,7 +69,9 @@ __lock_open(dbenv)
 	DB_LOCKREGION *region;
 	DB_LOCKTAB *lt;
 	size_t size;
-	int ret;
+	int region_locked, ret;
+
+	region_locked = 0;
 
 	/* Create the lock table structure. */
 	if ((ret = __os_calloc(dbenv, 1, sizeof(DB_LOCKTAB), &lt)) != 0)
@@ -83,9 +79,9 @@ __lock_open(dbenv)
 	lt->dbenv = dbenv;
 
 	/* Join/create the lock region. */
+	lt->reginfo.dbenv = dbenv;
 	lt->reginfo.type = REGION_TYPE_LOCK;
 	lt->reginfo.id = INVALID_REGION_ID;
-	lt->reginfo.mode = dbenv->db_mode;
 	lt->reginfo.flags = REGION_JOIN_OK;
 	if (F_ISSET(dbenv, DB_ENV_CREATE))
 		F_SET(&lt->reginfo, REGION_CREATE_OK);
@@ -95,15 +91,33 @@ __lock_open(dbenv)
 
 	/* If we created the region, initialize it. */
 	if (F_ISSET(&lt->reginfo, REGION_CREATE))
-		if ((ret = __lock_init(dbenv, lt)) != 0)
+		if ((ret = __lock_region_init(dbenv, lt)) != 0)
 			goto err;
 
 	/* Set the local addresses. */
 	region = lt->reginfo.primary =
 	    R_ADDR(&lt->reginfo, lt->reginfo.rp->primary);
 
-	/* Check for incompatible automatic deadlock detection requests. */
+	/* Set remaining pointers into region. */
+	lt->conflicts = R_ADDR(&lt->reginfo, region->conf_off);
+	lt->obj_tab = R_ADDR(&lt->reginfo, region->obj_off);
+	lt->locker_tab = R_ADDR(&lt->reginfo, region->locker_off);
+
+	dbenv->lk_handle = lt;
+
+	LOCK_SYSTEM_LOCK(dbenv);
+	region_locked = 1;
+
 	if (dbenv->lk_detect != DB_LOCK_NORUN) {
+		/*
+		 * Check for incompatible automatic deadlock detection requests.
+		 * There are scenarios where changing the detector configuration
+		 * is reasonable, but we disallow them guessing it is likely to
+		 * be an application error.
+		 *
+		 * We allow applications to turn on the lock detector, and we
+		 * ignore attempts to set it to the default or current value.
+		 */
 		if (region->detect != DB_LOCK_NORUN &&
 		    dbenv->lk_detect != DB_LOCK_DEFAULT &&
 		    region->detect != dbenv->lk_detect) {
@@ -112,12 +126,6 @@ __lock_open(dbenv)
 			ret = EINVAL;
 			goto err;
 		}
-
-		/*
-		 * Upgrade if our caller wants automatic detection, and it
-		 * was not currently being done, whether or not we created
-		 * the region.
-		 */
 		if (region->detect == DB_LOCK_NORUN)
 			region->detect = dbenv->lk_detect;
 	}
@@ -126,37 +134,33 @@ __lock_open(dbenv)
 	 * A process joining the region may have reset the lock and transaction
 	 * timeouts.
 	 */
-	 if (dbenv->lk_timeout != 0)
+	if (dbenv->lk_timeout != 0)
 		region->lk_timeout = dbenv->lk_timeout;
-	 if (dbenv->tx_timeout != 0)
+	if (dbenv->tx_timeout != 0)
 		region->tx_timeout = dbenv->tx_timeout;
 
-	/* Set remaining pointers into region. */
-	lt->conflicts = (u_int8_t *)R_ADDR(&lt->reginfo, region->conf_off);
-	lt->obj_tab = (DB_HASHTAB *)R_ADDR(&lt->reginfo, region->obj_off);
-	lt->locker_tab = (DB_HASHTAB *)R_ADDR(&lt->reginfo, region->locker_off);
+	LOCK_SYSTEM_UNLOCK(dbenv);
+	region_locked = 0;
 
-	R_UNLOCK(dbenv, &lt->reginfo);
-
-	dbenv->lk_handle = lt;
 	return (0);
 
-err:	if (lt->reginfo.addr != NULL) {
-		if (F_ISSET(&lt->reginfo, REGION_CREATE))
-			ret = __db_panic(dbenv, ret);
-		R_UNLOCK(dbenv, &lt->reginfo);
+err:	dbenv->lk_handle = NULL;
+	if (lt->reginfo.addr != NULL) {
+		if (region_locked)
+			LOCK_SYSTEM_UNLOCK(dbenv);
 		(void)__db_r_detach(dbenv, &lt->reginfo, 0);
 	}
+
 	__os_free(dbenv, lt);
 	return (ret);
 }
 
 /*
- * __lock_init --
+ * __lock_region_init --
  *	Initialize the lock region.
  */
 static int
-__lock_init(dbenv, lt)
+__lock_region_init(dbenv, lt)
 	DB_ENV *dbenv;
 	DB_LOCKTAB *lt;
 {
@@ -165,20 +169,21 @@ __lock_init(dbenv, lt)
 	DB_LOCKER *lidp;
 	DB_LOCKOBJ *op;
 	DB_LOCKREGION *region;
-#ifdef HAVE_MUTEX_SYSTEM_RESOURCES
-	size_t maint_size;
-#endif
-	u_int32_t i, lk_modes;
+	u_int32_t i;
 	u_int8_t *addr;
-	int ret;
+	int lk_modes, ret;
 
-	if ((ret = __db_shalloc(lt->reginfo.addr,
+	if ((ret = __db_shalloc(&lt->reginfo,
 	    sizeof(DB_LOCKREGION), 0, &lt->reginfo.primary)) != 0)
 		goto mem_err;
 	lt->reginfo.rp->primary = R_OFFSET(&lt->reginfo, lt->reginfo.primary);
 	region = lt->reginfo.primary;
 	memset(region, 0, sizeof(*region));
 
+	if ((ret = __mutex_alloc(
+	    dbenv, MTX_LOCK_REGION, 0, &region->mtx_region)) != 0)
+		return (ret);
+
 	/* Select a conflict matrix if none specified. */
 	if (dbenv->lk_modes == 0)
 		if (CDB_LOCKING(dbenv)) {
@@ -209,53 +214,35 @@ __lock_init(dbenv, lt)
 	region->stat.st_nmodes = lk_modes;
 
 	/* Allocate room for the conflict matrix and initialize it. */
-	if ((ret =
-	    __db_shalloc(lt->reginfo.addr, lk_modes * lk_modes, 0, &addr)) != 0)
+	if ((ret = __db_shalloc(
+	    &lt->reginfo, (size_t)(lk_modes * lk_modes), 0, &addr)) != 0)
 		goto mem_err;
-	memcpy(addr, lk_conflicts, lk_modes * lk_modes);
+	memcpy(addr, lk_conflicts, (size_t)(lk_modes * lk_modes));
 	region->conf_off = R_OFFSET(&lt->reginfo, addr);
 
 	/* Allocate room for the object hash table and initialize it. */
-	if ((ret = __db_shalloc(lt->reginfo.addr,
+	if ((ret = __db_shalloc(&lt->reginfo,
 	    region->object_t_size * sizeof(DB_HASHTAB), 0, &addr)) != 0)
 		goto mem_err;
 	__db_hashinit(addr, region->object_t_size);
 	region->obj_off = R_OFFSET(&lt->reginfo, addr);
 
 	/* Allocate room for the locker hash table and initialize it. */
-	if ((ret = __db_shalloc(lt->reginfo.addr,
+	if ((ret = __db_shalloc(&lt->reginfo,
 	    region->locker_t_size * sizeof(DB_HASHTAB), 0, &addr)) != 0)
 		goto mem_err;
 	__db_hashinit(addr, region->locker_t_size);
 	region->locker_off = R_OFFSET(&lt->reginfo, addr);
 
-#ifdef	HAVE_MUTEX_SYSTEM_RESOURCES
-	maint_size = __lock_region_maint(dbenv);
-	/* Allocate room for the locker maintenance info and initialize it. */
-	if ((ret = __db_shalloc(lt->reginfo.addr,
-	    sizeof(REGMAINT) + maint_size, 0, &addr)) != 0)
-		goto mem_err;
-	__db_maintinit(&lt->reginfo, addr, maint_size);
-	region->maint_off = R_OFFSET(&lt->reginfo, addr);
-#endif
-
-	/*
-	 * Initialize locks onto a free list. Initialize and lock the mutex
-	 * so that when we need to block, all we need do is try to acquire
-	 * the mutex.
-	 */
+	/* Initialize locks onto a free list. */
 	SH_TAILQ_INIT(&region->free_locks);
 	for (i = 0; i < region->stat.st_maxlocks; ++i) {
-		if ((ret = __db_shalloc(lt->reginfo.addr,
-		    sizeof(struct __db_lock), MUTEX_ALIGN, &lp)) != 0)
+		if ((ret = __db_shalloc(&lt->reginfo,
+		    sizeof(struct __db_lock), 0, &lp)) != 0)
 			goto mem_err;
-		lp->status = DB_LSTAT_FREE;
+		lp->mtx_lock = MUTEX_INVALID;
 		lp->gen = 0;
-		if ((ret = __db_mutex_setup(dbenv, &lt->reginfo, &lp->mutex,
-		    MUTEX_LOGICAL_LOCK | MUTEX_NO_RLOCK | MUTEX_SELF_BLOCK))
-		    != 0)
-			return (ret);
-		MUTEX_LOCK(dbenv, &lp->mutex);
+		lp->status = DB_LSTAT_FREE;
 		SH_TAILQ_INSERT_HEAD(&region->free_locks, lp, links, __db_lock);
 	}
 
@@ -263,7 +250,7 @@ __lock_init(dbenv, lt)
 	SH_TAILQ_INIT(&region->dd_objs);
 	SH_TAILQ_INIT(&region->free_objs);
 	for (i = 0; i < region->stat.st_maxobjects; ++i) {
-		if ((ret = __db_shalloc(lt->reginfo.addr,
+		if ((ret = __db_shalloc(&lt->reginfo,
 		    sizeof(DB_LOCKOBJ), 0, &op)) != 0)
 			goto mem_err;
 		SH_TAILQ_INSERT_HEAD(
@@ -274,7 +261,7 @@ __lock_init(dbenv, lt)
 	SH_TAILQ_INIT(&region->lockers);
 	SH_TAILQ_INIT(&region->free_lockers);
 	for (i = 0; i < region->stat.st_maxlockers; ++i) {
-		if ((ret = __db_shalloc(lt->reginfo.addr,
+		if ((ret = __db_shalloc(&lt->reginfo,
 		    sizeof(DB_LOCKER), 0, &lidp)) != 0) {
 mem_err:		__db_err(dbenv,
 			    "Unable to allocate memory for the lock table");
@@ -289,8 +276,7 @@ mem_err:		__db_err(dbenv,
 
 /*
  * __lock_dbenv_refresh --
- *	Clean up after the lock system on a close or failed open.  Called
- * only from __dbenv_refresh.  (Formerly called __lock_close.)
+ *	Clean up after the lock system on a close or failed open.
  *
  * PUBLIC: int __lock_dbenv_refresh __P((DB_ENV *));
  */
@@ -298,20 +284,80 @@ int
 __lock_dbenv_refresh(dbenv)
 	DB_ENV *dbenv;
 {
+	struct __db_lock *lp;
+	DB_LOCKER *locker;
+	DB_LOCKOBJ *lockobj;
+	DB_LOCKREGION *lr;
 	DB_LOCKTAB *lt;
+	REGINFO *reginfo;
 	int ret;
 
 	lt = dbenv->lk_handle;
+	reginfo = &lt->reginfo;
+	lr = reginfo->primary;
+
+	/*
+	 * If a private region, return the memory to the heap.  Not needed for
+	 * filesystem-backed or system shared memory regions, that memory isn't
+	 * owned by any particular process.
+	 */
+	if (F_ISSET(dbenv, DB_ENV_PRIVATE)) {
+		/* Discard the conflict matrix. */
+		__db_shalloc_free(reginfo, R_ADDR(reginfo, lr->conf_off));
+
+		/* Discard the object hash table. */
+		__db_shalloc_free(reginfo, R_ADDR(reginfo, lr->obj_off));
+
+		/* Discard the locker hash table. */
+		__db_shalloc_free(reginfo, R_ADDR(reginfo, lr->locker_off));
+
+		/* Discard locks. */
+		while ((lp =
+		    SH_TAILQ_FIRST(&lr->free_locks, __db_lock)) != NULL) {
+			SH_TAILQ_REMOVE(&lr->free_locks, lp, links, __db_lock);
+			__db_shalloc_free(reginfo, lp);
+		}
+
+		/* Discard objects. */
+		while ((lockobj =
+		    SH_TAILQ_FIRST(&lr->free_objs, __db_lockobj)) != NULL) {
+			SH_TAILQ_REMOVE(
+			    &lr->free_objs, lockobj, links, __db_lockobj);
+			__db_shalloc_free(reginfo, lockobj);
+		}
+
+		/* Discard lockers. */
+		while ((locker =
+		    SH_TAILQ_FIRST(&lr->free_lockers, __db_locker)) != NULL) {
+			SH_TAILQ_REMOVE(
+			    &lr->free_lockers, locker, links, __db_locker);
+			__db_shalloc_free(reginfo, locker);
+		}
+	}
 
 	/* Detach from the region. */
-	ret = __db_r_detach(dbenv, &lt->reginfo, 0);
+	ret = __db_r_detach(dbenv, reginfo, 0);
 
+	/* Discard DB_LOCKTAB. */
 	__os_free(dbenv, lt);
-
 	dbenv->lk_handle = NULL;
+
 	return (ret);
 }
 
+/*
+ * __lock_region_mutex_count --
+ *	Return the number of mutexes the lock region will need.
+ *
+ * PUBLIC: u_int32_t __lock_region_mutex_count __P((DB_ENV *));
+ */
+u_int32_t
+__lock_region_mutex_count(dbenv)
+	DB_ENV *dbenv;
+{
+	return (dbenv->lk_max);
+}
+
 /*
  * __lock_region_size --
  *	Return the region size.
@@ -324,25 +370,22 @@ __lock_region_size(dbenv)
 
 	/*
 	 * Figure out how much space we're going to need.  This list should
-	 * map one-to-one with the __db_shalloc calls in __lock_init.
+	 * map one-to-one with the __db_shalloc calls in __lock_region_init.
 	 */
 	retval = 0;
-	retval += __db_shalloc_size(sizeof(DB_LOCKREGION), 1);
-	retval += __db_shalloc_size(dbenv->lk_modes * dbenv->lk_modes, 1);
+	retval += __db_shalloc_size(sizeof(DB_LOCKREGION), 0);
 	retval += __db_shalloc_size(
-	    __db_tablesize(dbenv->lk_max_lockers) * (sizeof(DB_HASHTAB)), 1);
+	    (size_t)(dbenv->lk_modes * dbenv->lk_modes), 0);
 	retval += __db_shalloc_size(
-	    __db_tablesize(dbenv->lk_max_objects) * (sizeof(DB_HASHTAB)), 1);
-#ifdef HAVE_MUTEX_SYSTEM_RESOURCES
+	    __db_tablesize(dbenv->lk_max_objects) * (sizeof(DB_HASHTAB)), 0);
+	retval += __db_shalloc_size(
+	    __db_tablesize(dbenv->lk_max_lockers) * (sizeof(DB_HASHTAB)), 0);
 	retval +=
-	    __db_shalloc_size(sizeof(REGMAINT) + __lock_region_maint(dbenv), 1);
-#endif
-	retval += __db_shalloc_size(
-	    sizeof(struct __db_lock), MUTEX_ALIGN) * dbenv->lk_max;
-	retval += __db_shalloc_size(
-	    sizeof(DB_LOCKOBJ), 1) * dbenv->lk_max_objects;
-	retval += __db_shalloc_size(
-	    sizeof(DB_LOCKER), 1) * dbenv->lk_max_lockers;
+	    __db_shalloc_size(sizeof(struct __db_lock), 0) * dbenv->lk_max;
+	retval +=
+	    __db_shalloc_size(sizeof(DB_LOCKOBJ), 0) * dbenv->lk_max_objects;
+	retval +=
+	    __db_shalloc_size(sizeof(DB_LOCKER), 0) * dbenv->lk_max_lockers;
 
 	/*
 	 * Include 16 bytes of string space per lock.  DB doesn't use it
@@ -355,63 +398,3 @@ __lock_region_size(dbenv)
 
 	return (retval);
 }
-
-#ifdef HAVE_MUTEX_SYSTEM_RESOURCES
-/*
- * __lock_region_maint --
- *	Return the amount of space needed for region maintenance info.
- */
-static size_t
-__lock_region_maint(dbenv)
-	DB_ENV *dbenv;
-{
-	size_t s;
-
-	s = sizeof(DB_MUTEX *) * dbenv->lk_max;
-	return (s);
-}
-#endif
-
-/*
- * __lock_region_destroy
- *	Destroy any region maintenance info.
- *
- * PUBLIC: void __lock_region_destroy __P((DB_ENV *, REGINFO *));
- */
-void
-__lock_region_destroy(dbenv, infop)
-	DB_ENV *dbenv;
-	REGINFO *infop;
-{
-	__db_shlocks_destroy(infop, (REGMAINT *)R_ADDR(infop,
-	    ((DB_LOCKREGION *)R_ADDR(infop, infop->rp->primary))->maint_off));
-
-	COMPQUIET(dbenv, NULL);
-	COMPQUIET(infop, NULL);
-}
-
-/*
- * __lock_id_set --
- *	Set the current locker ID and current maximum unused ID (for
- *	testing purposes only).
- *
- * PUBLIC: int __lock_id_set __P((DB_ENV *, u_int32_t, u_int32_t));
- */
-int
-__lock_id_set(dbenv, cur_id, max_id)
-	DB_ENV *dbenv;
-	u_int32_t cur_id, max_id;
-{
-	DB_LOCKTAB *lt;
-	DB_LOCKREGION *region;
-
-	ENV_REQUIRES_CONFIG(dbenv,
-	    dbenv->lk_handle, "lock_id_set", DB_INIT_LOCK);
-
-	lt = dbenv->lk_handle;
-	region = lt->reginfo.primary;
-	region->stat.st_id = cur_id;
-	region->stat.st_cur_maxid = max_id;
-
-	return (0);
-}
