@@ -1,16 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: lock_util.c,v 12.2 2005/06/16 20:23:11 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: lock_util.c,v 11.9 2003/01/08 05:22:12 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -123,16 +121,16 @@ __lock_lhash(lock_obj)
 }
 
 /*
- * __lock_locker_hash --
- *	Hash function for entering lockers into the locker hash table.
- *	Since these are simply 32-bit unsigned integers, just return
- *	the locker value.
+ * __lock_nomem --
+ *	Report a lack of some resource.
  *
- * PUBLIC: u_int32_t __lock_locker_hash __P((u_int32_t));
+ * PUBLIC: int __lock_nomem __P((DB_ENV *, const char *));
  */
-u_int32_t
-__lock_locker_hash(locker)
-	u_int32_t locker;
+int
+__lock_nomem(dbenv, res)
+	DB_ENV *dbenv;
+	const char *res;
 {
-	return (locker);
+	__db_err(dbenv, "Lock table is out of available %s", res);
+	return (ENOMEM);
 }
