@@ -1,22 +1,19 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2002
+ * Copyright (c) 1997-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: log_archive.c,v 12.9 2005/11/04 17:27:58 ubell Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: log_archive.c,v 11.39 2002/08/06 05:00:31 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
 #include <stdlib.h>
 #include <string.h>
-#include <unistd.h>
 #endif
 
 #include "db_int.h"
@@ -28,38 +25,29 @@ static const char revid[] = "$Id: log_archive.c,v 11.39 2002/08/06 05:00:31 bost
 static int __absname __P((DB_ENV *, char *, char *, char **));
 static int __build_data __P((DB_ENV *, char *, char ***));
 static int __cmpfunc __P((const void *, const void *));
+static int __log_archive __P((DB_ENV *, char **[], u_int32_t));
 static int __usermem __P((DB_ENV *, char ***));
 
 /*
- * __log_archive --
- *	Supporting function for db_archive(1).
+ * __log_archive_pp --
+ *	DB_ENV->log_archive pre/post processing.
  *
- * PUBLIC: int __log_archive __P((DB_ENV *, char **[], u_int32_t));
+ * PUBLIC: int __log_archive_pp __P((DB_ENV *, char **[], u_int32_t));
  */
 int
-__log_archive(dbenv, listp, flags)
+__log_archive_pp(dbenv, listp, flags)
 	DB_ENV *dbenv;
 	char ***listp;
 	u_int32_t flags;
 {
-	DBT rec;
-	DB_LOG *dblp;
-	DB_LOGC *logc;
-	DB_LSN stable_lsn;
-	__txn_ckp_args *ckp_args;
-	char **array, **arrayp, *name, *p, *pref, buf[MAXPATHLEN];
-	int array_size, db_arch_abs, n, ret;
-	u_int32_t fnum;
+	DB_THREAD_INFO *ip;
+	int ret;
 
 	PANIC_CHECK(dbenv);
 	ENV_REQUIRES_CONFIG(dbenv,
 	    dbenv->lg_handle, "DB_ENV->log_archive", DB_INIT_LOG);
 
-	name = NULL;
-	dblp = dbenv->lg_handle;
-	COMPQUIET(fnum, 0);
-
-#define	OKFLAGS	(DB_ARCH_ABS | DB_ARCH_DATA | DB_ARCH_LOG)
+#define	OKFLAGS	(DB_ARCH_ABS | DB_ARCH_DATA | DB_ARCH_LOG | DB_ARCH_REMOVE)
 	if (flags != 0) {
 		if ((ret = __db_fchk(
 		    dbenv, "DB_ENV->log_archive", flags, OKFLAGS)) != 0)
@@ -67,89 +55,111 @@ __log_archive(dbenv, listp, flags)
 		if ((ret = __db_fcchk(dbenv, "DB_ENV->log_archive",
 		    flags, DB_ARCH_DATA, DB_ARCH_LOG)) != 0)
 			return (ret);
+		if ((ret = __db_fcchk(dbenv, "DB_ENV->log_archive",
+		    flags, DB_ARCH_REMOVE,
+		    DB_ARCH_ABS | DB_ARCH_DATA | DB_ARCH_LOG)) != 0)
+			return (ret);
 	}
 
-	if (LF_ISSET(DB_ARCH_ABS)) {
-		db_arch_abs = 1;
-		LF_CLR(DB_ARCH_ABS);
-	} else
-		db_arch_abs = 0;
+	ENV_ENTER(dbenv, ip);
+	REPLICATION_WRAP(dbenv, (__log_archive(dbenv, listp, flags)), ret);
+	ENV_LEAVE(dbenv, ip);
+	return (ret);
+}
 
-	if (flags == 0 || flags == DB_ARCH_DATA)
-		ENV_REQUIRES_CONFIG(dbenv,
-		    dbenv->tx_handle, "DB_ENV->log_archive", DB_INIT_TXN);
+/*
+ * __log_archive --
+ *	DB_ENV->log_archive.  Internal.
+ */
+static int
+__log_archive(dbenv, listp, flags)
+	DB_ENV *dbenv;
+	char ***listp;
+	u_int32_t flags;
+{
+	DBT rec;
+	DB_LOG *dblp;
+	LOG *lp;
+	DB_LOGC *logc;
+	DB_LSN stable_lsn;
+	u_int array_size, n;
+	u_int32_t fnum;
+	int ret, t_ret;
+	char **array, **arrayp, *name, *p, *pref;
+
+	dblp = dbenv->lg_handle;
+	lp = (LOG *)dblp->reginfo.primary;
+	array = NULL;
+	name = NULL;
+	ret = 0;
+	COMPQUIET(fnum, 0);
+
+	if (flags != DB_ARCH_REMOVE)
+		*listp = NULL;
+
+	/* There are no log files if logs are in memory. */
+	if (lp->db_log_inmemory) {
+		LF_CLR(~DB_ARCH_DATA);
+		if (flags == 0)
+			return (0);
+	}
 
 	/*
-	 * Get the absolute pathname of the current directory.  It would
-	 * be nice to get the shortest pathname of the database directory,
-	 * but that's just not possible.
-	 *
-	 * XXX
-	 * Can't trust getcwd(3) to set a valid errno.  If it doesn't, just
-	 * guess that we ran out of memory.
+	 * If the user wants the list of log files to remove and we're
+	 * at a bad time in replication initialization, just return.
 	 */
-	if (db_arch_abs) {
-		__os_set_errno(0);
-		if ((pref = getcwd(buf, sizeof(buf))) == NULL) {
-			if (__os_get_errno() == 0)
-				__os_set_errno(ENOMEM);
-			return (__os_get_errno());
-		}
-	} else
-		pref = NULL;
+	if (!LF_ISSET(DB_ARCH_DATA) &&
+	    !LF_ISSET(DB_ARCH_LOG) && __rep_noarchive(dbenv))
+		return (0);
 
+	/*
+	 * Prepend the original absolute pathname if the user wants an
+	 * absolute path to the database environment directory.
+	 */
+	pref = LF_ISSET(DB_ARCH_ABS) ? dbenv->db_abshome : NULL;
+
+	LF_CLR(DB_ARCH_ABS);
 	switch (flags) {
 	case DB_ARCH_DATA:
-		return (__build_data(dbenv, pref, listp));
+		ret = __build_data(dbenv, pref, listp);
+		goto err;
 	case DB_ARCH_LOG:
 		memset(&rec, 0, sizeof(rec));
-		if ((ret = dbenv->log_cursor(dbenv, &logc, 0)) != 0)
-			return (ret);
+		if ((ret = __log_cursor(dbenv, &logc)) != 0)
+			goto err;
 #ifdef UMRW
 		ZERO_LSN(stable_lsn);
 #endif
-		ret = logc->get(logc, &stable_lsn, &rec, DB_LAST);
-		(void)logc->close(logc, 0);
+		ret = __log_c_get(logc, &stable_lsn, &rec, DB_LAST);
+		if ((t_ret = __log_c_close(logc)) != 0 && ret == 0)
+			ret = t_ret;
 		if (ret != 0)
-			return (ret);
+			goto err;
 		fnum = stable_lsn.file;
 		break;
+	case DB_ARCH_REMOVE:
+		__log_autoremove(dbenv);
+		goto err;
 	case 0:
-		memset(&rec, 0, sizeof(rec));
-		if (__txn_getckp(dbenv, &stable_lsn) != 0) {
-			/*
-			 * A failure return means that there's no checkpoint
-			 * in the log (so we are not going to be deleting
-			 * any log files).
-			 */
-			*listp = NULL;
-			return (0);
-		}
-		if ((ret = dbenv->log_cursor(dbenv, &logc, 0)) != 0)
-			return (ret);
-		if ((ret = logc->get(logc, &stable_lsn, &rec, DB_SET)) != 0 ||
-		    (ret = __txn_ckp_read(dbenv, rec.data, &ckp_args)) != 0) {
-			/*
-			 * A return of DB_NOTFOUND may only mean that the
-			 * checkpoint LSN is before the beginning of the
-			 * log files that we still have.  This is not
-			 * an error;  it just means our work is done.
-			 */
-			if (ret == DB_NOTFOUND) {
-				*listp = NULL;
-				ret = 0;
-			}
-			(void)logc->close(logc, 0);
-			return (ret);
-		}
-		if ((ret = logc->close(logc, 0)) != 0)
-			return (ret);
-		stable_lsn = ckp_args->ckp_lsn;
-		__os_free(dbenv, ckp_args);
 
+		ret = __log_get_stable_lsn(dbenv, &stable_lsn);
+		/*
+		 * A return of DB_NOTFOUND means the checkpoint LSN
+		 * is before the beginning of the log files we have.
+		 * This is not an error; it just means we're done.
+		 */
+		if (ret != 0) {
+			if (ret == DB_NOTFOUND)
+				ret = 0;
+			goto err;
+		}
 		/* Remove any log files before the last stable LSN. */
 		fnum = stable_lsn.file - 1;
 		break;
+	default:
+		DB_ASSERT(0);
+		ret = EINVAL;
+		goto err;
 	}
 
 #define	LIST_INCREMENT	64
@@ -157,7 +167,7 @@ __log_archive(dbenv, listp, flags)
 	array_size = 64;
 	if ((ret = __os_malloc(dbenv,
 	    sizeof(char *) * array_size, &array)) != 0)
-		return (ret);
+		goto err;
 	array[0] = NULL;
 
 	/* Build an array of the file names. */
@@ -179,9 +189,9 @@ __log_archive(dbenv, listp, flags)
 				goto err;
 		}
 
-		if (db_arch_abs) {
-			if ((ret = __absname(dbenv,
-			    pref, name, &array[n])) != 0)
+		if (pref != NULL) {
+			if ((ret =
+			    __absname(dbenv, pref, name, &array[n])) != 0)
 				goto err;
 			__os_free(dbenv, name);
 		} else if ((p = __db_rpath(name)) != NULL) {
@@ -196,11 +206,8 @@ __log_archive(dbenv, listp, flags)
 	}
 
 	/* If there's nothing to return, we're done. */
-	if (n == 0) {
-		*listp = NULL;
-		ret = 0;
+	if (n == 0)
 		goto err;
-	}
 
 	/* Sort the list. */
 	qsort(array, (size_t)n, sizeof(char *), __cmpfunc);
@@ -209,19 +216,116 @@ __log_archive(dbenv, listp, flags)
 	if ((ret = __usermem(dbenv, &array)) != 0)
 		goto err;
 
-	*listp = array;
-	return (0);
+	if (listp != NULL)
+		*listp = array;
 
-err:	if (array != NULL) {
-		for (arrayp = array; *arrayp != NULL; ++arrayp)
-			__os_free(dbenv, *arrayp);
-		__os_free(dbenv, array);
+	if (0) {
+err:		if (array != NULL) {
+			for (arrayp = array; *arrayp != NULL; ++arrayp)
+				__os_free(dbenv, *arrayp);
+			__os_free(dbenv, array);
+		}
+		if (name != NULL)
+			__os_free(dbenv, name);
 	}
-	if (name != NULL)
-		__os_free(dbenv, name);
+
 	return (ret);
 }
 
+/*
+ * __log_get_stable_lsn --
+ *	Get the stable lsn based on where checkpoints are.
+ *
+ * PUBLIC: int __log_get_stable_lsn __P((DB_ENV *, DB_LSN *));
+ */
+int
+__log_get_stable_lsn(dbenv, stable_lsn)
+	DB_ENV *dbenv;
+	DB_LSN *stable_lsn;
+{
+	DB_LOGC *logc;
+	DBT rec;
+	__txn_ckp_args *ckp_args;
+	int ret, t_ret;
+
+	ret = 0;
+	memset(&rec, 0, sizeof(rec));
+	if (!TXN_ON(dbenv)) {
+		if ((ret = __log_get_cached_ckp_lsn(dbenv, stable_lsn)) != 0)
+			goto err;
+		/*
+		 * No need to check for a return value of DB_NOTFOUND;
+		 * __txn_findlastckp returns 0 if no checkpoint record
+		 * is found.  Instead of checking the return value, we
+		 * check to see if the return LSN has been filled in.
+		 */
+		if (IS_ZERO_LSN(*stable_lsn) && (ret =
+		     __txn_findlastckp(dbenv, stable_lsn, NULL)) != 0)
+			goto err;
+		/*
+		 * If the LSN has not been filled in return DB_NOTFOUND
+		 * so that the caller knows it may be done.
+		 */
+		if (IS_ZERO_LSN(*stable_lsn)) {
+			ret = DB_NOTFOUND;
+			goto err;
+		}
+	} else if ((ret = __txn_getckp(dbenv, stable_lsn)) != 0)
+		goto err;
+	if ((ret = __log_cursor(dbenv, &logc)) != 0)
+		goto err;
+	/*
+	 * If we can read it, set the stable_lsn to the ckp_lsn in the
+	 * checkpoint record.
+	 */
+	if ((ret = __log_c_get(logc, stable_lsn, &rec, DB_SET)) == 0 &&
+	    (ret = __txn_ckp_read(dbenv, rec.data, &ckp_args)) == 0) {
+		*stable_lsn = ckp_args->ckp_lsn;
+		__os_free(dbenv, ckp_args);
+	}
+	if ((t_ret = __log_c_close(logc)) != 0 && ret == 0)
+		ret = t_ret;
+err:
+	return (ret);
+}
+
+/*
+ * __log_autoremove --
+ *	Delete any non-essential log files.
+ *
+ * PUBLIC: void __log_autoremove __P((DB_ENV *));
+ */
+void
+__log_autoremove(dbenv)
+	DB_ENV *dbenv;
+{
+	int ret;
+	char **begin, **list;
+
+	/*
+	 * Complain if there's an error, but don't return the error to our
+	 * caller.  Auto-remove is done when writing a log record, and we
+	 * don't want to fail a write, which could fail the corresponding
+	 * committing transaction, for a permissions error.
+	 */
+	if ((ret = __log_archive(dbenv, &list, DB_ARCH_ABS)) != 0) {
+		if (ret != DB_NOTFOUND)
+			__db_err(dbenv,
+			    "log file auto-remove: %s", db_strerror(ret));
+		return;
+	}
+
+	/*
+	 * Remove the files.  No error message needed for __os_unlink failure,
+	 * the underlying OS layer has its own error handling.
+	 */
+	if (list != NULL) {
+		for (begin = list; *list != NULL; ++list)
+			(void)__os_unlink(dbenv, *list);
+		__os_ufree(dbenv, begin);
+	}
+}
+
 /*
  * __build_data --
  *	Build a list of datafiles for return.
@@ -235,8 +339,9 @@ __build_data(dbenv, pref, listp)
 	DB_LOGC *logc;
 	DB_LSN lsn;
 	__dbreg_register_args *argp;
+	u_int array_size, last, n, nxt;
 	u_int32_t rectype;
-	int array_size, last, n, nxt, ret, t_ret;
+	int ret, t_ret;
 	char **array, **arrayp, **list, **lp, *p, *real_name;
 
 	/* Get some initial space. */
@@ -247,13 +352,13 @@ __build_data(dbenv, pref, listp)
 	array[0] = NULL;
 
 	memset(&rec, 0, sizeof(rec));
-	if ((ret = dbenv->log_cursor(dbenv, &logc, 0)) != 0)
+	if ((ret = __log_cursor(dbenv, &logc)) != 0)
 		return (ret);
-	for (n = 0; (ret = logc->get(logc, &lsn, &rec, DB_PREV)) == 0;) {
+	for (n = 0; (ret = __log_c_get(logc, &lsn, &rec, DB_PREV)) == 0;) {
 		if (rec.size < sizeof(rectype)) {
 			ret = EINVAL;
 			__db_err(dbenv, "DB_ENV->log_archive: bad log record");
-			goto free_continue;
+			break;
 		}
 
 		memcpy(&rectype, rec.data, sizeof(rectype));
@@ -264,7 +369,7 @@ __build_data(dbenv, pref, listp)
 			ret = EINVAL;
 			__db_err(dbenv,
 			    "DB_ENV->log_archive: unable to read log record");
-			goto free_continue;
+			break;
 		}
 
 		if (n >= array_size - 2) {
@@ -306,7 +411,7 @@ free_continue:	__os_free(dbenv, argp);
 	}
 	if (ret == DB_NOTFOUND)
 		ret = 0;
-	if ((t_ret = logc->close(logc, 0)) != 0 && ret == 0)
+	if ((t_ret = __log_c_close(logc)) != 0 && ret == 0)
 		ret = t_ret;
 	if (ret != 0)
 		goto err1;
