@@ -1,14 +1,13 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: log_get.c,v 12.16 2005/10/21 17:13:42 bostic Exp $
  */
-#include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: log_get.c,v 11.98 2003/09/13 19:20:38 bostic Exp $";
-#endif /* not lint */
+#include "db_config.h"
 
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
@@ -31,11 +30,11 @@ static int __log_c_get_int __P((DB_LOGC *, DB_LSN *, DBT *, u_int32_t));
 static int __log_c_hdrchk __P((DB_LOGC *, DB_LSN *, HDR *, int *));
 static int __log_c_incursor __P((DB_LOGC *, DB_LSN *, HDR *, u_int8_t **));
 static int __log_c_inregion __P((DB_LOGC *,
-	       DB_LSN *, RLOCK *, DB_LSN *, HDR *, u_int8_t **));
+	       DB_LSN *, RLOCK *, DB_LSN *, HDR *, u_int8_t **, int *));
 static int __log_c_io __P((DB_LOGC *,
 	       u_int32_t, u_int32_t, void *, size_t *, int *));
 static int __log_c_ondisk __P((DB_LOGC *,
-	       DB_LSN *, DB_LSN *, int, HDR *, u_int8_t **, int *));
+	       DB_LSN *, DB_LSN *, u_int32_t, HDR *, u_int8_t **, int *));
 static int __log_c_set_maxrec __P((DB_LOGC *, char *));
 static int __log_c_shortread __P((DB_LOGC *, DB_LSN *, int));
 
@@ -51,7 +50,8 @@ __log_cursor_pp(dbenv, logcp, flags)
 	DB_LOGC **logcp;
 	u_int32_t flags;
 {
-	int rep_check, ret;
+	DB_THREAD_INFO *ip;
+	int ret;
 
 	PANIC_CHECK(dbenv);
 	ENV_REQUIRES_CONFIG(dbenv,
@@ -61,12 +61,9 @@ __log_cursor_pp(dbenv, logcp, flags)
 	if ((ret = __db_fchk(dbenv, "DB_ENV->log_cursor", flags, 0)) != 0)
 		return (ret);
 
-	rep_check = IS_ENV_REPLICATED(dbenv) ? 1 : 0;
-	if (rep_check)
-		__env_rep_enter(dbenv);
-	ret = __log_cursor(dbenv, logcp);
-	if (rep_check)
-		__env_rep_exit(dbenv);
+	ENV_ENTER(dbenv, ip);
+	REPLICATION_WRAP(dbenv, (__log_cursor(dbenv, logcp)), ret);
+	ENV_LEAVE(dbenv, ip);
 	return (ret);
 }
 
@@ -117,8 +114,9 @@ __log_c_close_pp(logc, flags)
 	DB_LOGC *logc;
 	u_int32_t flags;
 {
+	DB_THREAD_INFO *ip;
 	DB_ENV *dbenv;
-	int rep_check, ret;
+	int ret;
 
 	dbenv = logc->dbenv;
 
@@ -126,12 +124,9 @@ __log_c_close_pp(logc, flags)
 	if ((ret = __db_fchk(dbenv, "DB_LOGC->close", flags, 0)) != 0)
 		return (ret);
 
-	rep_check = IS_ENV_REPLICATED(dbenv) ? 1 : 0;
-	if (rep_check)
-		__env_rep_enter(dbenv);
-	ret = __log_c_close(logc);
-	if (rep_check)
-		__env_rep_exit(dbenv);
+	ENV_ENTER(dbenv, ip);
+	REPLICATION_WRAP(dbenv, (__log_c_close(logc)), ret);
+	ENV_LEAVE(dbenv, ip);
 	return (ret);
 }
 
@@ -175,7 +170,8 @@ __log_c_get_pp(logc, alsn, dbt, flags)
 	u_int32_t flags;
 {
 	DB_ENV *dbenv;
-	int rep_check, ret;
+	DB_THREAD_INFO *ip;
+	int ret;
 
 	dbenv = logc->dbenv;
 
@@ -200,12 +196,9 @@ __log_c_get_pp(logc, alsn, dbt, flags)
 		return (__db_ferr(dbenv, "DB_LOGC->get", 1));
 	}
 
-	rep_check = IS_ENV_REPLICATED(dbenv) ? 1 : 0;
-	if (rep_check)
-		__env_rep_enter(dbenv);
-	ret = __log_c_get(logc, alsn, dbt, flags);
-	if (rep_check)
-		__env_rep_exit(dbenv);
+	ENV_ENTER(dbenv, ip);
+	REPLICATION_WRAP(dbenv, (__log_c_get(logc, alsn, dbt, flags)), ret);
+	ENV_LEAVE(dbenv, ip);
 	return (ret);
 }
 
@@ -262,6 +255,10 @@ __log_c_get(logc, alsn, dbt, flags)
 		case DB_LAST:
 			flags = DB_PREV;
 			break;
+		case DB_NEXT:
+		case DB_PREV:
+		default:
+			break;
 		}
 		if (F_ISSET(dbt, DB_DBT_MALLOC)) {
 			__os_free(dbenv, dbt->data);
@@ -297,9 +294,10 @@ __log_c_get_int(logc, alsn, dbt, flags)
 	logfile_validity status;
 	u_int32_t cnt;
 	u_int8_t *rp;
-	int eof, is_hmac, ret;
+	int eof, is_hmac, need_cksum, ret;
 
 	dbenv = logc->dbenv;
+	db_cipher = dbenv->crypto_handle;
 	dblp = dbenv->lg_handle;
 	lp = dblp->reginfo.primary;
 	is_hmac = 0;
@@ -366,14 +364,15 @@ __log_c_get_int(logc, alsn, dbt, flags)
 		if (!IS_ZERO_LSN(nlsn)) {
 			/* If at start-of-file, move to the previous file. */
 			if (nlsn.offset == 0) {
-				if (nlsn.file == 1 || __log_valid(dblp,
-				    nlsn.file - 1, 0, NULL, 0, &status) != 0) {
+				if (nlsn.file == 1) {
 					ret = DB_NOTFOUND;
 					goto err;
 				}
-
-				if (status != DB_LV_NORMAL &&
-				    status != DB_LV_OLD_READABLE) {
+				if ((!lp->db_log_inmemory &&
+				    (__log_valid(dblp, nlsn.file - 1, 0, NULL,
+				    0, &status) != 0 ||
+				    (status != DB_LV_NORMAL &&
+				    status != DB_LV_OLD_READABLE)))) {
 					ret = DB_NOTFOUND;
 					goto err;
 				}
@@ -387,7 +386,7 @@ __log_c_get_int(logc, alsn, dbt, flags)
 	case DB_LAST:				/* Last log record. */
 		if (rlock == L_NONE) {
 			rlock = L_ACQUIRED;
-			R_LOCK(dbenv, &dblp->reginfo);
+			LOG_SYSTEM_LOCK(dbenv);
 		}
 		nlsn.file = lp->lsn.file;
 		nlsn.offset = lp->lsn.offset - lp->len;
@@ -395,6 +394,10 @@ __log_c_get_int(logc, alsn, dbt, flags)
 	case DB_SET:				/* Set log record. */
 		nlsn = *alsn;
 		break;
+	default:
+		DB_ASSERT(0);
+		ret = EINVAL;
+		goto err;
 	}
 
 	if (0) {				/* Move to the next file. */
@@ -414,7 +417,11 @@ next_file:	++nlsn.file;
 		hdr.size = HDR_NORMAL_SZ;
 		is_hmac = 0;
 	}
-	/* Check to see if the record is in the cursor's buffer. */
+
+	/*
+	 * Check to see if the record is in the cursor's buffer -- if so,
+	 * we'll need to checksum it.
+	 */
 	if ((ret = __log_c_incursor(logc, &nlsn, &hdr, &rp)) != 0)
 		goto err;
 	if (rp != NULL)
@@ -425,12 +432,13 @@ next_file:	++nlsn.file;
 	 * coming from the disk -- it means the record can't be in the region's
 	 * buffer.  Else, check the region's buffer.
 	 *
-	 * If the record isn't in the region's buffer, we're going to have to
-	 * read the record from disk.  We want to make a point of not reading
-	 * past the end of the logical log (after recovery, there may be data
-	 * after the end of the logical log, not to mention the log file may
-	 * have been pre-allocated).  So, zero out last_lsn, and initialize it
-	 * inside __log_c_inregion -- if it's still zero when we check it in
+	 * If the record isn't in the region's buffer, then either logs are
+	 * in-memory, and we're done, or we're going to have to read the
+	 * record from disk.  We want to make a point of not reading past the
+	 * end of the logical log (after recovery, there may be data after the
+	 * end of the logical log, not to mention the log file may have been
+	 * pre-allocated).  So, zero out last_lsn, and initialize it inside
+	 * __log_c_inregion -- if it's still zero when we check it in
 	 * __log_c_ondisk, that's OK, it just means the logical end of the log
 	 * isn't an issue for this request.
 	 */
@@ -440,10 +448,20 @@ next_file:	++nlsn.file;
 		F_CLR(logc, DB_LOG_DISK);
 
 		if ((ret = __log_c_inregion(logc,
-		    &nlsn, &rlock, &last_lsn, &hdr, &rp)) != 0)
+		    &nlsn, &rlock, &last_lsn, &hdr, &rp, &need_cksum)) != 0)
 			goto err;
-		if (rp != NULL)
-			goto cksum;
+		if (rp != NULL) {
+			/*
+			 * If we read the entire record from the in-memory log
+			 * buffer, we don't need to checksum it, nor do we need
+			 * to worry about vtruncate issues.
+			 */
+			if (need_cksum)
+				goto cksum;
+			goto from_memory;
+		}
+		if (lp->db_log_inmemory)
+			goto nohdr;
 	}
 
 	/*
@@ -456,12 +474,12 @@ next_file:	++nlsn.file;
 	 */
 	if (rlock == L_ACQUIRED) {
 		rlock = L_NONE;
-		R_UNLOCK(dbenv, &dblp->reginfo);
+		LOG_SYSTEM_UNLOCK(dbenv);
 	}
 	if ((ret = __log_c_ondisk(
 	    logc, &nlsn, &last_lsn, flags, &hdr, &rp, &eof)) != 0)
 		goto err;
-	if (eof == 1) {
+	if (eof) {
 		/*
 		 * Only DB_NEXT automatically moves to the next file, and
 		 * it only happens once.
@@ -474,13 +492,13 @@ next_file:	++nlsn.file;
 
 cksum:	/*
 	 * Discard the region lock if we're still holding it.  (The path to
-	 * get here is that we acquired the lock because of the caller's
-	 * flag argument, but we found the record in the cursor's buffer.
-	 * Improbable, but it's easy to avoid.
+	 * get here is we acquired the region lock because of the caller's
+	 * flag argument, but we found the record in the in-memory or cursor
+	 * buffers.  Improbable, but it's easy to avoid.)
 	 */
 	if (rlock == L_ACQUIRED) {
 		rlock = L_NONE;
-		R_UNLOCK(dbenv, &dblp->reginfo);
+		LOG_SYSTEM_UNLOCK(dbenv);
 	}
 
 	/*
@@ -489,7 +507,6 @@ cksum:	/*
 	 * OK if we're searching for the end of the log, and very, very bad
 	 * if we're reading random log records.
 	 */
-	db_cipher = dbenv->crypto_handle;
 	if ((ret = __db_check_chksum(dbenv, db_cipher,
 	    hdr.chksum, rp + hdr.size, hdr.len - hdr.size, is_hmac)) != 0) {
 		if (F_ISSET(logc, DB_LOG_SILENT_ERR)) {
@@ -512,12 +529,11 @@ cksum:	/*
 	 * going to have to retry.
 	 */
 	if (hdr.len == 0) {
-		switch (flags) {
+nohdr:		switch (flags) {
 		case DB_FIRST:
 		case DB_NEXT:
 			/* Zero'd records always indicate the end of a file. */
 			goto next_file;
-
 		case DB_LAST:
 		case DB_PREV:
 			/*
@@ -528,9 +544,11 @@ cksum:	/*
 			 * the first record in that new file should have its
 			 * prev field set correctly.
 			 */
-			 __db_err(dbenv,
+			__db_err(dbenv,
 		"Encountered zero length records while traversing backwards");
-			 DB_ASSERT(0);
+			DB_ASSERT(0);
+			ret = __db_panic(dbenv, DB_RUNRECOVERY);
+			goto err;
 		case DB_SET:
 		default:
 			/* Return the 0-length record. */
@@ -538,6 +556,18 @@ cksum:	/*
 		}
 	}
 
+from_memory:
+	/*
+	 * Discard the region lock if we're still holding it.  (The path to
+	 * get here is we acquired the region lock because of the caller's
+	 * flag argument, but we found the record in the in-memory or cursor
+	 * buffers.  Improbable, but it's easy to avoid.)
+	 */
+	if (rlock == L_ACQUIRED) {
+		rlock = L_NONE;
+		LOG_SYSTEM_UNLOCK(dbenv);
+	}
+
 	/* Copy the record into the user's DBT. */
 	if ((ret = __db_retcopy(dbenv, dbt, rp + hdr.size,
 	    (u_int32_t)(hdr.len - hdr.size),
@@ -567,7 +597,7 @@ cksum:	/*
 	logc->c_prev = hdr.prev;
 
 err:	if (rlock == L_ACQUIRED)
-		R_UNLOCK(dbenv, &dblp->reginfo);
+		LOG_SYSTEM_UNLOCK(dbenv);
 
 	return (ret);
 }
@@ -584,7 +614,7 @@ __log_c_incursor(logc, lsn, hdr, pp)
 	u_int8_t **pp;
 {
 	u_int8_t *p;
-	int	eof;
+	int eof;
 
 	*pp = NULL;
 
@@ -632,19 +662,20 @@ __log_c_incursor(logc, lsn, hdr, pp)
  *	Check to see if the requested record is in the region's buffer.
  */
 static int
-__log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
+__log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp, need_cksump)
 	DB_LOGC *logc;
 	DB_LSN *lsn, *last_lsn;
 	RLOCK *rlockp;
 	HDR *hdr;
 	u_int8_t **pp;
+	int *need_cksump;
 {
 	DB_ENV *dbenv;
 	DB_LOG *dblp;
 	LOG *lp;
-	size_t len, nr;
-	u_int32_t b_disk, b_region;
-	int ret;
+	size_t b_region, len, nr;
+	u_int32_t b_disk;
+	int eof, ret;
 	u_int8_t *p;
 
 	dbenv = logc->dbenv;
@@ -652,12 +683,14 @@ __log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
 	lp = ((DB_LOG *)logc->dbenv->lg_handle)->reginfo.primary;
 
 	ret = 0;
+	b_region = 0;
 	*pp = NULL;
+	*need_cksump = 0;
 
 	/* If we haven't yet acquired the log region lock, do so. */
 	if (*rlockp == L_NONE) {
 		*rlockp = L_ACQUIRED;
-		R_LOCK(dbenv, &dblp->reginfo);
+		LOG_SYSTEM_LOCK(dbenv);
 	}
 
 	/*
@@ -671,7 +704,7 @@ __log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
 	 * come later than this point if the log buffer isn't empty.
 	 */
 	*last_lsn = lp->lsn;
-	if (last_lsn->offset > lp->w_off)
+	if (!lp->db_log_inmemory && last_lsn->offset > lp->w_off)
 		last_lsn->offset = lp->w_off;
 
 	/*
@@ -698,52 +731,62 @@ __log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
 	 */
 	if (IS_ZERO_LSN(lp->lsn))
 		return (0);
-	if (lsn->file > lp->lsn.file ||
-	    (lsn->file == lp->lsn.file && lsn->offset >= lp->lsn.offset))
+	if (log_compare(lsn, &lp->lsn) >= 0)
 		return (DB_NOTFOUND);
-	if (lp->b_off == 0)
-		return (0);
-	if (lsn->file < lp->f_lsn.file || lsn->offset < lp->f_lsn.offset)
+	else if (lp->db_log_inmemory) {
+		if ((ret = __log_inmem_lsnoff(dblp, lsn, &b_region)) != 0)
+			return (ret);
+	} else if (lp->b_off == 0 || log_compare(lsn, &lp->f_lsn) < 0)
 		return (0);
 
 	/*
 	 * The current contents of the cursor's buffer will be useless for a
-	 * future call -- trash it rather than try and make it look correct.
+	 * future call, we're about to overwrite it -- trash it rather than
+	 * try and make it look correct.
 	 */
-	ZERO_LSN(logc->bp_lsn);
+	logc->bp_rlen = 0;
 
 	/*
 	 * If the requested LSN is greater than the region buffer's first
 	 * byte, we know the entire record is in the buffer on a good LSN.
 	 *
-	 * If we're given a bad LSN, the "entire" record might
-	 * not be in our buffer in order to fail at the chksum.
-	 * __log_c_hdrchk made sure our dest buffer fits, via
-	 * bp_maxrec, but we also need to make sure we don't run off
-	 * the end of this buffer, the src.
+	 * If we're given a bad LSN, the "entire" record might not be in
+	 * our buffer in order to fail at the chksum.  __log_c_hdrchk made
+	 * sure our dest buffer fits, via bp_maxrec, but we also need to
+	 * make sure we don't run off the end of this buffer, the src.
 	 *
-	 * If the header check fails for any reason, it must be because the
-	 * LSN is bogus.  Fail hard.
+	 * There is one case where the header check can fail: on a scan through
+	 * in-memory logs, when we reach the end of a file we can read an empty
+	 * header.  In that case, it's safe to return zero, here: it will be
+	 * caught in our caller.  Otherwise, the LSN is bogus.  Fail hard.
 	 */
-	if (lsn->offset > lp->f_lsn.offset) {
-		p = dblp->bufp + (lsn->offset - lp->w_off);
-		memcpy(hdr, p, hdr->size);
-		if (__log_c_hdrchk(logc, lsn, hdr, NULL))
+	if (lp->db_log_inmemory || log_compare(lsn, &lp->f_lsn) > 0) {
+		if (!lp->db_log_inmemory)
+			b_region = lsn->offset - lp->w_off;
+		__log_inmem_copyout(dblp, b_region, hdr, hdr->size);
+		if (__log_c_hdrchk(logc, lsn, hdr, &eof) != 0)
 			return (DB_NOTFOUND);
-		if (lsn->offset + hdr->len > lp->w_off + lp->buffer_size)
+		if (eof)
+			return (0);
+		if (lp->db_log_inmemory) {
+			if (RINGBUF_LEN(lp, b_region, lp->b_off) < hdr->len)
+				return (DB_NOTFOUND);
+		} else if (lsn->offset + hdr->len > lp->w_off + lp->buffer_size)
 			return (DB_NOTFOUND);
 		if (logc->bp_size <= hdr->len) {
-			len = ALIGN(hdr->len * 2, 128);
+			len = (size_t)DB_ALIGN((uintmax_t)hdr->len * 2, 128);
 			if ((ret =
 			    __os_realloc(logc->dbenv, len, &logc->bp)) != 0)
 				 return (ret);
 			logc->bp_size = (u_int32_t)len;
 		}
-		memcpy(logc->bp, p, hdr->len);
+		__log_inmem_copyout(dblp, b_region, logc->bp, hdr->len);
 		*pp = logc->bp;
 		return (0);
 	}
 
+	DB_ASSERT(!lp->db_log_inmemory);
+
 	/*
 	 * There's a partial record, that is, the requested record starts
 	 * in a log file and finishes in the region buffer.  We have to
@@ -780,7 +823,7 @@ __log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
 	 * of waiting.
 	 */
 	if (logc->bp_size <= b_region + b_disk) {
-		len = ALIGN((b_region + b_disk) * 2, 128);
+		len = (size_t)DB_ALIGN((uintmax_t)(b_region + b_disk) * 2, 128);
 		if ((ret = __os_realloc(logc->dbenv, len, &logc->bp)) != 0)
 			return (ret);
 		logc->bp_size = (u_int32_t)len;
@@ -793,7 +836,7 @@ __log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
 	/* Release the region lock. */
 	if (*rlockp == L_ACQUIRED) {
 		*rlockp = L_NONE;
-		R_UNLOCK(dbenv, &dblp->reginfo);
+		LOG_SYSTEM_UNLOCK(dbenv);
 	}
 
 	/*
@@ -808,6 +851,9 @@ __log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
 			return (ret);
 		if (nr < b_disk)
 			return (__log_c_shortread(logc, lsn, 0));
+
+		/* We read bytes from the disk, we'll need to checksum them. */
+		*need_cksump = 1;
 	}
 
 	/* Copy the header information into the caller's structure. */
@@ -825,7 +871,8 @@ static int
 __log_c_ondisk(logc, lsn, last_lsn, flags, hdr, pp, eofp)
 	DB_LOGC *logc;
 	DB_LSN *lsn, *last_lsn;
-	int flags, *eofp;
+	u_int32_t flags;
+	int *eofp;
 	HDR *hdr;
 	u_int8_t **pp;
 {
@@ -844,8 +891,14 @@ __log_c_ondisk(logc, lsn, last_lsn, flags, hdr, pp, eofp)
 	if (*eofp)
 		return (0);
 
-	/* If we read 0 bytes, assume we've hit EOF. */
-	if (nr == 0) {
+	/*
+	 * If the read was successful, but we can't read a full header, assume
+	 * we've hit EOF.  We can't check that the header has been partially
+	 * zeroed out, but it's unlikely that this is caused by a write failure
+	 * since the header is written as a single write call and it's less
+	 * than sector.
+	 */
+	if (nr < hdr->size) {
 		*eofp = 1;
 		return (0);
 	}
@@ -856,15 +909,11 @@ __log_c_ondisk(logc, lsn, last_lsn, flags, hdr, pp, eofp)
 	if (*eofp)
 		return (0);
 
-	/* Otherwise, we should have gotten the bytes we wanted. */
-	if (nr < hdr->size)
-		return (__log_c_shortread(logc, lsn, 1));
-
 	/*
 	 * Regardless of how we return, the previous contents of the cursor's
 	 * buffer are useless -- trash it.
 	 */
-	ZERO_LSN(logc->bp_lsn);
+	logc->bp_rlen = 0;
 
 	/*
 	 * Otherwise, we now (finally!) know how big the record is.  (Maybe
@@ -872,7 +921,7 @@ __log_c_ondisk(logc, lsn, last_lsn, flags, hdr, pp, eofp)
 	 * Make sure we have enough space.
 	 */
 	if (logc->bp_size <= hdr->len) {
-		len = ALIGN(hdr->len * 2, 128);
+		len = (size_t)DB_ALIGN((uintmax_t)hdr->len * 2, 128);
 		if ((ret = __os_realloc(dbenv, len, &logc->bp)) != 0)
 			return (ret);
 		logc->bp_size = (u_int32_t)len;
@@ -911,9 +960,13 @@ __log_c_ondisk(logc, lsn, last_lsn, flags, hdr, pp, eofp)
 	if (nr < (lsn->offset + hdr->len) - offset)
 		return (__log_c_shortread(logc, lsn, 1));
 
-	/* Set up the return information. */
+	/*
+	 * Set up the return information.
+	 *
+	 * !!!
+	 * No need to set the bp_lsn.file field, __log_c_io set it for us.
+	 */
 	logc->bp_rlen = (u_int32_t)nr;
-	logc->bp_lsn.file = lsn->file;
 	logc->bp_lsn.offset = offset;
 
 	*pp = logc->bp + (lsn->offset - offset);
@@ -933,16 +986,6 @@ __log_c_ondisk(logc, lsn, last_lsn, flags, hdr, pp, eofp)
  * and one made it to disk before a different one that logically precedes
  * it in the log file.
  *
- * XXX
- * I think there's a potential pre-allocation recovery flaw here -- if we
- * fail to write a buffer at the end of a log file (by scheduling its
- * write asynchronously, and it never making it to disk), then succeed in
- * writing a log file block to a subsequent log file, I don't think we will
- * detect that the buffer of 0's should have marked the end of the log files
- * during recovery.  I think we may need to always write some garbage after
- * each block write if we pre-allocate log files.  (At the moment, we do not
- * pre-allocate, so this isn't currently an issue.)
- *
  * Check for impossibly large records.  The malloc should fail later, but we
  * have customers that run mallocs that treat all allocation failures as fatal
  * errors.
@@ -1021,11 +1064,13 @@ __log_c_io(logc, fnum, offset, p, nrp, eofp)
 {
 	DB_ENV *dbenv;
 	DB_LOG *dblp;
+	LOG *lp;
 	int ret;
 	char *np;
 
 	dbenv = logc->dbenv;
 	dblp = dbenv->lg_handle;
+	lp = dblp->reginfo.primary;
 
 	/*
 	 * If we've switched files, discard the current file handle and acquire
@@ -1034,6 +1079,8 @@ __log_c_io(logc, fnum, offset, p, nrp, eofp)
 	if (logc->c_fhp != NULL && logc->bp_lsn.file != fnum) {
 		ret = __os_closehandle(dbenv, logc->c_fhp);
 		logc->c_fhp = NULL;
+		logc->bp_lsn.file = 0;
+
 		if (ret != 0)
 			return (ret);
 	}
@@ -1061,6 +1108,8 @@ __log_c_io(logc, fnum, offset, p, nrp, eofp)
 			return (ret);
 		}
 		__os_free(dbenv, np);
+
+		logc->bp_lsn.file = fnum;
 	}
 
 	/* Seek to the record's offset. */
@@ -1074,6 +1123,7 @@ __log_c_io(logc, fnum, offset, p, nrp, eofp)
 	}
 
 	/* Read the data. */
+	++lp->stat.st_rcount;
 	if ((ret = __os_read(dbenv, logc->c_fhp, p, *nrp, nrp)) != 0) {
 		if (!F_ISSET(logc, DB_LOG_SILENT_ERR))
 			__db_err(dbenv,
@@ -1154,3 +1204,134 @@ __log_c_set_maxrec(logc, np)
 
 	return (0);
 }
+
+#ifdef HAVE_REPLICATION
+/*
+ * __log_rep_split --
+ *	- Split a log buffer into individual records.
+ *
+ * This is used by a replication client to process a bulk log message from the
+ * master and convert it into individual __rep_apply requests.
+ *
+ * PUBLIC: int __log_rep_split __P((DB_ENV *, REP_CONTROL *, DBT *, DB_LSN *));
+ */
+int
+__log_rep_split(dbenv, rp, rec, ret_lsnp)
+	DB_ENV *dbenv;
+	REP_CONTROL *rp;
+	DBT *rec;
+	DB_LSN *ret_lsnp;
+{
+	DB_LSN save_lsn, tmp_lsn;
+	DB_REP *db_rep;
+	DBT logrec;
+	REP *rep;
+	REP_CONTROL tmprp;
+	u_int32_t len;
+	int is_dup, is_perm, ret, save_ret;
+	u_int8_t *p, *ep;
+#ifdef DIAGNOSTIC
+	DB_MSGBUF mb;
+#endif
+
+	memset(&logrec, 0, sizeof(logrec));
+	memset(&save_lsn, 0, sizeof(save_lsn));
+	memset(&tmp_lsn, 0, sizeof(tmp_lsn));
+	/*
+	 * We're going to be modifying the rp LSN contents so make
+	 * our own private copy to play with.
+	 */
+	memcpy(&tmprp, rp, sizeof(tmprp));
+	/*
+	 * We send the bulk buffer on a PERM record, so often we will have
+	 * DB_LOG_PERM set.  However, we only want to mark the last LSN
+	 * we have as a PERM record.  So clear it here, and when we're on
+	 * the last record below, set it.
+	 */
+	is_perm = F_ISSET(rp, DB_LOG_PERM);
+	F_CLR(&tmprp, DB_LOG_PERM);
+	ret = save_ret = 0;
+	db_rep = dbenv->rep_handle;
+	rep = db_rep->region;
+	for (ep = (u_int8_t *)rec->data + rec->size, p = (u_int8_t *)rec->data;
+	    p < ep; ) {
+		/*
+		 * First thing in the buffer is the length.  Then the LSN
+		 * of this record, then the record itself.
+		 */
+		/*
+		 * XXX
+		 * If/when we add architecture neutral log files we may want
+		 * to send/receive these lengths in network byte order.
+		 */
+		memcpy(&len, p, sizeof(len));
+		p += sizeof(len);
+		memcpy(&tmprp.lsn, p, sizeof(DB_LSN));
+		p += sizeof(DB_LSN);
+		logrec.data = p;
+		logrec.size = len;
+		RPRINT(dbenv, rep, (dbenv, &mb,
+		    "log_rep_split: Processing LSN [%lu][%lu]",
+		    (u_long)tmprp.lsn.file, (u_long)tmprp.lsn.offset));
+		RPRINT(dbenv, rep, (dbenv, &mb,
+    "log_rep_split: p %#lx ep %#lx logrec data %#lx, size %lu (%#lx)",
+		    P_TO_ULONG(p), P_TO_ULONG(ep), P_TO_ULONG(logrec.data),
+		    (u_long)logrec.size, (u_long)logrec.size));
+		is_dup = 0;
+		p += len;
+		if (p >= ep && is_perm)
+			F_SET(&tmprp, DB_LOG_PERM);
+		ret = __rep_apply(dbenv, &tmprp, &logrec, &tmp_lsn, &is_dup);
+		RPRINT(dbenv, rep, (dbenv, &mb,
+		    "log_split: rep_apply ret %d, tmp_lsn [%lu][%lu]",
+		    ret, (u_long)tmp_lsn.file, (u_long)tmp_lsn.offset));
+#if 0
+		/*
+		 * This buffer may be old and we've already gotten these
+		 * records.  Short-circuit processing this buffer.
+		 */
+		if (is_dup)
+			goto out;
+#endif
+		switch (ret) {
+		/*
+		 * If we received the pieces we need for running recovery,
+		 * short-circuit because recovery will truncate the log to
+		 * the LSN we want anyway.
+		 */
+		case DB_REP_LOGREADY:
+			goto out;
+		/*
+		 * If we just handled a special record, retain that information.
+		 */
+		case DB_REP_ISPERM:
+		case DB_REP_NOTPERM:
+		case DB_REP_STARTUPDONE:
+			save_ret = ret;
+			save_lsn = tmp_lsn;
+			ret = 0;
+			break;
+		/*
+		 * Normal processing, do nothing, just continue.
+		 */
+		case 0:
+			break;
+		/*
+		 * If we get an error, then stop immediately.
+		 */
+		default:
+			goto out;
+		}
+	}
+out:
+	/*
+	 * If we finish processing successfully, set our return values
+	 * based on what we saw.
+	 */
+	if (ret == 0) {
+		ret = save_ret;
+		*ret_lsnp = save_lsn;
+	}
+	return (ret);
+}
+#endif
