@@ -1,14 +1,13 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: mp_sync.c,v 12.11 2005/10/07 20:21:33 ubell Exp $
  */
-#include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: mp_sync.c,v 11.80 2003/09/13 19:20:41 bostic Exp $";
-#endif /* not lint */
+#include "db_config.h"
 
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
@@ -29,7 +28,7 @@ typedef struct {
 } BH_TRACK;
 
 static int __bhcmp __P((const void *, const void *));
-static int __memp_close_flush_files __P((DB_ENV *, DB_MPOOL *));
+static int __memp_close_flush_files __P((DB_ENV *, DB_MPOOL *, int));
 static int __memp_sync_files __P((DB_ENV *, DB_MPOOL *));
 
 /*
@@ -43,7 +42,8 @@ __memp_sync_pp(dbenv, lsnp)
 	DB_ENV *dbenv;
 	DB_LSN *lsnp;
 {
-	int rep_check, ret;
+	DB_THREAD_INFO *ip;
+	int ret;
 
 	PANIC_CHECK(dbenv);
 	ENV_REQUIRES_CONFIG(dbenv,
@@ -57,12 +57,9 @@ __memp_sync_pp(dbenv, lsnp)
 		ENV_REQUIRES_CONFIG(dbenv,
 		    dbenv->lg_handle, "memp_sync", DB_INIT_LOG);
 
-	rep_check = IS_ENV_REPLICATED(dbenv) ? 1 : 0;
-	if (rep_check)
-		__env_rep_enter(dbenv);
-	ret = __memp_sync(dbenv, lsnp);
-	if (rep_check)
-		__env_rep_exit(dbenv);
+	ENV_ENTER(dbenv, ip);
+	REPLICATION_WRAP(dbenv, (__memp_sync(dbenv, lsnp)), ret);
+	ENV_LEAVE(dbenv, ip);
 	return (ret);
 }
 
@@ -86,24 +83,24 @@ __memp_sync(dbenv, lsnp)
 
 	/* If we've flushed to the requested LSN, return that information. */
 	if (lsnp != NULL) {
-		R_LOCK(dbenv, dbmp->reginfo);
+		MPOOL_SYSTEM_LOCK(dbenv);
 		if (log_compare(lsnp, &mp->lsn) <= 0) {
 			*lsnp = mp->lsn;
 
-			R_UNLOCK(dbenv, dbmp->reginfo);
+			MPOOL_SYSTEM_UNLOCK(dbenv);
 			return (0);
 		}
-		R_UNLOCK(dbenv, dbmp->reginfo);
+		MPOOL_SYSTEM_UNLOCK(dbenv);
 	}
 
 	if ((ret = __memp_sync_int(dbenv, NULL, 0, DB_SYNC_CACHE, NULL)) != 0)
 		return (ret);
 
 	if (lsnp != NULL) {
-		R_LOCK(dbenv, dbmp->reginfo);
+		MPOOL_SYSTEM_LOCK(dbenv);
 		if (log_compare(lsnp, &mp->lsn) > 0)
 			mp->lsn = *lsnp;
-		R_UNLOCK(dbenv, dbmp->reginfo);
+		MPOOL_SYSTEM_UNLOCK(dbenv);
 	}
 
 	return (0);
@@ -120,18 +117,17 @@ __memp_fsync_pp(dbmfp)
 	DB_MPOOLFILE *dbmfp;
 {
 	DB_ENV *dbenv;
-	int rep_check, ret;
+	DB_THREAD_INFO *ip;
+	int ret;
 
 	dbenv = dbmfp->dbenv;
 
 	PANIC_CHECK(dbenv);
 	MPF_ILLEGAL_BEFORE_OPEN(dbmfp, "DB_MPOOLFILE->sync");
 
-	if ((rep_check = IS_ENV_REPLICATED(dbenv)) != 0)
-		__env_rep_enter(dbenv);
-	ret = __memp_fsync(dbmfp);
-	if (rep_check)
-		__env_rep_exit(dbenv);
+	ENV_ENTER(dbenv, ip);
+	REPLICATION_WRAP(dbenv, (__memp_fsync(dbmfp)), ret);
+	ENV_LEAVE(dbenv, ip);
 	return (ret);
 }
 
@@ -145,15 +141,23 @@ int
 __memp_fsync(dbmfp)
 	DB_MPOOLFILE *dbmfp;
 {
+	MPOOLFILE *mfp;
+
+	mfp = dbmfp->mfp;
+
 	/*
 	 * If this handle doesn't have a file descriptor that's open for
-	 * writing, or if the file is a temporary, there's no reason to
-	 * proceed further.
+	 * writing, or if the file is a temporary, or if the file hasn't
+	 * been written since it was flushed, there's no reason to proceed
+	 * further.
 	 */
 	if (F_ISSET(dbmfp, MP_READONLY))
 		return (0);
 
-	if (F_ISSET(dbmfp->mfp, MP_TEMP))
+	if (F_ISSET(dbmfp->mfp, MP_TEMP) || dbmfp->mfp->no_backing_file)
+		return (0);
+
+	if (mfp->file_written == 0)
 		return (0);
 
 	return (__memp_sync_int(dbmfp->dbenv, dbmfp, 0, DB_SYNC_FILE, NULL));
@@ -194,30 +198,39 @@ __mp_xxx_fh(dbmfp, fhp)
  * __memp_sync_int --
  *	Mpool sync internal function.
  *
- * PUBLIC: int __memp_sync_int
- * PUBLIC:     __P((DB_ENV *, DB_MPOOLFILE *, int, db_sync_op, int *));
+ * PUBLIC: int __memp_sync_int __P((DB_ENV *,
+ * PUBLIC:     DB_MPOOLFILE *, u_int32_t, db_sync_op, u_int32_t *));
  */
 int
 __memp_sync_int(dbenv, dbmfp, trickle_max, op, wrotep)
 	DB_ENV *dbenv;
 	DB_MPOOLFILE *dbmfp;
-	int trickle_max, *wrotep;
+	u_int32_t trickle_max, *wrotep;
 	db_sync_op op;
 {
 	BH *bhp;
 	BH_TRACK *bharray;
 	DB_MPOOL *dbmp;
 	DB_MPOOL_HASH *hp;
-	DB_MUTEX *mutexp;
 	MPOOL *c_mp, *mp;
 	MPOOLFILE *mfp;
-	u_int32_t n_cache;
-	int ar_cnt, ar_max, hb_lock, i, pass, remaining, ret, t_ret;
-	int wait_cnt, write_cnt, wrote;
+	db_mutex_t mutex;
+	roff_t last_mf_offset;
+	u_int32_t ar_cnt, ar_max, i, n_cache, remaining, wrote;
+	int filecnt, hb_lock, maxopenfd, maxwrite, maxwrite_sleep;
+	int pass, ret, t_ret, wait_cnt, write_cnt;
 
 	dbmp = dbenv->mp_handle;
 	mp = dbmp->reginfo[0].primary;
-	pass = wrote = 0;
+	last_mf_offset = INVALID_ROFF;
+	filecnt = pass = wrote = 0;
+
+	/* Get shared configuration information. */
+	MPOOL_SYSTEM_LOCK(dbenv);
+	maxopenfd = mp->mp_maxopenfd;
+	maxwrite = mp->mp_maxwrite;
+	maxwrite_sleep = mp->mp_maxwrite_sleep;
+	MPOOL_SYSTEM_UNLOCK(dbenv);
 
 	/* Assume one dirty page per bucket. */
 	ar_max = mp->nreg * mp->htab_buckets;
@@ -244,7 +257,7 @@ __memp_sync_int(dbenv, dbmfp, trickle_max, op, wrotep)
 			if (SH_TAILQ_FIRST(&hp->hash_bucket, __bh) == NULL)
 				continue;
 
-			MUTEX_LOCK(dbenv, &hp->hash_mutex);
+			MUTEX_LOCK(dbenv, hp->mtx_hash);
 			for (bhp = SH_TAILQ_FIRST(&hp->hash_bucket, __bh);
 			    bhp != NULL; bhp = SH_TAILQ_NEXT(bhp, hq, __bh)) {
 				/* Always ignore unreferenced, clean pages. */
@@ -269,14 +282,12 @@ __memp_sync_int(dbenv, dbmfp, trickle_max, op, wrotep)
 				mfp = R_ADDR(dbmp->reginfo, bhp->mf_offset);
 
 				/*
-				 * Ignore temporary files -- this means you
-				 * can't even flush temporary files by handle.
-				 * (Checkpoint doesn't require temporary files
-				 * be flushed and the underlying buffer write
-				 * write routine may not be able to write it
-				 * anyway.)
+				 * Ignore in-memory files, even if they are
+				 * temp files to whom a backing file has been
+				 * allocated.
 				 */
-				if (F_ISSET(mfp, MP_TEMP))
+				if (mfp->no_backing_file ||
+				    F_ISSET(mfp, MP_TEMP))
 					continue;
 
 				/*
@@ -313,7 +324,7 @@ __memp_sync_int(dbenv, dbmfp, trickle_max, op, wrotep)
 					ar_max *= 2;
 				}
 			}
-			MUTEX_UNLOCK(dbenv, &hp->hash_mutex);
+			MUTEX_UNLOCK(dbenv, hp->mtx_hash);
 
 			if (ret != 0)
 				goto err;
@@ -336,7 +347,7 @@ __memp_sync_int(dbenv, dbmfp, trickle_max, op, wrotep)
 	 * If we're trickling buffers, only write enough to reach the correct
 	 * percentage.
 	 */
-	 if (op == DB_SYNC_TRICKLE && ar_cnt > trickle_max)
+	if (op == DB_SYNC_TRICKLE && ar_cnt > trickle_max)
 		ar_cnt = trickle_max;
 
 	/*
@@ -359,14 +370,14 @@ __memp_sync_int(dbenv, dbmfp, trickle_max, op, wrotep)
 		if (i >= ar_cnt) {
 			i = 0;
 			++pass;
-			(void)__os_sleep(dbenv, 1, 0);
+			__os_sleep(dbenv, 1, 0);
 		}
 		if ((hp = bharray[i].track_hp) == NULL)
 			continue;
 
 		/* Lock the hash bucket and find the buffer. */
-		mutexp = &hp->hash_mutex;
-		MUTEX_LOCK(dbenv, mutexp);
+		mutex = hp->mtx_hash;
+		MUTEX_LOCK(dbenv, mutex);
 		for (bhp = SH_TAILQ_FIRST(&hp->hash_bucket, __bh);
 		    bhp != NULL; bhp = SH_TAILQ_NEXT(bhp, hq, __bh))
 			if (bhp->pgno == bharray[i].track_pgno &&
@@ -381,7 +392,7 @@ __memp_sync_int(dbenv, dbmfp, trickle_max, op, wrotep)
 		 * no work needed.
 		 */
 		if (bhp == NULL || (bhp->ref == 0 && !F_ISSET(bhp, BH_DIRTY))) {
-			MUTEX_UNLOCK(dbenv, mutexp);
+			MUTEX_UNLOCK(dbenv, mutex);
 			--remaining;
 			bharray[i].track_hp = NULL;
 			continue;
@@ -399,7 +410,7 @@ __memp_sync_int(dbenv, dbmfp, trickle_max, op, wrotep)
 		 * write it.
 		 */
 		if (F_ISSET(bhp, BH_LOCKED) || (bhp->ref != 0 && pass < 2)) {
-			MUTEX_UNLOCK(dbenv, mutexp);
+			MUTEX_UNLOCK(dbenv, mutex);
 			if (op != DB_SYNC_CACHE && op != DB_SYNC_FILE) {
 				--remaining;
 				bharray[i].track_hp = NULL;
@@ -418,7 +429,7 @@ __memp_sync_int(dbenv, dbmfp, trickle_max, op, wrotep)
 		/* Pin the buffer into memory and lock it. */
 		++bhp->ref;
 		F_SET(bhp, BH_LOCKED);
-		MUTEX_LOCK(dbenv, &bhp->mutex);
+		MUTEX_LOCK(dbenv, bhp->mtx_bh);
 
 		/*
 		 * Unlock the hash bucket and wait for the wait-for count to
@@ -433,13 +444,27 @@ __memp_sync_int(dbenv, dbmfp, trickle_max, op, wrotep)
 		 * If, when the wait-for count goes to 0, the buffer is found
 		 * to be dirty, write it.
 		 */
-		MUTEX_UNLOCK(dbenv, mutexp);
+		MUTEX_UNLOCK(dbenv, mutex);
 		for (wait_cnt = 1;
 		    bhp->ref_sync != 0 && wait_cnt < 4; ++wait_cnt)
-			(void)__os_sleep(dbenv, 1, 0);
-		MUTEX_LOCK(dbenv, mutexp);
+			__os_sleep(dbenv, 1, 0);
+		MUTEX_LOCK(dbenv, mutex);
 		hb_lock = 1;
 
+		/*
+		 * If we've switched files, check to see if we're configured
+		 * to close file descriptors.
+		 */
+		if (maxopenfd != 0 && bhp->mf_offset != last_mf_offset) {
+			if (++filecnt >= maxopenfd) {
+				filecnt = 0;
+				if ((ret = __memp_close_flush_files(
+				    dbenv, dbmp, 1)) != 0)
+					break;
+			}
+			last_mf_offset = bhp->mf_offset;
+		}
+
 		/*
 		 * If the ref_sync count has gone to 0, we're going to be done
 		 * with this buffer no matter what happens.
@@ -452,33 +477,25 @@ __memp_sync_int(dbenv, dbmfp, trickle_max, op, wrotep)
 		/*
 		 * If the ref_sync count has gone to 0 and the buffer is still
 		 * dirty, we write it.  We only try to write the buffer once.
-		 * Any process checkpointing or trickle-flushing the pool
-		 * must be able to write any underlying file -- if the write
-		 * fails, error out.  It would be very strange if file sync
-		 * failed to write, but we don't care if it happens.
 		 */
 		if (bhp->ref_sync == 0 && F_ISSET(bhp, BH_DIRTY)) {
 			hb_lock = 0;
-			MUTEX_UNLOCK(dbenv, mutexp);
+			MUTEX_UNLOCK(dbenv, mutex);
 
 			mfp = R_ADDR(dbmp->reginfo, bhp->mf_offset);
 			if ((ret = __memp_bhwrite(dbmp, hp, mfp, bhp, 1)) == 0)
 				++wrote;
-			else if (op == DB_SYNC_CACHE || op == DB_SYNC_TRICKLE)
+			else
 				__db_err(dbenv, "%s: unable to flush page: %lu",
 				    __memp_fns(dbmp, mfp), (u_long)bhp->pgno);
-			else
-				ret = 0;
 
 			/*
 			 * Avoid saturating the disk, sleep once we've done
 			 * some number of writes.
 			 */
-			 if (dbenv->mp_maxwrite != 0 &&
-			     ++write_cnt >= dbenv->mp_maxwrite) {
+			if (maxwrite != 0 && ++write_cnt >= maxwrite) {
 				write_cnt = 0;
-				(void)__os_sleep(
-				    dbenv, 0, dbenv->mp_maxwrite_sleep);
+				__os_sleep(dbenv, 0, (u_long)maxwrite_sleep);
 			}
 		}
 
@@ -500,10 +517,10 @@ __memp_sync_int(dbenv, dbmfp, trickle_max, op, wrotep)
 		 */
 		if (F_ISSET(bhp, BH_LOCKED)) {
 			F_CLR(bhp, BH_LOCKED);
-			MUTEX_UNLOCK(dbenv, &bhp->mutex);
+			MUTEX_UNLOCK(dbenv, bhp->mtx_bh);
 
 			if (!hb_lock)
-				MUTEX_LOCK(dbenv, mutexp);
+				MUTEX_LOCK(dbenv, mutex);
 		}
 
 		/*
@@ -514,7 +531,7 @@ __memp_sync_int(dbenv, dbmfp, trickle_max, op, wrotep)
 
 		/* Discard our reference and unlock the bucket. */
 		--bhp->ref;
-		MUTEX_UNLOCK(dbenv, mutexp);
+		MUTEX_UNLOCK(dbenv, mutex);
 
 		if (ret != 0)
 			break;
@@ -535,7 +552,7 @@ done:	/*
 	}
 
 	/* If we've opened files to flush pages, close them. */
-	if ((t_ret = __memp_close_flush_files(dbenv, dbmp)) != 0 && ret == 0)
+	if ((t_ret = __memp_close_flush_files(dbenv, dbmp, 0)) != 0 && ret == 0)
 		ret = t_ret;
 
 err:	__os_free(dbenv, bharray);
@@ -556,49 +573,201 @@ int __memp_sync_files(dbenv, dbmp)
 {
 	DB_MPOOLFILE *dbmfp;
 	MPOOL *mp;
-	MPOOLFILE *mfp;
-	int final_ret, ret;
+	MPOOLFILE *mfp, *next_mfp;
+	int need_discard_pass, ret, t_ret;
 
-	final_ret = 0;
+	need_discard_pass = ret = 0;
 	mp = dbmp->reginfo[0].primary;
 
-	R_LOCK(dbenv, dbmp->reginfo);
+	MPOOL_SYSTEM_LOCK(dbenv);
 	for (mfp = SH_TAILQ_FIRST(&mp->mpfq, __mpoolfile);
 	    mfp != NULL; mfp = SH_TAILQ_NEXT(mfp, q, __mpoolfile)) {
-		if (!mfp->file_written ||
+		if (!mfp->file_written || mfp->no_backing_file ||
 		    mfp->deadfile || F_ISSET(mfp, MP_TEMP))
 			continue;
+		/*
+		 * Pin the MPOOLFILE structure into memory, and release the
+		 * region mutex allowing us to walk the linked list.  We'll
+		 * re-acquire that mutex to move to the next entry in the list.
+		 *
+		 * This works because we only need to flush current entries,
+		 * we don't care about new entries being added, and the linked
+		 * list is never re-ordered, a single pass is sufficient.  It
+		 * requires MPOOLFILE structures removed before we get to them
+		 * be flushed to disk, but that's nothing new, they could have
+		 * been removed while checkpoint was running, too.
+		 *
+		 * Once we have the MPOOLFILE lock, re-check the MPOOLFILE is
+		 * not being discarded.  (A thread removing the MPOOLFILE
+		 * will: hold the MPOOLFILE mutex, set deadfile, drop the
+		 * MPOOLFILE mutex and then acquire the region MUTEX to walk
+		 * the linked list and remove the MPOOLFILE structure.  Make
+		 * sure the MPOOLFILE wasn't marked dead while we waited for
+		 * the mutex.
+		 */
+		MUTEX_LOCK(dbenv, mfp->mutex);
+		if (!mfp->file_written || mfp->deadfile) {
+			MUTEX_UNLOCK(dbenv, mfp->mutex);
+			continue;
+		}
+		MPOOL_SYSTEM_UNLOCK(dbenv);
+		++mfp->mpf_cnt;
+		MUTEX_UNLOCK(dbenv, mfp->mutex);
 
 		/*
 		 * Look for an already open, writeable handle (fsync doesn't
 		 * work on read-only Windows handles).
 		 */
-		ret = 0;
-		MUTEX_THREAD_LOCK(dbenv, dbmp->mutexp);
+		MUTEX_LOCK(dbenv, dbmp->mutex);
 		for (dbmfp = TAILQ_FIRST(&dbmp->dbmfq);
 		    dbmfp != NULL; dbmfp = TAILQ_NEXT(dbmfp, q)) {
 			if (dbmfp->mfp != mfp || F_ISSET(dbmfp, MP_READONLY))
 				continue;
-			ret = __os_fsync(dbenv, dbmfp->fhp);
+			/*
+			 * We don't want to hold the mutex while calling sync.
+			 * Increment the DB_MPOOLFILE handle ref count to pin
+			 * it into memory.
+			 */
+			++dbmfp->ref;
 			break;
 		}
-		MUTEX_THREAD_UNLOCK(dbenv, dbmp->mutexp);
+		MUTEX_UNLOCK(dbenv, dbmp->mutex);
 
-		/* If we don't find one, open one. */
-		if (dbmfp == NULL)
-			ret = __memp_mf_sync(dbmp, mfp);
-		if (ret != 0) {
-			__db_err(dbenv, "%s: unable to flush: %s",
-			    (char *)R_ADDR(dbmp->reginfo, mfp->path_off),
-			    db_strerror(ret));
-			if (final_ret == 0)
-				final_ret = ret;
+		/* If we don't find a handle we can use, open one. */
+		if (dbmfp == NULL) {
+			if ((t_ret = __memp_mf_sync(dbmp, mfp, 0)) != 0) {
+				__db_err(dbenv,
+				    "%s: unable to flush: %s", (char *)
+				    R_ADDR(dbmp->reginfo, mfp->path_off),
+				    db_strerror(t_ret));
+				if (ret == 0)
+					ret = t_ret;
+			}
+		} else {
+			if ((t_ret =
+			    __os_fsync(dbenv, dbmfp->fhp)) != 0 && ret == 0)
+				ret = t_ret;
+
+			if ((t_ret = __memp_fclose(dbmfp, 0)) != 0 && ret == 0)
+				ret = t_ret;
 		}
+
+		/*
+		 * Re-acquire the region lock, we need it to move to the next
+		 * MPOOLFILE.
+		 *
+		 * Re-acquire the MPOOLFILE mutex, we need it to modify the
+		 * reference count.
+		 */
+		MPOOL_SYSTEM_LOCK(dbenv);
+		MUTEX_LOCK(dbenv, mfp->mutex);
+		--mfp->mpf_cnt;
+
+		/*
+		 * If we wrote the file and there are no open handles (or there
+		 * is a single open handle, and it's the one we opened to write
+		 * buffers during checkpoint), clear the file_written flag.  We
+		 * do this so that applications opening thousands of files don't
+		 * loop here opening and flushing those files during checkpoint.
+		 *
+		 * The danger here is if a buffer were to be written as part of
+		 * a checkpoint, and then not be flushed to disk.  This cannot
+		 * happen because we only clear file_written when there are no
+		 * other users of the MPOOLFILE in the system, and, as we hold
+		 * the region lock, no possibility of another thread of control
+		 * racing with us to open a MPOOLFILE.
+		 */
+		if (mfp->mpf_cnt == 0 || (mfp->mpf_cnt == 1 &&
+		    dbmfp != NULL && F_ISSET(dbmfp, MP_FLUSH))) {
+			mfp->file_written = 0;
+
+			/*
+			 * We may be the last reference for a MPOOLFILE, as we
+			 * weren't holding the MPOOLFILE mutex when flushing
+			 * it's buffers to disk.  If we can discard it, set
+			 * a flag to schedule a clean-out pass.   (Not likely,
+			 * I mean, what are the chances that there aren't any
+			 * buffers in the pool?  Regardless, it might happen.)
+			 */
+			if (mfp->mpf_cnt == 0 && mfp->block_cnt == 0)
+				need_discard_pass = 1;
+		}
+
+		/* Unlock the MPOOLFILE, and move to the next entry. */
+		MUTEX_UNLOCK(dbenv, mfp->mutex);
 	}
 
-	R_UNLOCK(dbenv, dbmp->reginfo);
+	/*
+	 * We exit the loop holding the region lock.
+	 *
+	 * We may need to do a last pass through the MPOOLFILE list -- if we
+	 * were the last reference to an MPOOLFILE, we need to clean it out.
+	 */
+	if (need_discard_pass)
+		for (mfp = SH_TAILQ_FIRST(
+		    &mp->mpfq, __mpoolfile); mfp != NULL; mfp = next_mfp) {
+			next_mfp = SH_TAILQ_NEXT(mfp, q, __mpoolfile);
 
-	return (final_ret);
+			/*
+			 * Do a fast check -- we can check for zero/non-zero
+			 * without a mutex on the MPOOLFILE.  If likely to
+			 * succeed, lock the MPOOLFILE down and look for real.
+			 */
+			if (mfp->block_cnt != 0 || mfp->mpf_cnt != 0)
+				continue;
+
+			MUTEX_LOCK(dbenv, mfp->mutex);
+			if (mfp->block_cnt == 0 && mfp->mpf_cnt == 0)
+				(void)__memp_mf_discard(dbmp, mfp);
+			else
+				MUTEX_UNLOCK(dbenv, mfp->mutex);
+		}
+	MPOOL_SYSTEM_UNLOCK(dbenv);
+
+	return (ret);
+}
+
+/*
+ * __memp_mf_sync --
+ *	Flush an MPOOLFILE, when no currently open handle is available.
+ *
+ * PUBLIC: int __memp_mf_sync __P((DB_MPOOL *, MPOOLFILE *, int));
+ */
+int
+__memp_mf_sync(dbmp, mfp, region_locked)
+	DB_MPOOL *dbmp;
+	MPOOLFILE *mfp;
+	int region_locked;
+{
+	DB_ENV *dbenv;
+	DB_FH *fhp;
+	int ret, t_ret;
+	char *rpath;
+
+	dbenv = dbmp->dbenv;
+
+	/*
+	 * We need to be holding the region lock: we're using the path name
+	 * and __memp_nameop might try and rename the file.
+	 */
+	if (!region_locked)
+		MPOOL_SYSTEM_LOCK(dbenv);
+
+	if ((ret = __db_appname(dbenv, DB_APP_DATA,
+	    R_ADDR(dbmp->reginfo, mfp->path_off), 0, NULL, &rpath)) == 0) {
+		if ((ret = __os_open(dbenv, rpath, 0, 0, &fhp)) == 0) {
+			ret = __os_fsync(dbenv, fhp);
+			if ((t_ret =
+			    __os_closehandle(dbenv, fhp)) != 0 && ret == 0)
+				ret = t_ret;
+		}
+		__os_free(dbenv, rpath);
+	}
+
+	if (!region_locked)
+		MPOOL_SYSTEM_UNLOCK(dbenv);
+
+	return (ret);
 }
 
 /*
@@ -606,11 +775,13 @@ int __memp_sync_files(dbenv, dbmp)
  *	Close files opened only to flush buffers.
  */
 static int
-__memp_close_flush_files(dbenv, dbmp)
+__memp_close_flush_files(dbenv, dbmp, dosync)
 	DB_ENV *dbenv;
 	DB_MPOOL *dbmp;
+	int dosync;
 {
 	DB_MPOOLFILE *dbmfp;
+	MPOOLFILE *mfp;
 	int ret;
 
 	/*
@@ -618,7 +789,7 @@ __memp_close_flush_files(dbenv, dbmp)
 	 * flush buffers.  There are two cases: first, extent files have to
 	 * be closed so they may be removed when empty.  Second, regular
 	 * files have to be closed so we don't run out of descriptors (for
-	 * example, and application partitioning its data into databases
+	 * example, an application partitioning its data into databases
 	 * based on timestamps, so there's a continually increasing set of
 	 * files).
 	 *
@@ -626,17 +797,35 @@ __memp_close_flush_files(dbenv, dbmp)
 	 * MP_FLUSH flag.  Here we walk through our file descriptor list,
 	 * and, if a file was opened by __memp_bhwrite(), we close it.
 	 */
-retry:	MUTEX_THREAD_LOCK(dbenv, dbmp->mutexp);
+retry:	MUTEX_LOCK(dbenv, dbmp->mutex);
 	for (dbmfp = TAILQ_FIRST(&dbmp->dbmfq);
 	    dbmfp != NULL; dbmfp = TAILQ_NEXT(dbmfp, q))
 		if (F_ISSET(dbmfp, MP_FLUSH)) {
 			F_CLR(dbmfp, MP_FLUSH);
-			MUTEX_THREAD_UNLOCK(dbenv, dbmp->mutexp);
+			MUTEX_UNLOCK(dbenv, dbmp->mutex);
+			if (dosync) {
+				/*
+				 * If we have the only open handle on the file,
+				 * clear the dirty flag so we don't re-open and
+				 * sync it again when discarding the MPOOLFILE
+				 * structure.  Clear the flag before the sync
+				 * so can't race with a thread writing the file.
+				 */
+				mfp = dbmfp->mfp;
+				if (mfp->mpf_cnt == 1) {
+					MUTEX_LOCK(dbenv, mfp->mutex);
+					if (mfp->mpf_cnt == 1)
+						mfp->file_written = 0;
+					MUTEX_UNLOCK(dbenv, mfp->mutex);
+				}
+				if ((ret = __os_fsync(dbenv, dbmfp->fhp)) != 0)
+					return (ret);
+			}
 			if ((ret = __memp_fclose(dbmfp, 0)) != 0)
 				return (ret);
 			goto retry;
 		}
-	MUTEX_THREAD_UNLOCK(dbenv, dbmp->mutexp);
+	MUTEX_UNLOCK(dbenv, dbmp->mutex);
 
 	return (0);
 }
