@@ -1,21 +1,18 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2003
+ * Copyright (c) 1997-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: os_fsync.c,v 12.3 2005/09/07 17:30:20 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: os_fsync.c,v 11.18 2003/02/16 15:53:55 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
-#include <fcntl.h>			/* XXX: Required by __hp3000s900 */
-#include <unistd.h>
+#include <fcntl.h>			/* Required on some platforms. */
 #include <string.h>
 #endif
 
@@ -24,7 +21,7 @@ static const char revid[] = "$Id: os_fsync.c,v 11.18 2003/02/16 15:53:55 bostic
 #ifdef	HAVE_VXWORKS
 #include "ioLib.h"
 
-#define	fsync(fd)	__vx_fsync(fd);
+#define	fsync(fd)	__vx_fsync(fd)
 
 int
 __vx_fsync(fd)
@@ -44,7 +41,7 @@ __vx_fsync(fd)
 #endif
 
 #ifdef __hp3000s900
-#define	fsync(fd)	__mpe_fsync(fd);
+#define	fsync(fd)	__mpe_fsync(fd)
 
 int
 __mpe_fsync(fd)
@@ -69,7 +66,10 @@ __os_fsync(dbenv, fhp)
 	DB_ENV *dbenv;
 	DB_FH *fhp;
 {
-	int ret, retries;
+	int ret;
+
+	/* Check for illegal usage. */
+	DB_ASSERT(F_ISSET(fhp, DB_FH_OPENED) && fhp->fd != -1);
 
 	/*
 	 * Do nothing if the file descriptor has been marked as not requiring
@@ -78,16 +78,16 @@ __os_fsync(dbenv, fhp)
 	if (F_ISSET(fhp, DB_FH_NOSYNC))
 		return (0);
 
-	/* Check for illegal usage. */
-	DB_ASSERT(F_ISSET(fhp, DB_FH_OPENED) && fhp->fd != -1);
-
-	retries = 0;
-	do {
-		ret = DB_GLOBAL(j_fsync) != NULL ?
-		    DB_GLOBAL(j_fsync)(fhp->fd) : fsync(fhp->fd);
-	} while (ret != 0 &&
-	    ((ret = __os_get_errno()) == EINTR || ret == EBUSY) &&
-	    ++retries < DB_RETRY);
+	if (DB_GLOBAL(j_fsync) != NULL)
+		ret = DB_GLOBAL(j_fsync)(fhp->fd);
+	else
+#if defined(F_FULLFSYNC)
+		RETRY_CHK((fcntl(fhp->fd, F_FULLFSYNC, 0)), ret);
+#elif defined(HAVE_FDATASYNC)
+		RETRY_CHK((fdatasync(fhp->fd)), ret);
+#else
+		RETRY_CHK((fsync(fhp->fd)), ret);
+#endif
 
 	if (ret != 0)
 		__db_err(dbenv, "fsync %s", strerror(ret));
