@@ -1,20 +1,16 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1999-2003
+ * Copyright (c) 1999-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: os_root.c,v 12.2 2005/08/10 15:47:26 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: os_root.c,v 11.7 2003/01/08 05:29:32 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
-
-#include <unistd.h>
 #endif
 
 #include "db_int.h"
