@@ -1,16 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2003
+ * Copyright (c) 1997-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: os_stat.c,v 12.1 2005/06/16 20:23:26 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: os_stat.c,v 11.24 2003/02/16 15:54:06 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -31,25 +29,17 @@ __os_exists(path, isdirp)
 	const char *path;
 	int *isdirp;
 {
-	int ret, retries;
 	struct stat sb;
+	int ret;
 
 	if (DB_GLOBAL(j_exists) != NULL)
 		return (DB_GLOBAL(j_exists)(path, isdirp));
 
-	retries = 0;
-	do {
-		ret =
 #ifdef HAVE_VXWORKS
-		    stat((char *)path, &sb);
+	RETRY_CHK((stat((char *)path, &sb)), ret);
 #else
-		    stat(path, &sb);
+	RETRY_CHK((stat(path, &sb)), ret);
 #endif
-		if (ret != 0)
-			ret = __os_get_errno();
-		} while ((ret == EINTR || ret == EBUSY) &&
-		    ++retries < DB_RETRY);
-
 	if (ret != 0)
 		return (ret);
 
@@ -82,8 +72,8 @@ __os_ioinfo(dbenv, path, fhp, mbytesp, bytesp, iosizep)
 	DB_FH *fhp;
 	u_int32_t *mbytesp, *bytesp, *iosizep;
 {
-	int ret, retries;
 	struct stat sb;
+	int ret;
 
 	if (DB_GLOBAL(j_ioinfo) != NULL)
 		return (DB_GLOBAL(j_ioinfo)(path,
@@ -92,12 +82,8 @@ __os_ioinfo(dbenv, path, fhp, mbytesp, bytesp, iosizep)
 	/* Check for illegal usage. */
 	DB_ASSERT(F_ISSET(fhp, DB_FH_OPENED) && fhp->fd != -1);
 
-	retries = 0;
-retry:
-	if (fstat(fhp->fd, &sb) == -1) {
-		if (((ret = __os_get_errno()) == EINTR || ret == EBUSY) &&
-		    ++retries < DB_RETRY)
-			goto retry;
+	RETRY_CHK((fstat(fhp->fd, &sb)), ret);
+	if (ret != 0) {
 		__db_err(dbenv, "fstat: %s", strerror(ret));
 		return (ret);
 	}
