@@ -1,16 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2001
+ * Copyright (c) 1997-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: os_abs.c,v 12.1 2005/06/16 20:23:28 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: os_abs.c,v 11.4 2001/01/25 18:23:00 bostic Exp $";
-#endif /* not lint */
-
 #include "db_int.h"
 
 /*
