@@ -1,21 +1,19 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: os_map.c,v 12.2 2005/06/16 20:23:29 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: os_map.c,v 11.43 2003/02/17 16:05:45 bostic Exp $";
-#endif /* not lint */
-
 #include "db_int.h"
 
 static int __os_map
   __P((DB_ENV *, char *, REGINFO *, DB_FH *, size_t, int, int, int, void **));
-static int __os_unique_name __P((char *, HANDLE, char *, size_t));
+static int __os_unique_name __P((_TCHAR *, HANDLE, _TCHAR *, size_t));
 
 /*
  * __os_r_sysattach --
@@ -36,9 +34,8 @@ __os_r_sysattach(dbenv, infop, rp)
 	 * properly ordered, our caller has already taken care of that.
 	 */
 	if ((ret = __os_open(dbenv, infop->name,
-	    DB_OSO_DIRECT |
 	    F_ISSET(infop, REGION_CREATE_OK) ? DB_OSO_CREATE: 0,
-	    infop->mode, &fhp)) != 0) {
+	    dbenv->db_mode, &fhp)) != 0) {
 		__db_err(dbenv, "%s: %s", infop->name, db_strerror(ret));
 		return (ret);
 	}
@@ -85,13 +82,13 @@ __os_r_sysdetach(dbenv, infop, destroy)
 		infop->wnt_handle = NULL;
 	}
 
-	ret = !UnmapViewOfFile(infop->addr) ? __os_win32_errno() : 0;
+	ret = !UnmapViewOfFile(infop->addr) ? __os_get_errno() : 0;
 	if (ret != 0)
 		__db_err(dbenv, "UnmapViewOfFile: %s", strerror(ret));
 
 	if (!F_ISSET(dbenv, DB_ENV_SYSTEM_MEM) && destroy) {
 		if (F_ISSET(dbenv, DB_ENV_OVERWRITE))
-			(void)__db_overwrite(dbenv, infop->name);
+			(void)__db_file_multi_write(dbenv, infop->name);
 		if ((t_ret = __os_unlink(dbenv, infop->name)) != 0 && ret == 0)
 			ret = t_ret;
 	}
@@ -133,7 +130,7 @@ __os_unmapfile(dbenv, addr, len)
 	if (DB_GLOBAL(j_unmap) != NULL)
 		return (DB_GLOBAL(j_unmap)(addr, len));
 
-	return (!UnmapViewOfFile(addr) ? __os_win32_errno() : 0);
+	return (!UnmapViewOfFile(addr) ? __os_get_errno() : 0);
 }
 
 /*
@@ -145,12 +142,12 @@ __os_unmapfile(dbenv, addr, len)
  *	names), and repeatable (same files, map to same names).  It's not
  *	so easy to do by name.  Should handle not only:
  *
- *		foo.bar  ==  ./foo.bar  ==  c:/whatever_path/foo.bar
+ *		foo.bar == ./foo.bar == c:/whatever_path/foo.bar
  *
  *	but also understand that:
  *
- *		foo.bar  ==  Foo.Bar	(FAT file system)
- *		foo.bar  !=  Foo.Bar	(NTFS)
+ *		foo.bar == Foo.Bar	(FAT file system)
+ *		foo.bar != Foo.Bar	(NTFS)
  *
  *	The best solution is to use the file index, found in the file
  *	information structure (similar to UNIX inode #).
@@ -165,24 +162,24 @@ __os_unmapfile(dbenv, addr, len)
  */
 static int
 __os_unique_name(orig_path, hfile, result_path, result_path_len)
-	char *orig_path, *result_path;
+	_TCHAR *orig_path, *result_path;
 	HANDLE hfile;
 	size_t result_path_len;
 {
 	BY_HANDLE_FILE_INFORMATION fileinfo;
-	char *basename, *p;
+	_TCHAR *basename, *p;
 
 	/*
 	 * In Windows, pathname components are delimited by '/' or '\', and
 	 * if neither is present, we need to strip off leading drive letter
 	 * (e.g. c:foo.txt).
 	 */
-	basename = strrchr(orig_path, '/');
-	p = strrchr(orig_path, '\\');
+	basename = _tcsrchr(orig_path, '/');
+	p = _tcsrchr(orig_path, '\\');
 	if (basename == NULL || (p != NULL && p > basename))
 		basename = p;
 	if (basename == NULL)
-		basename = strrchr(orig_path, ':');
+		basename = _tcsrchr(orig_path, ':');
 
 	if (basename == NULL)
 		basename = orig_path;
@@ -190,10 +187,10 @@ __os_unique_name(orig_path, hfile, result_path, result_path_len)
 		basename++;
 
 	if (!GetFileInformationByHandle(hfile, &fileinfo))
-		return (__os_win32_errno());
+		return (__os_get_errno());
 
-	(void)snprintf(result_path, result_path_len,
-	    "__db_shmem.%8.8lx.%8.8lx.%8.8lx.%8.8lx.%8.8lx.%s",
+	(void)_sntprintf(result_path, result_path_len,
+	    _T("__db_shmem.%8.8lx.%8.8lx.%8.8lx.%8.8lx.%8.8lx.%s"),
 	    fileinfo.dwVolumeSerialNumber,
 	    fileinfo.nFileIndexHigh,
 	    fileinfo.nFileIndexLow,
@@ -220,7 +217,7 @@ __os_map(dbenv, path, infop, fhp, len, is_region, is_system, is_rdonly, addr)
 {
 	HANDLE hMemory;
 	int ret, use_pagefile;
-	char shmem_name[MAXPATHLEN];
+	_TCHAR *tpath, shmem_name[MAXPATHLEN];
 	void *pMemory;
 
 	ret = 0;
@@ -233,9 +230,16 @@ __os_map(dbenv, path, infop, fhp, len, is_region, is_system, is_rdonly, addr)
 	 * If creating a region in system space, get a matching name in the
 	 * paging file namespace.
 	 */
-	if (use_pagefile && (ret = __os_unique_name(
-	    path, fhp->handle, shmem_name, sizeof(shmem_name))) != 0)
-		return (ret);
+	if (use_pagefile) {
+		TO_TSTRING(dbenv, path, tpath, ret);
+		if (ret != 0)
+			return (ret);
+		ret = __os_unique_name(tpath, fhp->handle,
+		    shmem_name, sizeof(shmem_name));
+		FREE_STRING(dbenv, tpath);
+		if (ret != 0)
+			return (ret);
+	}
 
 	/*
 	 * XXX
@@ -275,7 +279,7 @@ __os_map(dbenv, path, infop, fhp, len, is_region, is_system, is_rdonly, addr)
 		    0, (DWORD)len, NULL);
 
 	if (hMemory == NULL) {
-		ret = __os_win32_errno();
+		ret = __os_get_errno();
 		__db_err(dbenv, "OpenFileMapping: %s", strerror(ret));
 		return (__db_panic(dbenv, ret));
 	}
@@ -283,7 +287,7 @@ __os_map(dbenv, path, infop, fhp, len, is_region, is_system, is_rdonly, addr)
 	pMemory = MapViewOfFile(hMemory,
 	    (is_rdonly ? FILE_MAP_READ : FILE_MAP_ALL_ACCESS), 0, 0, len);
 	if (pMemory == NULL) {
-		ret = __os_win32_errno();
+		ret = __os_get_errno();
 		__db_err(dbenv, "MapViewOfFile: %s", strerror(ret));
 		return (__db_panic(dbenv, ret));
 	}
