@@ -1,16 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2003
+ * Copyright (c) 1997-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: os_rename.c,v 12.1 2005/06/16 20:23:30 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: os_rename.c,v 1.15 2003/09/09 15:44:36 mjc Exp $";
-#endif /* not lint */
-
 #include "db_int.h"
 
 /*
@@ -24,6 +22,7 @@ __os_rename(dbenv, oldname, newname, flags)
 	u_int32_t flags;
 {
 	int ret;
+	_TCHAR *toldname, *tnewname;
 
 	ret = 0;
 	if (DB_GLOBAL(j_rename) != NULL) {
@@ -32,15 +31,22 @@ __os_rename(dbenv, oldname, newname, flags)
 		goto done;
 	}
 
-	if (!MoveFile(oldname, newname))
-		ret = __os_win32_errno();
+	TO_TSTRING(dbenv, oldname, toldname, ret);
+	if (ret != 0)
+		goto done;
+	TO_TSTRING(dbenv, newname, tnewname, ret);
+	if (ret != 0)
+		goto err;
+
+	if (!MoveFile(toldname, tnewname))
+		ret = __os_get_errno();
 
 	if (ret == EEXIST) {
 		ret = 0;
 		if (__os_is_winnt()) {
 			if (!MoveFileEx(
-			    oldname, newname, MOVEFILE_REPLACE_EXISTING))
-				ret = __os_win32_errno();
+			    toldname, tnewname, MOVEFILE_REPLACE_EXISTING))
+				ret = __os_get_errno();
 		} else {
 			/*
 			 * There is no MoveFileEx for Win9x/Me, so we have to
@@ -49,12 +55,15 @@ __os_rename(dbenv, oldname, newname, flags)
 			 * refer to the same file, so we don't need to check
 			 * that here.
 			 */
-			(void)DeleteFile(newname);
-			if (!MoveFile(oldname, newname))
-				ret = __os_win32_errno();
+			(void)DeleteFile(tnewname);
+			if (!MoveFile(toldname, tnewname))
+				ret = __os_get_errno();
 		}
 	}
 
+	FREE_STRING(dbenv, tnewname);
+err:	FREE_STRING(dbenv, toldname);
+
 done:	if (ret != 0 && flags == 0)
 		__db_err(dbenv,
 		    "Rename %s %s: %s", oldname, newname, strerror(ret));
