@@ -1,40 +1,31 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1999-2003
+ * Copyright (c) 1999-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: qam_method.c,v 12.2 2005/09/28 17:45:01 margo Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: qam_method.c,v 11.64 2003/10/01 20:03:43 ubell Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
-
-#include <string.h>
 #endif
 
 #include "db_int.h"
 #include "dbinc/db_page.h"
 #include "dbinc/db_shash.h"
 #include "dbinc/db_am.h"
-#include "dbinc/fop.h"
 #include "dbinc/lock.h"
 #include "dbinc/mp.h"
 #include "dbinc/qam.h"
 #include "dbinc/txn.h"
 
-static int __qam_get_extentsize __P((DB *, u_int32_t *));
+static int __qam_rr __P((DB *, DB_TXN *,
+	       const char *, const char *, const char *, qam_name_op));
 static int __qam_set_extentsize __P((DB *, u_int32_t));
 
-struct __qam_cookie {
-	DB_LSN lsn;
-	QUEUE_FILELIST *filelist;
-};
-
 /*
  * __qam_db_create --
  *	Queue specific initialization of the DB structure.
@@ -116,7 +107,13 @@ again:
 	return (ret);
 }
 
-static int
+/*
+ * __qam_get_extentsize --
+ *	The DB->q_get_extentsize method.
+ *
+ * PUBLIC: int __qam_get_extentsize __P((DB *, u_int32_t *));
+ */
+int
 __qam_get_extentsize(dbp, q_extentsizep)
 	DB *dbp;
 	u_int32_t *q_extentsizep;
@@ -143,22 +140,24 @@ __qam_set_extentsize(dbp, extentsize)
 }
 
 /*
- * __db_prqueue --
- *	Print out a queue
+ * __queue_pageinfo -
+ *	Given a dbp, get first/last page information about a queue.
  *
- * PUBLIC: int __db_prqueue __P((DB *, FILE *, u_int32_t));
+ * PUBLIC: int __queue_pageinfo __P((DB *, db_pgno_t *, db_pgno_t *,
+ * PUBLIC:       int *, int, u_int32_t));
  */
 int
-__db_prqueue(dbp, fp, flags)
+__queue_pageinfo(dbp, firstp, lastp, emptyp, prpage, flags)
 	DB *dbp;
-	FILE *fp;
+	db_pgno_t *firstp, *lastp;
+	int *emptyp;
+	int prpage;
 	u_int32_t flags;
 {
 	DB_MPOOLFILE *mpf;
-	PAGE *h;
 	QMETA *meta;
-	db_pgno_t first, i, last, pg_ext, stop;
-	int ret, t_ret;
+	db_pgno_t first, i, last;
+	int empty, ret, t_ret;
 
 	mpf = dbp->mpf;
 
@@ -168,18 +167,55 @@ __db_prqueue(dbp, fp, flags)
 		return (ret);
 
 	first = QAM_RECNO_PAGE(dbp, meta->first_recno);
-	last = QAM_RECNO_PAGE(dbp, meta->cur_recno);
+	last = QAM_RECNO_PAGE(
+	    dbp, meta->cur_recno == 1 ? 1 : meta->cur_recno - 1);
+
+	empty = meta->cur_recno == meta->first_recno;
+	if (firstp != NULL)
+		*firstp = first;
+	if (lastp != NULL)
+		*lastp = last;
+	if (emptyp != NULL)
+		*emptyp = empty;
+#ifdef HAVE_STATISTICS
+	if (prpage)
+		ret = __db_prpage(dbp, (PAGE *)meta, flags);
+#else
+	COMPQUIET(prpage, 0);
+	COMPQUIET(flags, 0);
+#endif
 
-	ret = __db_prpage(dbp, (PAGE *)meta, fp, flags);
 	if ((t_ret = __memp_fput(mpf, meta, 0)) != 0 && ret == 0)
 		ret = t_ret;
 
-	if (ret != 0)
+	return (ret);
+}
+
+#ifdef HAVE_STATISTICS
+/*
+ * __db_prqueue --
+ *	Print out a queue
+ *
+ * PUBLIC: int __db_prqueue __P((DB *, u_int32_t));
+ */
+int
+__db_prqueue(dbp, flags)
+	DB *dbp;
+	u_int32_t flags;
+{
+	PAGE *h;
+	db_pgno_t first, i, last, pg_ext, stop;
+	int empty, ret;
+
+	if ((ret = __queue_pageinfo(dbp, &first, &last, &empty, 1, flags)) != 0)
+		return (ret);
+
+	if (empty || ret != 0)
 		return (ret);
 
 	i = first;
 	if (first > last)
-		stop = QAM_RECNO_PAGE(dbp, UINT32_T_MAX);
+		stop = QAM_RECNO_PAGE(dbp, UINT32_MAX);
 	else
 		stop = last;
 
@@ -194,12 +230,12 @@ begin:
 				return (ret);
 			}
 			if (ret == ENOENT || ret == DB_PAGE_NOTFOUND) {
-				i += pg_ext - ((i - 1) % pg_ext) - 1;
+				i += (pg_ext - ((i - 1) % pg_ext)) - 1;
 				continue;
 			}
 			return (ret);
 		}
-		(void)__db_prpage(dbp, h, fp, flags);
+		(void)__db_prpage(dbp, h, flags);
 		if ((ret = __qam_fput(dbp, i, h, 0)) != 0)
 			return (ret);
 	}
@@ -212,126 +248,64 @@ begin:
 	}
 	return (0);
 }
+#endif
 
 /*
- * __qam_remove
+ * __qam_remove --
  *	Remove method for a Queue.
  *
- * PUBLIC: int __qam_remove __P((DB *,
- * PUBLIC:     DB_TXN *, const char *, const char *, DB_LSN *));
+ * PUBLIC: int __qam_remove __P((DB *, DB_TXN *, const char *, const char *));
  */
 int
-__qam_remove(dbp, txn, name, subdb, lsnp)
+__qam_remove(dbp, txn, name, subdb)
 	DB *dbp;
 	DB_TXN *txn;
 	const char *name, *subdb;
-	DB_LSN *lsnp;
 {
-	DB_ENV *dbenv;
-	DB *tmpdbp;
-	QUEUE *qp;
-	int ret, needclose, t_ret;
-
-	COMPQUIET(lsnp, NULL);
-
-	dbenv = dbp->dbenv;
-	ret = 0;
-	needclose = 0;
-
-	PANIC_CHECK(dbenv);
-
-	/*
-	 * Subdatabases.
-	 */
-	if (subdb != NULL) {
-		__db_err(dbenv,
-		    "Queue does not support multiple databases per file");
-		ret = EINVAL;
-		goto err;
-	}
-
-	/*
-	 * Since regular remove no longer opens the database, we may have
-	 * to do it here.
-	 */
-	if (F_ISSET(dbp, DB_AM_OPEN_CALLED))
-		tmpdbp = dbp;
-	else {
-		if ((ret = db_create(&tmpdbp, dbenv, 0)) != 0)
-			return (ret);
-		/*
-		 * We need to make sure we don't self-deadlock, so give
-		 * this dbp the same locker as the incoming one.
-		 */
-		tmpdbp->lid = dbp->lid;
-
-		/*
-		 * If this is a transactional dbp and the open fails, then
-		 * the transactional abort will close the dbp.  If it's not
-		 * a transactional open, then we always have to close it
-		 * even if the open fails.  Once the open has succeeded,
-		 * then we will always want to close it.
-		 */
-		if (txn == NULL)
-			needclose = 1;
-		if ((ret = __db_open(tmpdbp,
-		    txn, name, NULL, DB_QUEUE, 0, 0, PGNO_BASE_MD)) != 0)
-			goto err;
-		needclose = 1;
-	}
-
-	qp = (QUEUE *)tmpdbp->q_internal;
-
-	if (qp->page_ext != 0) 
-		ret = __qam_nameop(tmpdbp, txn, NULL, QAM_NAME_REMOVE);
-
-err:	if (needclose) {
-		/*
-		 * Since we copied the lid from the dbp, we'd better not
-		 * free it here.
-		 */
-		tmpdbp->lid = DB_LOCK_INVALIDID;
-
-		/* We need to remove the lockevent we associated with this. */
-		if (txn != NULL)
-			__txn_remlock(dbenv,
-			    txn, &tmpdbp->handle_lock, DB_LOCK_INVALIDID);
-
-		if ((t_ret =
-		    __db_close(tmpdbp, txn, DB_NOSYNC)) != 0 && ret == 0)
-			ret = t_ret;
-	}
-
-	return (ret);
+	return (__qam_rr(dbp, txn, name, subdb, NULL, QAM_NAME_REMOVE));
 }
 
 /*
- * __qam_rename
- *	Rename method for Queue.
+ * __qam_rename --
+ *	Rename method for a Queue.
  *
- * PUBLIC: int __qam_rename __P((DB *, DB_TXN *,
- * PUBLIC:     const char *, const char *, const char *));
+ * PUBLIC: int __qam_rename __P((DB *,
+ * PUBLIC:         DB_TXN *, const char *, const char *, const char *));
  */
 int
-__qam_rename(dbp, txn, filename, subdb, newname)
+__qam_rename(dbp, txn, name, subdb, newname)
 	DB *dbp;
 	DB_TXN *txn;
-	const char *filename, *subdb, *newname;
+	const char *name, *subdb, *newname;
+{
+	return (__qam_rr(dbp, txn, name, subdb, newname, QAM_NAME_RENAME));
+}
+
+/*
+ * __qam_rr --
+ *	Remove/Rename method for a Queue.
+ */
+static int
+__qam_rr(dbp, txn, name, subdb, newname, op)
+	DB *dbp;
+	DB_TXN *txn;
+	const char *name, *subdb, *newname;
+	qam_name_op op;
 {
 	DB_ENV *dbenv;
 	DB *tmpdbp;
 	QUEUE *qp;
-	int ret, needclose, t_ret;
+	int ret, t_ret;
 
 	dbenv = dbp->dbenv;
 	ret = 0;
-	needclose = 0;
 
-	if (subdb != NULL) {
+	PANIC_CHECK(dbenv);
+
+	if (subdb != NULL && name != NULL) {
 		__db_err(dbenv,
 		    "Queue does not support multiple databases per file");
-		ret = EINVAL;
-		goto err;
+		return (EINVAL);
 	}
 
 	/*
@@ -343,27 +317,32 @@ __qam_rename(dbp, txn, filename, subdb, newname)
 	else {
 		if ((ret = db_create(&tmpdbp, dbenv, 0)) != 0)
 			return (ret);
-		/* Copy the incoming locker so we don't self-deadlock. */
+
+		/*
+		 * We need to make sure we don't self-deadlock, so give
+		 * this dbp the same locker as the incoming one.
+		 */
 		tmpdbp->lid = dbp->lid;
-		needclose = 1;
-		if ((ret = __db_open(tmpdbp,
-		    txn, filename, NULL, DB_QUEUE, 0, 0, PGNO_BASE_MD)) != 0)
+		if ((ret = __db_open(tmpdbp, txn,
+		    name, NULL, DB_QUEUE, DB_RDONLY, 0, PGNO_BASE_MD)) != 0)
 			goto err;
 	}
 
 	qp = (QUEUE *)tmpdbp->q_internal;
-
 	if (qp->page_ext != 0)
-		ret = __qam_nameop(tmpdbp, txn, newname, QAM_NAME_RENAME);
+		ret = __qam_nameop(tmpdbp, txn, newname, op);
 
-err:	if (needclose) {
-		/* We copied this, so we mustn't free it. */
+	if (!F_ISSET(dbp, DB_AM_OPEN_CALLED)) {
+err:		/*
+		 * Since we copied the locker ID from the dbp, we'd better not
+		 * free it here.
+		 */
 		tmpdbp->lid = DB_LOCK_INVALIDID;
 
-		/* We need to remove the lockevent we associated with this. */
+		/* We need to remove the lock event we associated with this. */
 		if (txn != NULL)
-		    __txn_remlock(dbenv,
-			txn, &tmpdbp->handle_lock, DB_LOCK_INVALIDID);
+			__txn_remlock(dbenv,
+			    txn, &tmpdbp->handle_lock, DB_LOCK_INVALIDID);
 
 		if ((t_ret =
 		    __db_close(tmpdbp, txn, DB_NOSYNC)) != 0 && ret == 0)
@@ -371,3 +350,38 @@ err:	if (needclose) {
 	}
 	return (ret);
 }
+
+/*
+ * __qam_map_flags --
+ *	Map queue-specific flags from public to the internal values.
+ *
+ * PUBLIC: void __qam_map_flags __P((DB *, u_int32_t *, u_int32_t *));
+ */
+void
+__qam_map_flags(dbp, inflagsp, outflagsp)
+	DB *dbp;
+	u_int32_t *inflagsp, *outflagsp;
+{
+	COMPQUIET(dbp, NULL);
+
+	if (FLD_ISSET(*inflagsp, DB_INORDER)) {
+		FLD_SET(*outflagsp, DB_AM_INORDER);
+		FLD_CLR(*inflagsp, DB_INORDER);
+	}
+}
+
+/*
+ * __qam_set_flags --
+ *	Set queue-specific flags.
+ *
+ * PUBLIC: int __qam_set_flags __P((DB *, u_int32_t *flagsp));
+ */
+int
+__qam_set_flags(dbp, flagsp)
+	DB *dbp;
+	u_int32_t *flagsp;
+{
+
+	__qam_map_flags(dbp, flagsp, &dbp->flags);
+	return (0);
+}
