@@ -1,16 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1999-2003
+ * Copyright (c) 1999-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: qam_rec.c,v 12.8 2005/10/20 18:57:13 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: qam_rec.c,v 11.75 2003/08/17 23:38:14 ubell Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -25,6 +23,20 @@ static const char revid[] = "$Id: qam_rec.c,v 11.75 2003/08/17 23:38:14 ubell Ex
 #include "dbinc/log.h"
 #include "dbinc/mp.h"
 #include "dbinc/qam.h"
+#include "dbinc/txn.h"
+
+/*
+ * LSNs in queue data pages are advisory.  They do not have to be accurate
+ * as all operations are idempotent on records.  They should not be rolled
+ * forward during recovery as committed transaction may obscure updates from
+ * an incomplete transaction that updates the same page.  The incomplete
+ * transaction may be completed during a later hot backup cycle.
+ */
+
+/* Determine if we are restoring prepared transactions from __txn_recover. */
+#define	IS_IN_RESTORE(dbenv)						 \
+	(((DB_TXNREGION *)((DB_TXNMGR *)				 \
+	     (dbenv)->tx_handle)->reginfo.primary)->stat.st_nrestores != 0)
 
 /*
  * __qam_incfirst_recover --
@@ -50,10 +62,13 @@ __qam_incfirst_recover(dbenv, dbtp, lsnp, op, info)
 	QMETA *meta;
 	QUEUE_CURSOR *cp;
 	db_pgno_t metapg;
-	int exact, modified, ret, rec_ext;
+	u_int32_t rec_ext;
+	int exact, modified, ret, t_ret;
 
+	LOCK_INIT(lock);
+	COMPQUIET(meta, NULL);
 	REC_PRINT(__qam_incfirst_print);
-	REC_INTRO(__qam_incfirst_read, 1);
+	REC_INTRO(__qam_incfirst_read, 1, 1);
 
 	metapg = ((QUEUE *)file_dbp->q_internal)->q_meta;
 
@@ -71,8 +86,7 @@ __qam_incfirst_recover(dbenv, dbtp, lsnp, op, info)
 			meta->dbmeta.type = P_QAMMETA;
 		} else {
 			*lsnp = argp->prev_lsn;
-			ret = 0;
-			(void)__LPUT(dbc, lock);
+			ret = __LPUT(dbc, lock);
 			goto out;
 		}
 	}
@@ -102,8 +116,9 @@ __qam_incfirst_recover(dbenv, dbtp, lsnp, op, info)
 			LSN(meta) = *lsnp;
 			modified = 1;
 		}
-		rec_ext = 0;
-		if (meta->page_ext != 0)
+		if (meta->page_ext == 0)
+			rec_ext = 0;
+		else
 			rec_ext = meta->page_ext * meta->rec_page;
 		cp = (QUEUE_CURSOR *)dbc->internal;
 		if (meta->first_recno == RECNO_OOB)
@@ -113,8 +128,9 @@ __qam_incfirst_recover(dbenv, dbtp, lsnp, op, info)
 			if ((ret = __qam_position(dbc,
 			    &meta->first_recno, QAM_READ, &exact)) != 0)
 				goto err;
-			if (cp->page != NULL)
-				__qam_fput(file_dbp, cp->pgno, cp->page, 0);
+			if (cp->page != NULL && (ret =
+			    __qam_fput(file_dbp, cp->pgno, cp->page, 0)) != 0)
+				goto err;
 
 			if (exact == 1)
 				break;
@@ -130,17 +146,18 @@ __qam_incfirst_recover(dbenv, dbtp, lsnp, op, info)
 		}
 	}
 
-	if ((ret = __memp_fput(mpf, meta, modified ? DB_MPOOL_DIRTY : 0)) != 0)
-		goto err1;
-
-	(void)__LPUT(dbc, lock);
+	ret = __memp_fput(mpf, meta, modified ? DB_MPOOL_DIRTY : 0);
+	if ((t_ret = __LPUT(dbc, lock)) != 0 && ret == 0)
+		ret = t_ret;
+	if (ret != 0)
+		goto out;
 
 done:	*lsnp = argp->prev_lsn;
 	ret = 0;
 
 	if (0) {
 err:		(void)__memp_fput(mpf, meta, 0);
-err1:		(void)__LPUT(dbc, lock);
+		(void)__LPUT(dbc, lock);
 	}
 
 out:	REC_CLOSE;
@@ -168,11 +185,12 @@ __qam_mvptr_recover(dbenv, dbtp, lsnp, op, info)
 	DB_LOCK lock;
 	DB_MPOOLFILE *mpf;
 	QMETA *meta;
+	QUEUE_CURSOR *cp;
 	db_pgno_t metapg;
-	int cmp_n, cmp_p, modified, ret;
+	int cmp_n, cmp_p, exact, modified, ret;
 
 	REC_PRINT(__qam_mvptr_print);
-	REC_INTRO(__qam_mvptr_read, 1);
+	REC_INTRO(__qam_mvptr_read, 1, 1);
 
 	metapg = ((QUEUE *)file_dbp->q_internal)->q_meta;
 
@@ -190,8 +208,7 @@ __qam_mvptr_recover(dbenv, dbtp, lsnp, op, info)
 			meta->dbmeta.type = P_QAMMETA;
 		} else {
 			*lsnp = argp->prev_lsn;
-			ret = 0;
-			(void)__LPUT(dbc, lock);
+			ret = __LPUT(dbc, lock);
 			goto out;
 		}
 	}
@@ -203,6 +220,9 @@ __qam_mvptr_recover(dbenv, dbtp, lsnp, op, info)
 	/*
 	 * Under normal circumstances, we never undo a movement of one of
 	 * the pointers.  Just move them along regardless of abort/commit.
+	 * When going forward we need to verify that this is really where
+	 * the pointer belongs.  A transaction may roll back and reinsert
+	 * a record that was missing at the time of this action.
 	 *
 	 * If we're undoing a truncate, we need to reset the pointers to
 	 * their state before the truncate.
@@ -222,11 +242,40 @@ __qam_mvptr_recover(dbenv, dbtp, lsnp, op, info)
 			modified = 1;
 		}
 	} else if (op == DB_TXN_APPLY || cmp_p == 0) {
-		if (argp->opcode & QAM_SETFIRST)
-			meta->first_recno = argp->new_first;
+		cp = (QUEUE_CURSOR *)dbc->internal;
+		if ((argp->opcode & QAM_SETFIRST) &&
+		    meta->first_recno == argp->old_first) {
+			if (argp->old_first > argp->new_first)
+				meta->first_recno = argp->new_first;
+			else {
+				if ((ret = __qam_position(dbc,
+				    &meta->first_recno, QAM_READ, &exact)) != 0)
+					goto err;
+				if (!exact)
+					meta->first_recno = argp->new_first;
+				if (cp->page != NULL &&
+				    (ret = __qam_fput(file_dbp,
+				    cp->pgno, cp->page, 0)) != 0)
+					goto err;
+			}
+		}
 
-		if (argp->opcode & QAM_SETCUR)
-			meta->cur_recno = argp->new_cur;
+		if ((argp->opcode & QAM_SETCUR) &&
+		    meta->cur_recno == argp->old_cur) {
+			if (argp->old_cur < argp->new_cur)
+				meta->cur_recno = argp->new_cur;
+			else {
+				if ((ret = __qam_position(dbc,
+				    &meta->cur_recno, QAM_READ, &exact)) != 0)
+					goto err;
+				if (!exact)
+					meta->cur_recno = argp->new_cur;
+				if (cp->page != NULL
+				    && (ret = __qam_fput(file_dbp,
+				    cp->pgno, cp->page, 0)) != 0)
+					goto err;
+			}
+		}
 
 		modified = 1;
 		meta->dbmeta.lsn = *lsnp;
@@ -235,11 +284,17 @@ __qam_mvptr_recover(dbenv, dbtp, lsnp, op, info)
 	if ((ret = __memp_fput(mpf, meta, modified ? DB_MPOOL_DIRTY : 0)) != 0)
 		goto out;
 
-	(void)__LPUT(dbc, lock);
+	if ((ret = __LPUT(dbc, lock)) != 0)
+		goto out;
 
 done:	*lsnp = argp->prev_lsn;
 	ret = 0;
 
+	if (0) {
+err:		(void)__memp_fput(mpf, meta, 0);
+		(void)__LPUT(dbc, lock);
+	}
+
 out:	REC_CLOSE;
 }
 
@@ -268,11 +323,12 @@ __qam_del_recover(dbenv, dbtp, lsnp, op, info)
 	QMETA *meta;
 	QPAGE *pagep;
 	db_pgno_t metapg;
-	int cmp_n, modified, ret;
+	int cmp_n, modified, ret, t_ret;
 
 	COMPQUIET(info, NULL);
+	COMPQUIET(pagep, NULL);
 	REC_PRINT(__qam_del_print);
-	REC_INTRO(__qam_del_read, 1);
+	REC_INTRO(__qam_del_read, 1, 1);
 
 	if ((ret = __qam_fget(file_dbp,
 	    &argp->pgno, DB_MPOOL_CREATE, &pagep)) != 0)
@@ -303,10 +359,13 @@ __qam_del_recover(dbenv, dbtp, lsnp, op, info)
 		    meta->first_recno -
 		    argp->recno < argp->recno - meta->cur_recno))) {
 			meta->first_recno = argp->recno;
-			(void)__memp_fput(mpf, meta, DB_MPOOL_DIRTY);
+			ret = __memp_fput(mpf, meta, DB_MPOOL_DIRTY);
 		} else
-			(void)__memp_fput(mpf, meta, 0);
-		(void)__LPUT(dbc, lock);
+			ret = __memp_fput(mpf, meta, 0);
+		if ((t_ret = __LPUT(dbc, lock)) != 0 && ret == 0)
+			ret = t_ret;
+		if (ret != 0)
+			goto err;
 
 		/* Need to undo delete - mark the record as present */
 		qp = QAM_GET_RECORD(file_dbp, pagep, argp->indx);
@@ -319,15 +378,26 @@ __qam_del_recover(dbenv, dbtp, lsnp, op, info)
 		 * foul up a concurrent put.  Having too late an LSN
 		 * is harmless in queue except when we're determining
 		 * what we need to roll forward during recovery.  [#2588]
+		 * If we are aborting a restored transaction then it
+		 * might get rolled forward later so the LSN needs to
+		 * be correct in that case too. [#12181]
 		 */
-		if (op == DB_TXN_BACKWARD_ROLL && cmp_n <= 0)
+		if (cmp_n <= 0 &&
+		      (op == DB_TXN_BACKWARD_ROLL || IS_IN_RESTORE(dbenv)))
 			LSN(pagep) = argp->lsn;
 		modified = 1;
 	} else if (op == DB_TXN_APPLY || (cmp_n > 0 && DB_REDO(op))) {
 		/* Need to redo delete - clear the valid bit */
 		qp = QAM_GET_RECORD(file_dbp, pagep, argp->indx);
 		F_CLR(qp, QAM_VALID);
-		LSN(pagep) = *lsnp;
+		/*
+		 * We only move the LSN forward during replication.
+		 * During recovery we could obsucre an update from
+		 * a partially completed transaction while processing
+		 * a hot backup.  [#13823]
+		 */
+		if (op == DB_TXN_APPLY)
+			LSN(pagep) = *lsnp;
 		modified = 1;
 	}
 	if ((ret = __qam_fput(file_dbp,
@@ -367,11 +437,12 @@ __qam_delext_recover(dbenv, dbtp, lsnp, op, info)
 	QMETA *meta;
 	QPAGE *pagep;
 	db_pgno_t metapg;
-	int cmp_n, modified, ret;
+	int cmp_n, modified, ret, t_ret;
 
 	COMPQUIET(info, NULL);
+	COMPQUIET(pagep, NULL);
 	REC_PRINT(__qam_delext_print);
-	REC_INTRO(__qam_delext_read, 1);
+	REC_INTRO(__qam_delext_read, 1, 1);
 
 	if ((ret = __qam_fget(file_dbp, &argp->pgno, 0, &pagep)) != 0) {
 		if (ret != DB_PAGE_NOTFOUND && ret != ENOENT)
@@ -412,10 +483,13 @@ __qam_delext_recover(dbenv, dbtp, lsnp, op, info)
 		    meta->first_recno -
 		    argp->recno < argp->recno - meta->cur_recno))) {
 			meta->first_recno = argp->recno;
-			(void)__memp_fput(mpf, meta, DB_MPOOL_DIRTY);
+			ret = __memp_fput(mpf, meta, DB_MPOOL_DIRTY);
 		} else
-			(void)__memp_fput(mpf, meta, 0);
-		(void)__LPUT(dbc, lock);
+			ret = __memp_fput(mpf, meta, 0);
+		if ((t_ret = __LPUT(dbc, lock)) != 0 && ret == 0)
+			ret = t_ret;
+		if (ret != 0)
+			goto err;
 
 		if ((ret = __qam_pitem(dbc, pagep,
 		    argp->indx, argp->recno, &argp->data)) != 0)
@@ -429,14 +503,22 @@ __qam_delext_recover(dbenv, dbtp, lsnp, op, info)
 		 * is harmless in queue except when we're determining
 		 * what we need to roll forward during recovery.  [#2588]
 		 */
-		if (op == DB_TXN_BACKWARD_ROLL && cmp_n <= 0)
+		if (cmp_n <= 0 &&
+		      (op == DB_TXN_BACKWARD_ROLL || IS_IN_RESTORE(dbenv)))
 			LSN(pagep) = argp->lsn;
 		modified = 1;
 	} else if (op == DB_TXN_APPLY || (cmp_n > 0 && DB_REDO(op))) {
 		/* Need to redo delete - clear the valid bit */
 		qp = QAM_GET_RECORD(file_dbp, pagep, argp->indx);
 		F_CLR(qp, QAM_VALID);
-		LSN(pagep) = *lsnp;
+		/*
+		 * We only move the LSN forward during replication.
+		 * During recovery we could obsucre an update from
+		 * a partially completed transaction while processing
+		 * a hot backup.  [#13823]
+		 */
+		if (op == DB_TXN_APPLY)
+			LSN(pagep) = *lsnp;
 		modified = 1;
 	}
 	if ((ret = __qam_fput(file_dbp,
@@ -478,8 +560,9 @@ __qam_add_recover(dbenv, dbtp, lsnp, op, info)
 	int cmp_n, meta_dirty, modified, ret;
 
 	COMPQUIET(info, NULL);
+	COMPQUIET(pagep, NULL);
 	REC_PRINT(__qam_add_print);
-	REC_INTRO(__qam_add_read, 1);
+	REC_INTRO(__qam_add_read, 1, 1);
 
 	modified = 0;
 	if ((ret = __qam_fget(file_dbp, &argp->pgno, 0, &pagep)) != 0) {
@@ -529,7 +612,14 @@ __qam_add_recover(dbenv, dbtp, lsnp, op, info)
 			if ((ret = __qam_pitem(dbc,
 			    pagep, argp->indx, argp->recno, &argp->data)) != 0)
 				goto err;
-			LSN(pagep) = *lsnp;
+			/*
+			 * We only move the LSN forward during replication.
+			 * During recovery we could obsucre an update from
+			 * a partially completed transaction while processing
+			 * a hot backup.  [#13823]
+			 */
+			if (op == DB_TXN_APPLY)
+				LSN(pagep) = *lsnp;
 			modified = 1;
 		}
 	} else if (DB_UNDO(op)) {
@@ -563,7 +653,8 @@ __qam_add_recover(dbenv, dbtp, lsnp, op, info)
 		 * is harmless in queue except when we're determining
 		 * what we need to roll forward during recovery.  [#2588]
 		 */
-		if (op == DB_TXN_BACKWARD_ROLL && cmp_n <= 0)
+		if (cmp_n <= 0 &&
+		      (op == DB_TXN_BACKWARD_ROLL || IS_IN_RESTORE(dbenv)))
 			LSN(pagep) = argp->lsn;
 	}
 
