@@ -1,17 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2000-2003
+ * Copyright (c) 2000-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: gen_client_ret.c,v 12.2 2005/07/21 18:21:30 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: gen_client_ret.c,v 1.59 2003/06/14 17:56:02 bostic Exp $";
-#endif /* not lint */
-
-#ifdef HAVE_RPC
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -20,14 +17,21 @@ static const char revid[] = "$Id: gen_client_ret.c,v 1.59 2003/06/14 17:56:02 bo
 #include <string.h>
 #endif
 
+#include "db_server.h"
+
 #include "db_int.h"
 #include "dbinc/db_page.h"
 #include "dbinc/db_am.h"
 #include "dbinc/txn.h"
-
-#include "dbinc_auto/db_server.h"
 #include "dbinc_auto/rpc_client_ext.h"
 
+#define	FREE_IF_CHANGED(dbtp, orig)	do {				\
+	if ((dbtp)->data != NULL && (dbtp)->data != orig) {		\
+		__os_free(dbenv, (dbtp)->data);				\
+		(dbtp)->data = NULL;					\
+	}								\
+} while (0)
+
 /*
  * PUBLIC: int __dbcl_env_create_ret
  * PUBLIC:     __P((DB_ENV *, long, __env_create_reply *));
@@ -125,15 +129,15 @@ __dbcl_txn_abort_ret(txnp, replyp)
 }
 
 /*
- * PUBLIC: int __dbcl_txn_begin_ret __P((DB_ENV *,
- * PUBLIC:     DB_TXN *, DB_TXN **, u_int32_t, __txn_begin_reply *));
+ * PUBLIC: int __dbcl_env_txn_begin_ret __P((DB_ENV *,
+ * PUBLIC:     DB_TXN *, DB_TXN **, u_int32_t, __env_txn_begin_reply *));
  */
 int
-__dbcl_txn_begin_ret(envp, parent, txnpp, flags, replyp)
+__dbcl_env_txn_begin_ret(envp, parent, txnpp, flags, replyp)
 	DB_ENV *envp;
 	DB_TXN *parent, **txnpp;
 	u_int32_t flags;
-	__txn_begin_reply *replyp;
+	__env_txn_begin_reply *replyp;
 {
 	DB_TXN *txn;
 	int ret;
@@ -145,7 +149,13 @@ __dbcl_txn_begin_ret(envp, parent, txnpp, flags, replyp)
 
 	if ((ret = __os_calloc(envp, 1, sizeof(DB_TXN), &txn)) != 0)
 		return (ret);
-	__dbcl_txn_setup(envp, txn, parent, replyp->txnidcl_id);
+	/*
+	 * !!!
+	 * Cast the txnidcl_id to 32-bits.  We don't want to change the
+	 * size of the txn structure.  But if we're running on 64-bit
+	 * machines, we could overflow.  Ignore for now.
+	 */
+	__dbcl_txn_setup(envp, txn, parent, (u_int32_t)replyp->txnidcl_id);
 	*txnpp = txn;
 	return (replyp->status);
 }
@@ -183,17 +193,17 @@ __dbcl_txn_discard_ret(txnp, flags, replyp)
 }
 
 /*
- * PUBLIC: int __dbcl_txn_recover_ret __P((DB_ENV *, DB_PREPLIST *, long,
- * PUBLIC:      long *, u_int32_t, __txn_recover_reply *));
+ * PUBLIC: int __dbcl_env_txn_recover_ret __P((DB_ENV *, DB_PREPLIST *, long,
+ * PUBLIC:      long *, u_int32_t, __env_txn_recover_reply *));
  */
 int
-__dbcl_txn_recover_ret(dbenv, preplist, count, retp, flags, replyp)
+__dbcl_env_txn_recover_ret(dbenv, preplist, count, retp, flags, replyp)
 	DB_ENV * dbenv;
 	DB_PREPLIST * preplist;
 	long count;
 	long * retp;
 	u_int32_t flags;
-	__txn_recover_reply *replyp;
+	__env_txn_recover_reply *replyp;
 {
 	DB_PREPLIST *prep;
 	DB_TXN *txnarray, *txn;
@@ -227,7 +237,7 @@ __dbcl_txn_recover_ret(dbenv, preplist, count, retp, flags, replyp)
 	while (i++ < replyp->retcount) {
 		__dbcl_txn_setup(dbenv, txn, NULL, *txnid);
 		prep->txn = txn;
-		memcpy(&prep->gid, gid, DB_XIDDATASIZE);
+		memcpy(prep->gid, gid, DB_XIDDATASIZE);
 		/*
 		 * Now increment all our array pointers.
 		 */
@@ -319,8 +329,8 @@ __dbcl_db_get_ret(dbp, txnp, key, data, flags, replyp)
 	 * If an error on copying 'data' and we allocated for 'key'
 	 * free it before returning the error.
 	 */
-	if (ret && oldkey != NULL)
-		__os_free(dbenv, key->data);
+	if (ret)
+		FREE_IF_CHANGED(key, oldkey);
 	return (ret);
 }
 
@@ -367,13 +377,13 @@ __dbcl_db_open_ret(dbp, txn, name, subdb, type, flags, mode, replyp)
 	COMPQUIET(txn, NULL);
 	COMPQUIET(name, NULL);
 	COMPQUIET(subdb, NULL);
-	COMPQUIET(type, 0);
+	COMPQUIET(type, DB_UNKNOWN);
 	COMPQUIET(flags, 0);
 	COMPQUIET(mode, 0);
 
 	if (replyp->status == 0) {
 		dbp->cl_id = replyp->dbcl_id;
-		dbp->type = replyp->type;
+		dbp->type = (DBTYPE)replyp->type;
 		/*
 		 * We get back the database's byteorder on the server.
 		 * Determine if our byteorder is the same or not by
@@ -387,11 +397,10 @@ __dbcl_db_open_ret(dbp, txn, name, subdb, type, flags, mode, replyp)
 		(void)__db_set_lorder(dbp, replyp->lorder);
 
 		/*
-		 * XXX
-		 * This is only for Tcl which peeks at the dbp flags.
-		 * When dbp->get_flags exists, this should go away.
+		 * Explicitly set DB_AM_OPEN_CALLED since open is now
+		 * successfully completed.
 		 */
-		dbp->flags = replyp->dbflags;
+		F_SET(dbp, DB_AM_OPEN_CALLED);
 	}
 	return (replyp->status);
 }
@@ -431,25 +440,17 @@ __dbcl_db_pget_ret(dbp, txnp, skey, pkey, data, flags, replyp)
 		return (ret);
 
 	oldpkey = pkey->data;
-	ret = __dbcl_retcopy(dbenv, pkey, replyp->pkeydata.pkeydata_val,
+	if ((ret = __dbcl_retcopy(dbenv, pkey, replyp->pkeydata.pkeydata_val,
 	    replyp->pkeydata.pkeydata_len, &dbp->my_rkey.data,
-	    &dbp->my_rkey.ulen);
-	if (ret && oldskey != NULL) {
-		__os_free(dbenv, skey->data);
-		return (ret);
-	}
+	    &dbp->my_rkey.ulen)) != 0)
+		goto err;
 	ret = __dbcl_retcopy(dbenv, data, replyp->datadata.datadata_val,
 	    replyp->datadata.datadata_len, &dbp->my_rdata.data,
 	    &dbp->my_rdata.ulen);
-	/*
-	 * If an error on copying 'data' and we allocated for '*key'
-	 * free it before returning the error.
-	 */
+
 	if (ret) {
-		if (oldskey != NULL)
-			__os_free(dbenv, skey->data);
-		if (oldpkey != NULL)
-			__os_free(dbenv, pkey->data);
+err:		FREE_IF_CHANGED(skey, oldskey);
+		FREE_IF_CHANGED(pkey, oldpkey);
 	}
 	return (ret);
 }
@@ -532,19 +533,22 @@ __dbcl_db_rename_ret(dbp, name, subdb, newname, flags, replyp)
 
 /*
  * PUBLIC: int __dbcl_db_stat_ret
- * PUBLIC:     __P((DB *, void *, u_int32_t, __db_stat_reply *));
+ * PUBLIC:     __P((DB *, DB_TXN *, void *, u_int32_t, __db_stat_reply *));
  */
 int
-__dbcl_db_stat_ret(dbp, sp, flags, replyp)
+__dbcl_db_stat_ret(dbp, txnp, sp, flags, replyp)
 	DB *dbp;
+	DB_TXN *txnp;
 	void *sp;
 	u_int32_t flags;
 	__db_stat_reply *replyp;
 {
-	int len, ret;
+	size_t len;
 	u_int32_t i, *q, *p, *retsp;
+	int ret;
 
 	COMPQUIET(flags, 0);
+	COMPQUIET(txnp, NULL);
 
 	if (replyp->status != 0 || sp == NULL)
 		return (replyp->status);
@@ -630,27 +634,27 @@ __dbcl_db_join_ret(dbp, curs, dbcp, flags, replyp)
 }
 
 /*
- * PUBLIC: int __dbcl_dbc_close_ret __P((DBC *, __dbc_close_reply *));
+ * PUBLIC: int __dbcl_dbc_c_close_ret __P((DBC *, __dbc_c_close_reply *));
  */
 int
-__dbcl_dbc_close_ret(dbc, replyp)
+__dbcl_dbc_c_close_ret(dbc, replyp)
 	DBC *dbc;
-	__dbc_close_reply *replyp;
+	__dbc_c_close_reply *replyp;
 {
 	__dbcl_c_refresh(dbc);
 	return (replyp->status);
 }
 
 /*
- * PUBLIC: int __dbcl_dbc_count_ret
- * PUBLIC:     __P((DBC *, db_recno_t *, u_int32_t, __dbc_count_reply *));
+ * PUBLIC: int __dbcl_dbc_c_count_ret
+ * PUBLIC:     __P((DBC *, db_recno_t *, u_int32_t, __dbc_c_count_reply *));
  */
 int
-__dbcl_dbc_count_ret(dbc, countp, flags, replyp)
+__dbcl_dbc_c_count_ret(dbc, countp, flags, replyp)
 	DBC *dbc;
 	db_recno_t *countp;
 	u_int32_t flags;
-	__dbc_count_reply *replyp;
+	__dbc_c_count_reply *replyp;
 {
 	COMPQUIET(dbc, NULL);
 	COMPQUIET(flags, 0);
@@ -663,14 +667,14 @@ __dbcl_dbc_count_ret(dbc, countp, flags, replyp)
 }
 
 /*
- * PUBLIC: int __dbcl_dbc_dup_ret
- * PUBLIC:     __P((DBC *, DBC **, u_int32_t, __dbc_dup_reply *));
+ * PUBLIC: int __dbcl_dbc_c_dup_ret
+ * PUBLIC:     __P((DBC *, DBC **, u_int32_t, __dbc_c_dup_reply *));
  */
 int
-__dbcl_dbc_dup_ret(dbc, dbcp, flags, replyp)
+__dbcl_dbc_c_dup_ret(dbc, dbcp, flags, replyp)
 	DBC *dbc, **dbcp;
 	u_int32_t flags;
-	__dbc_dup_reply *replyp;
+	__dbc_c_dup_reply *replyp;
 {
 	COMPQUIET(flags, 0);
 
@@ -681,15 +685,15 @@ __dbcl_dbc_dup_ret(dbc, dbcp, flags, replyp)
 }
 
 /*
- * PUBLIC: int __dbcl_dbc_get_ret
- * PUBLIC:     __P((DBC *, DBT *, DBT *, u_int32_t, __dbc_get_reply *));
+ * PUBLIC: int __dbcl_dbc_c_get_ret
+ * PUBLIC:     __P((DBC *, DBT *, DBT *, u_int32_t, __dbc_c_get_reply *));
  */
 int
-__dbcl_dbc_get_ret(dbc, key, data, flags, replyp)
+__dbcl_dbc_c_get_ret(dbc, key, data, flags, replyp)
 	DBC *dbc;
 	DBT *key, *data;
 	u_int32_t flags;
-	__dbc_get_reply *replyp;
+	__dbc_c_get_reply *replyp;
 {
 	DB_ENV *dbenv;
 	int ret;
@@ -716,23 +720,23 @@ __dbcl_dbc_get_ret(dbc, key, data, flags, replyp)
 	 * If an error on copying 'data' and we allocated for 'key'
 	 * free it before returning the error.
 	 */
-	if (ret && oldkey != NULL)
-		__os_free(dbenv, key->data);
+	if (ret)
+		FREE_IF_CHANGED(key, oldkey);
 	return (ret);
 }
 
 /*
- * PUBLIC: int __dbcl_dbc_pget_ret __P((DBC *, DBT *, DBT *, DBT *, u_int32_t,
- * PUBLIC:      __dbc_pget_reply *));
+ * PUBLIC: int __dbcl_dbc_c_pget_ret __P((DBC *, DBT *, DBT *, DBT *, u_int32_t,
+ * PUBLIC:      __dbc_c_pget_reply *));
  */
 int
-__dbcl_dbc_pget_ret(dbc, skey, pkey, data, flags, replyp)
+__dbcl_dbc_c_pget_ret(dbc, skey, pkey, data, flags, replyp)
 	DBC * dbc;
 	DBT * skey;
 	DBT * pkey;
 	DBT * data;
 	u_int32_t flags;
-	__dbc_pget_reply *replyp;
+	__dbc_c_pget_reply *replyp;
 {
 	DB_ENV *dbenv;
 	int ret;
@@ -754,39 +758,35 @@ __dbcl_dbc_pget_ret(dbc, skey, pkey, data, flags, replyp)
 		return (ret);
 
 	oldpkey = pkey->data;
-	ret = __dbcl_retcopy(dbenv, pkey, replyp->pkeydata.pkeydata_val,
+	if ((ret = __dbcl_retcopy(dbenv, pkey, replyp->pkeydata.pkeydata_val,
 	    replyp->pkeydata.pkeydata_len, &dbc->my_rkey.data,
-	    &dbc->my_rkey.ulen);
-	if (ret && oldskey != NULL) {
-		__os_free(dbenv, skey->data);
-		return (ret);
-	}
+	    &dbc->my_rkey.ulen)) != 0)
+		goto err;
 	ret = __dbcl_retcopy(dbenv, data, replyp->datadata.datadata_val,
 	    replyp->datadata.datadata_len, &dbc->my_rdata.data,
 	    &dbc->my_rdata.ulen);
+
 	/*
 	 * If an error on copying 'data' and we allocated for '*key'
 	 * free it before returning the error.
 	 */
 	if (ret) {
-		if (oldskey != NULL)
-			__os_free(dbenv, skey->data);
-		if (oldpkey != NULL)
-			__os_free(dbenv, pkey->data);
+err:		FREE_IF_CHANGED(skey, oldskey);
+		FREE_IF_CHANGED(pkey, oldpkey);
 	}
 	return (ret);
 }
 
 /*
- * PUBLIC: int __dbcl_dbc_put_ret
- * PUBLIC:     __P((DBC *, DBT *, DBT *, u_int32_t, __dbc_put_reply *));
+ * PUBLIC: int __dbcl_dbc_c_put_ret
+ * PUBLIC:     __P((DBC *, DBT *, DBT *, u_int32_t, __dbc_c_put_reply *));
  */
 int
-__dbcl_dbc_put_ret(dbc, key, data, flags, replyp)
+__dbcl_dbc_c_put_ret(dbc, key, data, flags, replyp)
 	DBC *dbc;
 	DBT *key, *data;
 	u_int32_t flags;
-	__dbc_put_reply *replyp;
+	__dbc_c_put_reply *replyp;
 {
 	COMPQUIET(data, NULL);
 
@@ -799,4 +799,3 @@ __dbcl_dbc_put_ret(dbc, key, data, flags, replyp)
 		    *(db_recno_t *)replyp->keydata.keydata_val;
 	return (replyp->status);
 }
-#endif /* HAVE_RPC */
