@@ -1,16 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2000-2003
+ * Copyright (c) 2000-2005
  *      Sleepycat Software.  All rights reserved.
+ *
+ * $Id: db_server_util.c,v 12.3 2005/06/16 20:23:39 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: db_server_util.c,v 1.65 2003/04/24 18:54:13 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -34,15 +32,15 @@ static const char revid[] = "$Id: db_server_util.c,v 1.65 2003/04/24 18:54:13 bo
 #include <string.h>
 #include <unistd.h>
 #endif
-#include "dbinc_auto/db_server.h"
+
+#include "db_server.h"
 
 #include "db_int.h"
 #include "dbinc_auto/clib_ext.h"
 #include "dbinc/db_server_int.h"
-#include "dbinc_auto/rpc_server_ext.h"
 #include "dbinc_auto/common_ext.h"
+#include "dbinc_auto/rpc_server_ext.h"
 
-extern int __dbsrv_main	 __P((void));
 static int add_home __P((char *));
 static int add_passwd __P((char *));
 static int env_recover __P((char *));
@@ -66,6 +64,7 @@ main(argc, argv)
 	int argc;
 	char **argv;
 {
+	extern int __dbsrv_main();
 	extern char *optarg;
 	CLIENT *cl;
 	int ch, ret;
@@ -325,7 +324,7 @@ __dbsrv_timeout(force)
 		if (to < t || force) {
 			if (__dbsrv_verbose)
 				printf("Timing out env id %ld\n", ctp->ct_id);
-			(void)__dbenv_close_int(ctp->ct_id, 0, 1);
+			(void)__env_close_int(ctp->ct_id, 0, 1);
 			/*
 			 * If we timed out an env, we may have closed
 			 * all sorts of ctp's (maybe even all of them.
@@ -417,7 +416,7 @@ new_ct_ent(errp)
 	octp = LIST_FIRST(&__dbsrv_head);
 	if (octp != NULL && octp->ct_id >= t)
 		t = octp->ct_id + 1;
-	ctp->ct_id = t;
+	ctp->ct_id = (long)t;
 	ctp->ct_idle = __dbsrv_idleto;
 	ctp->ct_activep = &ctp->ct_active;
 	ctp->ct_origp = NULL;
@@ -602,6 +601,10 @@ __db_close_int(id, flags)
 		printf("Closing dbp id %ld\n", id);
 
 	ret = dbp->close(dbp, flags);
+	if (ctp->ct_dbdp.db != NULL)
+		__os_free(NULL, ctp->ct_dbdp.db);
+	if (ctp->ct_dbdp.subdb != NULL)
+		__os_free(NULL, ctp->ct_dbdp.subdb);
 	__dbdel_ctp(ctp);
 	return (ret);
 }
@@ -644,17 +647,17 @@ __dbc_close_int(dbc_ctp)
 }
 
 /*
- * PUBLIC: int __dbenv_close_int __P((long, u_int32_t, int));
+ * PUBLIC: int __env_close_int __P((long, u_int32_t, int));
  */
 int
-__dbenv_close_int(id, flags, force)
+__env_close_int(id, flags, force)
 	long id;
 	u_int32_t flags;
 	int force;
 {
 	DB_ENV *dbenv;
 	int ret;
-	ct_entry *ctp;
+	ct_entry *ctp, *dbctp, *nextctp;
 
 	ret = 0;
 	ctp = get_tableent(id);
@@ -674,6 +677,31 @@ __dbenv_close_int(id, flags, force)
 	if (__dbsrv_verbose)
 		printf("Closing env id %ld\n", id);
 
+	/*
+	 * If we're timing out an env, we want to close all of its
+	 * database handles as well.  All of the txns and cursors
+	 * must have been timed out prior to timing out the env.
+	 */
+	if (force)
+		for (dbctp = LIST_FIRST(&__dbsrv_head);
+		    dbctp != NULL; dbctp = nextctp) {
+			nextctp = LIST_NEXT(dbctp, entries);
+			if (dbctp->ct_type != CT_DB)
+				continue;
+			if (dbctp->ct_envparent != ctp)
+				continue;
+			/*
+			 * We found a DB handle that is part of this
+			 * environment.  Close it.
+			 */
+			__db_close_int(dbctp->ct_id, 0);
+			/*
+			 * If we timed out a dbp, we may have removed
+			 * multiple ctp entries.  Start over with a
+			 * guaranteed good ctp.
+			 */
+			nextctp = LIST_FIRST(&__dbsrv_head);
+		}
 	ret = dbenv->close(dbenv, flags);
 	__dbdel_ctp(ctp);
 	return (ret);
@@ -784,10 +812,8 @@ env_recover(progname)
 			    progname, db_strerror(ret));
 			exit(EXIT_FAILURE);
 		}
-		if (__dbsrv_verbose == 1) {
+		if (__dbsrv_verbose == 1)
 			(void)dbenv->set_verbose(dbenv, DB_VERB_RECOVERY, 1);
-			(void)dbenv->set_verbose(dbenv, DB_VERB_CHKPOINT, 1);
-		}
 		dbenv->set_errfile(dbenv, stderr);
 		dbenv->set_errpfx(dbenv, progname);
 		if (hp->passwd != NULL)
