@@ -1,17 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2001-2003
+ * Copyright (c) 2001-2005
  *      Sleepycat Software.  All rights reserved.
+ *
+ * $Id: db_server_cxxproc.cpp,v 12.6 2005/07/21 18:21:34 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifdef HAVE_RPC
-#ifndef lint
-static const char revid[] = "$Id: db_server_cxxproc.cpp,v 1.15 2003/04/23 20:43:09 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -19,7 +16,8 @@ static const char revid[] = "$Id: db_server_cxxproc.cpp,v 1.15 2003/04/23 20:43:
 
 #include <string.h>
 #endif
-#include "dbinc_auto/db_server.h"
+
+#include "db_server.h"
 
 #include "db_int.h"
 #include "db_cxx.h"
@@ -31,7 +29,7 @@ extern "C" {
 
 extern "C" void
 __env_get_cachesize_proc(
-	long dbenvcl_id,
+	u_int dbenvcl_id,
 	__env_get_cachesize_reply *replyp)
 {
 	DbEnv *dbenv;
@@ -45,12 +43,12 @@ __env_get_cachesize_proc(
 }
 
 extern "C" void
-__env_cachesize_proc(
-	long dbenvcl_id,
+__env_set_cachesize_proc(
+	u_int dbenvcl_id,
 	u_int32_t gbytes,
 	u_int32_t bytes,
 	u_int32_t ncache,
-	__env_cachesize_reply *replyp)
+	__env_set_cachesize_reply *replyp)
 {
 	DbEnv *dbenv;
 	ct_entry *dbenv_ctp;
@@ -67,14 +65,14 @@ __env_cachesize_proc(
 
 extern "C" void
 __env_close_proc(
-	long dbenvcl_id,
+	u_int dbenvcl_id,
 	u_int32_t flags,
 	__env_close_reply *replyp)
 {
 	ct_entry *dbenv_ctp;
 
 	ACTIVATE_CTP(dbenv_ctp, dbenvcl_id, CT_ENV);
-	replyp->status = __dbenv_close_int(dbenvcl_id, flags, 0);
+	replyp->status = __env_close_int(dbenvcl_id, flags, 0);
 	return;
 }
 
@@ -105,8 +103,8 @@ __env_create_proc(
 
 extern "C" void
 __env_dbremove_proc(
-	long dbenvcl_id,
-	long txnpcl_id,
+	u_int dbenvcl_id,
+	u_int txnpcl_id,
 	char *name,
 	char *subdb,
 	u_int32_t flags,
@@ -134,8 +132,8 @@ __env_dbremove_proc(
 
 void
 __env_dbrename_proc(
-	long dbenvcl_id,
-	long txnpcl_id,
+	u_int dbenvcl_id,
+	u_int txnpcl_id,
 	char *name,
 	char *subdb,
 	char *newname,
@@ -164,7 +162,7 @@ __env_dbrename_proc(
 
 extern "C" void
 __env_get_encrypt_flags_proc(
-	long dbenvcl_id,
+	u_int dbenvcl_id,
 	__env_get_encrypt_flags_reply *replyp)
 {
 	DbEnv *dbenv;
@@ -177,11 +175,11 @@ __env_get_encrypt_flags_proc(
 }
 
 extern "C" void
-__env_encrypt_proc(
-	long dbenvcl_id,
+__env_set_encrypt_proc(
+	u_int dbenvcl_id,
 	char *passwd,
 	u_int32_t flags,
-	__env_encrypt_reply *replyp)
+	__env_set_encrypt_reply *replyp)
 {
 	DbEnv *dbenv;
 	ct_entry *dbenv_ctp;
@@ -198,7 +196,7 @@ __env_encrypt_proc(
 
 extern "C" void
 __env_get_flags_proc(
-	long dbenvcl_id,
+	u_int dbenvcl_id,
 	__env_get_flags_reply *replyp)
 {
 	DbEnv *dbenv;
@@ -211,11 +209,11 @@ __env_get_flags_proc(
 }
 
 extern "C" void
-__env_flags_proc(
-	long dbenvcl_id,
+__env_set_flags_proc(
+	u_int dbenvcl_id,
 	u_int32_t flags,
 	u_int32_t onoff,
-	__env_flags_reply *replyp)
+	__env_set_flags_reply *replyp)
 {
 	DbEnv *dbenv;
 	ct_entry *dbenv_ctp;
@@ -236,7 +234,7 @@ __env_flags_proc(
 
 extern "C" void
 __env_get_home_proc(
-	long dbenvcl_id,
+	u_int dbenvcl_id,
 	__env_get_home_reply *replyp)
 {
 	DbEnv *dbenv;
@@ -250,7 +248,7 @@ __env_get_home_proc(
 
 extern "C" void
 __env_get_open_flags_proc(
-	long dbenvcl_id,
+	u_int dbenvcl_id,
 	__env_get_open_flags_reply *replyp)
 {
 	DbEnv *dbenv;
@@ -264,7 +262,7 @@ __env_get_open_flags_proc(
 
 extern "C" void
 __env_open_proc(
-	long dbenvcl_id,
+	u_int dbenvcl_id,
 	char *home,
 	u_int32_t flags,
 	u_int32_t mode,
@@ -314,7 +312,7 @@ __env_open_proc(
 		if (__dbsrv_verbose)
 			printf("Sharing env ID %ld\n", new_ctp->ct_id);
 		replyp->envcl_id = new_ctp->ct_id;
-		ret = __dbenv_close_int(dbenvcl_id, 0, 0);
+		ret = __env_close_int(dbenvcl_id, 0, 0);
 	} else {
 		ret = dbenv->open(fullhome->home, newflags, mode);
 		dbenv_ctp->ct_envdp.home = fullhome;
@@ -326,7 +324,7 @@ out:	replyp->status = ret;
 
 extern "C" void
 __env_remove_proc(
-	long dbenvcl_id,
+	u_int dbenvcl_id,
 	char *home,
 	u_int32_t flags,
 	__env_remove_reply *replyp)
@@ -352,7 +350,7 @@ __env_remove_proc(
 
 extern "C" void
 __txn_abort_proc(
-	long txnpcl_id,
+	u_int txnpcl_id,
 	__txn_abort_reply *replyp)
 {
 	DbTxn *txnp;
@@ -369,11 +367,11 @@ __txn_abort_proc(
 }
 
 extern "C" void
-__txn_begin_proc(
-	long dbenvcl_id,
-	long parentcl_id,
+__env_txn_begin_proc(
+	u_int dbenvcl_id,
+	u_int parentcl_id,
 	u_int32_t flags,
-	__txn_begin_reply *replyp)
+	__env_txn_begin_reply *replyp)
 {
 	DbEnv *dbenv;
 	DbTxn *parent, *txnp;
@@ -395,7 +393,7 @@ __txn_begin_proc(
 	} else
 		parent = NULL;
 
-	ret = dbenv->txn_begin(parent, &txnp, flags);
+	ret = dbenv->txn_begin(parent, &txnp, flags | DB_TXN_NOWAIT);
 	if (ret == 0) {
 		ctp->ct_txnp = txnp;
 		ctp->ct_type = CT_TXN;
@@ -413,7 +411,7 @@ __txn_begin_proc(
 
 extern "C" void
 __txn_commit_proc(
-	long txnpcl_id,
+	u_int txnpcl_id,
 	u_int32_t flags,
 	__txn_commit_reply *replyp)
 {
@@ -433,7 +431,7 @@ __txn_commit_proc(
 
 extern "C" void
 __txn_discard_proc(
-	long txnpcl_id,
+	u_int txnpcl_id,
 	u_int32_t flags,
 	__txn_discard_reply *replyp)
 {
@@ -453,7 +451,7 @@ __txn_discard_proc(
 
 extern "C" void
 __txn_prepare_proc(
-	long txnpcl_id,
+	u_int txnpcl_id,
 	u_int8_t *gid,
 	__txn_prepare_reply *replyp)
 {
@@ -470,11 +468,11 @@ __txn_prepare_proc(
 }
 
 extern "C" void
-__txn_recover_proc(
-	long dbenvcl_id,
+__env_txn_recover_proc(
+	u_int dbenvcl_id,
 	u_int32_t count,
 	u_int32_t flags,
-	__txn_recover_reply *replyp,
+	__env_txn_recover_reply *replyp,
 	int * freep)
 {
 	DbEnv *dbenv;
@@ -489,11 +487,10 @@ __txn_recover_proc(
 	dbenv = (DbEnv *)dbenv_ctp->ct_anyp;
 	*freep = 0;
 
-	if ((ret =
-	    __os_malloc(dbenv->get_DB_ENV(), count * sizeof(DbPreplist), &dbprep)) != 0)
+	if ((ret = __os_malloc(
+	    dbenv->get_DB_ENV(), count * sizeof(DbPreplist), &dbprep)) != 0)
 		goto out;
-	if ((ret =
-	    dbenv->txn_recover(dbprep, count, &retcount, flags)) != 0)
+	if ((ret = dbenv->txn_recover(dbprep, count, &retcount, flags)) != 0)
 		goto out;
 	/*
 	 * If there is nothing, success, but it's easy.
@@ -577,30 +574,11 @@ out2:
 	return;
 }
 
-extern "C" void
-__db_bt_maxkey_proc(
-	long dbpcl_id,
-	u_int32_t maxkey,
-	__db_bt_maxkey_reply *replyp)
-{
-	Db *dbp;
-	ct_entry *dbp_ctp;
-	int ret;
-
-	ACTIVATE_CTP(dbp_ctp, dbpcl_id, CT_DB);
-	dbp = (Db *)dbp_ctp->ct_anyp;
-
-	ret = dbp->set_bt_maxkey(maxkey);
-
-	replyp->status = ret;
-	return;
-}
-
 extern "C" void
 __db_associate_proc(
-	long dbpcl_id,
-	long txnpcl_id,
-	long sdbpcl_id,
+	u_int dbpcl_id,
+	u_int txnpcl_id,
+	u_int sdbpcl_id,
 	u_int32_t flags,
 	__db_associate_reply *replyp)
 {
@@ -620,11 +598,12 @@ __db_associate_proc(
 		txnp = NULL;
 
 	/*
-	 * We do not support DB_CREATE for associate.   Users
-	 * can only access secondary indices on a read-only basis,
-	 * so whatever they are looking for needs to be there already.
+	 * We do not support DB_CREATE for associate or the callbacks
+	 * implemented in the Java and JE RPC servers.   Users can only
+	 * access secondary indices on a read-only basis, so whatever they
+	 * are looking for needs to be there already.
 	 */
-	if (flags != 0)
+	if (LF_ISSET(DB_RPC2ND_MASK | DB_CREATE))
 		ret = EINVAL;
 	else
 		ret = dbp->associate(txnp, sdbp, NULL, flags);
@@ -635,7 +614,7 @@ __db_associate_proc(
 
 extern "C" void
 __db_get_bt_minkey_proc(
-	long dbpcl_id,
+	u_int dbpcl_id,
 	__db_get_bt_minkey_reply *replyp)
 {
 	Db *dbp;
@@ -648,10 +627,10 @@ __db_get_bt_minkey_proc(
 }
 
 extern "C" void
-__db_bt_minkey_proc(
-	long dbpcl_id,
+__db_set_bt_minkey_proc(
+	u_int dbpcl_id,
 	u_int32_t minkey,
-	__db_bt_minkey_reply *replyp)
+	__db_set_bt_minkey_reply *replyp)
 {
 	Db *dbp;
 	ct_entry *dbp_ctp;
@@ -668,7 +647,7 @@ __db_bt_minkey_proc(
 
 extern "C" void
 __db_close_proc(
-	long dbpcl_id,
+	u_int dbpcl_id,
 	u_int32_t flags,
 	__db_close_reply *replyp)
 {
@@ -681,7 +660,7 @@ __db_close_proc(
 
 extern "C" void
 __db_create_proc(
-	long dbenvcl_id,
+	u_int dbenvcl_id,
 	u_int32_t flags,
 	__db_create_reply *replyp)
 {
@@ -712,8 +691,8 @@ __db_create_proc(
 
 extern "C" void
 __db_del_proc(
-	long dbpcl_id,
-	long txnpcl_id,
+	u_int dbpcl_id,
+	u_int txnpcl_id,
 	u_int32_t keydlen,
 	u_int32_t keydoff,
 	u_int32_t keyulen,
@@ -751,7 +730,7 @@ __db_del_proc(
 
 extern "C" void
 __db_get_encrypt_flags_proc(
-	long dbpcl_id,
+	u_int dbpcl_id,
 	__db_get_encrypt_flags_reply *replyp)
 {
 	Db *dbp;
@@ -764,11 +743,11 @@ __db_get_encrypt_flags_proc(
 }
 
 extern "C" void
-__db_encrypt_proc(
-	long dbpcl_id,
+__db_set_encrypt_proc(
+	u_int dbpcl_id,
 	char *passwd,
 	u_int32_t flags,
-	__db_encrypt_reply *replyp)
+	__db_set_encrypt_reply *replyp)
 {
 	Db *dbp;
 	ct_entry *dbp_ctp;
@@ -783,9 +762,9 @@ __db_encrypt_proc(
 }
 
 extern "C" void
-__db_get_extentsize_proc(
-	long dbpcl_id,
-	__db_get_extentsize_reply *replyp)
+__db_get_q_extentsize_proc(
+	u_int dbpcl_id,
+	__db_get_q_extentsize_reply *replyp)
 {
 	Db *dbp;
 	ct_entry *dbp_ctp;
@@ -797,10 +776,10 @@ __db_get_extentsize_proc(
 }
 
 extern "C" void
-__db_extentsize_proc(
-	long dbpcl_id,
+__db_set_q_extentsize_proc(
+	u_int dbpcl_id,
 	u_int32_t extentsize,
-	__db_extentsize_reply *replyp)
+	__db_set_q_extentsize_reply *replyp)
 {
 	Db *dbp;
 	ct_entry *dbp_ctp;
@@ -817,7 +796,7 @@ __db_extentsize_proc(
 
 extern "C" void
 __db_get_flags_proc(
-	long dbpcl_id,
+	u_int dbpcl_id,
 	__db_get_flags_reply *replyp)
 {
 	Db *dbp;
@@ -830,10 +809,10 @@ __db_get_flags_proc(
 }
 
 extern "C" void
-__db_flags_proc(
-	long dbpcl_id,
+__db_set_flags_proc(
+	u_int dbpcl_id,
 	u_int32_t flags,
-	__db_flags_reply *replyp)
+	__db_set_flags_reply *replyp)
 {
 	Db *dbp;
 	ct_entry *dbp_ctp;
@@ -851,8 +830,8 @@ __db_flags_proc(
 
 extern "C" void
 __db_get_proc(
-	long dbpcl_id,
-	long txnpcl_id,
+	u_int dbpcl_id,
+	u_int txnpcl_id,
 	u_int32_t keydlen,
 	u_int32_t keydoff,
 	u_int32_t keyulen,
@@ -872,7 +851,7 @@ __db_get_proc(
 	Db *dbp;
 	DbTxn *txnp;
 	ct_entry *dbp_ctp, *txnp_ctp;
-	int key_alloc, bulk_alloc, ret;
+	int ret;
 	void *tmpdata;
 
 	ACTIVATE_CTP(dbp_ctp, dbpcl_id, CT_DB);
@@ -883,8 +862,9 @@ __db_get_proc(
 	} else
 		txnp = NULL;
 
+	replyp->keydata.keydata_val = NULL;
+	replyp->datadata.datadata_val = NULL;
 	*freep = 0;
-	bulk_alloc = 0;
 
 	/* Set up key and data */
 	Dbt key(keydata, keysize);
@@ -911,7 +891,6 @@ __db_get_proc(
 			if (ret != 0)
 				goto err;
 			data.set_data(tmpdata);
-			bulk_alloc = 1;
 		}
 		dataflags |= DB_DBT_USERMEM;
 	} else
@@ -936,17 +915,13 @@ __db_get_proc(
 		/*
 		 * Key
 		 */
-		key_alloc = 0;
 		if (key.get_data() == keydata) {
 			ret = __os_umalloc(dbp->get_DB()->dbenv,
 			    key.get_size(), &replyp->keydata.keydata_val);
-			if (ret != 0) {
-				__os_ufree(dbp->get_DB()->dbenv, key.get_data());
-				__os_ufree(dbp->get_DB()->dbenv, data.get_data());
+			if (ret != 0)
 				goto err;
-			}
-			key_alloc = 1;
-			memcpy(replyp->keydata.keydata_val, key.get_data(), key.get_size());
+			memcpy(replyp->keydata.keydata_val,
+			    key.get_data(), key.get_size());
 		} else
 			replyp->keydata.keydata_val = (char *)key.get_data();
 
@@ -958,27 +933,25 @@ __db_get_proc(
 		if (data.get_data() == datadata) {
 			ret = __os_umalloc(dbp->get_DB()->dbenv,
 			     data.get_size(), &replyp->datadata.datadata_val);
-			if (ret != 0) {
-				__os_ufree(dbp->get_DB()->dbenv, key.get_data());
-				__os_ufree(dbp->get_DB()->dbenv, data.get_data());
-				if (key_alloc)
-					__os_ufree(dbp->get_DB()->dbenv,
-					    replyp->keydata.keydata_val);
+			if (ret != 0)
 				goto err;
-			}
 			memcpy(replyp->datadata.datadata_val, data.get_data(),
 			    data.get_size());
 		} else
 			replyp->datadata.datadata_val = (char *)data.get_data();
 		replyp->datadata.datadata_len = data.get_size();
 	} else {
-err:		replyp->keydata.keydata_val = NULL;
+err:		FREE_IF_CHANGED(dbp->get_DB()->dbenv,
+		    key.get_data(), keydata);
+		FREE_IF_CHANGED(dbp->get_DB()->dbenv,
+		    data.get_data(), datadata);
+		FREE_IF_CHANGED(dbp->get_DB()->dbenv,
+		    replyp->keydata.keydata_val, key.get_data());
+		replyp->keydata.keydata_val = NULL;
 		replyp->keydata.keydata_len = 0;
 		replyp->datadata.datadata_val = NULL;
 		replyp->datadata.datadata_len = 0;
 		*freep = 0;
-		if (bulk_alloc)
-			__os_ufree(dbp->get_DB()->dbenv, data.get_data());
 	}
 	replyp->status = ret;
 	return;
@@ -986,7 +959,7 @@ err:		replyp->keydata.keydata_val = NULL;
 
 extern "C" void
 __db_get_h_ffactor_proc(
-	long dbpcl_id,
+	u_int dbpcl_id,
 	__db_get_h_ffactor_reply *replyp)
 {
 	Db *dbp;
@@ -999,10 +972,10 @@ __db_get_h_ffactor_proc(
 }
 
 extern "C" void
-__db_h_ffactor_proc(
-	long dbpcl_id,
+__db_set_h_ffactor_proc(
+	u_int dbpcl_id,
 	u_int32_t ffactor,
-	__db_h_ffactor_reply *replyp)
+	__db_set_h_ffactor_reply *replyp)
 {
 	Db *dbp;
 	ct_entry *dbp_ctp;
@@ -1019,7 +992,7 @@ __db_h_ffactor_proc(
 
 extern "C" void
 __db_get_h_nelem_proc(
-	long dbpcl_id,
+	u_int dbpcl_id,
 	__db_get_h_nelem_reply *replyp)
 {
 	Db *dbp;
@@ -1032,10 +1005,10 @@ __db_get_h_nelem_proc(
 }
 
 extern "C" void
-__db_h_nelem_proc(
-	long dbpcl_id,
+__db_set_h_nelem_proc(
+	u_int dbpcl_id,
 	u_int32_t nelem,
-	__db_h_nelem_reply *replyp)
+	__db_set_h_nelem_reply *replyp)
 {
 	Db *dbp;
 	ct_entry *dbp_ctp;
@@ -1052,8 +1025,8 @@ __db_h_nelem_proc(
 
 extern "C" void
 __db_key_range_proc(
-	long dbpcl_id,
-	long txnpcl_id,
+	u_int dbpcl_id,
+	u_int txnpcl_id,
 	u_int32_t keydlen,
 	u_int32_t keydoff,
 	u_int32_t keyulen,
@@ -1095,7 +1068,7 @@ __db_key_range_proc(
 
 extern "C" void
 __db_get_lorder_proc(
-	long dbpcl_id,
+	u_int dbpcl_id,
 	__db_get_lorder_reply *replyp)
 {
 	Db *dbp;
@@ -1108,10 +1081,10 @@ __db_get_lorder_proc(
 }
 
 extern "C" void
-__db_lorder_proc(
-	long dbpcl_id,
+__db_set_lorder_proc(
+	u_int dbpcl_id,
 	u_int32_t lorder,
-	__db_lorder_reply *replyp)
+	__db_set_lorder_reply *replyp)
 {
 	Db *dbp;
 	ct_entry *dbp_ctp;
@@ -1127,9 +1100,9 @@ __db_lorder_proc(
 }
 
 extern "C" void
-__db_get_name_proc(
-	long dbpcl_id,
-	__db_get_name_reply *replyp)
+__db_get_dbname_proc(
+	u_int dbpcl_id,
+	__db_get_dbname_reply *replyp)
 {
 	Db *dbp;
 	ct_entry *dbp_ctp;
@@ -1143,7 +1116,7 @@ __db_get_name_proc(
 
 extern "C" void
 __db_get_open_flags_proc(
-	long dbpcl_id,
+	u_int dbpcl_id,
 	__db_get_open_flags_reply *replyp)
 {
 	Db *dbp;
@@ -1157,8 +1130,8 @@ __db_get_open_flags_proc(
 
 extern "C" void
 __db_open_proc(
-	long dbpcl_id,
-	long txnpcl_id,
+	u_int dbpcl_id,
+	u_int txnpcl_id,
 	char *name,
 	char *subdb,
 	u_int32_t type,
@@ -1181,8 +1154,8 @@ __db_open_proc(
 		txnp = NULL;
 
 	replyp->dbcl_id = dbpcl_id;
-	if ((new_ctp = __dbsrv_sharedb(dbp_ctp, name, subdb, (DBTYPE)type, flags))
-	    != NULL) {
+	if ((new_ctp = __dbsrv_sharedb(
+	    dbp_ctp, name, subdb, (DBTYPE)type, flags)) != NULL) {
 		/*
 		 * We can share, clean up old ID, set new one.
 		 */
@@ -1196,11 +1169,6 @@ __db_open_proc(
 	if (ret == 0) {
 		(void)dbp->get_type(&dbtype);
 		replyp->type = dbtype;
-		/* XXX
-		 * Tcl needs to peek at dbp->flags for DB_AM_DUP.  Send
-		 * this dbp's flags back.
-		 */
-		replyp->dbflags = (int) dbp->get_DB()->flags;
 		/*
 		 * We need to determine the byte order of the database
 		 * and send it back to the client.  Determine it by
@@ -1239,7 +1207,7 @@ out:
 
 extern "C" void
 __db_get_pagesize_proc(
-	long dbpcl_id,
+	u_int dbpcl_id,
 	__db_get_pagesize_reply *replyp)
 {
 	Db *dbp;
@@ -1252,10 +1220,10 @@ __db_get_pagesize_proc(
 }
 
 extern "C" void
-__db_pagesize_proc(
-	long dbpcl_id,
+__db_set_pagesize_proc(
+	u_int dbpcl_id,
 	u_int32_t pagesize,
-	__db_pagesize_reply *replyp)
+	__db_set_pagesize_reply *replyp)
 {
 	Db *dbp;
 	ct_entry *dbp_ctp;
@@ -1272,8 +1240,8 @@ __db_pagesize_proc(
 
 extern "C" void
 __db_pget_proc(
-	long dbpcl_id,
-	long txnpcl_id,
+	u_int dbpcl_id,
+	u_int txnpcl_id,
 	u_int32_t skeydlen,
 	u_int32_t skeydoff,
 	u_int32_t skeyulen,
@@ -1299,7 +1267,7 @@ __db_pget_proc(
 	Db *dbp;
 	DbTxn *txnp;
 	ct_entry *dbp_ctp, *txnp_ctp;
-	int key_alloc, ret;
+	int ret;
 
 	ACTIVATE_CTP(dbp_ctp, dbpcl_id, CT_DB);
 	dbp = (Db *)dbp_ctp->ct_anyp;
@@ -1309,6 +1277,9 @@ __db_pget_proc(
 	} else
 		txnp = NULL;
 
+	replyp->skeydata.skeydata_val = NULL;
+	replyp->pkeydata.pkeydata_val = NULL;
+	replyp->datadata.datadata_val = NULL;
 	*freep = 0;
 
 	/*
@@ -1351,17 +1322,11 @@ __db_pget_proc(
 		/*
 		 * Key
 		 */
-		key_alloc = 0;
 		if (skey.get_data() == skeydata) {
 			ret = __os_umalloc(dbp->get_DB()->dbenv,
 			    skey.get_size(), &replyp->skeydata.skeydata_val);
-			if (ret != 0) {
-				__os_ufree(dbp->get_DB()->dbenv, skey.get_data());
-				__os_ufree(dbp->get_DB()->dbenv, pkey.get_data());
-				__os_ufree(dbp->get_DB()->dbenv, data.get_data());
+			if (ret != 0)
 				goto err;
-			}
-			key_alloc = 1;
 			memcpy(replyp->skeydata.skeydata_val, skey.get_data(),
 			    skey.get_size());
 		} else
@@ -1375,22 +1340,8 @@ __db_pget_proc(
 		if (pkey.get_data() == pkeydata) {
 			ret = __os_umalloc(dbp->get_DB()->dbenv,
 			     pkey.get_size(), &replyp->pkeydata.pkeydata_val);
-			if (ret != 0) {
-				__os_ufree(dbp->get_DB()->dbenv, skey.get_data());
-				__os_ufree(dbp->get_DB()->dbenv, pkey.get_data());
-				__os_ufree(dbp->get_DB()->dbenv, data.get_data());
-				if (key_alloc)
-					__os_ufree(dbp->get_DB()->dbenv,
-					    replyp->skeydata.skeydata_val);
+			if (ret != 0)
 				goto err;
-			}
-			/*
-			 * We can set it to 2, because they cannot send the
-			 * pkey over without sending the skey over too.
-			 * So if they did send a pkey, they must have sent
-			 * the skey as well.
-			 */
-			key_alloc = 2;
 			memcpy(replyp->pkeydata.pkeydata_val, pkey.get_data(),
 			    pkey.get_size());
 		} else
@@ -1403,30 +1354,27 @@ __db_pget_proc(
 		if (data.get_data() == datadata) {
 			ret = __os_umalloc(dbp->get_DB()->dbenv,
 			     data.get_size(), &replyp->datadata.datadata_val);
-			if (ret != 0) {
-				__os_ufree(dbp->get_DB()->dbenv, skey.get_data());
-				__os_ufree(dbp->get_DB()->dbenv, pkey.get_data());
-				__os_ufree(dbp->get_DB()->dbenv, data.get_data());
-				/*
-				 * If key_alloc is 1, just skey needs to be
-				 * freed, if key_alloc is 2, both skey and pkey
-				 * need to be freed.
-				 */
-				if (key_alloc--)
-					__os_ufree(dbp->get_DB()->dbenv,
-					    replyp->skeydata.skeydata_val);
-				if (key_alloc)
-					__os_ufree(dbp->get_DB()->dbenv,
-					    replyp->pkeydata.pkeydata_val);
+			if (ret != 0)
 				goto err;
-			}
 			memcpy(replyp->datadata.datadata_val, data.get_data(),
 			    data.get_size());
 		} else
 			replyp->datadata.datadata_val = (char *)data.get_data();
 		replyp->datadata.datadata_len = data.get_size();
 	} else {
-err:		replyp->skeydata.skeydata_val = NULL;
+err:		FREE_IF_CHANGED(dbp->get_DB()->dbenv,
+		    skey.get_data(), skeydata);
+		FREE_IF_CHANGED(dbp->get_DB()->dbenv,
+		    pkey.get_data(), pkeydata);
+		FREE_IF_CHANGED(dbp->get_DB()->dbenv,
+		    data.get_data(), datadata);
+		FREE_IF_CHANGED(dbp->get_DB()->dbenv,
+		    replyp->skeydata.skeydata_val, skey.get_data());
+		FREE_IF_CHANGED(dbp->get_DB()->dbenv,
+		    replyp->pkeydata.pkeydata_val, pkey.get_data());
+		FREE_IF_CHANGED(dbp->get_DB()->dbenv,
+		    replyp->datadata.datadata_val, data.get_data());
+		replyp->skeydata.skeydata_val = NULL;
 		replyp->skeydata.skeydata_len = 0;
 		replyp->pkeydata.pkeydata_val = NULL;
 		replyp->pkeydata.pkeydata_len = 0;
@@ -1440,8 +1388,8 @@ err:		replyp->skeydata.skeydata_val = NULL;
 
 extern "C" void
 __db_put_proc(
-	long dbpcl_id,
-	long txnpcl_id,
+	u_int dbpcl_id,
+	u_int txnpcl_id,
 	u_int32_t keydlen,
 	u_int32_t keydoff,
 	u_int32_t keyulen,
@@ -1471,6 +1419,7 @@ __db_put_proc(
 	} else
 		txnp = NULL;
 
+	replyp->keydata.keydata_val = NULL;
 	*freep = 0;
 
 	/* Set up key and data */
@@ -1508,17 +1457,20 @@ __db_put_proc(
 		if (key.get_data() == keydata) {
 			ret = __os_umalloc(dbp->get_DB()->dbenv,
 			    key.get_size(), &replyp->keydata.keydata_val);
-			if (ret != 0) {
-				__os_ufree(dbp->get_DB()->dbenv, key.get_data());
+			if (ret != 0)
 				goto err;
-			}
-			memcpy(replyp->keydata.keydata_val, key.get_data(), key.get_size());
+			memcpy(replyp->keydata.keydata_val,
+			    key.get_data(), key.get_size());
 		} else
 			replyp->keydata.keydata_val = (char *)key.get_data();
 
 		replyp->keydata.keydata_len = key.get_size();
 	} else {
-err:		replyp->keydata.keydata_val = NULL;
+err:		FREE_IF_CHANGED(dbp->get_DB()->dbenv,
+		    key.get_data(), keydata);
+		FREE_IF_CHANGED(dbp->get_DB()->dbenv,
+		    replyp->keydata.keydata_val, key.get_data());
+		replyp->keydata.keydata_val = NULL;
 		replyp->keydata.keydata_len = 0;
 		*freep = 0;
 	}
@@ -1528,7 +1480,7 @@ err:		replyp->keydata.keydata_val = NULL;
 
 extern "C" void
 __db_get_re_delim_proc(
-	long dbpcl_id,
+	u_int dbpcl_id,
 	__db_get_re_delim_reply *replyp)
 {
 	Db *dbp;
@@ -1541,10 +1493,10 @@ __db_get_re_delim_proc(
 }
 
 extern "C" void
-__db_re_delim_proc(
-	long dbpcl_id,
+__db_set_re_delim_proc(
+	u_int dbpcl_id,
 	u_int32_t delim,
-	__db_re_delim_reply *replyp)
+	__db_set_re_delim_reply *replyp)
 {
 	Db *dbp;
 	ct_entry *dbp_ctp;
@@ -1561,7 +1513,7 @@ __db_re_delim_proc(
 
 extern "C" void
 __db_get_re_len_proc(
-	long dbpcl_id,
+	u_int dbpcl_id,
 	__db_get_re_len_reply *replyp)
 {
 	Db *dbp;
@@ -1574,10 +1526,10 @@ __db_get_re_len_proc(
 }
 
 extern "C" void
-__db_re_len_proc(
-	long dbpcl_id,
+__db_set_re_len_proc(
+	u_int dbpcl_id,
 	u_int32_t len,
-	__db_re_len_reply *replyp)
+	__db_set_re_len_reply *replyp)
 {
 	Db *dbp;
 	ct_entry *dbp_ctp;
@@ -1594,7 +1546,7 @@ __db_re_len_proc(
 
 void
 __db_get_re_pad_proc(
-	long dbpcl_id,
+	u_int dbpcl_id,
 	__db_get_re_pad_reply *replyp)
 {
 	Db *dbp;
@@ -1607,10 +1559,10 @@ __db_get_re_pad_proc(
 }
 
 extern "C" void
-__db_re_pad_proc(
-	long dbpcl_id,
+__db_set_re_pad_proc(
+	u_int dbpcl_id,
 	u_int32_t pad,
-	__db_re_pad_reply *replyp)
+	__db_set_re_pad_reply *replyp)
 {
 	Db *dbp;
 	ct_entry *dbp_ctp;
@@ -1627,7 +1579,7 @@ __db_re_pad_proc(
 
 extern "C" void
 __db_remove_proc(
-	long dbpcl_id,
+	u_int dbpcl_id,
 	char *name,
 	char *subdb,
 	u_int32_t flags,
@@ -1649,7 +1601,7 @@ __db_remove_proc(
 
 extern "C" void
 __db_rename_proc(
-	long dbpcl_id,
+	u_int dbpcl_id,
 	char *name,
 	char *subdb,
 	char *newname,
@@ -1672,22 +1624,29 @@ __db_rename_proc(
 
 extern "C" void
 __db_stat_proc(
-	long dbpcl_id,
+	u_int dbpcl_id,
+	u_int txnpcl_id,
 	u_int32_t flags,
 	__db_stat_reply *replyp,
 	int * freep)
 {
 	Db *dbp;
+	DbTxn *txnp;
 	DBTYPE type;
-	ct_entry *dbp_ctp;
+	ct_entry *dbp_ctp, *txnp_ctp;
 	u_int32_t *q, *p, *retsp;
 	int i, len, ret;
 	void *sp;
 
 	ACTIVATE_CTP(dbp_ctp, dbpcl_id, CT_DB);
 	dbp = (Db *)dbp_ctp->ct_anyp;
+	if (txnpcl_id != 0) {
+		ACTIVATE_CTP(txnp_ctp, txnpcl_id, CT_TXN);
+		txnp = (DbTxn *)txnp_ctp->ct_anyp;
+	} else
+		txnp = NULL;
 
-	ret = dbp->stat(&sp, flags);
+	ret = dbp->stat(txnp, &sp, flags);
 	replyp->status = ret;
 	if (ret != 0)
 		return;
@@ -1726,7 +1685,7 @@ out:
 
 extern "C" void
 __db_sync_proc(
-	long dbpcl_id,
+	u_int dbpcl_id,
 	u_int32_t flags,
 	__db_sync_reply *replyp)
 {
@@ -1745,8 +1704,8 @@ __db_sync_proc(
 
 extern "C" void
 __db_truncate_proc(
-	long dbpcl_id,
-	long txnpcl_id,
+	u_int dbpcl_id,
+	u_int txnpcl_id,
 	u_int32_t flags,
 	__db_truncate_reply *replyp)
 {
@@ -1773,8 +1732,8 @@ __db_truncate_proc(
 
 extern "C" void
 __db_cursor_proc(
-	long dbpcl_id,
-	long txnpcl_id,
+	u_int dbpcl_id,
+	u_int txnpcl_id,
 	u_int32_t flags,
 	__db_cursor_reply *replyp)
 {
@@ -1815,7 +1774,7 @@ __db_cursor_proc(
 
 extern "C" void
 __db_join_proc(
-	long dbpcl_id,
+	u_int dbpcl_id,
 	u_int32_t *curs,
 	u_int32_t curslen,
 	u_int32_t flags,
@@ -1913,9 +1872,9 @@ out:
 }
 
 extern "C" void
-__dbc_close_proc(
-	long dbccl_id,
-	__dbc_close_reply *replyp)
+__dbc_c_close_proc(
+	u_int dbccl_id,
+	__dbc_c_close_reply *replyp)
 {
 	ct_entry *dbc_ctp;
 
@@ -1925,10 +1884,10 @@ __dbc_close_proc(
 }
 
 extern "C" void
-__dbc_count_proc(
-	long dbccl_id,
+__dbc_c_count_proc(
+	u_int dbccl_id,
 	u_int32_t flags,
-	__dbc_count_reply *replyp)
+	__dbc_c_count_reply *replyp)
 {
 	Dbc *dbc;
 	ct_entry *dbc_ctp;
@@ -1946,10 +1905,10 @@ __dbc_count_proc(
 }
 
 extern "C" void
-__dbc_del_proc(
-	long dbccl_id,
+__dbc_c_del_proc(
+	u_int dbccl_id,
 	u_int32_t flags,
-	__dbc_del_reply *replyp)
+	__dbc_c_del_reply *replyp)
 {
 	Dbc *dbc;
 	ct_entry *dbc_ctp;
@@ -1965,10 +1924,10 @@ __dbc_del_proc(
 }
 
 extern "C" void
-__dbc_dup_proc(
-	long dbccl_id,
+__dbc_c_dup_proc(
+	u_int dbccl_id,
 	u_int32_t flags,
-	__dbc_dup_reply *replyp)
+	__dbc_c_dup_reply *replyp)
 {
 	Dbc *dbc, *newdbc;
 	ct_entry *dbc_ctp, *new_ctp;
@@ -2002,8 +1961,8 @@ __dbc_dup_proc(
 }
 
 extern "C" void
-__dbc_get_proc(
-	long dbccl_id,
+__dbc_c_get_proc(
+	u_int dbccl_id,
 	u_int32_t keydlen,
 	u_int32_t keydoff,
 	u_int32_t keyulen,
@@ -2017,21 +1976,22 @@ __dbc_get_proc(
 	void *datadata,
 	u_int32_t datasize,
 	u_int32_t flags,
-	__dbc_get_reply *replyp,
+	__dbc_c_get_reply *replyp,
 	int * freep)
 {
 	Dbc *dbc;
 	DbEnv *dbenv;
 	ct_entry *dbc_ctp;
-	int key_alloc, bulk_alloc, ret;
+	int ret;
 	void *tmpdata;
 
 	ACTIVATE_CTP(dbc_ctp, dbccl_id, CT_CURSOR);
 	dbc = (Dbc *)dbc_ctp->ct_anyp;
 	dbenv = DbEnv::get_DbEnv(((DBC *)dbc)->dbp->dbenv);
 
+	replyp->keydata.keydata_val = NULL;
+	replyp->datadata.datadata_val = NULL;
 	*freep = 0;
-	bulk_alloc = 0;
 
 	/* Set up key and data */
 	Dbt key(keydata, keysize);
@@ -2052,7 +2012,6 @@ __dbc_get_proc(
 			if (ret != 0)
 				goto err;
 			data.set_data(tmpdata);
-			bulk_alloc = 1;
 		}
 		dataflags |= DB_DBT_USERMEM;
 	} else
@@ -2078,17 +2037,13 @@ __dbc_get_proc(
 		/*
 		 * Key
 		 */
-		key_alloc = 0;
 		if (key.get_data() == keydata) {
 			ret = __os_umalloc(dbenv->get_DB_ENV(), key.get_size(),
 			    &replyp->keydata.keydata_val);
-			if (ret != 0) {
-				__os_ufree(dbenv->get_DB_ENV(), key.get_data());
-				__os_ufree(dbenv->get_DB_ENV(), data.get_data());
+			if (ret != 0)
 				goto err;
-			}
-			key_alloc = 1;
-			memcpy(replyp->keydata.keydata_val, key.get_data(), key.get_size());
+			memcpy(replyp->keydata.keydata_val,
+			    key.get_data(), key.get_size());
 		} else
 			replyp->keydata.keydata_val = (char *)key.get_data();
 
@@ -2100,35 +2055,35 @@ __dbc_get_proc(
 		if (data.get_data() == datadata) {
 			ret = __os_umalloc(dbenv->get_DB_ENV(), data.get_size(),
 			    &replyp->datadata.datadata_val);
-			if (ret != 0) {
-				__os_ufree(dbenv->get_DB_ENV(), key.get_data());
-				__os_ufree(dbenv->get_DB_ENV(), data.get_data());
-				if (key_alloc)
-					__os_ufree(dbenv->get_DB_ENV(),
-					    replyp->keydata.keydata_val);
+			if (ret != 0)
 				goto err;
-			}
 			memcpy(replyp->datadata.datadata_val, data.get_data(),
 			    data.get_size());
 		} else
 			replyp->datadata.datadata_val = (char *)data.get_data();
 		replyp->datadata.datadata_len = data.get_size();
 	} else {
-err:		replyp->keydata.keydata_val = NULL;
+err:		FREE_IF_CHANGED(dbenv->get_DB_ENV(),
+		    key.get_data(), keydata);
+		FREE_IF_CHANGED(dbenv->get_DB_ENV(),
+		    data.get_data(), datadata);
+		FREE_IF_CHANGED(dbenv->get_DB_ENV(),
+		    replyp->keydata.keydata_val, key.get_data());
+		FREE_IF_CHANGED(dbenv->get_DB_ENV(),
+		    replyp->datadata.datadata_val, data.get_data());
+		replyp->keydata.keydata_val = NULL;
 		replyp->keydata.keydata_len = 0;
 		replyp->datadata.datadata_val = NULL;
 		replyp->datadata.datadata_len = 0;
 		*freep = 0;
-		if (bulk_alloc)
-			__os_ufree(dbenv->get_DB_ENV(), data.get_data());
 	}
 	replyp->status = ret;
 	return;
 }
 
 extern "C" void
-__dbc_pget_proc(
-	long dbccl_id,
+__dbc_c_pget_proc(
+	u_int dbccl_id,
 	u_int32_t skeydlen,
 	u_int32_t skeydoff,
 	u_int32_t skeyulen,
@@ -2148,18 +2103,21 @@ __dbc_pget_proc(
 	void *datadata,
 	u_int32_t datasize,
 	u_int32_t flags,
-	__dbc_pget_reply *replyp,
+	__dbc_c_pget_reply *replyp,
 	int * freep)
 {
 	Dbc *dbc;
 	DbEnv *dbenv;
 	ct_entry *dbc_ctp;
-	int key_alloc, ret;
+	int ret;
 
 	ACTIVATE_CTP(dbc_ctp, dbccl_id, CT_CURSOR);
 	dbc = (Dbc *)dbc_ctp->ct_anyp;
 	dbenv = DbEnv::get_DbEnv(((DBC *)dbc)->dbp->dbenv);
 
+	replyp->skeydata.skeydata_val = NULL;
+	replyp->pkeydata.pkeydata_val = NULL;
+	replyp->datadata.datadata_val = NULL;
 	*freep = 0;
 
 	/*
@@ -2202,17 +2160,11 @@ __dbc_pget_proc(
 		/*
 		 * Key
 		 */
-		key_alloc = 0;
 		if (skey.get_data() == skeydata) {
 			ret = __os_umalloc(dbenv->get_DB_ENV(),
 			    skey.get_size(), &replyp->skeydata.skeydata_val);
-			if (ret != 0) {
-				__os_ufree(dbenv->get_DB_ENV(), skey.get_data());
-				__os_ufree(dbenv->get_DB_ENV(), pkey.get_data());
-				__os_ufree(dbenv->get_DB_ENV(), data.get_data());
+			if (ret != 0)
 				goto err;
-			}
-			key_alloc = 1;
 			memcpy(replyp->skeydata.skeydata_val, skey.get_data(),
 			    skey.get_size());
 		} else
@@ -2225,22 +2177,8 @@ __dbc_pget_proc(
 		if (pkey.get_data() == pkeydata) {
 			ret = __os_umalloc(dbenv->get_DB_ENV(),
 			     pkey.get_size(), &replyp->pkeydata.pkeydata_val);
-			if (ret != 0) {
-				__os_ufree(dbenv->get_DB_ENV(), skey.get_data());
-				__os_ufree(dbenv->get_DB_ENV(), pkey.get_data());
-				__os_ufree(dbenv->get_DB_ENV(), data.get_data());
-				if (key_alloc)
-					__os_ufree(dbenv->get_DB_ENV(),
-					    replyp->skeydata.skeydata_val);
+			if (ret != 0)
 				goto err;
-			}
-			/*
-			 * We can set it to 2, because they cannot send the
-			 * pkey over without sending the skey over too.
-			 * So if they did send a pkey, they must have sent
-			 * the skey as well.
-			 */
-			key_alloc = 2;
 			memcpy(replyp->pkeydata.pkeydata_val, pkey.get_data(),
 			    pkey.get_size());
 		} else
@@ -2253,30 +2191,27 @@ __dbc_pget_proc(
 		if (data.get_data() == datadata) {
 			ret = __os_umalloc(dbenv->get_DB_ENV(),
 			     data.get_size(), &replyp->datadata.datadata_val);
-			if (ret != 0) {
-				__os_ufree(dbenv->get_DB_ENV(), skey.get_data());
-				__os_ufree(dbenv->get_DB_ENV(), pkey.get_data());
-				__os_ufree(dbenv->get_DB_ENV(), data.get_data());
-				/*
-				 * If key_alloc is 1, just skey needs to be
-				 * freed, if key_alloc is 2, both skey and pkey
-				 * need to be freed.
-				 */
-				if (key_alloc--)
-					__os_ufree(dbenv->get_DB_ENV(),
-					    replyp->skeydata.skeydata_val);
-				if (key_alloc)
-					__os_ufree(dbenv->get_DB_ENV(),
-					    replyp->pkeydata.pkeydata_val);
+			if (ret != 0)
 				goto err;
-			}
 			memcpy(replyp->datadata.datadata_val, data.get_data(),
 			    data.get_size());
 		} else
 			replyp->datadata.datadata_val = (char *)data.get_data();
 		replyp->datadata.datadata_len = data.get_size();
 	} else {
-err:		replyp->skeydata.skeydata_val = NULL;
+err:		FREE_IF_CHANGED(dbenv->get_DB_ENV(),
+		    skey.get_data(), skeydata);
+		FREE_IF_CHANGED(dbenv->get_DB_ENV(),
+		    pkey.get_data(), pkeydata);
+		FREE_IF_CHANGED(dbenv->get_DB_ENV(),
+		    data.get_data(), datadata);
+		FREE_IF_CHANGED(dbenv->get_DB_ENV(),
+		    replyp->skeydata.skeydata_val, skey.get_data());
+		FREE_IF_CHANGED(dbenv->get_DB_ENV(),
+		    replyp->pkeydata.pkeydata_val, pkey.get_data());
+		FREE_IF_CHANGED(dbenv->get_DB_ENV(),
+		    replyp->datadata.datadata_val, data.get_data());
+		replyp->skeydata.skeydata_val = NULL;
 		replyp->skeydata.skeydata_len = 0;
 		replyp->pkeydata.pkeydata_val = NULL;
 		replyp->pkeydata.pkeydata_len = 0;
@@ -2289,8 +2224,8 @@ err:		replyp->skeydata.skeydata_val = NULL;
 }
 
 extern "C" void
-__dbc_put_proc(
-	long dbccl_id,
+__dbc_c_put_proc(
+	u_int dbccl_id,
 	u_int32_t keydlen,
 	u_int32_t keydoff,
 	u_int32_t keyulen,
@@ -2304,7 +2239,7 @@ __dbc_put_proc(
 	void *datadata,
 	u_int32_t datasize,
 	u_int32_t flags,
-	__dbc_put_reply *replyp,
+	__dbc_c_put_reply *replyp,
 	int * freep)
 {
 	Db *dbp;
@@ -2343,7 +2278,8 @@ __dbc_put_proc(
 		ret = dbp->get_type(&dbtype);
 		if (ret == 0 && dbtype == DB_RECNO) {
 			/*
-			 * We need to xdr_free whatever we are returning, next time.
+			 * We need to xdr_free whatever we are returning, next
+			 * time.
 			 */
 			replyp->keydata.keydata_val = (char *)key.get_data();
 			replyp->keydata.keydata_len = key.get_size();
@@ -2352,4 +2288,3 @@ __dbc_put_proc(
 	replyp->status = ret;
 	return;
 }
-#endif /* HAVE_RPC */
