@@ -1,10 +1,10 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2001-2003
- *      Sleepycat Software.  All rights reserved.
+ * Copyright (c) 2001-2005
+ *	Sleepycat Software.  All rights reserved.
  *
- * $Id: Timer.java,v 1.2 2003/01/08 05:45:53 bostic Exp $
+ * $Id: Timer.java,v 12.1 2005/06/16 20:23:43 bostic Exp $
  */
 
 package com.sleepycat.db.rpcserver;
@@ -15,8 +15,7 @@ package com.sleepycat.db.rpcserver;
  * used to group/share access times. This is done to keep the Java code
  * close to the canonical C implementation of the RPC server.
  */
-public class Timer
-{
-	Timer timer = this;
-	long last_access;
+public class Timer {
+    Timer timer = this;
+    long last_access;
 }
