@@ -1,16 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1999-2003
+ * Copyright (c) 1999-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: tcl_dbcursor.c,v 12.2 2005/06/16 20:23:46 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: tcl_dbcursor.c,v 11.57 2003/05/17 15:15:45 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -43,7 +41,7 @@ dbc_Cmd(clientData, interp, objc, objv)
 	Tcl_Obj *CONST objv[];		/* The argument objects */
 {
 	static const char *dbccmds[] = {
-#if CONFIG_TEST
+#ifdef CONFIG_TEST
 		"pget",
 #endif
 		"close",
@@ -54,7 +52,7 @@ dbc_Cmd(clientData, interp, objc, objv)
 		NULL
 	};
 	enum dbccmds {
-#if CONFIG_TEST
+#ifdef CONFIG_TEST
 		DBCPGET,
 #endif
 		DBCCLOSE,
@@ -93,7 +91,7 @@ dbc_Cmd(clientData, interp, objc, objv)
 	    TCL_EXACT, &cmdindex) != TCL_OK)
 		return (IS_HELP(objv[1]));
 	switch ((enum dbccmds)cmdindex) {
-#if CONFIG_TEST
+#ifdef CONFIG_TEST
 	case DBCPGET:
 		result = tcl_DbcGet(interp, objc, objv, dbc, 1);
 		break;
@@ -152,7 +150,7 @@ tcl_DbcPut(interp, objc, objv, dbc)
 	DBC *dbc;			/* Cursor pointer */
 {
 	static const char *dbcutopts[] = {
-#if CONFIG_TEST
+#ifdef CONFIG_TEST
 		"-nodupdata",
 #endif
 		"-after",
@@ -164,7 +162,7 @@ tcl_DbcPut(interp, objc, objv, dbc)
 		NULL
 	};
 	enum dbcutopts {
-#if CONFIG_TEST
+#ifdef CONFIG_TEST
 		DBCPUT_NODUPDATA,
 #endif
 		DBCPUT_AFTER,
@@ -184,6 +182,9 @@ tcl_DbcPut(interp, objc, objv, dbc)
 	u_int32_t flag;
 	int elemc, freekey, freedata, i, optindex, result, ret;
 
+	COMPQUIET(dtmp, NULL);
+	COMPQUIET(ktmp, NULL);
+
 	result = TCL_OK;
 	flag = 0;
 	freekey = freedata = 0;
@@ -217,7 +218,7 @@ tcl_DbcPut(interp, objc, objv, dbc)
 		}
 		i++;
 		switch ((enum dbcutopts)optindex) {
-#if CONFIG_TEST
+#ifdef CONFIG_TEST
 		case DBCPUT_NODUPDATA:
 			FLAG_CHECK(flag);
 			flag = DB_NODUPDATA;
@@ -363,9 +364,9 @@ tcl_DbcPut(interp, objc, objv, dbc)
 	}
 out:
 	if (freedata)
-		(void)__os_free(NULL, dtmp);
+		__os_free(NULL, dtmp);
 	if (freekey)
-		(void)__os_free(NULL, ktmp);
+		__os_free(NULL, ktmp);
 	return (result);
 }
 
@@ -381,11 +382,12 @@ tcl_DbcGet(interp, objc, objv, dbc, ispget)
 	int ispget;			/* 1 for pget, 0 for get */
 {
 	static const char *dbcgetopts[] = {
-#if CONFIG_TEST
-		"-dirty",
+#ifdef CONFIG_TEST
 		"-get_both_range",
 		"-multi",
 		"-multi_key",
+		"-read_committed",
+		"-read_uncommitted",
 #endif
 		"-current",
 		"-first",
@@ -406,11 +408,12 @@ tcl_DbcGet(interp, objc, objv, dbc, ispget)
 		NULL
 	};
 	enum dbcgetopts {
-#if CONFIG_TEST
-		DBCGET_DIRTY,
+#ifdef CONFIG_TEST
 		DBCGET_BOTH_RANGE,
 		DBCGET_MULTI,
 		DBCGET_MULTI_KEY,
+		DBCGET_READ_COMMITTED,
+		DBCGET_READ_UNCOMMITTED,
 #endif
 		DBCGET_CURRENT,
 		DBCGET_FIRST,
@@ -438,9 +441,13 @@ tcl_DbcGet(interp, objc, objv, dbc, ispget)
 	db_recno_t precno, recno;
 	u_int32_t flag, op;
 	int elemc, freekey, freedata, i, optindex, result, ret;
-#if CONFIG_TEST
+#ifdef CONFIG_TEST
 	int bufsize;
+
+	bufsize = 0;
 #endif
+	COMPQUIET(dtmp, NULL);
+	COMPQUIET(ktmp, NULL);
 
 	result = TCL_OK;
 	flag = 0;
@@ -473,14 +480,14 @@ tcl_DbcGet(interp, objc, objv, dbc, ispget)
 			break;
 		}
 		i++;
+
+#define	FLAG_CHECK2_STDARG	\
+	(DB_RMW | DB_MULTIPLE | DB_MULTIPLE_KEY | DB_READ_UNCOMMITTED)
+
 		switch ((enum dbcgetopts)optindex) {
-#if CONFIG_TEST
-		case DBCGET_DIRTY:
-			flag |= DB_DIRTY_READ;
-			break;
+#ifdef CONFIG_TEST
 		case DBCGET_BOTH_RANGE:
-			FLAG_CHECK2(flag,
-			    DB_RMW|DB_MULTIPLE|DB_MULTIPLE_KEY|DB_DIRTY_READ);
+			FLAG_CHECK2(flag, FLAG_CHECK2_STDARG);
 			flag |= DB_GET_BOTH_RANGE;
 			break;
 		case DBCGET_MULTI:
@@ -497,78 +504,70 @@ tcl_DbcGet(interp, objc, objv, dbc, ispget)
 				goto out;
 			i++;
 			break;
+		case DBCGET_READ_COMMITTED:
+			flag |= DB_READ_COMMITTED;
+			break;
+		case DBCGET_READ_UNCOMMITTED:
+			flag |= DB_READ_UNCOMMITTED;
+			break;
 #endif
 		case DBCGET_RMW:
 			flag |= DB_RMW;
 			break;
 		case DBCGET_CURRENT:
-			FLAG_CHECK2(flag,
-			    DB_RMW|DB_MULTIPLE|DB_MULTIPLE_KEY|DB_DIRTY_READ);
+			FLAG_CHECK2(flag, FLAG_CHECK2_STDARG);
 			flag |= DB_CURRENT;
 			break;
 		case DBCGET_FIRST:
-			FLAG_CHECK2(flag,
-			    DB_RMW|DB_MULTIPLE|DB_MULTIPLE_KEY|DB_DIRTY_READ);
+			FLAG_CHECK2(flag, FLAG_CHECK2_STDARG);
 			flag |= DB_FIRST;
 			break;
 		case DBCGET_LAST:
-			FLAG_CHECK2(flag,
-			    DB_RMW|DB_MULTIPLE|DB_MULTIPLE_KEY|DB_DIRTY_READ);
+			FLAG_CHECK2(flag, FLAG_CHECK2_STDARG);
 			flag |= DB_LAST;
 			break;
 		case DBCGET_NEXT:
-			FLAG_CHECK2(flag,
-			    DB_RMW|DB_MULTIPLE|DB_MULTIPLE_KEY|DB_DIRTY_READ);
+			FLAG_CHECK2(flag, FLAG_CHECK2_STDARG);
 			flag |= DB_NEXT;
 			break;
 		case DBCGET_PREV:
-			FLAG_CHECK2(flag,
-			    DB_RMW|DB_MULTIPLE|DB_MULTIPLE_KEY|DB_DIRTY_READ);
+			FLAG_CHECK2(flag, FLAG_CHECK2_STDARG);
 			flag |= DB_PREV;
 			break;
 		case DBCGET_PREVNODUP:
-			FLAG_CHECK2(flag,
-			    DB_RMW|DB_MULTIPLE|DB_MULTIPLE_KEY|DB_DIRTY_READ);
+			FLAG_CHECK2(flag, FLAG_CHECK2_STDARG);
 			flag |= DB_PREV_NODUP;
 			break;
 		case DBCGET_NEXTNODUP:
-			FLAG_CHECK2(flag,
-			    DB_RMW|DB_MULTIPLE|DB_MULTIPLE_KEY|DB_DIRTY_READ);
+			FLAG_CHECK2(flag, FLAG_CHECK2_STDARG);
 			flag |= DB_NEXT_NODUP;
 			break;
 		case DBCGET_NEXTDUP:
-			FLAG_CHECK2(flag,
-			    DB_RMW|DB_MULTIPLE|DB_MULTIPLE_KEY|DB_DIRTY_READ);
+			FLAG_CHECK2(flag, FLAG_CHECK2_STDARG);
 			flag |= DB_NEXT_DUP;
 			break;
 		case DBCGET_BOTH:
-			FLAG_CHECK2(flag,
-			    DB_RMW|DB_MULTIPLE|DB_MULTIPLE_KEY|DB_DIRTY_READ);
+			FLAG_CHECK2(flag, FLAG_CHECK2_STDARG);
 			flag |= DB_GET_BOTH;
 			break;
 		case DBCGET_RECNO:
-			FLAG_CHECK2(flag,
-			    DB_RMW|DB_MULTIPLE|DB_MULTIPLE_KEY|DB_DIRTY_READ);
+			FLAG_CHECK2(flag, FLAG_CHECK2_STDARG);
 			flag |= DB_GET_RECNO;
 			break;
 		case DBCGET_JOIN:
-			FLAG_CHECK2(flag,
-			    DB_RMW|DB_MULTIPLE|DB_MULTIPLE_KEY|DB_DIRTY_READ);
+			FLAG_CHECK2(flag, FLAG_CHECK2_STDARG);
 			flag |= DB_JOIN_ITEM;
 			break;
 		case DBCGET_SET:
-			FLAG_CHECK2(flag,
-			    DB_RMW|DB_MULTIPLE|DB_MULTIPLE_KEY|DB_DIRTY_READ);
+			FLAG_CHECK2(flag, FLAG_CHECK2_STDARG);
 			flag |= DB_SET;
 			break;
 		case DBCGET_SETRANGE:
-			FLAG_CHECK2(flag,
-			    DB_RMW|DB_MULTIPLE|DB_MULTIPLE_KEY|DB_DIRTY_READ);
+			FLAG_CHECK2(flag, FLAG_CHECK2_STDARG);
 			flag |= DB_SET_RANGE;
 			break;
 		case DBCGET_SETRECNO:
-			FLAG_CHECK2(flag,
-			    DB_RMW|DB_MULTIPLE|DB_MULTIPLE_KEY|DB_DIRTY_READ);
+			FLAG_CHECK2(flag, FLAG_CHECK2_STDARG);
 			flag |= DB_SET_RECNO;
 			break;
 		case DBCGET_PART:
@@ -643,7 +642,7 @@ tcl_DbcGet(interp, objc, objv, dbc, ispget)
 	op = flag & DB_OPFLAGS_MASK;
 	switch (op) {
 	case DB_GET_BOTH:
-#if CONFIG_TEST
+#ifdef CONFIG_TEST
 	case DB_GET_BOTH_RANGE:
 #endif
 		if (i != (objc - 2)) {
@@ -703,10 +702,10 @@ tcl_DbcGet(interp, objc, objv, dbc, ispget)
 			result = TCL_ERROR;
 			goto out;
 		}
-#if CONFIG_TEST
+#ifdef CONFIG_TEST
 		if (flag & (DB_MULTIPLE|DB_MULTIPLE_KEY)) {
-			(void)__os_malloc(NULL, bufsize, &data.data);
-			data.ulen = bufsize;
+			(void)__os_malloc(NULL, (size_t)bufsize, &data.data);
+			data.ulen = (u_int32_t)bufsize;
 			data.flags |= DB_DBT_USERMEM;
 		} else
 #endif
@@ -739,10 +738,10 @@ tcl_DbcGet(interp, objc, objv, dbc, ispget)
 			goto out;
 		}
 		key.flags |= DB_DBT_MALLOC;
-#if CONFIG_TEST
+#ifdef CONFIG_TEST
 		if (flag & (DB_MULTIPLE|DB_MULTIPLE_KEY)) {
-			(void)__os_malloc(NULL, bufsize, &data.data);
-			data.ulen = bufsize;
+			(void)__os_malloc(NULL, (size_t)bufsize, &data.data);
+			data.ulen = (u_int32_t)bufsize;
 			data.flags |= DB_DBT_USERMEM;
 		} else
 #endif
@@ -761,7 +760,7 @@ tcl_DbcGet(interp, objc, objv, dbc, ispget)
 		goto out;
 
 	retlist = Tcl_NewListObj(0, NULL);
-	if (ret == DB_NOTFOUND)
+	if (ret != 0)
 		goto out1;
 	if (op == DB_GET_RECNO) {
 		recno = *((db_recno_t *)data.data);
@@ -806,9 +805,9 @@ out:
 	if (data.data != NULL && flag & (DB_MULTIPLE|DB_MULTIPLE_KEY))
 		__os_free(dbc->dbp->dbenv, data.data);
 	if (freedata)
-		(void)__os_free(NULL, dtmp);
+		__os_free(NULL, dtmp);
 	if (freekey)
-		(void)__os_free(NULL, ktmp);
+		__os_free(NULL, ktmp);
 	return (result);
 
 }
@@ -901,7 +900,6 @@ tcl_DbcDup(interp, objc, objv, dbc)
 	 * Now duplicate the cursor.  If successful, we need to create
 	 * a new cursor command.
 	 */
-
 	snprintf(newname, sizeof(newname),
 	    "%s.c%d", dbip->i_name, dbip->i_dbdbcid);
 	newdbcip = _NewInfo(interp, NULL, newname, I_DBC);
@@ -910,10 +908,10 @@ tcl_DbcDup(interp, objc, objv, dbc)
 		if (ret == 0) {
 			dbip->i_dbdbcid++;
 			newdbcip->i_parent = dbip;
-			Tcl_CreateObjCommand(interp, newname,
+			(void)Tcl_CreateObjCommand(interp, newname,
 			    (Tcl_ObjCmdProc *)dbc_Cmd,
 			    (ClientData)newdbc, NULL);
-			res = Tcl_NewStringObj(newname, strlen(newname));
+			res = NewStringObj(newname, strlen(newname));
 			_SetInfoData(newdbcip, newdbc);
 			Tcl_SetObjResult(interp, res);
 		} else {
