@@ -1,16 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1999-2003
+ * Copyright (c) 1999-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: tcl_mp.c,v 12.1 2005/06/16 20:23:48 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: tcl_mp.c,v 11.50 2003/09/04 20:45:45 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -25,7 +23,7 @@ static const char revid[] = "$Id: tcl_mp.c,v 11.50 2003/09/04 20:45:45 bostic Ex
 /*
  * Prototypes for procedures defined later in this file:
  */
-#if CONFIG_TEST
+#ifdef CONFIG_TEST
 static int      mp_Cmd __P((ClientData, Tcl_Interp *, int, Tcl_Obj * CONST*));
 static int      pg_Cmd __P((ClientData, Tcl_Interp *, int, Tcl_Obj * CONST*));
 static int      tcl_MpGet __P((Tcl_Interp *, int, Tcl_Obj * CONST*,
@@ -58,14 +56,14 @@ _MpInfoDelete(interp, mpip)
 		 * mp.  Remove its commands and info structure.
 		 */
 		nextp = LIST_NEXT(p, entries);
-		 if (p->i_parent == mpip && p->i_type == I_PG) {
+		if (p->i_parent == mpip && p->i_type == I_PG) {
 			(void)Tcl_DeleteCommand(interp, p->i_name);
 			_DeleteInfo(p);
 		}
 	}
 }
 
-#if CONFIG_TEST
+#ifdef CONFIG_TEST
 /*
  * tcl_MpSync --
  *
@@ -100,8 +98,7 @@ tcl_MpSync(interp, objc, objv, envp)
 
 	_debug_check();
 	ret = envp->memp_sync(envp, lsnp);
-	result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret), "memp sync");
-	return (result);
+	return (_ReturnSetup(interp, ret, DB_RETOK_STD(ret), "memp sync"));
 }
 
 /*
@@ -118,11 +115,8 @@ tcl_MpTrickle(interp, objc, objv, envp)
 	DB_ENV *envp;			/* Environment pointer */
 {
 
-	int pages;
-	int percent;
-	int result;
-	int ret;
 	Tcl_Obj *res;
+	int pages, percent, result, ret;
 
 	result = TCL_OK;
 	/*
@@ -298,9 +292,9 @@ tcl_Mp(interp, objc, objv, envp, envip)
 	ip->i_parent = envip;
 	ip->i_pgsz = pgsize;
 	_SetInfoData(ip, mpf);
-	Tcl_CreateObjCommand(interp, newname,
+	(void)Tcl_CreateObjCommand(interp, newname,
 	    (Tcl_ObjCmdProc *)mp_Cmd, (ClientData)mpf, NULL);
-	res = Tcl_NewStringObj(newname, strlen(newname));
+	res = NewStringObj(newname, strlen(newname));
 	Tcl_SetObjResult(interp, res);
 
 error:
@@ -353,6 +347,11 @@ tcl_MpStat(interp, objc, objv, envp)
 	MAKE_STAT_LIST("Cache size (bytes)", sp->st_bytes);
 	MAKE_STAT_LIST("Number of caches", sp->st_ncache);
 	MAKE_STAT_LIST("Region size", sp->st_regsize);
+	MAKE_STAT_LIST("Maximum memory-mapped file size", sp->st_mmapsize);
+	MAKE_STAT_LIST("Maximum open file descriptors", sp->st_maxopenfd);
+	MAKE_STAT_LIST("Maximum sequential buffer writes", sp->st_maxwrite);
+	MAKE_STAT_LIST(
+	    "Sleep after writing maximum buffers", sp->st_maxwrite_sleep);
 	MAKE_STAT_LIST("Pages mapped into address space", sp->st_map);
 	MAKE_STAT_LIST("Cache hits", sp->st_cache_hit);
 	MAKE_STAT_LIST("Cache misses", sp->st_cache_miss);
@@ -415,9 +414,9 @@ tcl_MpStat(interp, objc, objv, envp)
 	}
 	Tcl_SetObjResult(interp, res1);
 error:
-	(void)__os_ufree(envp, sp);
+	__os_ufree(envp, sp);
 	if (savefsp != NULL)
-		(void)__os_ufree(envp, savefsp);
+		__os_ufree(envp, savefsp);
 	return (result);
 }
 
@@ -521,7 +520,7 @@ mp_Cmd(clientData, interp, objc, objv)
 		ret = mp->get_clear_len(mp, &value);
 		if ((result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
 		    "mp get_clear_len")) == TCL_OK)
-			res = Tcl_NewIntObj(value);
+			res = Tcl_NewIntObj((int)value);
 		break;
 	case MPGETFILEID:
 		if (objc != 2) {
@@ -531,8 +530,7 @@ mp_Cmd(clientData, interp, objc, objv)
 		ret = mp->get_fileid(mp, fileid);
 		if ((result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
 		    "mp get_fileid")) == TCL_OK)
-			res = Tcl_NewStringObj((char *)fileid,
-			    (int)DB_FILE_ID_LEN);
+			res = NewStringObj((char *)fileid, DB_FILE_ID_LEN);
 		break;
 	case MPGETFTYPE:
 		if (objc != 2) {
@@ -564,7 +562,7 @@ mp_Cmd(clientData, interp, objc, objv)
 		if ((result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
 		    "mp get_pgcookie")) == TCL_OK)
 			res = Tcl_NewByteArrayObj((u_char *)cookie.data,
-			    cookie.size);
+			    (int)cookie.size);
 		break;
 	}
 	/*
@@ -663,7 +661,7 @@ tcl_MpGet(interp, objc, objv, mp, mpip)
 		return (TCL_ERROR);
 	}
 	_debug_check();
-	pgno = ipgno;
+	pgno = (db_pgno_t)ipgno;
 	ret = mp->get(mp, &pgno, flag, &page);
 	result = _ReturnSetup(interp, ret, DB_RETOK_MPGET(ret), "mpool get");
 	if (result == TCL_ERROR)
@@ -678,9 +676,9 @@ tcl_MpGet(interp, objc, objv, mp, mpip)
 		ip->i_pgno = pgno;
 		ip->i_pgsz = mpip->i_pgsz;
 		_SetInfoData(ip, page);
-		Tcl_CreateObjCommand(interp, newname,
+		(void)Tcl_CreateObjCommand(interp, newname,
 		    (Tcl_ObjCmdProc *)pg_Cmd, (ClientData)page, NULL);
-		res = Tcl_NewStringObj(newname, strlen(newname));
+		res = NewStringObj(newname, strlen(newname));
 		Tcl_SetObjResult(interp, res);
 	}
 error:
@@ -761,7 +759,7 @@ pg_Cmd(clientData, interp, objc, objv)
 	case PGSET:
 	case PGPUT:
 		result = tcl_Pg(interp, objc, objv, page, mp, pgip,
-		    cmdindex == PGSET ? 0 : 1);
+		    (enum pgcmds)cmdindex == PGSET ? 0 : 1);
 		break;
 	case PGINIT:
 		result = tcl_PgInit(interp, objc, objv, page, pgip);
@@ -849,9 +847,8 @@ tcl_PgInit(interp, objc, objv, page, pgip)
 	DBTCL_INFO *pgip;		/* Info pointer */
 {
 	Tcl_Obj *res;
-	size_t pgsz;
 	long *p, *endp, newval;
-	int length, result;
+	int length, pgsz, result;
 	u_char *s;
 
 	result = TCL_OK;
@@ -866,12 +863,11 @@ tcl_PgInit(interp, objc, objv, page, pgip)
 		s = Tcl_GetByteArrayFromObj(objv[2], &length);
 		if (s == NULL)
 			return (TCL_ERROR);
-		memcpy(page, s,
-		    ((size_t)length < pgsz) ? (size_t)length : pgsz);
+		memcpy(page, s, (size_t)((length < pgsz) ? length : pgsz));
 		result = TCL_OK;
 	} else {
 		p = (long *)page;
-		for (endp = p + (pgsz / sizeof(long)); p < endp; p++)
+		for (endp = p + ((u_int)pgsz / sizeof(long)); p < endp; p++)
 			*p = newval;
 	}
 	res = Tcl_NewIntObj(0);
@@ -888,9 +884,8 @@ tcl_PgIsset(interp, objc, objv, page, pgip)
 	DBTCL_INFO *pgip;		/* Info pointer */
 {
 	Tcl_Obj *res;
-	size_t pgsz;
 	long *p, *endp, newval;
-	int length, result;
+	int length, pgsz, result;
 	u_char *s;
 
 	result = TCL_OK;
@@ -907,7 +902,7 @@ tcl_PgIsset(interp, objc, objv, page, pgip)
 		result = TCL_OK;
 
 		if (memcmp(page, s,
-		    ((size_t)length < pgsz) ? (size_t)length : pgsz ) != 0) {
+		    (size_t)((length < pgsz) ? length : pgsz)) != 0) {
 			res = Tcl_NewIntObj(0);
 			Tcl_SetObjResult(interp, res);
 			return (result);
@@ -919,7 +914,7 @@ tcl_PgIsset(interp, objc, objv, page, pgip)
 		 * this value).  Otherwise, if we finish the loop, we return 1
 		 * (is set to this value).
 		 */
-		for (endp = p + (pgsz/sizeof(long)); p < endp; p++)
+		for (endp = p + ((u_int)pgsz / sizeof(long)); p < endp; p++)
 			if (*p != newval) {
 				res = Tcl_NewIntObj(0);
 				Tcl_SetObjResult(interp, res);
