@@ -1,9 +1,9 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 2001-2003
+# Copyright (c) 2001-2005
 #	Sleepycat Software.  All rights reserved.
 #
-# $Id: bigfile002.tcl,v 11.8 2003/01/08 05:49:21 bostic Exp $
+# $Id: bigfile002.tcl,v 12.2 2005/06/16 20:23:51 bostic Exp $
 #
 # TEST	bigfile002
 # TEST	This one should be faster and not require so much disk space,
@@ -12,8 +12,7 @@
 proc bigfile002 { args } {
 	source ./include.tcl
 
-	puts -nonewline \
-	    "Bigfile002: Creating large, sparse file through mpool..."
+	puts "Bigfile002: Creating large, sparse file through mpool..."
 	flush stdout
 
 	env_cleanup $testdir
@@ -37,8 +36,6 @@ proc bigfile002 { args } {
 	# Fsync.
 	error_check_good fsync [$file fsync] 0
 
-	puts "succeeded."
-
 	# Close.
 	error_check_good fclose [$file close] 0
 	error_check_good env_close [$env close] 0
