@@ -1,16 +1,18 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1999-2001
+# Copyright (c) 1999-2005
 #	Sleepycat Software.  All rights reserved.
 #
-# $Id: recd007.tcl,v 11.45 2001/06/19 18:30:47 sue Exp $
+# $Id: recd007.tcl,v 12.3 2005/09/28 18:17:04 carol Exp $
 #
-# Recovery Test 7.
-# This is a recovery test for create/delete of databases.  We have
-# hooks in the database so that we can abort the process at various
-# points and make sure that the transaction doesn't commit.  We
-# then need to recover and make sure the file is correctly existing
-# or not, as the case may be.
+# TEST	recd007
+# TEST	File create/delete tests.
+# TEST
+# TEST	This is a recovery test for create/delete of databases.  We have
+# TEST	hooks in the database so that we can abort the process at various
+# TEST	points and make sure that the transaction doesn't commit.  We
+# TEST	then need to recover and make sure the file is correctly existing
+# TEST	or not, as the case may be.
 proc recd007 { method args} {
 	global fixed_len
 	source ./include.tcl
@@ -28,10 +30,10 @@ proc recd007 { method args} {
 	set flags "-create -txn -home $testdir"
 
 	puts "\tRecd007.a: creating environment"
-	set env_cmd "berkdb env $flags"
+	set env_cmd "berkdb_env $flags"
 
 	set env [eval $env_cmd]
-	#
+
 	# We need to create a database to get the pagesize (either
 	# the default or whatever might have been specified).
 	# Then remove it so we can compute fixed_len and create the
@@ -53,8 +55,10 @@ proc recd007 { method args} {
 
 	# Convert the args again because fixed_len is now real.
 	set opts [convert_args $method ""]
+	set save_opts $opts
+	set moreopts {" -lorder 1234 " " -lorder 1234 -chksum " \
+	    " -lorder 4321 " " -lorder 4321 -chksum "}
 
-	#
 	# List of recovery tests: {HOOKS MSG} pairs
 	# Where each HOOK is a list of {COPY ABORT}
 	#
@@ -84,8 +88,17 @@ proc recd007 { method args} {
 	foreach pair $rlist {
 		set cmd [lindex $pair 0]
 		set msg [lindex $pair 1]
+		#
+		# Run natively
+		#
 		file_recover_create $testdir $env_cmd $omethod \
-		    $opts $testfile $cmd $msg
+		    $save_opts $testfile $cmd $msg
+		foreach o $moreopts {
+			set opts $save_opts
+			append opts $o
+			file_recover_create $testdir $env_cmd $omethod \
+			    $opts $testfile $cmd $msg
+		}
 	}
 
 	set rlist {
@@ -97,17 +110,25 @@ proc recd007 { method args} {
 	{ {"predestroy" "postdestroy"}	"Recd007.o: predestroy/postdestroy"}
 	{ {"postdestroy" "postdestroy"}	"Recd007.p: postdestroy/postdestroy"}
 	}
+
 	foreach op { dbremove dbrename dbtruncate } {
 		foreach pair $rlist {
 			set cmd [lindex $pair 0]
 			set msg [lindex $pair 1]
 			file_recover_delete $testdir $env_cmd $omethod \
-			$opts $testfile $cmd $msg $op
+			    $save_opts $testfile $cmd $msg $op
+			foreach o $moreopts {
+				set opts $save_opts
+				append opts $o
+				file_recover_delete $testdir $env_cmd $omethod \
+				    $opts $testfile $cmd $msg $op
+			}
 		}
 	}
 
 	if { $is_windows_test != 1 } {
-		do_file_recover_delmk $testdir $env_cmd $omethod $opts $testfile
+		set env_cmd "berkdb_env_noerr $flags"
+		do_file_recover_delmk $testdir $env_cmd $method $opts $testfile
 	}
 
 	puts "\tRecd007.r: Verify db_printlog can read logfile"
@@ -116,6 +137,8 @@ proc recd007 { method args} {
 	    > $tmpfile} ret]
 	error_check_good db_printlog $stat 0
 	fileremove $tmpfile
+	set fixed_len $orig_fixed_len
+	return
 }
 
 proc file_recover_create { dir env_cmd method opts dbfile cmd msg } {
@@ -124,17 +147,17 @@ proc file_recover_create { dir env_cmd method opts dbfile cmd msg } {
 	# 1.  Creating just a database
 	# 2.  Creating a database with a subdb
 	# 3.  Creating a 2nd subdb in a database
-	puts "\t$msg create with a database"
+	puts "\t$msg ($opts) create with a database"
 	do_file_recover_create $dir $env_cmd $method $opts $dbfile \
 	    0 $cmd $msg
 	if { [is_queue $method] == 1 } {
 		puts "\tSkipping subdatabase tests for method $method"
 		return
 	}
-	puts "\t$msg create with a database and subdb"
+	puts "\t$msg ($opts) create with a database and subdb"
 	do_file_recover_create $dir $env_cmd $method $opts $dbfile \
 	    1 $cmd $msg
-	puts "\t$msg create with a database and 2nd subdb"
+	puts "\t$msg ($opts) create with a database and 2nd subdb"
 	do_file_recover_create $dir $env_cmd $method $opts $dbfile \
 	    2 $cmd $msg
 
@@ -168,17 +191,16 @@ proc do_file_recover_create { dir env_cmd method opts dbfile sub cmd msg } {
 		return
 	}
 
-	#
 	# Basically non-existence is our initial state.  When we
 	# abort, it is also our final state.
 	#
 	switch $sub {
 		0 {
-			set oflags "-create $method -mode 0644 \
+			set oflags "-create $method -auto_commit -mode 0644 \
 			    -env $env $opts $dbfile"
 		}
 		1 {
-			set oflags "-create $method -mode 0644 \
+			set oflags "-create $method -auto_commit -mode 0644 \
 			    -env $env $opts $dbfile sub0"
 		}
 		2 {
@@ -186,14 +208,14 @@ proc do_file_recover_create { dir env_cmd method opts dbfile sub cmd msg } {
 			# If we are aborting here, then we need to
 			# create a first subdb, then create a second
 			#
-			set oflags "-create $method -mode 0644 \
+			set oflags "-create $method -auto_commit -mode 0644 \
 			    -env $env $opts $dbfile sub0"
 			set db [eval {berkdb_open} $oflags]
 			error_check_good db_open [is_valid_db $db] TRUE
 			error_check_good db_close [$db close] 0
 			set init_file $dir/$dbfile.init
 			catch { file copy -force $dir/$dbfile $init_file } res
-			set oflags "-create $method -mode 0644 \
+			set oflags "-create $method -auto_commit -mode 0644 \
 			    -env $env $opts $dbfile sub1"
 		}
 		default {
@@ -215,8 +237,7 @@ proc do_file_recover_create { dir env_cmd method opts dbfile sub cmd msg } {
 	# Sync the mpool so any changes to the file that are
 	# in mpool get written to the disk file before the
 	# diff.
-	puts "\t\tSyncing"
-	$env mpool_sync "0 0"
+	$env mpool_sync
 
 	#
 	# If we don't abort, then we expect success.
@@ -239,7 +260,7 @@ proc do_file_recover_create { dir env_cmd method opts dbfile sub cmd msg } {
 		} else {
 			error_check_good \
 			    diff(init,postcreate):diff($init_file,$dir/$dbfile)\
-			    [dbdump_diff $dflags $init_file $dir/$dbfile] 0
+                          [dbdump_diff $dflags $init_file $dir $dbfile] 0
 		}
 	} else {
 		#
@@ -290,7 +311,7 @@ proc do_file_recover_create { dir env_cmd method opts dbfile sub cmd msg } {
 		#
 		error_check_good \
 		    diff(initial,post-recover1):diff($init_file,$dir/$dbfile) \
-		    [dbdump_diff $dflags $init_file $dir/$dbfile] 0
+                  [dbdump_diff $dflags $init_file $dir $dbfile] 0
 		#
 		# Need a new copy to get the right LSN into the file.
 		#
@@ -301,7 +322,6 @@ proc do_file_recover_create { dir env_cmd method opts dbfile sub cmd msg } {
 		}
 	}
 
-	#
 	# If we didn't make a copy, then we are done.
 	#
 	if {[string first "none" $copy] != -1} {
@@ -311,11 +331,7 @@ proc do_file_recover_create { dir env_cmd method opts dbfile sub cmd msg } {
 	#
 	# Now move the .afterop file to $dbfile.  Run recovery again.
 	#
-	file copy -force $dir/$dbfile.afterop $dir/$dbfile
-
-	if { [is_queue $method] == 1 } {
-		move_file_extent $dir $dbfile afterop copy
-	}
+	copy_afterop $dir
 
 	berkdb debug_check
 	puts -nonewline "\t\tAbout to run recovery ... "
@@ -340,7 +356,7 @@ proc do_file_recover_create { dir env_cmd method opts dbfile sub cmd msg } {
 		#
 		error_check_good \
 		    diff(initial,post-recover2):diff($init_file,$dir/$dbfile) \
-		    [dbdump_diff $dflags $init_file $dir/$dbfile] 0
+                  [dbdump_diff $dflags $init_file $dir $dbfile] 0
 	}
 
 }
@@ -351,17 +367,17 @@ proc file_recover_delete { dir env_cmd method opts dbfile cmd msg op } {
 	# 1.  Deleting/Renaming just a database
 	# 2.  Deleting/Renaming a database with a subdb
 	# 3.  Deleting/Renaming a 2nd subdb in a database
-	puts "\t$msg $op with a database"
+	puts "\t$msg $op ($opts) with a database"
 	do_file_recover_delete $dir $env_cmd $method $opts $dbfile \
 	    0 $cmd $msg $op
 	if { [is_queue $method] == 1 } {
 		puts "\tSkipping subdatabase tests for method $method"
 		return
 	}
-	puts "\t$msg $op with a database and subdb"
+	puts "\t$msg $op ($opts) with a database and subdb"
 	do_file_recover_delete $dir $env_cmd $method $opts $dbfile \
 	    1 $cmd $msg $op
-	puts "\t$msg $op with a database and 2nd subdb"
+	puts "\t$msg $op ($opts) with a database and 2nd subdb"
 	do_file_recover_delete $dir $env_cmd $method $opts $dbfile \
 	    2 $cmd $msg $op
 
@@ -411,14 +427,14 @@ proc do_file_recover_delete { dir env_cmd method opts dbfile sub cmd msg op } {
 			set subdb ""
 			set new $dbfile.new
 			set dflags "-dar"
-			set oflags "-create $method -mode 0644 \
+			set oflags "-create $method -auto_commit -mode 0644 \
 			    -env \$env $opts $dbfile"
 		}
 		1 {
 			set subdb sub0
 			set new $subdb.new
 			set dflags ""
-			set oflags "-create $method -mode 0644 \
+			set oflags "-create $method -auto_commit -mode 0644 \
 			    -env \$env $opts $dbfile $subdb"
 		}
 		2 {
@@ -429,7 +445,7 @@ proc do_file_recover_delete { dir env_cmd method opts dbfile sub cmd msg op } {
 			set subdb sub1
 			set new $subdb.new
 			set dflags ""
-			set oflags "-create $method -mode 0644 \
+			set oflags "-create $method -auto_commit -mode 0644 \
 			    -env \$env $opts $dbfile sub0"
 			set db [eval {berkdb_open} $oflags]
 			error_check_good db_open [is_valid_db $db] TRUE
@@ -438,7 +454,7 @@ proc do_file_recover_delete { dir env_cmd method opts dbfile sub cmd msg op } {
 			error_check_good db_put $ret 0
 			error_check_good commit [$txn commit] 0
 			error_check_good db_close [$db close] 0
-			set oflags "-create $method -mode 0644 \
+			set oflags "-create $method -auto_commit -mode 0644 \
 			    -env \$env $opts $dbfile $subdb"
 		}
 		default {
@@ -469,6 +485,8 @@ proc do_file_recover_delete { dir env_cmd method opts dbfile sub cmd msg op } {
 	error_check_good commit [$txn commit] 0
 	error_check_good db_close [$db close] 0
 
+	$env mpool_sync
+
 	set init_file $dir/$dbfile.init
 	catch { file copy -force $dir/$dbfile $init_file } res
 
@@ -482,21 +500,23 @@ proc do_file_recover_delete { dir env_cmd method opts dbfile sub cmd msg op } {
 	#
 	switch $op {
 	"dbrename" {
-		set ret [catch { eval {berkdb} $op -env $env \
+		set ret [catch { eval {berkdb} $op -env $env -auto_commit \
 		    $dbfile $subdb $new } remret]
 	}
 	"dbremove" {
-		set ret [catch { eval {berkdb} $op -env $env \
+		set ret [catch { eval {berkdb} $op -env $env -auto_commit \
 		    $dbfile $subdb } remret]
 	}
 	"dbtruncate" {
 		set txn [$env txn]
-		set db [eval {berkdb_open_noerr -env} $env $dbfile $subdb]
+		set db [eval {berkdb_open_noerr -env} \
+		    $env -auto_commit $dbfile $subdb]
 		error_check_good dbopen [is_valid_db $db] TRUE
 		error_check_good txnbegin [is_valid_txn $txn $env] TRUE
 		set ret [catch {$db truncate -txn $txn} remret]
 	}
 	}
+	$env mpool_sync
 	if { $abort == "none" } {
 		if { $op == "dbtruncate" } {
 			error_check_good txncommit [$txn commit] 0
@@ -533,15 +553,16 @@ proc do_file_recover_delete { dir env_cmd method opts dbfile sub cmd msg op } {
 		error_check_good post$op.1 [file exists $dir/$dbfile] 1
 		error_check_good \
 		    diff(init,post$op.2):diff($init_file,$dir/$dbfile)\
-		    [dbdump_diff $dflags $init_file $dir/$dbfile] 0
+                  [dbdump_diff $dflags $init_file $dir $dbfile] 0
 	}
+	$env mpool_sync
 	error_check_good env_close [$env close] 0
 	catch { file copy -force $dir/$dbfile $init_file } res
-
 	if { [is_queue $method] == 1} {
 		copy_extent_file $dir $dbfile init
 	}
 
+
 	#
 	# Run recovery here.  Should be a no-op.  Verify that
 	# the file still doesn't exist or change (depending on abort)
@@ -556,7 +577,9 @@ proc do_file_recover_delete { dir env_cmd method opts dbfile sub cmd msg op } {
 		error "FAIL: Recovery error: $result."
 		return
 	}
+
 	puts "complete"
+
 	if { $abort == "none" } {
 		#
 		# Operate was committed.
@@ -568,9 +591,10 @@ proc do_file_recover_delete { dir env_cmd method opts dbfile sub cmd msg op } {
 		#
 		# Operation was aborted, verify it did not change.
 		#
+		berkdb debug_check
 		error_check_good \
 		    diff(initial,post-recover1):diff($init_file,$dir/$dbfile) \
-		    [dbdump_diff $dflags $init_file $dir/$dbfile] 0
+                  [dbdump_diff $dflags $init_file $dir $dbfile] 0
 	}
 
 	#
@@ -581,15 +605,10 @@ proc do_file_recover_delete { dir env_cmd method opts dbfile sub cmd msg op } {
 	}
 
 	#
-	# Now move the .afterop file to $dbfile.  Run recovery again.
+	# Now restore the .afterop file(s) to their original name.
+	# Run recovery again.
 	#
-	set filecopy [glob $dir/*.afterop]
-	set afterop [lindex $filecopy 0]
-#	set afterop [string range $afterop \
-#	    [expr [string last "/" $afterop] + 1]  \
-#	    [string last "." $afterop]]
-	catch { file rename -force $afterop $dir/$dbfile} res
-	move_file_extent $dir $dbfile afterop rename
+	copy_afterop $dir
 
 	berkdb debug_check
 	puts -nonewline "\t\tAbout to run recovery ... "
@@ -612,7 +631,7 @@ proc do_file_recover_delete { dir env_cmd method opts dbfile sub cmd msg op } {
 		#
 		error_check_good \
 		    diff(initial,post-recover2):diff($init_file,$dir/$dbfile) \
-		    [dbdump_diff $dflags $init_file $dir/$dbfile] 0
+                  [dbdump_diff $dflags $init_file $dir $dbfile] 0
 	}
 
 }
@@ -635,11 +654,13 @@ proc do_file_recover_delmk { dir env_cmd method opts dbfile } {
 	if { $log_log_record_types == 1} {
 		logtrack_read $dir
 	}
+	set omethod [convert_method $method]
 
 	puts "\tRecd007.q: Delete and recreate a database"
 	env_cleanup $dir
 	# Open the environment and set the copy/abort locations
 	set env [eval $env_cmd]
+	error_check_good env_open [is_valid_env $env] TRUE
 
 	if { [is_record_based $method] == 1 } {
 		set key 1
@@ -648,22 +669,27 @@ proc do_file_recover_delmk { dir env_cmd method opts dbfile } {
 	}
 	set data1 recd007_data
 	set data2 NEWrecd007_data2
+	set data3 LASTrecd007_data3
 
-	set oflags "-create $method -mode 0644 -env $env $opts $dbfile"
+	set oflags \
+	    "-create $omethod -auto_commit -mode 0644 $opts $dbfile"
 
 	#
 	# Open our db, add some data, close and copy as our
 	# init file.
 	#
-	set db [eval {berkdb_open} $oflags]
+	set db [eval {berkdb_open_noerr} -env $env $oflags]
 	error_check_good db_open [is_valid_db $db] TRUE
 	set txn [$env txn]
 	set ret [$db put -txn $txn $key $data1]
 	error_check_good db_put $ret 0
 	error_check_good commit [$txn commit] 0
 	error_check_good db_close [$db close] 0
+	file copy -force $testdir/$dbfile $testdir/${dbfile}.1
+
+	set ret \
+	    [catch { berkdb dbremove -env $env -auto_commit $dbfile } remret]
 
-	set ret [catch { berkdb dbremove -env $env $dbfile } remret]
 	#
 	# Operation was committed, verify it does
 	# not exist.
@@ -675,10 +701,10 @@ proc do_file_recover_delmk { dir env_cmd method opts dbfile } {
 	#
 	# Now create a new db with the same name.
 	#
-	set db [eval {berkdb_open} $oflags]
+	set db [eval {berkdb_open_noerr} -env $env $oflags]
 	error_check_good db_open [is_valid_db $db] TRUE
 	set txn [$env txn]
-	set ret [$db put -txn $txn $key $data1]
+	set ret [$db put -txn $txn $key [chop_data $method $data2]]
 	error_check_good db_put $ret 0
 	error_check_good commit [$txn commit] 0
 	error_check_good db_sync [$db sync] 0
@@ -694,16 +720,154 @@ proc do_file_recover_delmk { dir env_cmd method opts dbfile } {
 	}
 	puts "complete"
 	error_check_good db_recover $stat 0
-	error_check_good db_recover.1 [file exists $dir/$dbfile] 1
+	error_check_good file_exist [file exists $dir/$dbfile] 1
 	#
 	# Since we ran recovery on the open db/env, we need to
 	# catch these calls.  Basically they are there to clean
 	# up the Tcl widgets.
 	#
 	set stat [catch {$db close} ret]
+	error_check_bad dbclose_after_remove $stat 0
+	error_check_good dbclose_after_remove [is_substr $ret recovery] 1
 	set stat [catch {$env close} ret]
+	error_check_bad envclose_after_remove $stat 0
+	error_check_good envclose_after_remove [is_substr $ret recovery] 1
 
+	#
+	# Reopen env and db and verify 2nd database is there.
+	#
+	set env [eval $env_cmd]
+	error_check_good env_open [is_valid_env $env] TRUE
+	set db [eval {berkdb_open} -env $env $oflags]
+	error_check_good db_open [is_valid_db $db] TRUE
+	set ret [$db get $key]
+	error_check_good dbget [llength $ret] 1
+	set kd [lindex $ret 0]
+	error_check_good key [lindex $kd 0] $key
+	error_check_good data2 [lindex $kd 1] [pad_data $method $data2]
+
+	error_check_good dbclose [$db close] 0
+	error_check_good envclose [$env close] 0
+
+	#
+	# Copy back the original database and run recovery again.
+	# SR [#13026]
+	#
+	puts "\t\tRecover from first database"
+	file copy -force $testdir/${dbfile}.1 $testdir/$dbfile
+	berkdb debug_check
+	puts -nonewline "\t\tAbout to run recovery ... "
+	flush stdout
+
+	set stat [catch {exec $util_path/db_recover -h $dir -c} result]
+	if { $stat == 1 } {
+		error "FAIL: Recovery error: $result."
+		return
+	}
+	puts "complete"
+	error_check_good db_recover $stat 0
+	error_check_good db_recover.1 [file exists $dir/$dbfile] 1
+
+	#
+	# Reopen env and db and verify 2nd database is there.
+	#
+	set env [eval $env_cmd]
+	error_check_good env_open [is_valid_env $env] TRUE
+	set db [eval {berkdb_open_noerr} -env $env $oflags]
+	error_check_good db_open [is_valid_db $db] TRUE
+	set ret [$db get $key]
+	error_check_good dbget [llength $ret] 1
+	set kd [lindex $ret 0]
+	error_check_good key [lindex $kd 0] $key
+	error_check_good data2 [lindex $kd 1] [pad_data $method $data2]
+
+	error_check_good dbclose [$db close] 0
+
+	file copy -force $testdir/$dbfile $testdir/${dbfile}.2
+
+	puts "\t\tRemove second db"
+	set ret \
+	    [catch { berkdb dbremove -env $env -auto_commit $dbfile } remret]
+
+	#
+	# Operation was committed, verify it does
+	# not exist.
+	#
+	puts "\t\tCommand executed and committed."
+	error_check_good dbremove $ret 0
+	error_check_good dbremove.2 [file exists $dir/$dbfile] 0
+
+	#
+	# Now create a new db with the same name.
+	#
+	puts "\t\tAdd a third version of the database"
+	set db [eval {berkdb_open_noerr} -env $env $oflags]
+	error_check_good db_open [is_valid_db $db] TRUE
+	set txn [$env txn]
+	set ret [$db put -txn $txn $key [chop_data $method $data3]]
+	error_check_good db_put $ret 0
+	error_check_good commit [$txn commit] 0
+	error_check_good db_sync [$db sync] 0
+
+	berkdb debug_check
+	puts -nonewline "\t\tAbout to run recovery ... "
+	flush stdout
+
+	set stat [catch {exec $util_path/db_recover -h $dir -c} result]
+	if { $stat == 1 } {
+		error "FAIL: Recovery error: $result."
+		return
+	}
+	puts "complete"
+	error_check_good db_recover $stat 0
+	error_check_good file_exist [file exists $dir/$dbfile] 1
+
+	#
+	# Since we ran recovery on the open db/env, we need to
+	# catch these calls to clean up the Tcl widgets.
+	#
+	set stat [catch {$db close} ret]
+	error_check_bad dbclose_after_remove $stat 0
+	error_check_good dbclose_after_remove [is_substr $ret recovery] 1
+	set stat [catch {$env close} ret]
+	error_check_bad envclose_after_remove $stat 0
+	error_check_good envclose_after_remove [is_substr $ret recovery] 1
+
+	#
+	# Copy back the second database and run recovery again.
+	#
+	puts "\t\tRecover from second database"
+	file copy -force $testdir/${dbfile}.2 $testdir/$dbfile
+	berkdb debug_check
+	puts -nonewline "\t\tAbout to run recovery ... "
+	flush stdout
+
+	set stat [catch {exec $util_path/db_recover -h $dir -c} result]
+	if { $stat == 1 } {
+		error "FAIL: Recovery error: $result."
+		return
+	}
+	puts "complete"
+	error_check_good db_recover $stat 0
+	error_check_good file_exist.2 [file exists $dir/$dbfile] 1
+
+	#
+	# Reopen env and db and verify 3rd database is there.
+	#
+	set env [eval $env_cmd]
+	error_check_good env_open [is_valid_env $env] TRUE
+	set db [eval {berkdb_open} -env $env $oflags]
+	error_check_good db_open [is_valid_db $db] TRUE
+	set ret [$db get $key]
+	error_check_good dbget [llength $ret] 1
+	set kd [lindex $ret 0]
+	error_check_good key [lindex $kd 0] $key
+	error_check_good data2 [lindex $kd 1] [pad_data $method $data3]
+
+	error_check_good dbclose [$db close] 0
+	error_check_good envclose [$env close] 0
 }
+
 proc is_valid_create_loc { loc } {
 	switch $loc {
 		none		-
@@ -735,7 +899,7 @@ proc is_valid_delete_loc { loc } {
 # just a free/invalid page.
 # Return 1 if they are different, 0 if logically the same (or identical).
 #
-proc dbdump_diff { flags initfile dbfile } {
+proc dbdump_diff { flags initfile dir dbfile } {
 	source ./include.tcl
 
 	set initdump $initfile.dump
@@ -747,11 +911,11 @@ proc dbdump_diff { flags initfile dbfile } {
 
 	# Do a dump without the freelist which should eliminate any
 	# recovery differences.
-	set stat [catch {eval {exec $util_path/db_dump} $flags -f $dbdump \
-	    $dbfile} ret]
+	set stat [catch {eval {exec $util_path/db_dump} $flags -f $dir/$dbdump \
+          $dir/$dbfile} ret]
 	error_check_good dbdump.db $stat 0
 
-	set stat [filecmp $dbdump $initdump]
+	set stat [filecmp $dir/$dbdump $initdump]
 
 	if {$stat == 0} {
 		return 0
@@ -768,7 +932,7 @@ proc recd007_check { op sub dir dbfile subdb new env oflags } {
 	#
 	if { $sub == 0 } {
 		if { $op == "dbremove" } {
-			error_check_good $op:not-exist \
+			error_check_good $op:not-exist:$dir/$dbfile \
 			    [file exists $dir/$dbfile] 0
 		} elseif { $op == "dbrename"} {
 			error_check_good $op:exist \
@@ -850,3 +1014,14 @@ proc recd007_check { op sub dir dbfile subdb new env oflags } {
 	}
 }
 
+proc copy_afterop { dir } {
+	set r [catch { set filecopy [glob $dir/*.afterop] } res]
+	if { $r == 1 } {
+		return
+	}
+	foreach f $filecopy {
+		set orig [string range $f 0 \
+			[expr [string last "." $f] - 1]]
+		catch { file rename -force $f $orig} res
+	}
+}
