@@ -1,9 +1,9 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 2001-2003
+# Copyright (c) 2001-2005
 #	Sleepycat Software.  All rights reserved.
 #
-# $Id: si002.tcl,v 1.9 2003/01/08 05:53:25 bostic Exp $
+# $Id: si002.tcl,v 12.5 2005/10/26 18:44:03 carol Exp $
 #
 # TEST	si002
 # TEST	Basic cursor-based secondary index put/delete test
@@ -24,30 +24,57 @@ proc si002 { methods {nentries 200} {tnum "002"} args } {
 	set pargs [convert_args $pmethod $args]
 	set pomethod [convert_method $pmethod]
 
+	# Renumbering recno databases can't be used as primaries.
+	if { [is_rrecno $pmethod] == 1 } {
+		puts "Skipping si$tnum for method $pmethod"
+		return 
+	}
+
 	# Method/args for all the secondaries.  If only one method
-	# was specified, assume the same method and a standard N
-	# secondaries.
+	# was specified, assume the same method (for btree or hash) 
+	# and a standard number of secondaries.  If primary is not
+	# btree or hash, force secondaries to be one btree, one hash.
 	set methods [lrange $methods 1 end]
 	if { [llength $methods] == 0 } {
 		for { set i 0 } { $i < $nsecondaries } { incr i } {
-			lappend methods $pmethod
+			if { [is_btree $pmethod] || [is_hash $pmethod] } {
+				lappend methods $pmethod
+			} else {
+				if { [expr $i % 2] == 0 } {
+					lappend methods "-btree"
+				} else {
+					lappend methods "-hash"
+				}
+			}
 		}
 	}
 
 	set argses [convert_argses $methods $args]
 	set omethods [convert_methods $methods]
 
-	puts "Si$tnum ($pmethod/$methods) $nentries equal key/data pairs"
-	env_cleanup $testdir
+	# If we are given an env, use it.  Otherwise, open one.
+	set eindex [lsearch -exact $args "-env"]
+	if { $eindex == -1 } {
+		env_cleanup $testdir
+		set env [berkdb_env -create -home $testdir]
+		error_check_good env_open [is_valid_env $env] TRUE
+	} else {
+		incr eindex
+		set env [lindex $args $eindex]
+		set envflags [$env get_open_flags]
+		if { [lsearch -exact $envflags "-thread"] != -1 } {
+			puts "Skipping si$tnum for threaded env"
+			return 
+		}
+		set testdir [get_home $env]
+	}
+
+	puts "si$tnum \{\[ list $pmethod $methods \]\} $nentries"
+	cleanup $testdir $env
 
 	set pname "primary$tnum.db"
 	set snamebase "secondary$tnum"
 
-	# Open an environment
-	# XXX if one is not supplied!
-	set env [berkdb_env -create -home $testdir]
-	error_check_good env_open [is_valid_env $env] TRUE
-
 	# Open the primary.
 	set pdb [eval {berkdb_open -create -env} $env $pomethod $pargs $pname]
 	error_check_good primary_open [is_valid_db $pdb] TRUE
@@ -171,5 +198,9 @@ proc si002 { methods {nentries 200} {tnum "002"} args } {
 		error_check_good secondary_close [$sdb close] 0
 	}
 	error_check_good primary_close [$pdb close] 0
-	error_check_good env_close [$env close] 0
+
+	# Close the env if it was created within this test.
+	if { $eindex == -1 } {
+		error_check_good env_close [$env close] 0
+	}
 }
