@@ -1,16 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2003
+ * Copyright (c) 1996-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: txn_region.c,v 12.10 2005/10/14 21:12:18 ubell Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: txn_region.c,v 11.79 2003/07/23 13:13:12 mjc Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -32,7 +30,6 @@ static const char revid[] = "$Id: txn_region.c,v 11.79 2003/07/23 13:13:12 mjc E
 #include "dbinc/log.h"
 #include "dbinc/txn.h"
 
-static int __txn_findlastckp __P((DB_ENV *, DB_LSN *));
 static int __txn_init __P((DB_ENV *, DB_TXNMGR *));
 static size_t __txn_region_size __P((DB_ENV *));
 
@@ -46,56 +43,49 @@ int
 __txn_open(dbenv)
 	DB_ENV *dbenv;
 {
-	DB_TXNMGR *tmgrp;
+	DB_TXNMGR *mgr;
 	int ret;
 
 	/* Create/initialize the transaction manager structure. */
-	if ((ret = __os_calloc(dbenv, 1, sizeof(DB_TXNMGR), &tmgrp)) != 0)
+	if ((ret = __os_calloc(dbenv, 1, sizeof(DB_TXNMGR), &mgr)) != 0)
 		return (ret);
-	TAILQ_INIT(&tmgrp->txn_chain);
-	tmgrp->dbenv = dbenv;
+	TAILQ_INIT(&mgr->txn_chain);
+	mgr->dbenv = dbenv;
 
 	/* Join/create the txn region. */
-	tmgrp->reginfo.type = REGION_TYPE_TXN;
-	tmgrp->reginfo.id = INVALID_REGION_ID;
-	tmgrp->reginfo.mode = dbenv->db_mode;
-	tmgrp->reginfo.flags = REGION_JOIN_OK;
+	mgr->reginfo.dbenv = dbenv;
+	mgr->reginfo.type = REGION_TYPE_TXN;
+	mgr->reginfo.id = INVALID_REGION_ID;
+	mgr->reginfo.flags = REGION_JOIN_OK;
 	if (F_ISSET(dbenv, DB_ENV_CREATE))
-		F_SET(&tmgrp->reginfo, REGION_CREATE_OK);
+		F_SET(&mgr->reginfo, REGION_CREATE_OK);
 	if ((ret = __db_r_attach(dbenv,
-	    &tmgrp->reginfo, __txn_region_size(dbenv))) != 0)
+	    &mgr->reginfo, __txn_region_size(dbenv))) != 0)
 		goto err;
 
 	/* If we created the region, initialize it. */
-	if (F_ISSET(&tmgrp->reginfo, REGION_CREATE))
-		if ((ret = __txn_init(dbenv, tmgrp)) != 0)
+	if (F_ISSET(&mgr->reginfo, REGION_CREATE))
+		if ((ret = __txn_init(dbenv, mgr)) != 0)
 			goto err;
 
 	/* Set the local addresses. */
-	tmgrp->reginfo.primary =
-	    R_ADDR(&tmgrp->reginfo, tmgrp->reginfo.rp->primary);
+	mgr->reginfo.primary =
+	    R_ADDR(&mgr->reginfo, mgr->reginfo.rp->primary);
 
-	/* Acquire a mutex to protect the active TXN list. */
-	if (F_ISSET(dbenv, DB_ENV_THREAD) &&
-	    (ret = __db_mutex_setup(dbenv, &tmgrp->reginfo, &tmgrp->mutexp,
-	    MUTEX_ALLOC | MUTEX_NO_RLOCK | MUTEX_THREAD)) != 0)
+	/* If threaded, acquire a mutex to protect the active TXN list. */
+	if ((ret = __mutex_alloc(
+	    dbenv, MTX_TXN_ACTIVE, DB_MUTEX_THREAD, &mgr->mutex)) != 0)
 		goto err;
 
-	R_UNLOCK(dbenv, &tmgrp->reginfo);
-
-	dbenv->tx_handle = tmgrp;
+	dbenv->tx_handle = mgr;
 	return (0);
 
-err:	if (tmgrp->reginfo.addr != NULL) {
-		if (F_ISSET(&tmgrp->reginfo, REGION_CREATE))
-			ret = __db_panic(dbenv, ret);
-		R_UNLOCK(dbenv, &tmgrp->reginfo);
+err:	dbenv->tx_handle = NULL;
+	if (mgr->reginfo.addr != NULL)
+		(void)__db_r_detach(dbenv, &mgr->reginfo, 0);
 
-		(void)__db_r_detach(dbenv, &tmgrp->reginfo, 0);
-	}
-	if (tmgrp->mutexp != NULL)
-		__db_mutex_free(dbenv, &tmgrp->reginfo, tmgrp->mutexp);
-	__os_free(dbenv, tmgrp);
+	(void)__mutex_free(dbenv, &mgr->mutex);
+	__os_free(dbenv, mgr);
 	return (ret);
 }
 
@@ -104,16 +94,13 @@ err:	if (tmgrp->reginfo.addr != NULL) {
  *	Initialize a transaction region in shared memory.
  */
 static int
-__txn_init(dbenv, tmgrp)
+__txn_init(dbenv, mgr)
 	DB_ENV *dbenv;
-	DB_TXNMGR *tmgrp;
+	DB_TXNMGR *mgr;
 {
 	DB_LSN last_ckp;
 	DB_TXNREGION *region;
 	int ret;
-#ifdef	HAVE_MUTEX_SYSTEM_RESOURCES
-	u_int8_t *addr;
-#endif
 
 	/*
 	 * Find the last checkpoint in the log.
@@ -124,31 +111,40 @@ __txn_init(dbenv, tmgrp)
 		 * The log system has already walked through the last
 		 * file.  Get the LSN of a checkpoint it may have found.
 		 */
-		__log_get_cached_ckp_lsn(dbenv, &last_ckp);
+		if ((ret = __log_get_cached_ckp_lsn(dbenv, &last_ckp)) != 0)
+			return (ret);
 
 		/*
 		 * If that didn't work, look backwards from the beginning of
 		 * the last log file until we find the last checkpoint.
 		 */
 		if (IS_ZERO_LSN(last_ckp) &&
-		    (ret = __txn_findlastckp(dbenv, &last_ckp)) != 0)
+		    (ret = __txn_findlastckp(dbenv, &last_ckp, NULL)) != 0)
 			return (ret);
 	}
 
-	if ((ret = __db_shalloc(tmgrp->reginfo.addr,
-	    sizeof(DB_TXNREGION), 0, &tmgrp->reginfo.primary)) != 0) {
+	if ((ret = __db_shalloc(&mgr->reginfo,
+	    sizeof(DB_TXNREGION), 0, &mgr->reginfo.primary)) != 0) {
 		__db_err(dbenv,
 		    "Unable to allocate memory for the transaction region");
 		return (ret);
 	}
-	tmgrp->reginfo.rp->primary =
-	    R_OFFSET(&tmgrp->reginfo, tmgrp->reginfo.primary);
-	region = tmgrp->reginfo.primary;
+	mgr->reginfo.rp->primary =
+	    R_OFFSET(&mgr->reginfo, mgr->reginfo.primary);
+	region = mgr->reginfo.primary;
 	memset(region, 0, sizeof(*region));
 
+	if ((ret = __mutex_alloc(
+	    dbenv, MTX_TXN_REGION, 0, &region->mtx_region)) != 0)
+		return (ret);
+
 	region->maxtxns = dbenv->tx_max;
 	region->last_txnid = TXN_MINIMUM;
 	region->cur_maxid = TXN_MAXIMUM;
+
+	if ((ret = __mutex_alloc(
+	    dbenv, MTX_TXN_CHKPT, 0, &region->mtx_ckp)) != 0)
+		return (ret);
 	region->last_ckp = last_ckp;
 	region->time_ckp = time(NULL);
 
@@ -156,30 +152,22 @@ __txn_init(dbenv, tmgrp)
 	region->stat.st_maxtxns = region->maxtxns;
 
 	SH_TAILQ_INIT(&region->active_txn);
-#ifdef	HAVE_MUTEX_SYSTEM_RESOURCES
-	/* Allocate room for the txn maintenance info and initialize it. */
-	if ((ret = __db_shalloc(tmgrp->reginfo.addr,
-	    sizeof(REGMAINT) + TXN_MAINT_SIZE, 0, &addr)) != 0) {
-		__db_err(dbenv,
-		    "Unable to allocate memory for mutex maintenance");
-		return (ret);
-	}
-	__db_maintinit(&tmgrp->reginfo, addr, TXN_MAINT_SIZE);
-	region->maint_off = R_OFFSET(&tmgrp->reginfo, addr);
-#endif
-	return (0);
+	return (ret);
 }
 
 /*
  * __txn_findlastckp --
  *	Find the last checkpoint in the log, walking backwards from the
- *	beginning of the last log file.  (The log system looked through
- *	the last log file when it started up.)
+ *	max_lsn given or the beginning of the last log file.  (The
+ *	log system looked through the last log file when it started up.)
+ *
+ * PUBLIC: int __txn_findlastckp __P((DB_ENV *, DB_LSN *, DB_LSN *));
  */
-static int
-__txn_findlastckp(dbenv, lsnp)
+int
+__txn_findlastckp(dbenv, lsnp, max_lsn)
 	DB_ENV *dbenv;
 	DB_LSN *lsnp;
+	DB_LSN *max_lsn;
 {
 	DB_LOGC *logc;
 	DB_LSN lsn;
@@ -187,20 +175,27 @@ __txn_findlastckp(dbenv, lsnp)
 	int ret, t_ret;
 	u_int32_t rectype;
 
+	ZERO_LSN(*lsnp);
+
 	if ((ret = __log_cursor(dbenv, &logc)) != 0)
 		return (ret);
 
 	/* Get the last LSN. */
 	memset(&dbt, 0, sizeof(dbt));
-	if ((ret = __log_c_get(logc, &lsn, &dbt, DB_LAST)) != 0)
-		goto err;
-
-	/*
-	 * Twiddle the last LSN so it points to the beginning of the last
-	 * file;  we know there's no checkpoint after that, since the log
-	 * system already looked there.
-	 */
-	lsn.offset = 0;
+	if (max_lsn != NULL) {
+		lsn = *max_lsn;
+		if ((ret = __log_c_get(logc, &lsn, &dbt, DB_SET)) != 0)
+			goto err;
+	} else {
+		if ((ret = __log_c_get(logc, &lsn, &dbt, DB_LAST)) != 0)
+			goto err;
+		/*
+		 * Twiddle the last LSN so it points to the beginning of the
+		 * last file; we know there's no checkpoint after that, since
+		 * the log system already looked there.
+		 */
+		lsn.offset = 0;
+	}
 
 	/* Read backwards, looking for checkpoints. */
 	while ((ret = __log_c_get(logc, &lsn, &dbt, DB_PREV)) == 0) {
@@ -215,6 +210,7 @@ __txn_findlastckp(dbenv, lsnp)
 
 err:	if ((t_ret = __log_c_close(logc)) != 0 && ret == 0)
 		ret = t_ret;
+
 	/*
 	 * Not finding a checkpoint is not an error;  there may not exist
 	 * one in the log.
@@ -225,7 +221,6 @@ err:	if ((t_ret = __log_c_close(logc)) != 0 && ret == 0)
 /*
  * __txn_dbenv_refresh --
  *	Clean up after the transaction system on a close or failed open.
- * Called only from __dbenv_refresh.  (Formerly called __txn_close.)
  *
  * PUBLIC: int __txn_dbenv_refresh __P((DB_ENV *));
  */
@@ -233,14 +228,15 @@ int
 __txn_dbenv_refresh(dbenv)
 	DB_ENV *dbenv;
 {
-	DB_TXN *txnp;
-	DB_TXNMGR *tmgrp;
-	TXN_DETAIL *td;
+	DB_TXN *txn;
+	DB_TXNMGR *mgr;
+	REGINFO *reginfo;
 	u_int32_t txnid;
 	int aborted, ret, t_ret;
 
 	ret = 0;
-	tmgrp = dbenv->tx_handle;
+	mgr = dbenv->tx_handle;
+	reginfo = &mgr->reginfo;
 
 	/*
 	 * This function can only be called once per process (i.e., not
@@ -254,13 +250,12 @@ __txn_dbenv_refresh(dbenv)
 	 * to a known state.
 	 */
 	aborted = 0;
-	if (TAILQ_FIRST(&tmgrp->txn_chain) != NULL) {
-		while ((txnp = TAILQ_FIRST(&tmgrp->txn_chain)) != NULL) {
+	if (TAILQ_FIRST(&mgr->txn_chain) != NULL) {
+		while ((txn = TAILQ_FIRST(&mgr->txn_chain)) != NULL) {
 			/* Prepared transactions are OK. */
-			td = (TXN_DETAIL *)R_ADDR(&tmgrp->reginfo, txnp->off);
-			txnid = txnp->txnid;
-			if (td->status == TXN_PREPARED) {
-				if ((ret = __txn_discard(txnp, 0)) != 0) {
+			txnid = txn->txnid;
+			if (((TXN_DETAIL *)txn->td)->status == TXN_PREPARED) {
+				if ((ret = __txn_discard_int(txn, 0)) != 0) {
 					__db_err(dbenv,
 					    "Unable to discard txn 0x%x: %s",
 					    txnid, db_strerror(ret));
@@ -269,7 +264,7 @@ __txn_dbenv_refresh(dbenv)
 				continue;
 			}
 			aborted = 1;
-			if ((t_ret = __txn_abort(txnp)) != 0) {
+			if ((t_ret = __txn_abort(txn)) != 0) {
 				__db_err(dbenv,
 				    "Unable to abort transaction 0x%x: %s",
 				    txnid, db_strerror(t_ret));
@@ -291,14 +286,14 @@ __txn_dbenv_refresh(dbenv)
 		ret = t_ret;
 
 	/* Discard the per-thread lock. */
-	if (tmgrp->mutexp != NULL)
-		__db_mutex_free(dbenv, &tmgrp->reginfo, tmgrp->mutexp);
-
-	/* Detach from the region. */
-	if ((t_ret = __db_r_detach(dbenv, &tmgrp->reginfo, 0)) != 0 && ret == 0)
+	if ((t_ret = __mutex_free(dbenv, &mgr->mutex)) != 0 && ret == 0)
 		ret = t_ret;
 
-	__os_free(dbenv, tmgrp);
+	/* Detach from the region. */
+	if ((t_ret = __db_r_detach(dbenv, reginfo, 0)) != 0 && ret == 0)
+		ret = t_ret;
+
+	__os_free(dbenv, mgr);
 
 	dbenv->tx_handle = NULL;
 	return (ret);
@@ -320,31 +315,9 @@ __txn_region_size(dbenv)
 
 	s = sizeof(DB_TXNREGION) +
 	    dbenv->tx_max * sizeof(TXN_DETAIL) + 10 * 1024;
-#ifdef HAVE_MUTEX_SYSTEM_RESOURCES
-	if (F_ISSET(dbenv, DB_ENV_THREAD))
-		s += sizeof(REGMAINT) + TXN_MAINT_SIZE;
-#endif
 	return (s);
 }
 
-/*
- * __txn_region_destroy
- *	Destroy any region maintenance info.
- *
- * PUBLIC: void __txn_region_destroy __P((DB_ENV *, REGINFO *));
- */
-void
-__txn_region_destroy(dbenv, infop)
-	DB_ENV *dbenv;
-	REGINFO *infop;
-{
-	__db_shlocks_destroy(infop, (REGMAINT *)R_ADDR(infop,
-	    ((DB_TXNREGION *)R_ADDR(infop, infop->rp->primary))->maint_off));
-
-	COMPQUIET(dbenv, NULL);
-	COMPQUIET(infop, NULL);
-}
-
 /*
  * __txn_id_set --
  *	Set the current transaction ID and current maximum unused ID (for
