@@ -1,16 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1998-2003
+ * Copyright (c) 1998-2005
  *	Sleepycat Software.  All rights reserved.
+ *
+ * $Id: xa.c,v 12.5 2005/11/01 00:44:39 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: xa.c,v 11.29 2003/04/24 16:55:09 bostic Exp $";
-#endif /* not lint */
-
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
@@ -21,17 +19,17 @@ static const char revid[] = "$Id: xa.c,v 11.29 2003/04/24 16:55:09 bostic Exp $"
 #include "db_int.h"
 #include "dbinc/txn.h"
 
-static int  __db_xa_close __P((char *, int, long));
-static int  __db_xa_commit __P((XID *, int, long));
-static int  __db_xa_complete __P((int *, int *, int, long));
-static int  __db_xa_end __P((XID *, int, long));
-static int  __db_xa_forget __P((XID *, int, long));
-static int  __db_xa_open __P((char *, int, long));
-static int  __db_xa_prepare __P((XID *, int, long));
-static int  __db_xa_recover __P((XID *, long, int, long));
-static int  __db_xa_rollback __P((XID *, int, long));
-static int  __db_xa_start __P((XID *, int, long));
-static void __xa_put_txn __P((DB_ENV *, DB_TXN *));
+static int __db_xa_close __P((char *, int, long));
+static int __db_xa_commit __P((XID *, int, long));
+static int __db_xa_complete __P((int *, int *, int, long));
+static int __db_xa_end __P((XID *, int, long));
+static int __db_xa_forget __P((XID *, int, long));
+static int __db_xa_open __P((char *, int, long));
+static int __db_xa_prepare __P((XID *, int, long));
+static int __db_xa_recover __P((XID *, long, int, long));
+static int __db_xa_rollback __P((XID *, int, long));
+static int __db_xa_start __P((XID *, int, long));
+static int __xa_put_txn __P((DB_ENV *, DB_TXN *));
 
 /*
  * Possible flag values:
@@ -60,23 +58,12 @@ const struct xa_switch_t db_xa_switch = {
 };
 
 /*
- * If you want your XA server to be multi-threaded, then you must
+ * If you want your XA server to be multi-threaded, then you must (at least)
  * edit this file and change:
  *	#undef  XA_MULTI_THREAD
  * to:
  *	#define XA_MULTI_THREAD 1
- *
- * You must then modify the FILL ME IN; section below to assign a
- * 32-bit unsigned, unique thread ID to the variable tid.  If no
- * such thread ID is available, e.g., you're using pthreads on a POSIX
- * 1003.1 system where a pthread_t is declared as a structure, you
- * will probably want to change the tid field of the DB_TXN structure
- * to be the correct type in which to define a thread ID on the system,
- * and then modify both the FILL ME IN section and then subsequent
- * comparison of the thread IDs in the XA transaction list, as the
- * current simple equality tests may not work.
  */
-
 #undef XA_MULTI_THREAD
 
 /*
@@ -88,98 +75,108 @@ const struct xa_switch_t db_xa_switch = {
  * PUBLIC: int __xa_get_txn __P((DB_ENV *, DB_TXN **, int));
  */
 int
-__xa_get_txn(env, txnp, do_init)
-	DB_ENV *env;
+__xa_get_txn(dbenv, txnp, do_init)
+	DB_ENV *dbenv;
 	DB_TXN **txnp;
 	int do_init;
 {
-	int ret;
 #ifdef XA_MULTI_THREAD
 	DB_TXN *t;
-	u_int32_t tid;
 	DB_TXNMGR *mgr;
-#else
-	COMPQUIET(do_init, 0);
+	TXN_DETAIL *td;
+	db_threadid_t tid;
+	pid_t pid;
 #endif
+	int ret;
+
 	ret = 0;
 
 #ifdef XA_MULTI_THREAD
-	/* Specify Thread-ID retrieval here. */
-	tid  = FILL ME IN
+	dbenv->thread_id(dbenv, &pid, &tid);
 	*txnp = NULL;
-	mgr = (DB_TXNMGR *)env->tx_handle;
+
+	DB_ASSERT(dbenv->tx_handle != NULL);
+	mgr = (DB_TXNMGR *)dbenv->tx_handle;
 
 	/*
-	 * We need to protect the xa_txn linked list, but the
-	 * environment does not have a mutex.  Since we are in
-	 * an XA transaction environment, we know that there is
-	 * a transaction structure, so use its mutex.
+	 * We need to protect the xa_txn linked list, but the environment does
+	 * not have a mutex.  Since we are in an XA transaction environment,
+	 * we know there is a transaction structure, we can use its mutex.
 	 */
-	DB_ASSERT(env->tx_handle != NULL);
-	MUTEX_THREAD_LOCK(mgr->mutexp);
-	for (t = TAILQ_FIRST(&env->xa_txn);
+	MUTEX_LOCK(dbenv, mgr->mutex);
+	for (t = TAILQ_FIRST(&dbenv->xa_txn);
 	    t != NULL;
-	    t = TAILQ_NEXT(t, xalinks))
-		/*
-		 * FILL ME IN; if tids are not a 32-bit integral type;
-		 * put a comparison here that will work.
-		 */
+	    t = TAILQ_NEXT(t, xalinks)) {
+		td = t->td;
+		if (td->pid != pid)
+			continue;
+#ifdef HAVE_INTEGRAL_THREAD_TYPE
 		if (t->tid == tid) {
 			*txnp = t;
 			break;
 		}
-	MUTEX_THREAD_UNLOCK(mgr-&gt;mutexp);
+#else
+		if (memcmp(&t->tid, &tid, sizeof(tid)) == 0) {
+			*txnp = t;
+			break;
+		}
+#endif
+	}
+	MUTEX_UNLOCK(dbenv, mgr->mutex);
 
 	if (*txnp == NULL) {
 		if (!do_init)
 			ret = EINVAL;
 		else if ((ret =
-		    __os_malloc(env, sizeof(DB_TXN), NULL, txnp)) == 0) {
+		    __os_malloc(dbenv, sizeof(DB_TXN), txnp)) == 0) {
 			(*txnp)->tid = tid;
-			MUTEX_THREAD_LOCK(mgr->mutexp);
-			TAILQ_INSERT_HEAD(&env->xa_txn, *txnp, xalinks);
-			MUTEX_THREAD_UNLOCK(mgr->mutexp);
+			MUTEX_LOCK(dbenv, mgr->mutex);
+			TAILQ_INSERT_HEAD(&dbenv->xa_txn, *txnp, xalinks);
+			MUTEX_UNLOCK(dbenv, mgr->mutex);
 		}
 	}
 #else
-	*txnp = TAILQ_FIRST(&env->xa_txn);
+	COMPQUIET(do_init, 0);
+
+	*txnp = TAILQ_FIRST(&dbenv->xa_txn);
 	if (*txnp == NULL &&
-	    (ret = __os_calloc(env, 1, sizeof(DB_TXN), txnp)) == 0) {
+	    (ret = __os_calloc(dbenv, 1, sizeof(DB_TXN), txnp)) == 0) {
 		(*txnp)->txnid = TXN_INVALID;
-		TAILQ_INSERT_HEAD(&env->xa_txn, *txnp, xalinks);
+		TAILQ_INSERT_HEAD(&dbenv->xa_txn, *txnp, xalinks);
 	}
 #endif
 
 	return (ret);
 }
 
-static void
-__xa_put_txn(env, txnp)
-	DB_ENV *env;
+static int
+__xa_put_txn(dbenv, txnp)
+	DB_ENV *dbenv;
 	DB_TXN *txnp;
 {
 #ifdef XA_MULTI_THREAD
 	DB_TXNMGR *mgr;
-	mgr = (DB_TXNMGR *)env->tx_handle;
+	mgr = (DB_TXNMGR *)dbenv->tx_handle;
 
-	MUTEX_THREAD_LOCK(mgr->mutexp);
-	TAILQ_REMOVE(&env->xa_txn, txnp, xalinks);
-	MUTEX_THREAD_UNLOCK(mgr->mutexp);
-	__os_free(env, txnp);
+	MUTEX_LOCK(dbenv, mgr->mutex);
+	TAILQ_REMOVE(&dbenv->xa_txn, txnp, xalinks);
+	MUTEX_UNLOCK(dbenv, mgr->mutex);
+	__os_free(dbenv, txnp);
 #else
-	COMPQUIET(env, NULL);
+	COMPQUIET(dbenv, NULL);
 	txnp->txnid = TXN_INVALID;
 #endif
+	return (0);
 }
 
 #ifdef XA_MULTI_THREAD
 #define	XA_FLAGS \
-	DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_MPOOL | \
-	DB_INIT_TXN | DB_THREAD
+	(DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_MPOOL | \
+	DB_INIT_TXN | DB_THREAD)
 #else
 #define	XA_FLAGS \
-	DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_MPOOL | \
-	DB_INIT_TXN
+	(DB_CREATE | DB_INIT_LOCK | DB_INIT_LOG | DB_INIT_MPOOL | \
+	DB_INIT_TXN)
 #endif
 
 /*
@@ -197,12 +194,15 @@ __xa_put_txn(env, txnp)
  * call.
  */
 static int
-__db_xa_open(xa_info, rmid, flags)
+__db_xa_open(xa_info, rmid, arg_flags)
 	char *xa_info;
 	int rmid;
-	long flags;
+	long arg_flags;
 {
-	DB_ENV *env;
+	DB_ENV *dbenv;
+	u_long flags;
+
+	flags = (u_long)arg_flags;	/* Conversion for bit operations. */
 
 	if (LF_ISSET(TMASYNC))
 		return (XAER_ASYNC);
@@ -210,27 +210,25 @@ __db_xa_open(xa_info, rmid, flags)
 		return (XAER_INVAL);
 
 	/* Verify if we already have this environment open. */
-	if (__db_rmid_to_env(rmid, &env) == 0)
+	if (__db_rmid_to_env(rmid, &dbenv) == 0)
 		return (XA_OK);
-	if (__os_calloc(env, 1, sizeof(DB_ENV), &env) != 0)
-		return (XAER_RMERR);
 
 	/* Open a new environment. */
-	if (db_env_create(&env, 0) != 0)
+	if (db_env_create(&dbenv, 0) != 0)
 		return (XAER_RMERR);
-	if (env->open(env, xa_info, XA_FLAGS, 0) != 0)
+	if (dbenv->open(dbenv, xa_info, XA_FLAGS, 0) != 0)
 		goto err;
 
 	/* Create the mapping. */
-	if (__db_map_rmid(rmid, env) != 0)
+	if (__db_map_rmid(rmid, dbenv) != 0)
 		goto err;
 
 	/* Allocate space for the current transaction. */
-	TAILQ_INIT(&env->xa_txn);
+	TAILQ_INIT(&dbenv->xa_txn);
 
 	return (XA_OK);
 
-err:	(void)env->close(env, 0);
+err:	(void)dbenv->close(dbenv, 0);
 
 	return (XAER_RMERR);
 }
@@ -244,41 +242,45 @@ err:	(void)env->close(env, 0);
  * environment that has never been opened).
  */
 static int
-__db_xa_close(xa_info, rmid, flags)
+__db_xa_close(xa_info, rmid, arg_flags)
 	char *xa_info;
 	int rmid;
-	long flags;
+	long arg_flags;
 {
-	DB_ENV *env;
+	DB_ENV *dbenv;
 	DB_TXN *t;
 	int ret, t_ret;
+	u_long flags;
 
 	COMPQUIET(xa_info, NULL);
 
+	flags = (u_long)arg_flags;	/* Conversion for bit operations. */
+
 	if (LF_ISSET(TMASYNC))
 		return (XAER_ASYNC);
 	if (flags != TMNOFLAGS)
 		return (XAER_INVAL);
 
 	/* If the environment is closed, then we're done. */
-	if (__db_rmid_to_env(rmid, &env) != 0)
+	if (__db_rmid_to_env(rmid, &dbenv) != 0)
 		return (XA_OK);
 
 	/* Check if there are any pending transactions. */
-	if ((t = TAILQ_FIRST(&env->xa_txn)) != NULL && t->txnid != TXN_INVALID)
+	if ((t = TAILQ_FIRST(&dbenv->xa_txn)) != NULL &&
+	    t->txnid != TXN_INVALID)
 		return (XAER_PROTO);
 
 	/* Destroy the mapping. */
 	ret = __db_unmap_rmid(rmid);
 
 	/* Discard space held for the current transaction. */
-	while ((t = TAILQ_FIRST(&env->xa_txn)) != NULL) {
-		TAILQ_REMOVE(&env->xa_txn, t, xalinks);
-		__os_free(env, t);
+	while ((t = TAILQ_FIRST(&dbenv->xa_txn)) != NULL) {
+		TAILQ_REMOVE(&dbenv->xa_txn, t, xalinks);
+		__os_free(dbenv, t);
 	}
 
 	/* Close the environment. */
-	if ((t_ret = env->close(env, 0)) != 0 && ret == 0)
+	if ((t_ret = dbenv->close(dbenv, 0)) != 0 && ret == 0)
 		ret = t_ret;
 
 	return (ret == 0 ? XA_OK : XAER_RMERR);
@@ -289,17 +291,20 @@ __db_xa_close(xa_info, rmid, flags)
  *	Begin a transaction for the current resource manager.
  */
 static int
-__db_xa_start(xid, rmid, flags)
+__db_xa_start(xid, rmid, arg_flags)
 	XID *xid;
 	int rmid;
-	long flags;
+	long arg_flags;
 {
-	DB_ENV *env;
+	DB_ENV *dbenv;
 	DB_TXN *txnp;
 	TXN_DETAIL *td;
-	size_t off;
+	roff_t off;
+	u_long flags;
 	int is_known;
 
+	flags = (u_long)arg_flags;	/* Conversion for bit operations. */
+
 #define	OK_FLAGS	(TMJOIN | TMRESUME | TMNOWAIT | TMASYNC | TMNOFLAGS)
 	if (LF_ISSET(~OK_FLAGS))
 		return (XAER_INVAL);
@@ -310,10 +315,10 @@ __db_xa_start(xid, rmid, flags)
 	if (LF_ISSET(TMASYNC))
 		return (XAER_ASYNC);
 
-	if (__db_rmid_to_env(rmid, &env) != 0)
+	if (__db_rmid_to_env(rmid, &dbenv) != 0)
 		return (XAER_PROTO);
 
-	is_known = __db_xid_to_txn(env, xid, &off) == 0;
+	is_known = __db_xid_to_txn(dbenv, xid, &off) == 0;
 
 	if (is_known && !LF_ISSET(TMRESUME) && !LF_ISSET(TMJOIN))
 		return (XAER_DUPID);
@@ -327,8 +332,7 @@ __db_xa_start(xid, rmid, flags)
 	 * Other error conditions: RMERR, RMFAIL, OUTSIDE, PROTO, RB*
 	 */
 	if (is_known) {
-		td = (TXN_DETAIL *)
-		    R_ADDR(&((DB_TXNMGR *)env->tx_handle)->reginfo, off);
+		td = R_ADDR(&((DB_TXNMGR *)dbenv->tx_handle)->reginfo, off);
 		if (td->xa_status == TXN_XA_SUSPENDED &&
 		    !LF_ISSET(TMRESUME | TMJOIN))
 			return (XAER_PROTO);
@@ -338,18 +342,17 @@ __db_xa_start(xid, rmid, flags)
 			return (XA_RBOTHER);
 
 		/* Now, fill in the global transaction structure. */
-		if (__xa_get_txn(env, &txnp, 1) != 0)
+		if (__xa_get_txn(dbenv, &txnp, 1) != 0)
 			return (XAER_RMERR);
-		__txn_continue(env, txnp, td, off);
+		__txn_continue(dbenv, txnp, td);
 		td->xa_status = TXN_XA_STARTED;
 	} else {
-		if (__xa_get_txn(env, &txnp, 1) != 0)
+		if (__xa_get_txn(dbenv, &txnp, 1) != 0)
 			return (XAER_RMERR);
-		if (__txn_xa_begin(env, txnp))
+		if (__txn_xa_begin(dbenv, txnp))
 			return (XAER_RMERR);
-		(void)__db_map_xid(env, xid, txnp->off);
-		td = (TXN_DETAIL *)
-		    R_ADDR(&((DB_TXNMGR *)env->tx_handle)->reginfo, txnp->off);
+		(void)__db_map_xid(dbenv, xid, txnp->td);
+		td = txnp->td;
 		td->xa_status = TXN_XA_STARTED;
 	}
 	return (XA_OK);
@@ -365,27 +368,27 @@ __db_xa_end(xid, rmid, flags)
 	int rmid;
 	long flags;
 {
-	DB_ENV *env;
+	DB_ENV *dbenv;
 	DB_TXN *txn;
 	TXN_DETAIL *td;
-	size_t off;
+	roff_t off;
 
 	if (flags != TMNOFLAGS && !LF_ISSET(TMSUSPEND | TMSUCCESS | TMFAIL))
 		return (XAER_INVAL);
 
-	if (__db_rmid_to_env(rmid, &env) != 0)
+	if (__db_rmid_to_env(rmid, &dbenv) != 0)
 		return (XAER_PROTO);
 
-	if (__db_xid_to_txn(env, xid, &off) != 0)
+	if (__db_xid_to_txn(dbenv, xid, &off) != 0)
 		return (XAER_NOTA);
 
-	if (__xa_get_txn(env, &txn, 0) != 0)
+	if (__xa_get_txn(dbenv, &txn, 0) != 0)
 		return (XAER_RMERR);
 
-	if (off != txn->off)
+	td = R_ADDR(&((DB_TXNMGR *)dbenv->tx_handle)->reginfo, off);
+	if (td != txn->td)
 		return (XAER_PROTO);
 
-	td = (TXN_DETAIL *)R_ADDR(&((DB_TXNMGR *)env->tx_handle)->reginfo, off);
 	if (td->xa_status == TXN_XA_DEADLOCKED)
 		return (XA_RBDEADLOCK);
 
@@ -395,9 +398,6 @@ __db_xa_end(xid, rmid, flags)
 	if (td->xa_status != TXN_XA_STARTED)
 		return (XAER_PROTO);
 
-	/* Update the shared memory last_lsn field */
-	td->last_lsn = txn->last_lsn;
-
 	/*
 	 * If we ever support XA migration, we cannot keep SUSPEND/END
 	 * status in the shared region; it would have to be process local.
@@ -407,7 +407,11 @@ __db_xa_end(xid, rmid, flags)
 	else
 		td->xa_status = TXN_XA_ENDED;
 
-	__xa_put_txn(env, txn);
+	/*
+	 * XXX
+	 * This can fail in XA_MULTI_THREAD mode.
+	 */
+	(void)__xa_put_txn(dbenv, txn);
 	return (XA_OK);
 }
 
@@ -416,15 +420,18 @@ __db_xa_end(xid, rmid, flags)
  *	Sync the log to disk so we can guarantee recoverability.
  */
 static int
-__db_xa_prepare(xid, rmid, flags)
+__db_xa_prepare(xid, rmid, arg_flags)
 	XID *xid;
 	int rmid;
-	long flags;
+	long arg_flags;
 {
-	DB_ENV *env;
+	DB_ENV *dbenv;
 	DB_TXN *txnp;
 	TXN_DETAIL *td;
-	size_t off;
+	roff_t off;
+	u_long flags;
+
+	flags = (u_long)arg_flags;	/* Conversion for bit operations. */
 
 	if (LF_ISSET(TMASYNC))
 		return (XAER_ASYNC);
@@ -437,12 +444,12 @@ __db_xa_prepare(xid, rmid, flags)
 	 * reflect that fact that prepare has been called, and if
 	 * it's ever called again, it's an error.
 	 */
-	if (__db_rmid_to_env(rmid, &env) != 0)
+	if (__db_rmid_to_env(rmid, &dbenv) != 0)
 		return (XAER_PROTO);
 
-	if (__db_xid_to_txn(env, xid, &off) != 0)
+	if (__db_xid_to_txn(dbenv, xid, &off) != 0)
 		return (XAER_NOTA);
-	td = (TXN_DETAIL *)R_ADDR(&((DB_TXNMGR *)env->tx_handle)->reginfo, off);
+	td = R_ADDR(&((DB_TXNMGR *)dbenv->tx_handle)->reginfo, off);
 	if (td->xa_status == TXN_XA_DEADLOCKED)
 		return (XA_RBDEADLOCK);
 
@@ -450,17 +457,20 @@ __db_xa_prepare(xid, rmid, flags)
 		return (XAER_PROTO);
 
 	/* Now, fill in the global transaction structure. */
-	if (__xa_get_txn(env, &txnp, 0) != 0)
+	if (__xa_get_txn(dbenv, &txnp, 0) != 0)
 		return (XAER_PROTO);
-	__txn_continue(env, txnp, td, off);
+	__txn_continue(dbenv, txnp, td);
 
 	if (txnp->prepare(txnp, (u_int8_t *)xid->data) != 0)
 		return (XAER_RMERR);
 
 	td->xa_status = TXN_XA_PREPARED;
 
-	/* No fatal value that would require an XAER_RMFAIL. */
-	__xa_put_txn(env, txnp);
+	/*
+	 * XXX
+	 * This can fail in XA_MULTI_THREAD mode.
+	 */
+	(void)__xa_put_txn(dbenv, txnp);
 	return (XA_OK);
 }
 
@@ -469,15 +479,18 @@ __db_xa_prepare(xid, rmid, flags)
  *	Commit the transaction
  */
 static int
-__db_xa_commit(xid, rmid, flags)
+__db_xa_commit(xid, rmid, arg_flags)
 	XID *xid;
 	int rmid;
-	long flags;
+	long arg_flags;
 {
-	DB_ENV *env;
+	DB_ENV *dbenv;
 	DB_TXN *txnp;
 	TXN_DETAIL *td;
-	size_t off;
+	roff_t off;
+	u_long flags;
+
+	flags = (u_long)arg_flags;	/* Conversion for bit operations. */
 
 	if (LF_ISSET(TMASYNC))
 		return (XAER_ASYNC);
@@ -490,13 +503,13 @@ __db_xa_commit(xid, rmid, flags)
 	 * We need to know if we've ever called prepare on this.
 	 * We can verify this by examining the xa_status field.
 	 */
-	if (__db_rmid_to_env(rmid, &env) != 0)
+	if (__db_rmid_to_env(rmid, &dbenv) != 0)
 		return (XAER_PROTO);
 
-	if (__db_xid_to_txn(env, xid, &off) != 0)
+	if (__db_xid_to_txn(dbenv, xid, &off) != 0)
 		return (XAER_NOTA);
 
-	td = (TXN_DETAIL *)R_ADDR(&((DB_TXNMGR *)env->tx_handle)->reginfo, off);
+	td = R_ADDR(&((DB_TXNMGR *)dbenv->tx_handle)->reginfo, off);
 	if (td->xa_status == TXN_XA_DEADLOCKED)
 		return (XA_RBDEADLOCK);
 
@@ -511,15 +524,18 @@ __db_xa_commit(xid, rmid, flags)
 		return (XAER_PROTO);
 
 	/* Now, fill in the global transaction structure. */
-	if (__xa_get_txn(env, &txnp, 0) != 0)
+	if (__xa_get_txn(dbenv, &txnp, 0) != 0)
 		return (XAER_RMERR);
-	__txn_continue(env, txnp, td, off);
+	__txn_continue(dbenv, txnp, td);
 
 	if (txnp->commit(txnp, 0) != 0)
 		return (XAER_RMERR);
 
-	/* No fatal value that would require an XAER_RMFAIL. */
-	__xa_put_txn(env, txnp);
+	/*
+	 * XXX
+	 * This can fail in XA_MULTI_THREAD mode.
+	 */
+	(void)__xa_put_txn(dbenv, txnp);
 	return (XA_OK);
 }
 
@@ -537,12 +553,12 @@ __db_xa_recover(xids, count, rmid, flags)
 	long count, flags;
 	int rmid;
 {
-	DB_ENV *env;
+	DB_ENV *dbenv;
 	u_int32_t newflags;
 	long rval;
 
 	/* If the environment is closed, then we're done. */
-	if (__db_rmid_to_env(rmid, &env) != 0)
+	if (__db_rmid_to_env(rmid, &dbenv) != 0)
 		return (XAER_PROTO);
 
 	if (LF_ISSET(TMSTARTRSCAN))
@@ -553,7 +569,7 @@ __db_xa_recover(xids, count, rmid, flags)
 		newflags = DB_NEXT;
 
 	rval = 0;
-	if (__txn_get_prepared(env, xids, NULL, count, &rval, newflags) != 0)
+	if (__txn_get_prepared(dbenv, xids, NULL, count, &rval, newflags) != 0)
 		return (XAER_RMERR);
 	else
 		return (rval);
@@ -564,28 +580,31 @@ __db_xa_recover(xids, count, rmid, flags)
  *	Abort an XA transaction.
  */
 static int
-__db_xa_rollback(xid, rmid, flags)
+__db_xa_rollback(xid, rmid, arg_flags)
 	XID *xid;
 	int rmid;
-	long flags;
+	long arg_flags;
 {
-	DB_ENV *env;
+	DB_ENV *dbenv;
 	DB_TXN *txnp;
 	TXN_DETAIL *td;
-	size_t off;
+	roff_t off;
+	u_long flags;
+
+	flags = (u_long)arg_flags;	/* Conversion for bit operations. */
 
 	if (LF_ISSET(TMASYNC))
 		return (XAER_ASYNC);
 	if (flags != TMNOFLAGS)
 		return (XAER_INVAL);
 
-	if (__db_rmid_to_env(rmid, &env) != 0)
+	if (__db_rmid_to_env(rmid, &dbenv) != 0)
 		return (XAER_PROTO);
 
-	if (__db_xid_to_txn(env, xid, &off) != 0)
+	if (__db_xid_to_txn(dbenv, xid, &off) != 0)
 		return (XAER_NOTA);
 
-	td = (TXN_DETAIL *)R_ADDR(&((DB_TXNMGR *)env->tx_handle)->reginfo, off);
+	td = R_ADDR(&((DB_TXNMGR *)dbenv->tx_handle)->reginfo, off);
 	if (td->xa_status == TXN_XA_DEADLOCKED)
 		return (XA_RBDEADLOCK);
 
@@ -598,14 +617,17 @@ __db_xa_rollback(xid, rmid, flags)
 		return (XAER_PROTO);
 
 	/* Now, fill in the global transaction structure. */
-	if (__xa_get_txn(env, &txnp, 0) != 0)
+	if (__xa_get_txn(dbenv, &txnp, 0) != 0)
 		return (XAER_RMERR);
-	__txn_continue(env, txnp, td, off);
+	__txn_continue(dbenv, txnp, td);
 	if (txnp->abort(txnp) != 0)
 		return (XAER_RMERR);
 
-	/* No fatal value that would require an XAER_RMFAIL. */
-	__xa_put_txn(env, txnp);
+	/*
+	 * XXX
+	 * This can fail in XA_MULTI_THREAD mode.
+	 */
+	(void)__xa_put_txn(dbenv, txnp);
 	return (XA_OK);
 }
 
@@ -617,29 +639,32 @@ __db_xa_rollback(xid, rmid, flags)
  * that we reclaim the slots in the txnid table.
  */
 static int
-__db_xa_forget(xid, rmid, flags)
+__db_xa_forget(xid, rmid, arg_flags)
 	XID *xid;
 	int rmid;
-	long flags;
+	long arg_flags;
 {
-	DB_ENV *env;
-	size_t off;
+	DB_ENV *dbenv;
+	roff_t off;
+	u_long flags;
+
+	flags = (u_long)arg_flags;	/* Conversion for bit operations. */
 
 	if (LF_ISSET(TMASYNC))
 		return (XAER_ASYNC);
 	if (flags != TMNOFLAGS)
 		return (XAER_INVAL);
 
-	if (__db_rmid_to_env(rmid, &env) != 0)
+	if (__db_rmid_to_env(rmid, &dbenv) != 0)
 		return (XAER_PROTO);
 
 	/*
 	 * If mapping is gone, then we're done.
 	 */
-	if (__db_xid_to_txn(env, xid, &off) != 0)
+	if (__db_xid_to_txn(dbenv, xid, &off) != 0)
 		return (XA_OK);
 
-	__db_unmap_xid(env, xid, off);
+	__db_unmap_xid(dbenv, xid, off);
 
 	/* No fatal value that would require an XAER_RMFAIL. */
 	return (XA_OK);
